//------------------------------------------------------------------------
// File for project XCALC
//------------------------------------------------------------------------
// User draw button procedure
// Actions:
//	- draw border at window edge
//	- draw 3d button
//	- fixed text
//------------------------------------------------------------------------

#include "xcalc.h"

const char * const aButtonText[NBUTTONS] = {
	"Configure","Help","Exit"
};

int aButtonUnder[NBUTTONS] = {  // specifies which letter is underlined
	2,3,1
};

// ToolTip added July 00

char *aButDesc[NBUTTONS] = {
	"Configuration [Alt-N]",
	"Help [F1]",
	"Exit [Alt-X or Alt-F4]"
};

LRESULT CALLBACK ButtonProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	int iWindow;
	static HWND ButtonDownHWND = NULL; // the onscreen button
	static BOOL MouseDown = FALSE;	// TRUE when mousebutton down
	RECT rect;
	
	// find which window this is (iDispMode)
	iWindow = GetWindowLong(hwnd,GWL_ID);
	
	// Relay to XCALC ToolTip window
	GetClientRect(hwnd,&rect);
	rect.left += Border;
	rect.top += Border;
	rect.right -= Border;
	rect.bottom -= Border;
	ToolTipEvent(hwnd,wMsg,wParam,lParam,&rect,aButDesc[iWindow]);

	switch (wMsg) {
		
	case WM_LBUTTONDOWN:
		LButtonDown(hwnd,lParam,&ButtonDownHWND,&MouseDown);
		break;
		
	case WM_LBUTTONUP:
		if (LButtonUp(hwnd,lParam,&ButtonDownHWND,&MouseDown))
			PostMessage(GetParent(hwnd),PM_BUTTON+iWindow,0,0);
		break;
		
	case WM_MOUSEMOVE:
		if (!MouseDown) break;
		MouseMove(hwnd,lParam,&ButtonDownHWND);
		break;
		
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			RECT rect;
			int i;
			SIZE siz;
			
			BeginPaint(hwnd,&ps);
			
			GetClientRect(hwnd,&rect);
			FillRect(ps.hdc,&rect,hBackBrush);
			MakeBorder(ps.hdc,&rect);
			
			//3d look around dot
			
			DrawButton(ps.hdc,ButtonDownHWND==hwnd);
			
			// Draw fixed text
			// special treatment for Configure -> Config
			
			i = lstrlen(aButtonText[iWindow]);
			SelectObject(ps.hdc,hBigFont);
			GetTextExtentPoint32(ps.hdc,aButtonText[iWindow],i,&siz);
			if (lstrcmp(aButtonText[iWindow],"Configure")==0 && siz.cx+ButHeight+Border>=ButWidth)
				i = 6;
			
			DrawButtonText(ps.hdc,&rect,GetColour(RGBBUTTON),aButtonText[iWindow],aButtonUnder[iWindow],i);
			
			EndPaint(hwnd,&ps);
			
			break;
		} // end case WMPAINT
		
	default:
		return DefWindowProc(hwnd, wMsg, wParam, lParam);
	}
	return 0L;
}
