//------------------------------------------------------------------------
// File for project XCALC
//------------------------------------------------------------------------
// Dialogue box procedures
//------------------------------------------------------------------------

#include "xcalc.h"

// Colour item names

const char * const ItemTitle[] = {
	"Background",
	"Border",
	"Bright 3D edge",
	"Dark 3D edge",
	"Button text",
	"Gray button text",
	"Selected button",
	"Message",
	"Caption",
	"Keycap text",
	"Digits"
};

// Predefined colour values (may change if [PreDef] is set in .INI file)

COLORREF PreDefColours[16];

const char * const PreDefStr[16] = {
	"  0   0   0",
	" 20  20  20",
	" 64  64  64",
	"128 128 128",
	"192 192 192",
	"255 255 255",
	"255   0   0",
	"  0 255   0",
	"  0   0 255",
	"255 255   0",
	"255   0 255",
	"  0 255 255",
	"128   0   0",
	"  0 128   0",
	"  0   0 128",
	"  0   0 192"
};

// to restore default colours (from getcolou.c)

extern char * DfltCols[NCOLOURS];

// messages from constants/conversion sections

char ConstMsg[128];
char ConvMsg[128];

//------------------------------------------------------------------------
// Property sheets - i.e. config box
//------------------------------------------------------------------------

#define NPROPPAGES 4
enum Proppages {PP_GENERAL, PP_NORMAL, PP_MINI, PP_COLOURS};

static int dlgStartPage=PP_GENERAL; // per-session

const char * const ProppageName[NPROPPAGES] = {
	"General",
	"Normal",
	"Mini",
	"Colours",
};

const char * const IconName[NPROPPAGES] = {
	"XXGENERAL",
	"XXNORMAL",
	"XXMINI",
	"XXCOLOURS",
};

const DLGPROC ProppageProc[NPROPPAGES] = {
	GeneralDlgProc,
	NormalDlgProc,
	MiniDlgProc,
	ColoursDlgProc,
};

const char * const ProppageTemplate[NPROPPAGES] = {
	"PROPGENERAL",
	"PROPNORMAL",
	"PROPMINI",
	"PROPCOLOURS",
};

int DoPropertySheet(HWND hwndOwner)
{
    PROPSHEETPAGE psp[NPROPPAGES];
    PROPSHEETHEADER psh;
	int i;

	for (i=0;i<NPROPPAGES;i++) {
		psp[i].dwSize = sizeof(PROPSHEETPAGE);
		psp[i].dwFlags = PSP_USETITLE|PSP_HASHELP|PSP_USEHICON;
		psp[i].hInstance = hInst;
#ifdef __BORLANDC__
		psp[i].u2.hIcon = LoadIcon(hInst,IconName[i]);
		psp[i].u.pszTemplate = ProppageTemplate[i];
#else
		psp[i].hIcon = LoadIcon(hInst,IconName[i]);
		psp[i].pszTemplate = ProppageTemplate[i];
#endif
		psp[i].pfnDlgProc = ProppageProc[i];
		psp[i].pszTitle = ProppageName[i];
		psp[i].lParam = 0;
		psp[i].pfnCallback = NULL;
	}
	
    psh.dwSize = sizeof(PROPSHEETHEADER);
    psh.dwFlags = PSH_PROPSHEETPAGE|PSH_NOAPPLYNOW|PSH_HASHELP|PSH_USEHICON;
    psh.hwndParent = hwndOwner;
    psh.hInstance = hInst;
    psh.pszCaption = (LPSTR) "XCALC Configuration";
    psh.nPages = NPROPPAGES;
	// 19sep06: If normal or mini page was active in last config session,
	// switch to page for current mode.
	if (Mini && dlgStartPage==PP_NORMAL) dlgStartPage = PP_MINI;
	else if (!Mini && dlgStartPage==PP_MINI) dlgStartPage = PP_NORMAL;
#ifdef __BORLANDC__
    psh.u2.nStartPage = dlgStartPage;
#else
    psh.nStartPage = dlgStartPage;
#endif
#ifdef __BORLANDC__
    psh.u3.ppsp = (LPCPROPSHEETPAGE) &psp;
#else
    psh.ppsp = (LPCPROPSHEETPAGE) &psp;
#endif
    psh.pfnCallback = NULL;
	
	Restart = FALSE;
	RefreshNow = FALSE;
    return PropertySheet(&psh);
}

// These settings are shared between general property sheet and about box

LRESULT CALLBACK MailButtonProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam);
WNDPROC OldMailButtonProc;
HCURSOR HandCursor = NULL;
BOOL CommButtonIsPressed = FALSE;
BOOL MailVisited = FALSE;
BOOL WebVisited = FALSE;

//------------------------------------------------------------------------
// General proprty page procedure - sets global variables. Sets Restart if
// StkSiz, Size or Bitmap setting change. Sets RefreshNow if settings
// require new rendering of numbers (ThousandSep/ThousandthSep changes).
//------------------------------------------------------------------------
BOOL CALLBACK GeneralDlgProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	static BOOL _SoundOn=FALSE;
	static BOOL _SaveRegs;
	static BOOL _ThousandSep;
	static BOOL _ThousandthSep;
	static BOOL _CleanFrac;
	static BOOL _ToolTip;
	static BOOL _ZeroOverflow;
	static BOOL _CopyTop;
	static BOOL _InvTrigAsDMS;
	static BOOL _MessagePopup;
	static BOOL _AllowEdit;
	static int _MinStkSiz;
	static BOOL _Mini;
	static BOOL _Cstyle;
	static BOOL _k_useComma;
	static BOOL _d_useComma;
	static BOOL _c_useComma;
	static int _nKeyRows;
	static int _nKeyCols;

	switch (wMsg)
	{
	case WM_INITDIALOG: {
		DWORD dwStyle;
		HWND mailWnd, webWnd;
		// Subclass mailto and web buttons
		mailWnd = GetDlgItem(hwnd,IDD_MAIL);
		webWnd = GetDlgItem(hwnd,IDD_WEB);
		OldMailButtonProc = (WNDPROC)SetWindowLong(mailWnd,GWL_WNDPROC,(LONG)MailButtonProc);
		SetWindowLong(webWnd,GWL_WNDPROC,(LONG)MailButtonProc);
		// And let them get mouse clicks but no default cursor
		dwStyle = GetWindowLong(mailWnd, GWL_STYLE);
		SetWindowLong(mailWnd, GWL_STYLE, dwStyle | SS_NOTIFY);
		SetClassWord(mailWnd, GCL_HCURSOR, (WORD)NULL);
		dwStyle = GetWindowLong(webWnd, GWL_STYLE);
		SetWindowLong(webWnd, GWL_STYLE, dwStyle | SS_NOTIFY);
		SetClassWord(webWnd, GCL_HCURSOR, (WORD)NULL);
		// Load their cursor
		HandCursor = LoadCursor(hInst,"HANDCUR");
		// Setup local variables
		_SoundOn=SoundOn;
		_SaveRegs=SaveRegs;
		_ThousandSep=ThousandSep;
		_ThousandthSep=ThousandthSep;
		_CleanFrac=bCleanFrac;
		_ToolTip=ToolTip;
		_ZeroOverflow=ZeroOverflow;
		_CopyTop=CopyTop;
		_InvTrigAsDMS=InvTrigAsDMS;
		_MessagePopup=MessagePopup;
		_AllowEdit=AllowEdit;
		_MinStkSiz=MinStkSiz;
		_Mini=Mini;
		_Cstyle=Cstyle;
		_k_useComma=k_useComma;
		_d_useComma=d_useComma;
		_c_useComma=c_useComma;
		_nKeyRows=nKeyRows;
		_nKeyCols=nKeyCols;
		// Set buttons
		CheckDlgButton(hwnd,IDD_SNDON,_SoundOn);
		CheckDlgButton(hwnd,IDD_SAVEON,_SaveRegs);
		CheckDlgButton(hwnd,IDD_THOUSAND,_ThousandSep);
		CheckDlgButton(hwnd,IDD_THOUSANDTH,_ThousandthSep);
		CheckDlgButton(hwnd,IDD_CLEANFRAC,_CleanFrac);
		CheckDlgButton(hwnd,IDD_TOOLTIP,_ToolTip);
		CheckDlgButton(hwnd,IDD_ZEROOVERFLOW,_ZeroOverflow);
		CheckDlgButton(hwnd,IDD_COPYTOP,_CopyTop);
		CheckDlgButton(hwnd,IDD_INVTRIGASDMS,_InvTrigAsDMS);
		CheckDlgButton(hwnd,IDD_MESSAGEPOPUP,_MessagePopup);
		CheckDlgButton(hwnd,IDD_ALLOWEDIT,_AllowEdit);
		SetDlgItemInt(hwnd,IDD_STACKSIZE,_MinStkSiz,FALSE);
		CheckRadioButton(hwnd,IDD_NORM,IDD_MINI,_Mini?IDD_MINI:IDD_NORM);
		CheckRadioButton(hwnd,IDD_CSTYLE,IDD_BASIC,_Cstyle?IDD_CSTYLE:IDD_BASIC);
		CheckRadioButton(hwnd,IDD_K_POINT,IDD_K_COMMA,_k_useComma?IDD_K_COMMA:IDD_K_POINT);
		CheckRadioButton(hwnd,IDD_D_POINT,IDD_D_COMMA,_d_useComma?IDD_D_COMMA:IDD_D_POINT);
		CheckRadioButton(hwnd,IDD_C_POINT,IDD_C_COMMA,_c_useComma?IDD_C_COMMA:IDD_C_POINT);
		SetDlgItemInt(hwnd,IDD_KEYROWS,_nKeyRows,FALSE);
		SetDlgItemInt(hwnd,IDD_KEYCOLS,_nKeyCols,FALSE);
		SendMessage(hwnd,WM_COMMAND,IDD_ALLOWEDIT,0);
		return TRUE;
	}
	// end case WM_INITDIALOG:

	case WM_CTLCOLORSTATIC: {
		int id = GetDlgCtrlID((HWND)lParam);
		if (id == IDD_WEB || id == IDD_MAIL) {
			if (CommButtonIsPressed) {
				SetTextColor((HDC)wParam,URLPRESSEDCOL);
			}
			else {
				if (id == IDD_WEB && WebVisited || id == IDD_MAIL && MailVisited)
					SetTextColor((HDC)wParam,URLVISITEDCOL);
				else
					SetTextColor((HDC)wParam,URLCOL);
			}
			SetBkColor((HDC)wParam,GetSysColor(COLOR_BTNFACE));
			return (LONG)GetSysColorBrush(COLOR_BTNFACE);
		}
		return FALSE;
	}
	// end case WM_CTLCOLORSTATIC

	case WM_MOUSEMOVE:
		if (CommButtonIsPressed) {
			CommButtonIsPressed = FALSE;
			InvalidateRect(GetDlgItem(hwnd,IDD_WEB),NULL,TRUE);
			InvalidateRect(GetDlgItem(hwnd,IDD_MAIL),NULL,TRUE);
		}
		return TRUE;
	// end case WM_MOUSEMOVE:
		
	case WM_NOTIFY: {
		LPNMHDR pnmh = (LPNMHDR)lParam;
		char ach[32];
		char helpstr[_MAX_PATH+MAX_HELP];
		switch (pnmh->code)
		{
		case PSN_SETACTIVE:
			dlgStartPage=PP_GENERAL;
			return TRUE;
		case PSN_KILLACTIVE:
			SetWindowLong(hwnd,DWL_MSGRESULT,PSNRET_NOERROR);
			return TRUE;
		case PSN_APPLY:
			_SoundOn 		= IsDlgButtonChecked(hwnd,IDD_SNDON);
			_SaveRegs		= IsDlgButtonChecked(hwnd,IDD_SAVEON);
			_ThousandSep	= IsDlgButtonChecked(hwnd,IDD_THOUSAND);
			_ThousandthSep	= IsDlgButtonChecked(hwnd,IDD_THOUSANDTH);
			_CleanFrac		= IsDlgButtonChecked(hwnd,IDD_CLEANFRAC);
			_ToolTip		= IsDlgButtonChecked(hwnd,IDD_TOOLTIP);
			_ZeroOverflow	= IsDlgButtonChecked(hwnd,IDD_ZEROOVERFLOW);
			_CopyTop		= IsDlgButtonChecked(hwnd,IDD_COPYTOP);
			_InvTrigAsDMS   = IsDlgButtonChecked(hwnd,IDD_INVTRIGASDMS);
			_MessagePopup   = IsDlgButtonChecked(hwnd,IDD_MESSAGEPOPUP);
			_AllowEdit		= IsDlgButtonChecked(hwnd,IDD_ALLOWEDIT);
			_MinStkSiz      = GetDlgItemInt(hwnd,IDD_STACKSIZE,NULL,FALSE);
			_Mini			= IsDlgButtonChecked(hwnd,IDD_MINI);
			_Cstyle			= IsDlgButtonChecked(hwnd,IDD_CSTYLE);
			_k_useComma		= IsDlgButtonChecked(hwnd,IDD_K_COMMA);
			_d_useComma		= IsDlgButtonChecked(hwnd,IDD_D_COMMA);
			_c_useComma		= IsDlgButtonChecked(hwnd,IDD_C_COMMA);
			_nKeyRows       = GetDlgItemInt(hwnd,IDD_KEYROWS,NULL,FALSE);
			_nKeyCols       = GetDlgItemInt(hwnd,IDD_KEYCOLS,NULL,FALSE);
			// Check if changes require rebuild of xcalc
			RefreshNow |= (ThousandSep != _ThousandSep || ThousandthSep != _ThousandthSep || bCleanFrac != _CleanFrac);
			Restart |= (Mini!=_Mini || MinStkSiz!=_MinStkSiz || nKeyRows!=_nKeyRows || nKeyCols!=_nKeyCols);
			NewKB = nKeyRows!=_nKeyRows || nKeyCols!=_nKeyCols;
			SoundOn=_SoundOn;
			SaveRegs=_SaveRegs;
			ThousandSep=_ThousandSep;
			ThousandthSep=_ThousandthSep;
			bCleanFrac=_CleanFrac;
			ToolTip=_ToolTip;
			ZeroOverflow=_ZeroOverflow;
			CopyTop=_CopyTop;
			InvTrigAsDMS=_InvTrigAsDMS;
			MessagePopup=_MessagePopup;
			AllowEdit=_AllowEdit;
			MinStkSiz=_MinStkSiz;
			Mini=_Mini;
			Cstyle=_Cstyle;
			k_useComma=_k_useComma;
			d_useComma=_d_useComma;
			c_useComma=_c_useComma;
			nKeyRows=_nKeyRows;
			nKeyCols=_nKeyCols;
			WritePrivateProfileString(ClassName,"Sound",itoa(SoundOn,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"SaveRegs",itoa(SaveRegs,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"ThousandSep",itoa(ThousandSep,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"ThousandthSep",itoa(ThousandthSep,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"ToolTip",itoa(ToolTip,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"ZeroOverflow",itoa(ZeroOverflow,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"CopyTop",itoa(CopyTop,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"InvTrigAsDMS",itoa(InvTrigAsDMS,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"MessagePopup",itoa(MessagePopup,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"AllowEdit",itoa(AllowEdit,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"MinStkSiz",itoa(MinStkSiz,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"k_useComma",itoa(k_useComma,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"d_useComma",itoa(d_useComma,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"c_useComma",itoa(c_useComma,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"KeyRows",itoa(nKeyRows,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"KeyCols",itoa(nKeyCols,ach,10),IniFile);
			// Apply fudge to keyboard
			KeyboardDecimal(k_useComma);
			SetWindowLong(hwnd,DWL_MSGRESULT,PSNRET_NOERROR);
			return TRUE;
		case PSN_HELP:
            // Display help for the properties page.
			strcpy(helpstr,HelpFile);
			strcat(helpstr,GENERALHELP);
			HtmlHelp(GetParent(hwnd),helpstr,HH_DISPLAY_TOPIC,0);
            return TRUE;
		} // switch(pnmh->code)
		break;
	}
	// end case WM_NOTIFY

	case WM_COMMAND: {
		WORD wID = GET_WM_COMMAND_ID(wParam,lParam);
		//WORD cmd = GET_WM_COMMAND_CMD(wParam,lParam);
		int translated;
		switch (wID) {
		case IDD_STACKSIZE:
			// Check display stack size and change if illegal value
			_MinStkSiz = GetDlgItemInt(hwnd,IDD_STACKSIZE,&translated,FALSE);
			if (!translated || _MinStkSiz<4 || _MinStkSiz>MAXSTKSIZ) {
				if (_MinStkSiz<4) _MinStkSiz=4;
				if (_MinStkSiz>MAXSTKSIZ) _MinStkSiz=MAXSTKSIZ;
				SetDlgItemInt(hwnd,IDD_STACKSIZE,_MinStkSiz,FALSE);
				SendDlgItemMessage(hwnd,IDD_STACKSIZE,EM_SETSEL,0,MAKELONG(0,-1));
			}
			return TRUE;
		case IDD_KEYROWS:
			// Check rows and change if illegal value
			_nKeyRows = GetDlgItemInt(hwnd,IDD_KEYROWS,&translated,FALSE);
			if (!translated || _nKeyRows<1 || _nKeyRows>16) {
				if (_nKeyRows<1) _nKeyRows=1;
				if (_nKeyRows>16) _nKeyRows=16;
				SetDlgItemInt(hwnd,IDD_KEYROWS,_nKeyRows,FALSE);
				SendDlgItemMessage(hwnd,IDD_KEYROWS,EM_SETSEL,0,MAKELONG(0,-1));
			}
			return TRUE;
		case IDD_KEYCOLS:
			// Check rows and change if illegal value
			_nKeyRows = GetDlgItemInt(hwnd,IDD_KEYCOLS,&translated,FALSE);
			if (!translated || _nKeyRows<1 || _nKeyRows>12) {
				if (_nKeyRows<1) _nKeyRows=1;
				if (_nKeyRows>12) _nKeyRows=12;
				SetDlgItemInt(hwnd,IDD_KEYCOLS,_nKeyRows,FALSE);
				SendDlgItemMessage(hwnd,IDD_KEYCOLS,EM_SETSEL,0,MAKELONG(0,-1));
			}
			return TRUE;
		case IDD_ALLOWEDIT:
			// Enable/disable rows/cols fields
			EnableWindow(GetDlgItem(hwnd,IDD_KEYROWS),IsDlgButtonChecked(hwnd,IDD_ALLOWEDIT));
			EnableWindow(GetDlgItem(hwnd,IDD_KEYCOLS),IsDlgButtonChecked(hwnd,IDD_ALLOWEDIT));
			EnableWindow(GetDlgItem(hwnd,IDD_SCROLLROWS),IsDlgButtonChecked(hwnd,IDD_ALLOWEDIT));
			EnableWindow(GetDlgItem(hwnd,IDD_SCROLLCOLS),IsDlgButtonChecked(hwnd,IDD_ALLOWEDIT));
			return TRUE;
		} // switch (wID)
	}
	// end case WM_COMMAND

	case WM_VSCROLL: {
		int ScrollCode = GET_WM_VSCROLL_CODE(wParam,lParam);
		HWND hwndScrollBar = GET_WM_VSCROLL_HWND(wParam,lParam);
		int num;

		switch (GetDlgCtrlID(hwndScrollBar)) {
		case IDD_SCROLLSTACK:
			num = (int)GetDlgItemInt(hwnd,IDD_STACKSIZE,NULL,FALSE);
			switch (ScrollCode) {
			case SB_BOTTOM:		num = 4; break;
			case SB_TOP:		num = MAXSTKSIZ; break;
			case SB_LINEDOWN:	num -= 1; break;
			case SB_LINEUP:		num += 1; break;
			case SB_PAGEDOWN:	num -= 2; break;
			case SB_PAGEUP:		num += 2; break;
			}
			if (num<4) num=4;
			if (num>MAXSTKSIZ) num=MAXSTKSIZ;
			SetDlgItemInt(hwnd,IDD_STACKSIZE,num,FALSE);
			return TRUE;
		case IDD_SCROLLROWS:
			num = (int)GetDlgItemInt(hwnd,IDD_KEYROWS,NULL,FALSE);
			switch (ScrollCode) {
			case SB_BOTTOM:		num = 1; break;
			case SB_TOP:		num = 16; break;
			case SB_LINEDOWN:	num -= 1; break;
			case SB_LINEUP:		num += 1; break;
			case SB_PAGEDOWN:	num -= 2; break;
			case SB_PAGEUP:		num += 2; break;
			}
			if (num<1) num=1;
			if (num>16) num=16;
			SetDlgItemInt(hwnd,IDD_KEYROWS,num,FALSE);
			return TRUE;
		} // switch (GetDlgCtrlID(hwndScrollBar))
		break;
	}
	// end case WM_VSCROLL
	case WM_HSCROLL: {
		int ScrollCode = GET_WM_HSCROLL_CODE(wParam,lParam);
		HWND hwndScrollBar = GET_WM_HSCROLL_HWND(wParam,lParam);
		int num;

		switch (GetDlgCtrlID(hwndScrollBar)) {
		case IDD_SCROLLCOLS:
			num = (int)GetDlgItemInt(hwnd,IDD_KEYCOLS,NULL,FALSE);
			switch (ScrollCode) {
			case SB_LEFT:		num = 1; break;
			case SB_RIGHT:		num = 12; break;
			case SB_LINELEFT:	num -= 1; break;
			case SB_LINERIGHT:	num += 1; break;
			case SB_PAGELEFT:	num -= 2; break;
			case SB_PAGERIGHT:	num += 2; break;
			}
			if (num<1) num=1;
			if (num>12) num=12;
			SetDlgItemInt(hwnd,IDD_KEYCOLS,num,FALSE);
			return TRUE;
		} // switch (GetDlgCtrlID(hwndScrollBar))
		break;
	}
	// end case WM_HSCROLL

	} // switch (wMsg)

	return FALSE;
}

BOOL CALLBACK NormalDlgProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	static BOOL _n_InTray=FALSE;
	static BOOL _n_IconStart=FALSE;
	static BOOL _n_KeepInFront=FALSE;
	static BOOL _n_EnableMouseMove=FALSE;
	static BOOL _n_ShowMemoryPopup=FALSE;
	static int _n_MemCols=1;
	static BOOL _n_ShowTitle=FALSE;
	static BOOL _n_ShowPane1=FALSE;
	static BOOL _n_ShowPane2=FALSE;
	static BOOL _n_ShowPane3=FALSE;
	static BOOL _n_ShowPane4=FALSE;
	static BOOL _n_ShowPane5=FALSE;
	static int _n_DisplayStkSiz=4;
	static int _n_Size=100;
	static BOOL _n_UseFonts=FALSE;
	static BOOL _n_Bold=FALSE;
	static char _n_BitmapName[32];
	static char _n_FontName[64];

	switch (wMsg)
	{
	case WM_INITDIALOG: {
		char ach[32];
		int i;
		HBITMAP hBitmap;
		HDC hdc;
		char BitmapTmpName[32];

		_n_InTray=n_InTray;
		_n_IconStart=n_IconStart;
		_n_KeepInFront=n_KeepInFront;
		_n_EnableMouseMove=n_EnableMouseMove;
		_n_ShowMemoryPopup=n_ShowMemoryPopup;
		_n_MemCols=n_MemCols;
		_n_ShowTitle=n_ShowTitle;
		_n_ShowPane1=n_ShowPane1;
		_n_ShowPane2=n_ShowPane2;
		_n_ShowPane3=n_ShowPane3;
		_n_ShowPane4=n_ShowPane4;
		_n_ShowPane5=n_ShowPane5;
		_n_DisplayStkSiz=n_DisplayStkSiz;
		_n_Size=n_Size;
		_n_UseFonts=n_UseFonts;
		_n_Bold=n_Bold;
		lstrcpy(_n_BitmapName,n_BitmapName);
		lstrcpy(_n_FontName,n_FontName);

		// Fill the Font Name listbox with valid entries
		hdc = GetDC(hwnd);
		EnumFontFamilies(hdc,NULL,(FONTENUMPROC)EnumFontProc,(LPARAM)hwnd);
		ReleaseDC(hwnd,hdc);

		// Fill the Bitmap Name listbox with valid entries
		lstrcpy(BitmapTmpName,"DIGITS");
		i = 1;
		lstrcat(BitmapTmpName,itoa(i,ach,10));
		while ((hBitmap = LoadBitmap(hInst,BitmapTmpName)) != NULL) {
			DeleteObject(hBitmap);
			SendDlgItemMessage(hwnd,IDD_BITMAPNAME,CB_ADDSTRING,0,(LPARAM)(LPSTR)BitmapTmpName);
			i++;
			lstrcpy(BitmapTmpName,"DIGITS");
			lstrcat(BitmapTmpName,itoa(i,ach,10));
		}

		CheckDlgButton(hwnd,IDD_INTRAY,_n_InTray);
		CheckDlgButton(hwnd,IDD_ICONON,_n_IconStart);
		CheckDlgButton(hwnd,IDD_FRONTON,_n_KeepInFront);
		CheckDlgButton(hwnd,IDD_MOUSEMOVE,_n_EnableMouseMove);
		CheckDlgButton(hwnd,IDD_POPUPON,_n_ShowMemoryPopup);
		EnableWindow(GetDlgItem(hwnd,IDD_1COL),IsDlgButtonChecked(hwnd,IDD_POPUPON));
		EnableWindow(GetDlgItem(hwnd,IDD_2COL),IsDlgButtonChecked(hwnd,IDD_POPUPON));
		CheckRadioButton(hwnd,IDD_1COL,IDD_2COL,_n_MemCols==1?IDD_1COL:IDD_2COL);
		CheckDlgButton(hwnd,IDD_TITLEON,_n_ShowTitle);
		CheckDlgButton(hwnd,IDD_1ON,_n_ShowPane1);
		CheckDlgButton(hwnd,IDD_2ON,_n_ShowPane2);
		CheckDlgButton(hwnd,IDD_3ON,_n_ShowPane3);
		CheckDlgButton(hwnd,IDD_4ON,_n_ShowPane4);
		CheckDlgButton(hwnd,IDD_5ON,_n_ShowPane5);
		SetDlgItemInt(hwnd,IDD_DISPLAYSTKSIZ,_n_DisplayStkSiz,FALSE);
		SetDlgItemInt(hwnd,IDD_SIZE,_n_Size,FALSE);
		CheckRadioButton(hwnd,IDD_BITMAPS,IDD_FONTS,_n_UseFonts?IDD_FONTS:IDD_BITMAPS);
		CheckDlgButton(hwnd,IDD_BOLD,_n_Bold);
		// Select item in _n_FontName, or item zero if error
		if (SendDlgItemMessage(hwnd,IDD_FONTNAME,CB_SELECTSTRING,(WPARAM)-1,(LPARAM)(LPSTR)(_n_FontName)) == CB_ERR)
			SendDlgItemMessage(hwnd,IDD_FONTNAME,CB_SETCURSEL,0,0);
		// Select item in _BitmapName, or item zero if error
		if (SendDlgItemMessage(hwnd,IDD_BITMAPNAME,CB_SELECTSTRING,(WPARAM)-1,(LPARAM)(LPSTR)(_n_BitmapName)) == CB_ERR)
			SendDlgItemMessage(hwnd,IDD_BITMAPNAME,CB_SETCURSEL,0,0);
		// Display the correct items based on fonts/bitmap selection
		ShowWindow(GetDlgItem(hwnd,IDD_FONTNAME),_n_UseFonts?SW_SHOWNOACTIVATE:SW_HIDE);
		ShowWindow(GetDlgItem(hwnd,IDD_BOLD),_n_UseFonts?SW_SHOWNOACTIVATE:SW_HIDE);
		ShowWindow(GetDlgItem(hwnd,IDD_BITMAPNAME),_n_UseFonts?SW_HIDE:SW_SHOWNOACTIVATE);

		return TRUE;
	}
	// end case WM_INITDIALOG:

	case WM_NOTIFY: {
		int BitmapCurSel;
		char ach[32];
		char helpstr[_MAX_PATH+MAX_HELP];
		LPNMHDR pnmh = (LPNMHDR)lParam;
		int translated;
		switch (pnmh->code)
		{
		case PSN_SETACTIVE:
			dlgStartPage=PP_NORMAL;
			return TRUE;
		case PSN_KILLACTIVE:
			SetWindowLong(hwnd,DWL_MSGRESULT,PSNRET_NOERROR);
			return TRUE;
		case PSN_APPLY:
			_n_InTray			= IsDlgButtonChecked(hwnd,IDD_INTRAY);
			_n_IconStart		= IsDlgButtonChecked(hwnd,IDD_ICONON);
			_n_KeepInFront		= IsDlgButtonChecked(hwnd,IDD_FRONTON);
			_n_EnableMouseMove	= IsDlgButtonChecked(hwnd,IDD_MOUSEMOVE);
			_n_ShowMemoryPopup	= IsDlgButtonChecked(hwnd,IDD_POPUPON);
			_n_MemCols			= IsDlgButtonChecked(hwnd,IDD_1COL)?1:2;
			_n_ShowTitle		= IsDlgButtonChecked(hwnd,IDD_TITLEON);
			_n_ShowPane1		= IsDlgButtonChecked(hwnd,IDD_1ON);
			_n_ShowPane2		= IsDlgButtonChecked(hwnd,IDD_2ON);
			_n_ShowPane3		= IsDlgButtonChecked(hwnd,IDD_3ON);
			_n_ShowPane4		= IsDlgButtonChecked(hwnd,IDD_4ON);
			_n_ShowPane5		= IsDlgButtonChecked(hwnd,IDD_5ON);
			_n_DisplayStkSiz	= GetDlgItemInt(hwnd,IDD_DISPLAYSTKSIZ,NULL,FALSE);
			_n_Size = (int)GetDlgItemInt(hwnd,IDD_SIZE,(LPINT)&translated,FALSE);
			if (!translated) _n_Size = 100;
			if (_n_Size<MINSIZE) _n_Size=MINSIZE;
			if (_n_Size>MAXSIZE) _n_Size=MAXSIZE;
			_n_UseFonts			= IsDlgButtonChecked(hwnd,IDD_FONTS);
			_n_Bold				= IsDlgButtonChecked(hwnd,IDD_BOLD);
			BitmapCurSel = (int)SendDlgItemMessage(hwnd,IDD_BITMAPNAME,CB_GETCURSEL,0,0);
			SendDlgItemMessage(hwnd,IDD_BITMAPNAME,CB_GETLBTEXT,BitmapCurSel,(LPARAM)(LPSTR)_n_BitmapName);
			GetDlgItemText(hwnd,IDD_FONTNAME,_n_FontName,63);
			// Check for changes
			if (_n_MemCols != n_MemCols) { // change memory popup display layout
				int i;
				MemCols = _n_MemCols;
				MemRows = MEMSIZ/_n_MemCols;
				SendMessage(hMemoryPopup,PM_SIZEWINDOW,0,0);
				for (i=0;i<MEMSIZ;i++) SendMessage(hMemoryDisplay[i],PM_SIZEWINDOW,0,0);
			}
			ShowWindow(hMemoryPopup,_n_ShowMemoryPopup?SW_SHOWNOACTIVATE:SW_HIDE);
			Restart |= (
				   n_InTray != _n_InTray
				|| n_DisplayStkSiz != _n_DisplayStkSiz
				|| n_Size != _n_Size
				|| n_UseFonts != _n_UseFonts
				|| n_Bold != _n_Bold
				|| lstrcmp(n_BitmapName,_n_BitmapName)
				|| lstrcmp(n_FontName,_n_FontName)
				|| n_ShowTitle != _n_ShowTitle
				|| n_ShowPane1 != _n_ShowPane1
				|| n_ShowPane2 != _n_ShowPane2
				|| n_ShowPane3 != _n_ShowPane3
				|| n_ShowPane4 != _n_ShowPane4
				|| n_ShowPane5 != _n_ShowPane5
				);

			n_InTray=_n_InTray;
			n_IconStart=_n_IconStart;
			n_KeepInFront=_n_KeepInFront;
			n_EnableMouseMove=_n_EnableMouseMove;
			n_ShowMemoryPopup=_n_ShowMemoryPopup;
			n_MemCols=_n_MemCols;
			n_ShowTitle=_n_ShowTitle;
			n_ShowPane1=_n_ShowPane1;
			n_ShowPane2=_n_ShowPane2;
			n_ShowPane3=_n_ShowPane3;
			n_ShowPane4=_n_ShowPane4;
			n_ShowPane5=_n_ShowPane5;
			n_DisplayStkSiz=_n_DisplayStkSiz;
			n_Size=_n_Size;
			n_UseFonts=_n_UseFonts;
			n_Bold=_n_Bold;
			lstrcpy(n_BitmapName,_n_BitmapName);
			lstrcpy(n_FontName,_n_FontName);
			WritePrivateProfileString(ClassName,"InTray",itoa(n_InTray,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"IconStart",itoa(n_IconStart,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"KeepInFront",itoa(n_KeepInFront,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"EnableMouseMove",itoa(n_EnableMouseMove,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"ShowMemoryPopup",itoa(n_ShowMemoryPopup,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"MemCols",itoa(n_MemCols,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"ShowTitle",itoa(n_ShowTitle,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"ShowPane1",itoa(n_ShowPane1,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"ShowPane2",itoa(n_ShowPane2,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"ShowPane3",itoa(n_ShowPane3,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"ShowPane4",itoa(n_ShowPane4,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"ShowPane5",itoa(n_ShowPane5,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"DisplayStkSiz",itoa(n_DisplayStkSiz,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"Size",itoa(n_Size,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"UseFonts",itoa(n_UseFonts,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"Bold",itoa(n_Bold,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"BitmapName",n_BitmapName,IniFile);
			WritePrivateProfileString(ClassName,"FontName",n_FontName,IniFile);
			SetWindowLong(hwnd,DWL_MSGRESULT,PSNRET_NOERROR);
			return TRUE;
        case PSN_HELP:
            // Display help for the properties page.
			strcpy(helpstr,HelpFile);
			strcat(helpstr,NORMALHELP);
			HtmlHelp(GetParent(hwnd),helpstr,HH_DISPLAY_TOPIC,0);
            return TRUE;
		} // switch (pnmh->code)
		break;
	}
	// end case WM_NOTIFY:

	case WM_COMMAND: {
		WORD wID = GET_WM_COMMAND_ID(wParam,lParam);
		WORD cmd = GET_WM_COMMAND_CMD(wParam,lParam);
		int translated;
		switch (wID) {
		case IDD_POPUPON:
			EnableWindow(GetDlgItem(hwnd,IDD_1COL),IsDlgButtonChecked(hwnd,IDD_POPUPON));
			EnableWindow(GetDlgItem(hwnd,IDD_2COL),IsDlgButtonChecked(hwnd,IDD_POPUPON));
			return FALSE;
		case IDD_BITMAPS:
			ShowWindow(GetDlgItem(hwnd,IDD_FONTNAME),SW_HIDE);
			ShowWindow(GetDlgItem(hwnd,IDD_BOLD),SW_HIDE);
			ShowWindow(GetDlgItem(hwnd,IDD_BITMAPNAME),SW_SHOWNOACTIVATE);
			InvalidateRect(GetDlgItem(hwnd,IDD_TESTFIELD),NULL,FALSE);
			PropSheet_Changed(GetParent(hwnd),hwnd);
			return FALSE;
		case IDD_FONTS:
			ShowWindow(GetDlgItem(hwnd,IDD_FONTNAME),SW_SHOWNOACTIVATE);
			ShowWindow(GetDlgItem(hwnd,IDD_BOLD),SW_SHOWNOACTIVATE);
			ShowWindow(GetDlgItem(hwnd,IDD_BITMAPNAME),SW_HIDE);
			InvalidateRect(GetDlgItem(hwnd,IDD_TESTFIELD),NULL,FALSE);
			PropSheet_Changed(GetParent(hwnd),hwnd);
			return FALSE;
		case IDD_DISPLAYSTKSIZ:
			// Check display stack size and change if illegal value
			_n_DisplayStkSiz = GetDlgItemInt(hwnd,IDD_DISPLAYSTKSIZ,&translated,FALSE);
			if (!translated || _n_DisplayStkSiz<1 || _n_DisplayStkSiz>MAXSTKSIZ) {
				if (_n_DisplayStkSiz<1) _n_DisplayStkSiz=1;
				if (_n_DisplayStkSiz>MAXSTKSIZ) _n_DisplayStkSiz=MAXSTKSIZ;
				SetDlgItemInt(hwnd,IDD_DISPLAYSTKSIZ,_n_DisplayStkSiz,FALSE);
				SendDlgItemMessage(hwnd,IDD_DISPLAYSTKSIZ,EM_SETSEL,0,MAKELONG(0,-1));
			}
			return FALSE;
		case IDD_SIZE:
			if (cmd==EN_KILLFOCUS) {
				// Check size and change if illegal value
				_n_Size = (int)GetDlgItemInt(hwnd,IDD_SIZE,(LPINT)&translated,FALSE);
				if (!translated || _n_Size<MINSIZE || _n_Size>MAXSIZE) {
					if (!translated) _n_Size=100;
					if (_n_Size<MINSIZE) _n_Size=MINSIZE;
					if (_n_Size>MAXSIZE) _n_Size=MAXSIZE;
					SetDlgItemInt(hwnd,IDD_SIZE,_n_Size,FALSE);
					SendDlgItemMessage(hwnd,IDD_SIZE,EM_SETSEL,0,MAKELONG(0,-1));
				}
				return FALSE;
			}
			// follow through - no break
		// Follow these to update testfield display when changed
		case IDD_BOLD:
		case IDD_BITMAPNAME:
		case IDD_FONTNAME:
			{
				// We can only follow when the box has been created.
				// A few calls are made at first when there is
				// yet no IDD_TESTFIELD item. A check must be made to be
				// sure we don't invalidate NULL - the whole screen.
				// Return FALSE to ensure proper dialogue handling.
				HWND dlgitem = GetDlgItem(hwnd,IDD_TESTFIELD);
				if (dlgitem) InvalidateRect(dlgitem,NULL,FALSE);
				return FALSE;
			} // case IDD_SIZE..IDD_FONTNAME
		} // switch (wID)
		break;
	}
	// end case WM_COMMAND:

	// This message sent from enumfonts callback routine
	case PM_ADDFONTNAME:
		SendDlgItemMessage(hwnd,IDD_FONTNAME,CB_ADDSTRING,0,lParam);
		return TRUE;
	// end case PM_ADDFONTNAME:
		
	// WM_MEASUREITEM is sent for each owner draw item in bitmap combobox
	case WM_MEASUREITEM: {
		UINT idCtl = (UINT)wParam;
		LPMEASUREITEMSTRUCT sizes = (LPMEASUREITEMSTRUCT)lParam;
		HBITMAP hBitmap;
		BITMAP bitmap;
		char BitmapTmpName[32];
		
		// do not respond to other controls than IDD_BITMAPNAME combo box
		if (idCtl != IDD_BITMAPNAME)
			return FALSE;
		
		// itemID is -1 to measure edit field, >=0 for each item
		// The item text is here returned in itemData
		
		if (sizes->itemID==(UINT)-1) {
			sizes->itemHeight = 30; // this should really be max of all others...
		}
		else {
			// To get the item text here, a CB_GETLBTEXT message must be
			// sent to the combo box. For 16-bit Windows, the name could be
			// taken from sizes->itemData, but this does not seem to work
			// for Win95.
			SendDlgItemMessage(hwnd,IDD_BITMAPNAME,CB_GETLBTEXT,sizes->itemID,(LPARAM)(LPSTR)BitmapTmpName);
			hBitmap = LoadBitmap(hInst,BitmapTmpName);
			GetObject(hBitmap,sizeof(bitmap),(LPVOID)&bitmap);
			sizes->itemHeight = bitmap.bmHeight+2;
			DeleteObject(hBitmap);
		}
		
		return TRUE;
	}
	// end case WM_MEASUREITEM:

	// WM_DRAWITEM is sent to draw each item in bitmap combobox	
	case WM_DRAWITEM: {
		UINT idCtl = (UINT)wParam;
		LPDRAWITEMSTRUCT lpdis = (LPDRAWITEMSTRUCT)lParam;
		DWORD rop = 0;
		HBITMAP hBitmap,hbmpOld;
		BITMAP bitmap;
		HDC hMemDC;
		int bmpY, bmpX, rcHeight, rcWidth;
		char BitmapTmpName[32],FontTmpName[64];
		
		// do not respond to other controls than IDD_BITMAPNAME combo box
		// and the IDD_TESTFIELD ownerdraw button
		
		if (idCtl == IDD_BITMAPNAME) {
			// To get the item text here, a CB_GETLBTEXT message must be
			// sent to the combo box!
			SendDlgItemMessage(hwnd,IDD_BITMAPNAME,CB_GETLBTEXT,lpdis->itemID,(LPARAM)(LPSTR)BitmapTmpName);
			
			// Load and get bitmap from file
			hBitmap = LoadBitmap(hInst,(LPSTR)BitmapTmpName);
			GetObject(hBitmap,sizeof(bitmap),(LPVOID)&bitmap);
			
			// Save height and from where to copy bitmap
			bmpY = bitmap.bmHeight;
			bmpX = bitmap.bmWidth/NPRINTABLE;
			
			// For positioning bitmap when drawing
			rcHeight = lpdis->rcItem.bottom - lpdis->rcItem.top + 1;
			rcWidth = lpdis->rcItem.right - lpdis->rcItem.left + 1;
			
			// Create a memory DC and select the bitmap for subsequent copy
			hMemDC = CreateCompatibleDC(lpdis->hDC);
			hbmpOld = (HBITMAP)SelectObject(hMemDC,hBitmap);
			
			// Set raster operation based on action. Crude, but OK'ish.
			switch (lpdis->itemAction) {
			case ODA_DRAWENTIRE:
				rop = NOTSRCCOPY;
				break;
			case ODA_SELECT:
				rop = NOTSRCCOPY;
				break;
			case ODA_FOCUS:
				rop = SRCCOPY;
				break;
			}
			
			// Copy bitmap into combo box item
			BitBlt(lpdis->hDC,
				lpdis->rcItem.left,lpdis->rcItem.top+(rcHeight-bmpY)/2,
				rcWidth,rcHeight,hMemDC,bmpX,0,rop);
			
			SelectObject(hMemDC,hbmpOld);
			DeleteDC(hMemDC);
			
			DeleteObject(hBitmap);
			return TRUE;
		}
		else if (idCtl == IDD_TESTFIELD) {
			HBRUSH hBrush;
			COLORREF oldbk,oldtx;
			RECT rect;
			int Percent,yPos;
			SIZE siz;
			char OneChar[2] = "0";
			int translated;
			int i;
			hBrush = CreateSolidBrush(GetColour(RGBBACK));
			SetRect(&rect,
				lpdis->rcItem.left,lpdis->rcItem.top,
				lpdis->rcItem.right,lpdis->rcItem.bottom);
			FillRect(lpdis->hDC,&rect,hBrush);
			DeleteObject(hBrush);
			
			Percent = (int)GetDlgItemInt(hwnd,IDD_SIZE,(LPINT)&translated,FALSE);
			if (!translated) Percent = 100;
			
			if (IsDlgButtonChecked(hwnd,IDD_FONTS)) { // font display
				HFONT hFont,hOldFont;
				int MapModePrevious;
				
				// Get font name from IDD_FONTNAME box, create font and
				// display in button area
				GetDlgItemText(hwnd,IDD_FONTNAME,FontTmpName,63);
				
				MapModePrevious = SetMapMode(lpdis->hDC, MM_TEXT);
				
				hFont = CreateFont(
					-MulDiv(Percent,18,100),
					0,             // width
					0,             // escapement
					0,             // orientation
					IsDlgButtonChecked(hwnd,IDD_BOLD)?FW_BOLD:FW_NORMAL, // weight
					0,             // italic
					0,             // underline
					0,             // strikeout
					ANSI_CHARSET,  // character set
					OUT_DEFAULT_PRECIS,  // output precision
					CLIP_DEFAULT_PRECIS, // clip precision
					PROOF_QUALITY,       // output quality
					0,                   // pitch and family
					FontTmpName);      // typeface name
				
				hOldFont = (HFONT)SelectObject(lpdis->hDC, hFont);
				
				bmpX = 0;
				bmpY = 0;
				for (i=0;i<NPRINTABLE;i++) {
					OneChar[0] = Printable[i];
					GetTextExtentPoint32(lpdis->hDC,OneChar,1,&siz);
					bmpX = max((long)bmpX,siz.cx);
					bmpY = max((long)bmpY,siz.cy);
				}
				// Position text middle of box
				yPos = ((int)(lpdis->rcItem.bottom-lpdis->rcItem.top+1)-bmpY) / 2;
				
				// Copy text onto button, same as is done in main routine (but skip space)
				oldbk = SetBkColor(lpdis->hDC,GetColour(RGBBACK));
				oldtx = SetTextColor(lpdis->hDC,GetColour(RGBDIGITS));
				for (i=1;i<NPRINTABLE;i++) {
					SetRect(&rect,(i-1)*bmpX,yPos,i*bmpX,yPos+bmpY);
					ExtTextOut(lpdis->hDC,(i-1)*bmpX,yPos,ETO_CLIPPED,&rect,Printable+i,1,NULL);
				}
				SelectObject(lpdis->hDC,hOldFont);
				SetMapMode(lpdis->hDC, MapModePrevious);
				SetBkColor(lpdis->hDC,oldbk);
				SetTextColor(lpdis->hDC,oldtx);
				
			} // end font display
			
			else { // bitmap display
				
				int SBModePrevious;
				// Get bitmap name from IDD_BITMAPNAME box, size it and
				// display in button area
				GetDlgItemText(hwnd,IDD_BITMAPNAME,BitmapTmpName,31);
				
				// Load and get bitmap from file
				hBitmap = LoadBitmap(hInst,(LPSTR)BitmapTmpName);
				GetObject(hBitmap,sizeof(bitmap),(LPVOID)&bitmap);
				
				// Save height and from where to copy bitmap
				bmpY = bitmap.bmHeight;
				bmpX = bitmap.bmWidth/NPRINTABLE;
				
				// For positioning bitmap when drawing
				rcHeight = lpdis->rcItem.bottom - lpdis->rcItem.top + 1;
				yPos = (rcHeight - (int)MulDiv(bmpY,Percent,100) ) / 2;
				
				// Create a memory DC and select the bitmap for subsequent copy
				hMemDC = CreateCompatibleDC(lpdis->hDC);
				hbmpOld = (HBITMAP)SelectObject(hMemDC,hBitmap);
				
				// Copy bitmap onto button
				oldbk = SetBkColor(lpdis->hDC,GetColour(RGBDIGITS));  // use global background for digits
				oldtx = SetTextColor(lpdis->hDC,GetColour(RGBBACK));  // use global digit colour
				
				SBModePrevious = SetStretchBltMode(lpdis->hDC,STRETCH_ORSCANS);
				
				StretchBlt(lpdis->hDC,
					0,
					yPos,
					MulDiv(bmpX*8,Percent,100),
					MulDiv(bmpY,Percent,100),
					hMemDC,
					bmpX,
					0,
					bmpX*8,
					bmpY,
					SRCCOPY);
				
				SetStretchBltMode(lpdis->hDC,SBModePrevious);
				SetBkColor(lpdis->hDC,oldbk);
				SetTextColor(lpdis->hDC,oldtx);
				
				SelectObject(hMemDC,hbmpOld);
				DeleteDC(hMemDC);
				
				DeleteObject(hBitmap);
			} // end bitmap display
			
			DeleteObject(hBrush);
			
			return TRUE;
		} // IDD_TESTFIELD
  
		break;
	}
	// end case WM_DRAWITEM:

	case WM_VSCROLL: {
		int ScrollCode = GET_WM_VSCROLL_CODE(wParam,lParam);
		HWND hwndScrollBar = GET_WM_VSCROLL_HWND(wParam,lParam);
		int StkSiz,Siz;

		switch (GetDlgCtrlID(hwndScrollBar)) {
		case IDD_SCROLLSTACK:
			StkSiz = (int)GetDlgItemInt(hwnd,IDD_DISPLAYSTKSIZ,NULL,FALSE);
			switch (ScrollCode) {
			case SB_BOTTOM:		StkSiz = 1; break;
			case SB_TOP:		StkSiz = MAXSTKSIZ; break;
			case SB_LINEDOWN:	StkSiz -= 1; break;
			case SB_LINEUP:		StkSiz += 1; break;
			case SB_PAGEDOWN:	StkSiz -= 2; break;
			case SB_PAGEUP:		StkSiz += 2; break;
			}
			if (StkSiz<1) StkSiz=1;
			if (StkSiz>MAXSTKSIZ) StkSiz=MAXSTKSIZ;
			SetDlgItemInt(hwnd,IDD_DISPLAYSTKSIZ,StkSiz,FALSE);
			break;
		case IDD_SCROLLSIZE:
			Siz = (int)GetDlgItemInt(hwnd,IDD_SIZE,NULL,FALSE);
			switch (ScrollCode) {
			case SB_BOTTOM:		Siz = 1; break;
			case SB_TOP:		Siz = 1000; break;
			case SB_LINEDOWN:	Siz -= 1; break;
			case SB_LINEUP:		Siz += 1; break;
			case SB_PAGEDOWN:	Siz -= 10; break;
			case SB_PAGEUP:		Siz += 10; break;
			}
			if (Siz<MINSIZE) Siz=MINSIZE;
			if (Siz>MAXSIZE) Siz=MAXSIZE;
			SetDlgItemInt(hwnd,IDD_SIZE,Siz,FALSE);
			break;
		} // switch (GetDlgCtrlID(hwndScrollBar))
		break;
	}
	// end case WM_VSCROLL

	} // switch (wMsg)
	return FALSE;
}

BOOL CALLBACK MiniDlgProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	static BOOL _m_InTray=FALSE;
	static BOOL _m_IconStart=FALSE;
	static BOOL _m_KeepInFront=FALSE;
	static BOOL _m_EnableMouseMove=FALSE;
	static BOOL _m_ShowMemoryPopup=FALSE;
	static int _m_MemCols=1;
	static BOOL _m_ShowTitle=FALSE;
	static BOOL _m_ShowPane1=FALSE;
	static BOOL _m_ShowPane2=FALSE;
	static BOOL _m_ShowPane3=FALSE;
	static BOOL _m_ShowPane4=FALSE;
	static BOOL _m_ShowPane5=FALSE;
	static int _m_DisplayStkSiz=1;
	static int _m_Size=100;
	static BOOL _m_UseFonts=FALSE;
	static BOOL _m_Bold=FALSE;
	static char _m_BitmapName[32];
	static char _m_FontName[64];

	switch (wMsg)
	{
	case WM_INITDIALOG: {
		int i;
		char BitmapTmpName[32];
		HDC hdc;
		HBITMAP hBitmap;
		char ach[32];

		_m_InTray=m_InTray;
		_m_IconStart=m_IconStart;
		_m_KeepInFront=m_KeepInFront;
		_m_EnableMouseMove=m_EnableMouseMove;
		_m_ShowMemoryPopup=m_ShowMemoryPopup;
		_m_MemCols=m_MemCols;
		_m_ShowTitle=m_ShowTitle;
		_m_ShowPane1=m_ShowPane1;
		_m_ShowPane2=m_ShowPane2;
		_m_ShowPane3=m_ShowPane3;
		_m_ShowPane4=m_ShowPane4;
		_m_ShowPane5=m_ShowPane5;
		_m_DisplayStkSiz=m_DisplayStkSiz;
		_m_Size=m_Size;
		_m_UseFonts=m_UseFonts;
		_m_Bold=m_Bold;
		lstrcpy(_m_BitmapName,m_BitmapName);
		lstrcpy(_m_FontName,m_FontName);

		// Fill the Font Name listbox with valid entries
		hdc = GetDC(hwnd);
		EnumFontFamilies(hdc,NULL,(FONTENUMPROC)EnumFontProc,(LPARAM)hwnd);
		ReleaseDC(hwnd,hdc);

		// Fill the Bitmap Name listbox with valid entries
		lstrcpy(BitmapTmpName,"DIGITS");
		i = 1;
		lstrcat(BitmapTmpName,itoa(i,ach,10));
		while ((hBitmap = LoadBitmap(hInst,BitmapTmpName)) != NULL) {
			DeleteObject(hBitmap);
			SendDlgItemMessage(hwnd,IDD_BITMAPNAME,CB_ADDSTRING,0,(LPARAM)(LPSTR)BitmapTmpName);
			i++;
			lstrcpy(BitmapTmpName,"DIGITS");
			lstrcat(BitmapTmpName,itoa(i,ach,10));
		}

		CheckDlgButton(hwnd,IDD_INTRAY,_m_InTray);
		CheckDlgButton(hwnd,IDD_ICONON,_m_IconStart);
		CheckDlgButton(hwnd,IDD_FRONTON,_m_KeepInFront);
		CheckDlgButton(hwnd,IDD_MOUSEMOVE,_m_EnableMouseMove);
		CheckDlgButton(hwnd,IDD_POPUPON,_m_ShowMemoryPopup);
		EnableWindow(GetDlgItem(hwnd,IDD_1COL),IsDlgButtonChecked(hwnd,IDD_POPUPON));
		EnableWindow(GetDlgItem(hwnd,IDD_2COL),IsDlgButtonChecked(hwnd,IDD_POPUPON));
		CheckRadioButton(hwnd,IDD_1COL,IDD_2COL,_m_MemCols==1?IDD_1COL:IDD_2COL);
		CheckDlgButton(hwnd,IDD_TITLEON,_m_ShowTitle);
		CheckDlgButton(hwnd,IDD_1ON,_m_ShowPane1);
		CheckDlgButton(hwnd,IDD_2ON,_m_ShowPane2);
		CheckDlgButton(hwnd,IDD_3ON,_m_ShowPane3);
		CheckDlgButton(hwnd,IDD_4ON,_m_ShowPane4);
		CheckDlgButton(hwnd,IDD_5ON,_m_ShowPane5);
		SetDlgItemInt(hwnd,IDD_DISPLAYSTKSIZ,_m_DisplayStkSiz,FALSE);
		SetDlgItemInt(hwnd,IDD_SIZE,_m_Size,FALSE);
		CheckRadioButton(hwnd,IDD_BITMAPS,IDD_FONTS,_m_UseFonts?IDD_FONTS:IDD_BITMAPS);
		CheckDlgButton(hwnd,IDD_BOLD,_m_Bold);
		// Select item in _m_FontName, or item zero if error
		if (SendDlgItemMessage(hwnd,IDD_FONTNAME,CB_SELECTSTRING,(WPARAM)-1,(LPARAM)(LPSTR)(_m_FontName)) == CB_ERR)
			SendDlgItemMessage(hwnd,IDD_FONTNAME,CB_SETCURSEL,0,0);
		// Select item in _BitmapName, or item zero if error
		if (SendDlgItemMessage(hwnd,IDD_BITMAPNAME,CB_SELECTSTRING,(WPARAM)-1,(LPARAM)(LPSTR)(_m_BitmapName)) == CB_ERR)
			SendDlgItemMessage(hwnd,IDD_BITMAPNAME,CB_SETCURSEL,0,0);
		// Display the correct items based on fonts/bitmap selection
		ShowWindow(GetDlgItem(hwnd,IDD_FONTNAME),_m_UseFonts?SW_SHOWNOACTIVATE:SW_HIDE);
		ShowWindow(GetDlgItem(hwnd,IDD_BOLD),_m_UseFonts?SW_SHOWNOACTIVATE:SW_HIDE);
		ShowWindow(GetDlgItem(hwnd,IDD_BITMAPNAME),_m_UseFonts?SW_HIDE:SW_SHOWNOACTIVATE);
		return TRUE;
	}
	// end case WM_INITDIALOG:

	case WM_NOTIFY: {
		LPNMHDR pnmh = (LPNMHDR)lParam;
		int BitmapCurSel;
		int translated;
		char ach[32];
		char helpstr[_MAX_PATH+MAX_HELP];

		switch (pnmh->code)
		{
		case PSN_SETACTIVE:
			dlgStartPage=PP_MINI;
			return TRUE;
		case PSN_KILLACTIVE:
			SetWindowLong(hwnd,DWL_MSGRESULT,PSNRET_NOERROR);
			return TRUE;
		case PSN_APPLY:
			_m_InTray			= IsDlgButtonChecked(hwnd,IDD_INTRAY);
			_m_IconStart		= IsDlgButtonChecked(hwnd,IDD_ICONON);
			_m_KeepInFront		= IsDlgButtonChecked(hwnd,IDD_FRONTON);
			_m_EnableMouseMove	= IsDlgButtonChecked(hwnd,IDD_MOUSEMOVE);
			_m_ShowMemoryPopup	= IsDlgButtonChecked(hwnd,IDD_POPUPON);
			_m_MemCols			= IsDlgButtonChecked(hwnd,IDD_1COL)?1:2;
			_m_ShowTitle		= IsDlgButtonChecked(hwnd,IDD_TITLEON);
			_m_ShowPane1		= IsDlgButtonChecked(hwnd,IDD_1ON);
			_m_ShowPane2		= IsDlgButtonChecked(hwnd,IDD_2ON);
			_m_ShowPane3		= IsDlgButtonChecked(hwnd,IDD_3ON);
			_m_ShowPane4		= IsDlgButtonChecked(hwnd,IDD_4ON);
			_m_ShowPane5		= IsDlgButtonChecked(hwnd,IDD_5ON);
			_m_DisplayStkSiz	= GetDlgItemInt(hwnd,IDD_DISPLAYSTKSIZ,NULL,FALSE);
			// Check size and change if illegal value
			_m_Size = (int)GetDlgItemInt(hwnd,IDD_SIZE,(LPINT)&translated,FALSE);
			if (!translated) _m_Size = 100;
			if (_m_Size<MINSIZE) _m_Size=MINSIZE;
			if (_m_Size>MAXSIZE) _m_Size=MAXSIZE;
			_m_UseFonts			= IsDlgButtonChecked(hwnd,IDD_FONTS);
			_m_Bold				= IsDlgButtonChecked(hwnd,IDD_BOLD);
			BitmapCurSel = (int)SendDlgItemMessage(hwnd,IDD_BITMAPNAME,CB_GETCURSEL,0,0);
			SendDlgItemMessage(hwnd,IDD_BITMAPNAME,CB_GETLBTEXT,BitmapCurSel,(LPARAM)(LPSTR)_m_BitmapName);
			GetDlgItemText(hwnd,IDD_FONTNAME,_m_FontName,63);
			// Check for changes
			if (_m_MemCols != m_MemCols) { // change memory popup display layout
				int i;
				MemCols = _m_MemCols;
				MemRows = MEMSIZ/_m_MemCols;
				SendMessage(hMemoryPopup,PM_SIZEWINDOW,0,0);
				for (i=0;i<MEMSIZ;i++) SendMessage(hMemoryDisplay[i],PM_SIZEWINDOW,0,0);
			}
			ShowWindow(hMemoryPopup,_m_ShowMemoryPopup?SW_SHOWNOACTIVATE:SW_HIDE);
			Restart |= (
				   m_InTray != _m_InTray
				|| m_DisplayStkSiz != _m_DisplayStkSiz
				|| m_Size != _m_Size
				|| m_UseFonts != _m_UseFonts
				|| m_Bold != _m_Bold
				|| lstrcmp(m_BitmapName,_m_BitmapName)
				|| lstrcmp(m_FontName,_m_FontName)
				|| m_ShowTitle != _m_ShowTitle
				|| m_ShowPane1 != _m_ShowPane1
				|| m_ShowPane2 != _m_ShowPane2
				|| m_ShowPane3 != _m_ShowPane3
				|| m_ShowPane4 != _m_ShowPane4
				|| m_ShowPane5 != _m_ShowPane5
				);
							
			m_InTray=_m_InTray;
			m_IconStart=_m_IconStart;
			m_KeepInFront=_m_KeepInFront;
			m_EnableMouseMove=_m_EnableMouseMove;
			m_ShowMemoryPopup=_m_ShowMemoryPopup;
			m_MemCols=_m_MemCols;
			m_ShowTitle=_m_ShowTitle;
			m_ShowPane1=_m_ShowPane1;
			m_ShowPane2=_m_ShowPane2;
			m_ShowPane3=_m_ShowPane3;
			m_ShowPane4=_m_ShowPane4;
			m_ShowPane5=_m_ShowPane5;
			m_DisplayStkSiz=_m_DisplayStkSiz;
			m_Size=_m_Size;
			m_UseFonts=_m_UseFonts;
			m_Bold=_m_Bold;
			lstrcpy(m_BitmapName,_m_BitmapName);
			lstrcpy(m_FontName,_m_FontName);
			WritePrivateProfileString(ClassName,"m_InTray",itoa(m_InTray,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_IconStart",itoa(m_IconStart,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_KeepInFront",itoa(m_KeepInFront,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_EnableMouseMove",itoa(m_EnableMouseMove,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_ShowMemoryPopup",itoa(m_ShowMemoryPopup,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_MemCols",itoa(m_MemCols,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_ShowTitle",itoa(m_ShowTitle,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_ShowPane1",itoa(m_ShowPane1,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_ShowPane2",itoa(m_ShowPane2,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_ShowPane3",itoa(m_ShowPane3,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_ShowPane4",itoa(m_ShowPane4,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_ShowPane5",itoa(m_ShowPane5,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_DisplayStkSiz",itoa(m_DisplayStkSiz,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_Size",itoa(m_Size,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_UseFonts",itoa(m_UseFonts,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_Bold",itoa(m_Bold,ach,10),IniFile);
			WritePrivateProfileString(ClassName,"m_BitmapName",m_BitmapName,IniFile);
			WritePrivateProfileString(ClassName,"m_FontName",m_FontName,IniFile);
			SetWindowLong(hwnd,DWL_MSGRESULT,PSNRET_NOERROR);
			return TRUE;
        case PSN_HELP:
			// Display help for the mini properties page.
			strcpy(helpstr,HelpFile);
			strcat(helpstr,MINIHELP);
			HtmlHelp(GetParent(hwnd),helpstr,HH_DISPLAY_TOPIC,0);
			return TRUE;
		} // switch (pnmh->code)
		break;
	}
	// end case WM_NOTIFY:

	case WM_COMMAND: {
		WORD wID = GET_WM_COMMAND_ID(wParam,lParam);
		WORD cmd = GET_WM_COMMAND_CMD(wParam,lParam);
		int translated;
		switch (wID) {
		case IDD_POPUPON:
			EnableWindow(GetDlgItem(hwnd,IDD_1COL),IsDlgButtonChecked(hwnd,IDD_POPUPON));
			EnableWindow(GetDlgItem(hwnd,IDD_2COL),IsDlgButtonChecked(hwnd,IDD_POPUPON));
			return FALSE;
		case IDD_BITMAPS:
			ShowWindow(GetDlgItem(hwnd,IDD_FONTNAME),SW_HIDE);
			ShowWindow(GetDlgItem(hwnd,IDD_BOLD),SW_HIDE);
			ShowWindow(GetDlgItem(hwnd,IDD_BITMAPNAME),SW_SHOWNOACTIVATE);
			InvalidateRect(GetDlgItem(hwnd,IDD_TESTFIELD),NULL,FALSE);
			PropSheet_Changed(GetParent(hwnd),hwnd);
			return FALSE;
		case IDD_FONTS:
			ShowWindow(GetDlgItem(hwnd,IDD_FONTNAME),SW_SHOWNOACTIVATE);
			ShowWindow(GetDlgItem(hwnd,IDD_BOLD),SW_SHOWNOACTIVATE);
			ShowWindow(GetDlgItem(hwnd,IDD_BITMAPNAME),SW_HIDE);
			InvalidateRect(GetDlgItem(hwnd,IDD_TESTFIELD),NULL,FALSE);
			PropSheet_Changed(GetParent(hwnd),hwnd);
			return FALSE;
		// Follow these to update testfield display when changed
		case IDD_DISPLAYSTKSIZ:
			// Check display stack size and change if illegal value
			_m_DisplayStkSiz = GetDlgItemInt(hwnd,IDD_DISPLAYSTKSIZ,&translated,FALSE);
			if (!translated || _m_DisplayStkSiz<1 || _m_DisplayStkSiz>MAXSTKSIZ) {
				if (_m_DisplayStkSiz<1) _m_DisplayStkSiz=1;
				if (_m_DisplayStkSiz>MAXSTKSIZ) _m_DisplayStkSiz=MAXSTKSIZ;
				SetDlgItemInt(hwnd,IDD_DISPLAYSTKSIZ,_m_DisplayStkSiz,FALSE);
				SendDlgItemMessage(hwnd,IDD_DISPLAYSTKSIZ,EM_SETSEL,0,MAKELONG(0,-1));
			}
			return FALSE;
		case IDD_SIZE:
			if (cmd==EN_KILLFOCUS) {
				// Check size and change if illegal value
				int translated;
				_m_Size = (int)GetDlgItemInt(hwnd,IDD_SIZE,(LPINT)&translated,FALSE);
				if (!translated || _m_Size<MINSIZE || _m_Size>MAXSIZE) {
					if (!translated) _m_Size=100;
					if (_m_Size<MINSIZE) _m_Size=MINSIZE;
					if (_m_Size>MAXSIZE) _m_Size=MAXSIZE;
					SetDlgItemInt(hwnd,IDD_SIZE,_m_Size,FALSE);
					SendDlgItemMessage(hwnd,IDD_SIZE,EM_SETSEL,0,MAKELONG(0,-1));
				}
				return FALSE;
			}
			// follow through - no break
		// Follow these to update testfield display when changed
		case IDD_BOLD:
		case IDD_BITMAPNAME:
		case IDD_FONTNAME:
			{
				// We can only follow when the box has been created.
				// A few calls are made at first when there is
				// yet no IDD_TESTFIELD item. A check must be made to be
				// sure we don't invalidate NULL - the whole screen.
				// Return FALSE to ensure proper dialogue handling.
				HWND dlgitem = GetDlgItem(hwnd,IDD_TESTFIELD);
				if (dlgitem) InvalidateRect(dlgitem,NULL,FALSE);
				return FALSE;
			} // case IDD_SIZE..IDD_FONTNAME
		} // switch (wID)
		break;
	}
	// end case WM_COMMAND:

	// This message sent from enumfonts callback routine
	case PM_ADDFONTNAME:
		SendDlgItemMessage(hwnd,IDD_FONTNAME,CB_ADDSTRING,0,lParam);
		return TRUE;
	// end case PM_ADDFONTNAME:
		
	// WM_MEASUREITEM is sent for each owner draw item in bitmap combobox
	case WM_MEASUREITEM: {
		UINT idCtl = (UINT)wParam;
		LPMEASUREITEMSTRUCT sizes = (LPMEASUREITEMSTRUCT)lParam;
		char BitmapTmpName[32];
		HBITMAP hBitmap;
		BITMAP bitmap;

		// do not respond to other controls than IDD_BITMAPNAME combo box
		if (idCtl != IDD_BITMAPNAME)
			return FALSE;
		
		// itemID is -1 to measure edit field, >=0 for each item
		// The item text is here returned in itemData
		
		if (sizes->itemID==(UINT)-1) {
			sizes->itemHeight = 30; // this should really be max of all others...
		}
		else {
			// To get the item text here, a CB_GETLBTEXT message must be
			// sent to the combo box. For 16-bit Windows, the name could be
			// taken from sizes->itemData, but this does not seem to work
			// for Win95.
			SendDlgItemMessage(hwnd,IDD_BITMAPNAME,CB_GETLBTEXT,sizes->itemID,(LPARAM)(LPSTR)BitmapTmpName);
			hBitmap = LoadBitmap(hInst,BitmapTmpName);
			GetObject(hBitmap,sizeof(bitmap),(LPVOID)&bitmap);
			sizes->itemHeight = bitmap.bmHeight+2;
			DeleteObject(hBitmap);
		}	
		return TRUE;
	}
	// end case WM_MEASUREITEM:

	// WM_DRAWITEM is sent to draw each item in bitmap combobox	
	case WM_DRAWITEM: {
		UINT idCtl = (UINT)wParam;
		LPDRAWITEMSTRUCT lpdis = (LPDRAWITEMSTRUCT)lParam;
		DWORD rop = 0;
		char BitmapTmpName[32],FontTmpName[64];
		HBITMAP hBitmap,hbmpOld;
		BITMAP bitmap;
		HDC hMemDC;
		int bmpY, bmpX, rcHeight, rcWidth;
		
		// do not respond to other controls than IDD_BITMAPNAME combo box
		// and the IDD_TESTFIELD ownerdraw button
		if (idCtl == IDD_BITMAPNAME) {
			// To get the item text here, a CB_GETLBTEXT message must be
			// sent to the combo box!
			SendDlgItemMessage(hwnd,IDD_BITMAPNAME,CB_GETLBTEXT,lpdis->itemID,(LPARAM)(LPSTR)BitmapTmpName);
			
			// Load and get bitmap from file
			hBitmap = LoadBitmap(hInst,(LPSTR)BitmapTmpName);
			GetObject(hBitmap,sizeof(bitmap),(LPVOID)&bitmap);
			
			// Save height and from where to copy bitmap
			bmpY = bitmap.bmHeight;
			bmpX = bitmap.bmWidth/NPRINTABLE;
			
			// For positioning bitmap when drawing
			rcHeight = lpdis->rcItem.bottom - lpdis->rcItem.top + 1;
			rcWidth = lpdis->rcItem.right - lpdis->rcItem.left + 1;
			
			// Create a memory DC and select the bitmap for subsequent copy
			hMemDC = CreateCompatibleDC(lpdis->hDC);
			hbmpOld = (HBITMAP)SelectObject(hMemDC,hBitmap);
			
			// Set raster operation based on action. Crude, but OK'ish.
			switch (lpdis->itemAction) {
			case ODA_DRAWENTIRE:
				rop = NOTSRCCOPY;
				break;
			case ODA_SELECT:
				rop = NOTSRCCOPY;
				break;
			case ODA_FOCUS:
				rop = SRCCOPY;
				break;
			}
			
			// Copy bitmap into combo box item
			BitBlt(lpdis->hDC,
				lpdis->rcItem.left,lpdis->rcItem.top+(rcHeight-bmpY)/2,
				rcWidth,rcHeight,hMemDC,bmpX,0,rop);
			
			SelectObject(hMemDC,hbmpOld);
			DeleteDC(hMemDC);
			
			DeleteObject(hBitmap);
			return TRUE;
		}
		else if (idCtl == IDD_TESTFIELD) {
			HBRUSH hBrush;
			COLORREF oldbk,oldtx;
			RECT rect;
			int Percent,yPos;
			SIZE siz;
			char OneChar[2] = "0";
			int translated;
			hBrush = CreateSolidBrush(GetColour(RGBBACK));
			SetRect(&rect,
				lpdis->rcItem.left,lpdis->rcItem.top,
				lpdis->rcItem.right,lpdis->rcItem.bottom);
			FillRect(lpdis->hDC,&rect,hBrush);
			DeleteObject(hBrush);
			
			Percent = (int)GetDlgItemInt(hwnd,IDD_SIZE,(LPINT)&translated,FALSE);
			if (!translated) Percent = 100;
			
			if (IsDlgButtonChecked(hwnd,IDD_FONTS)) { // font display
				HFONT hFont,hOldFont;
				int MapModePrevious;
				int i;
				
				// Get font name from IDD_FONTNAME box, create font and
				// display in button area
				GetDlgItemText(hwnd,IDD_FONTNAME,FontTmpName,63);
				
				MapModePrevious = SetMapMode(lpdis->hDC, MM_TEXT);
				
				hFont = CreateFont(
					-MulDiv(Percent,18,100),
					0,             // width
					0,             // escapement
					0,             // orientation
					IsDlgButtonChecked(hwnd,IDD_BOLD)?FW_BOLD:FW_NORMAL, // weight
					0,             // italic
					0,             // underline
					0,             // strikeout
					ANSI_CHARSET,  // character set
					OUT_DEFAULT_PRECIS,  // output precision
					CLIP_DEFAULT_PRECIS, // clip precision
					PROOF_QUALITY,       // output quality
					0,                   // pitch and family
					FontTmpName);      // typeface name
				
				hOldFont = (HFONT)SelectObject(lpdis->hDC, hFont);
				
				bmpX = 0;
				bmpY = 0;
				for (i=0;i<NPRINTABLE;i++) {
					OneChar[0] = Printable[i];
					GetTextExtentPoint32(lpdis->hDC,OneChar,1,&siz);
					bmpX = max((long)bmpX,siz.cx);
					bmpY = max((long)bmpY,siz.cy);
				}
				// Position text middle of box
				yPos = ((int)(lpdis->rcItem.bottom-lpdis->rcItem.top+1)-bmpY) / 2;
				
				// Copy text onto button, same as is done in main routine (but skip space)
				oldbk = SetBkColor(lpdis->hDC,GetColour(RGBBACK));
				oldtx = SetTextColor(lpdis->hDC,GetColour(RGBDIGITS));
				for (i=1;i<NPRINTABLE;i++) {
					SetRect(&rect,(i-1)*bmpX,yPos,i*bmpX,yPos+bmpY);
					ExtTextOut(lpdis->hDC,(i-1)*bmpX,yPos,ETO_CLIPPED,&rect,Printable+i,1,NULL);
				}
				SelectObject(lpdis->hDC,hOldFont);
				SetMapMode(lpdis->hDC, MapModePrevious);
				SetBkColor(lpdis->hDC,oldbk);
				SetTextColor(lpdis->hDC,oldtx);
				
			} // end font display
			
			else { // bitmap display
				
				int SBModePrevious;

				// Get bitmap name from IDD_BITMAPNAME box, size it and
				// display in button area
				GetDlgItemText(hwnd,IDD_BITMAPNAME,BitmapTmpName,31);
				
				// Load and get bitmap from file
				hBitmap = LoadBitmap(hInst,(LPSTR)BitmapTmpName);
				GetObject(hBitmap,sizeof(bitmap),(LPVOID)&bitmap);
				
				// Save height and from where to copy bitmap
				bmpY = bitmap.bmHeight;
				bmpX = bitmap.bmWidth/NPRINTABLE;
				
				// For positioning bitmap when drawing
				rcHeight = lpdis->rcItem.bottom - lpdis->rcItem.top + 1;
				yPos = (rcHeight - (int)MulDiv(bmpY,Percent,100) ) / 2;
				
				// Create a memory DC and select the bitmap for subsequent copy
				hMemDC = CreateCompatibleDC(lpdis->hDC);
				hbmpOld = (HBITMAP)SelectObject(hMemDC,hBitmap);
				
				// Copy bitmap onto button
				oldbk = SetBkColor(lpdis->hDC,GetColour(RGBDIGITS));  // use global background for digits
				oldtx = SetTextColor(lpdis->hDC,GetColour(RGBBACK));  // use global digit colour
				
				SBModePrevious = SetStretchBltMode(lpdis->hDC,STRETCH_ORSCANS);
				
				StretchBlt(lpdis->hDC,
					0,
					yPos,
					MulDiv(bmpX*8,Percent,100),
					MulDiv(bmpY,Percent,100),
					hMemDC,
					bmpX,
					0,
					bmpX*8,
					bmpY,
					SRCCOPY);
				
				SetStretchBltMode(lpdis->hDC,SBModePrevious);
				SetBkColor(lpdis->hDC,oldbk);
				SetTextColor(lpdis->hDC,oldtx);
				
				SelectObject(hMemDC,hbmpOld);
				DeleteDC(hMemDC);
				
				DeleteObject(hBitmap);
				
			} // end bitmap display
			
			DeleteObject(hBrush);
			
			return TRUE;
		} // IDD_TESTFIELD
		return FALSE;
	}
	// end case WM_DRAWITEM:

	case WM_VSCROLL: {
		int ScrollCode = GET_WM_VSCROLL_CODE(wParam,lParam);
		HWND hwndScrollBar = GET_WM_VSCROLL_HWND(wParam,lParam);
		int Siz;

		switch (GetDlgCtrlID(hwndScrollBar)) {
		case IDD_SCROLLSTACK:
			StkSiz = (int)GetDlgItemInt(hwnd,IDD_DISPLAYSTKSIZ,NULL,FALSE);
			switch (ScrollCode) {
			case SB_BOTTOM:		StkSiz = 1; break;
			case SB_TOP:		StkSiz = MAXSTKSIZ; break;
			case SB_LINEDOWN:	StkSiz -= 1; break;
			case SB_LINEUP:		StkSiz += 1; break;
			case SB_PAGEDOWN:	StkSiz -= 2; break;
			case SB_PAGEUP:		StkSiz += 2; break;
			}
			if (StkSiz<1) StkSiz=1;
			if (StkSiz>MAXSTKSIZ) StkSiz=MAXSTKSIZ;
			SetDlgItemInt(hwnd,IDD_DISPLAYSTKSIZ,StkSiz,FALSE);
			break;
		case IDD_SCROLLSIZE:
			Siz = (int)GetDlgItemInt(hwnd,IDD_SIZE,NULL,FALSE);
			switch (ScrollCode) {
			case SB_BOTTOM:		Siz = 1; break;
			case SB_TOP:		Siz = 1000; break;
			case SB_LINEDOWN:	Siz -= 1; break;
			case SB_LINEUP:		Siz += 1; break;
			case SB_PAGEDOWN:	Siz -= 10; break;
			case SB_PAGEUP:		Siz += 10; break;
			}
			if (Siz<MINSIZE) Siz=MINSIZE;
			if (Siz>MAXSIZE) Siz=MAXSIZE;
			SetDlgItemInt(hwnd,IDD_SIZE,Siz,FALSE);
			break;
		}
		return TRUE;
	}
	// end case WM_VSCROLL

	} // switch (wMsg)
	return FALSE;
}

BOOL CALLBACK ColoursDlgProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	// Static members to keep between calls
	static BOOL ViewFirst;
	static COLORREF aclrCust[16];			// custom colours in common dialogue box
	static COLORREF NewColours[NCOLOURS];	// new colours kept temporarily here
	static COLORREF PrevColours[NCOLOURS];	// for fast update of IDD_VIEW ownerdraw button
	// Bitmap kept intact for the duration of the dialogue
	static const char * const ViewBitmapName = "VIEW16";// 11 colour test bitmap
	static BITMAP  SrcBitmap;							// we use only the size here
	static HBITMAP hSrcBitmap;							// contains VIEW16 bitmap
	static HBITMAP hMonoBitmap;							// same size, monochrome bitmap
	static HBITMAP hDestBitmap;							// memory representation of test view
	
	switch (wMsg)
	{
	// WM_INITDIALOG:
	// init colours from main storage, check first radio button,
	// set static RGB text control, set up bitmap displays...
	case WM_INITDIALOG: {
		HDC hdcTemp;
		HBITMAP hOldTempBitmap;
		int i;
		COLORREF rgb;
		UINT red,grn,blu;
		char ach[4];
		char str[64];

		ViewFirst = TRUE;
		
		for (i=0;i<NCOLOURS;i++) {
			NewColours[i] = GetColour(i);
			PrevColours[i] = RGB(0,0,0);
		}
		
		CheckRadioButton(hwnd,IDD_BACKGROUND,IDD_DIGITS,IDD_BACKGROUND);
		SetDlgItemText(hwnd,IDD_SELECTEDITEM,(LPSTR)ItemTitle[0]);
		rgb = GetColour(0);
		SetDlgItemText(hwnd,IDD_RVALUE,(LPSTR)itoa(GetRValue(rgb),str,10));
		SetDlgItemText(hwnd,IDD_GVALUE,(LPSTR)itoa(GetGValue(rgb),str,10));
		SetDlgItemText(hwnd,IDD_BVALUE,(LPSTR)itoa(GetBValue(rgb),str,10));
		
		// Read predefined colours from .INI file:
		for (i=0;i<16;i++) {
			GetPrivateProfileString("PreDef",itoa(i,ach,10),PreDefStr[i],str,64,IniFile);
			sscanf(str,"%u %u %u",&red,&grn,&blu);
			PreDefColours[i] = RGB(red,grn,blu);
		}
		
		// load display bitmap and create copies
		hSrcBitmap = LoadBitmap(hInst,ViewBitmapName);
		GetObject(hSrcBitmap,sizeof(SrcBitmap),(LPVOID)&SrcBitmap);
		
		hdcTemp = CreateCompatibleDC(NULL);
		
		// hMonoBitmap is monochrome, since default bitmap in hdcTemp is monochrome
		hMonoBitmap = CreateCompatibleBitmap(hdcTemp,SrcBitmap.bmWidth,SrcBitmap.bmHeight);
		
		// Create a colour bitmap
		hOldTempBitmap = (HBITMAP)SelectObject(hdcTemp,hSrcBitmap);
		hDestBitmap = CreateCompatibleBitmap(hdcTemp,SrcBitmap.bmWidth,SrcBitmap.bmHeight);
		
		SelectObject(hdcTemp,hOldTempBitmap);
		DeleteDC(hdcTemp);
		
		return TRUE;
	}
	// end case WM_INITDIALOG:

	case WM_NOTIFY: {
		LPNMHDR pnmh = (LPNMHDR)lParam;
		BOOL ColourChange = FALSE;
		int i;
		char ach[4];
		char str[64];
		char helpstr[_MAX_PATH+MAX_HELP];

		switch (pnmh->code)
		{
		case PSN_SETACTIVE:
			dlgStartPage=PP_COLOURS;
			return TRUE;
		case PSN_KILLACTIVE:
			SetWindowLong(hwnd,DWL_MSGRESULT,PSNRET_NOERROR);
			return TRUE;
		case PSN_RESET:
			// cancel is pressed - clean up
			// Delete bitmaps created in WM_INITDIALOG
			DeleteObject(hSrcBitmap);
			DeleteObject(hMonoBitmap);
			DeleteObject(hDestBitmap);
			return TRUE;
		case PSN_APPLY:
			// OK is pressed (no apply button) - save and clean up
			for (i=0;i<NCOLOURS;i++) {
				ColourChange |= NewColours[i]!=GetColour(i);
				SetColour(i,NewColours[i]);
			}
			if (ColourChange) WriteColour();
			// Write predefined colours
			for (i=0;i<16;i++) {
				sprintf(str,"%3d %3d %3d",GetRValue(PreDefColours[i]),GetGValue(PreDefColours[i]),GetBValue(PreDefColours[i]));
				WritePrivateProfileString("PreDef",itoa(i,ach,10),str,IniFile);
			}
			
			// Delete bitmaps created in WM_INITDIALOG
			DeleteObject(hSrcBitmap);
			DeleteObject(hMonoBitmap);
			DeleteObject(hDestBitmap);
			// Update colours in main displays
			DeletePensBrushes();
			CreatePensBrushes();
			DeleteDigitBitmap();
			CreateDigitBitmap(hwndXCALC);
			InvalidateRect(hwndXCALC,0,0);
			if (ShowMemoryPopup) InvalidateRect(hMemoryPopup,0,0);
			SetWindowLong(hwnd,DWL_MSGRESULT,PSNRET_NOERROR);
			return TRUE;
        case PSN_HELP:
			// Display help for the colours properties page.
			strcpy(helpstr,HelpFile);
			strcat(helpstr,COLOURSHELP);
			HtmlHelp(GetParent(hwnd),helpstr,HH_DISPLAY_TOPIC,0);
			return TRUE;
		} // switch (pnmh->code)
		break;
	}
	// end case WM_NOTIFY:

	case WM_COMMAND: {
		WORD wID = GET_WM_COMMAND_ID(wParam,lParam);
		int i;
		COLORREF rgb;
		UINT red,grn,blu;
		char str[64];
		CHOOSECOLOR cc;

		switch (wID)
		{
		// Restore default colours (rescue from screw-up)	
		case IDD_DEFAULT:
			// Default colours
			for (i=0;i<NCOLOURS;i++) {
				sscanf(DfltCols[i],"%u %u %u",&red,&grn,&blu);
				NewColours[i] = RGB(red,grn,blu);
			}
			// Predefined colours
			for (i=0;i<16;i++) {
				lstrcpy(str,PreDefStr[i]);
				sscanf(str,"%u %u %u",&red,&grn,&blu);
				PreDefColours[i] = RGB(red,grn,blu);
				aclrCust[i] = PreDefColours[i];
			}
			for (i=0;!IsDlgButtonChecked(hwnd,IDD_BACKGROUND+i);i++);
			rgb = NewColours[i];
			SetDlgItemText(hwnd,IDD_RVALUE,(LPSTR)itoa(GetRValue(rgb),str,10));
			SetDlgItemText(hwnd,IDD_GVALUE,(LPSTR)itoa(GetGValue(rgb),str,10));
			SetDlgItemText(hwnd,IDD_BVALUE,(LPSTR)itoa(GetBValue(rgb),str,10));
			InvalidateRect(hwnd,NULL,FALSE);
			return FALSE;
			
		// IDD_GETCOLOUR to call common dialogue box and update RGB strings
		case IDD_GETCOLOUR:
			// Set custom colours to these in the config box.
			for (i=0;i<16;i++) aclrCust[i] = PreDefColours[i];
			// Initialize the necessary CHOOSECOLOR members
			memset(&cc, 0, sizeof(CHOOSECOLOR));
			cc.lStructSize = sizeof(CHOOSECOLOR);
			cc.hwndOwner = hwnd;
			red = GetDlgItemInt(hwnd,IDD_RVALUE,NULL,FALSE);
			if (red>255) red = 255;
			grn = GetDlgItemInt(hwnd,IDD_GVALUE,NULL,FALSE);
			if (grn>255) grn = 255;
			blu = GetDlgItemInt(hwnd,IDD_BVALUE,NULL,FALSE);
			if (blu>255) blu = 255;
			cc.rgbResult = RGB(red,grn,blu);
			cc.lpCustColors = aclrCust;
			cc.Flags = CC_FULLOPEN|CC_RGBINIT;
			
			if (ChooseColor(&cc)) {
				// Colour is chosen. Update RGB strings
				rgb = cc.rgbResult;
				SetDlgItemText(hwnd,IDD_RVALUE,(LPSTR)itoa(GetRValue(rgb),str,10));
				SetDlgItemText(hwnd,IDD_GVALUE,(LPSTR)itoa(GetGValue(rgb),str,10));
				SetDlgItemText(hwnd,IDD_BVALUE,(LPSTR)itoa(GetBValue(rgb),str,10));
				InvalidateRect(GetDlgItem(hwnd,IDD_TESTFIELD),NULL,FALSE);
				for (i=0;i<16;i++) PreDefColours[i] = aclrCust[i];
				InvalidateRect(hwnd,NULL,TRUE);
			}
			
			return FALSE;
			
		// IDD_SETCOLOUR to read RGB strings and set selected item colour
		case IDD_SETCOLOUR:
			for (i=0;!IsDlgButtonChecked(hwnd,IDD_BACKGROUND+i);i++);
			red = GetDlgItemInt(hwnd,IDD_RVALUE,NULL,FALSE);
			if (red>255) red = 255;
			SetDlgItemText(hwnd,IDD_RVALUE,(LPSTR)itoa(red,str,10));
			grn = GetDlgItemInt(hwnd,IDD_GVALUE,NULL,FALSE);
			if (grn>255) grn = 255;
			SetDlgItemText(hwnd,IDD_GVALUE,(LPSTR)itoa(grn,str,10));
			blu = GetDlgItemInt(hwnd,IDD_BVALUE,NULL,FALSE);
			if (blu>255) blu = 255;
			SetDlgItemText(hwnd,IDD_BVALUE,(LPSTR)itoa(blu,str,10));
			NewColours[i] = RGB(red,grn,blu);
			InvalidateRect(GetDlgItem(hwnd,IDD_BACKGROUNDVAL+i),NULL,FALSE);
			InvalidateRect(GetDlgItem(hwnd,IDD_TESTFIELD),NULL,FALSE);
			InvalidateRect(GetDlgItem(hwnd,IDD_VIEW),NULL,FALSE);
			return FALSE;
			
		// Click on predefined colours:
		case IDD_PREDEF0:
		case IDD_PREDEF1:
		case IDD_PREDEF2:
		case IDD_PREDEF3:
		case IDD_PREDEF4:
		case IDD_PREDEF5:
		case IDD_PREDEF6:
		case IDD_PREDEF7:
		case IDD_PREDEF8:
		case IDD_PREDEF9:
		case IDD_PREDEF10:
		case IDD_PREDEF11:
		case IDD_PREDEF12:
		case IDD_PREDEF13:
		case IDD_PREDEF14:
		case IDD_PREDEF15:
			rgb = PreDefColours[wID-IDD_PREDEF0];
			red = GetRValue(rgb);
			SetDlgItemText(hwnd,IDD_RVALUE,(LPSTR)itoa(red,str,10));
			grn = GetGValue(rgb);
			SetDlgItemText(hwnd,IDD_GVALUE,(LPSTR)itoa(grn,str,10));
			blu = GetBValue(rgb);
			SetDlgItemText(hwnd,IDD_BVALUE,(LPSTR)itoa(blu,str,10));
			InvalidateRect(GetDlgItem(hwnd,IDD_TESTFIELD),NULL,FALSE);
			return FALSE;
			
		// Translate click on colour samples to clicks on radio button.
		case IDD_BACKGROUNDVAL:
		case IDD_BORDERVAL:
		case IDD_3DBRIGHTVAL:
		case IDD_3DDARKVAL:
		case IDD_BUTTONTEXTVAL:
		case IDD_GRAYBUTTONTEXTVAL:
		case IDD_SELECTEDBUTTONVAL:
		case IDD_MESSAGEVAL:
		case IDD_CAPTIONVAL:
		case IDD_KEYCAPSVAL:
		case IDD_DIGITSVAL:
			i = wID - IDD_BACKGROUNDVAL;
			CheckRadioButton(hwnd,IDD_BACKGROUND,IDD_DIGITS,IDD_BACKGROUND+i);
			// This message should be OK for both 16 and 32 bit Windows
			SendMessage(hwnd,WM_COMMAND,IDD_BACKGROUND+i,lParam);
			return FALSE;
			
		// Radio button click: update static string and RGB value strings
		case IDD_BACKGROUND:
		case IDD_BORDER:
		case IDD_3DBRIGHT:
		case IDD_3DDARK:
		case IDD_BUTTONTEXT:
		case IDD_GRAYBUTTONTEXT:
		case IDD_SELECTEDBUTTON:
		case IDD_MESSAGE:
		case IDD_CAPTION:
		case IDD_KEYCAPS:
		case IDD_DIGITS:
			i = wID - IDD_BACKGROUND;
			rgb = NewColours[i];
			
			SetDlgItemText(hwnd,IDD_SELECTEDITEM,(LPSTR)ItemTitle[i]);
			SetDlgItemText(hwnd,IDD_RVALUE,(LPSTR)itoa(GetRValue(rgb),str,10));
			SetDlgItemText(hwnd,IDD_GVALUE,(LPSTR)itoa(GetGValue(rgb),str,10));
			SetDlgItemText(hwnd,IDD_BVALUE,(LPSTR)itoa(GetBValue(rgb),str,10));
			InvalidateRect(GetDlgItem(hwnd,IDD_TESTFIELD),NULL,FALSE);
			return FALSE;
			
		// By catching these, we can automatically update the test colour
		// field whenever these controls receive a character. We also
		// make sure 255 is the maximum value entered.
		case IDD_RVALUE:
		case IDD_GVALUE:
		case IDD_BVALUE: {
			int translated;
			int val=GetDlgItemInt(hwnd,wID,&translated,FALSE);
			if (!translated || val>255) {
				if (val>255) val=255;
				SetDlgItemText(hwnd,wID,(LPSTR)itoa(val,str,10));
				SendDlgItemMessage(hwnd,wID,EM_SETSEL,0,MAKELONG(0,-1));
			}
			InvalidateRect(GetDlgItem(hwnd,IDD_TESTFIELD),NULL,FALSE);
			return FALSE;
		}	
		// end case IDD_RVALUE..IDD_BVALUE

		} // switch (wParam)
		return TRUE;
	}
	// end case WM_COMMAND:
	
	// WM_DRAWITEM is sent to each static ownerdraw button.
	// If button is IDD_VIEW, colour and show bitmap, otherwise
	// create the appropriate brush and fill rectangle.	
	case WM_DRAWITEM: {
		UINT idCtl = (UINT)wParam;
		LPDRAWITEMSTRUCT lpdis = (LPDRAWITEMSTRUCT)lParam;
		int i;
		HBRUSH hbr;
		UINT red,grn,blu;
		
		if (idCtl==IDD_VIEW) {
			RECT BitmapRect;
			HDC hSrcDC = CreateCompatibleDC(lpdis->hDC);
			HDC hMonoDC = CreateCompatibleDC(lpdis->hDC);
			HDC hDestDC = CreateCompatibleDC(lpdis->hDC);
			HBITMAP hOldSrcBitmap = (HBITMAP)SelectObject(hSrcDC,hSrcBitmap);
			HBITMAP hOldMonoBitmap = (HBITMAP)SelectObject(hMonoDC,hMonoBitmap);
			HBITMAP hOldDestBitmap = (HBITMAP)SelectObject(hDestDC,hDestBitmap);
			
			// ViewColours correspond to those in the VIEW16 bitmap
			static const COLORREF ViewColours[NCOLOURS] = {
				RGB(128,128,128),
				RGB(255,0,255),
				RGB(192,192,192),
				RGB(0,0,128),
				RGB(255,255,255),
				RGB(0,128,0),
				RGB(255,0,0),
				RGB(0,255,0),
				RGB(0,0,255),
				RGB(255,255,0),
				RGB(0,255,255)
			};
			
			// Find rectangle used by bitmap (first, size only)
			SetRect(&BitmapRect,0,0,SrcBitmap.bmWidth,SrcBitmap.bmHeight);
			
			// Clear destination bitmap to black on first call
			if (ViewFirst)
				FillRect(hDestDC,&BitmapRect,(HBRUSH)GetStockObject(BLACK_BRUSH));
			
			// Now retrieve all colours from hSrcBitmap, convert to correct display
			// colours via a monochrome bitmap, and send them to hDestBitmap.
			// We make the loop as fast as we can by updating only changed colours.
			
			for (i=0;i<NCOLOURS;i++) {
				if (ViewFirst || PrevColours[i] != NewColours[i]) {
					// Copying from colour bitmap to monochrome. Using the source DC
					// colour selection, background colour pixels->white, others->black.
					SetBkColor(hSrcDC,ViewColours[i]);
					BitBlt(hMonoDC,0,0,
						SrcBitmap.bmWidth,SrcBitmap.bmHeight,hSrcDC,0,0,SRCCOPY);
					// Copying from monochrome bitmap to colour. Using the destination DC
					// colour selection, white->background, black->foreground.
					// Lacking any possible ROP operation (?), we first XOR the current
					// item colour (making it black), then OR in the new colour.
					SetBkColor(hDestDC,PrevColours[i]);
					SetTextColor(hDestDC,RGB(0,0,0));
					BitBlt(hDestDC,BitmapRect.left,BitmapRect.top,
						SrcBitmap.bmWidth,SrcBitmap.bmHeight,hMonoDC,0,0,SRCINVERT);
					SetBkColor(hDestDC,NewColours[i]);
					SetTextColor(hDestDC,RGB(0,0,0));
					BitBlt(hDestDC,0,0,
						SrcBitmap.bmWidth,SrcBitmap.bmHeight,hMonoDC,0,0,SRCPAINT);
					PrevColours[i] = NewColours[i];
				}
			}
			
			// Set ViewFirst off to minimise later work.
			ViewFirst = FALSE;
			
			// Now offset rectangle for screen output
			OffsetRect(&BitmapRect,
				(lpdis->rcItem.right-BitmapRect.right)/2,
				(lpdis->rcItem.bottom-BitmapRect.bottom)/2);
			
			// Copy new bitmap to screen
			BitBlt(lpdis->hDC,BitmapRect.left,BitmapRect.top,
				SrcBitmap.bmWidth,SrcBitmap.bmHeight,hDestDC,0,0,SRCCOPY);
			
			SelectObject(hDestDC,hOldDestBitmap);
			SelectObject(hMonoDC,hOldMonoBitmap);
			SelectObject(hSrcDC,hOldSrcBitmap);
			DeleteDC(hDestDC);
			DeleteDC(hMonoDC);
			DeleteDC(hSrcDC);
			return TRUE;
		}
		
		else if (IDD_BACKGROUNDVAL<=idCtl && idCtl<=IDD_DIGITSVAL)
			hbr = CreateSolidBrush(NewColours[idCtl-IDD_BACKGROUNDVAL]);
		
		else if (IDD_PREDEF0<=idCtl && idCtl<=IDD_PREDEF15)
			hbr = CreateSolidBrush(PreDefColours[idCtl-IDD_PREDEF0]);
		
		else if (idCtl==IDD_TESTFIELD) {
			red = GetDlgItemInt(hwnd,IDD_RVALUE,NULL,FALSE);
			if (red>255) red = 255;
			grn = GetDlgItemInt(hwnd,IDD_GVALUE,NULL,FALSE);
			if (grn>255) grn = 255;
			blu = GetDlgItemInt(hwnd,IDD_BVALUE,NULL,FALSE);
			if (blu>255) blu = 255;
			hbr = CreateSolidBrush(RGB(red,grn,blu));
		}
		else // just for safety!
			hbr = CreateSolidBrush(RGB(0,0,0));
		
		FillRect(lpdis->hDC,&lpdis->rcItem,hbr);
		DeleteObject(hbr);
		
		FrameRect(lpdis->hDC,&lpdis->rcItem,(HBRUSH)GetStockObject(BLACK_BRUSH));
		return TRUE;
	}
	// end case WM_DRAWITEM:
		
	} // switch (wMsg)

	return FALSE;
}

//------------------------------------------------------------------------
// Fatal Dialog Box procedure - started only on unrecoverable
// maths error. May start Help prior to abort.
//------------------------------------------------------------------------

#ifdef __BORLANDC__
#pragma argsused
#endif
BOOL CALLBACK FatalDlgProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	switch (wMsg)
	{
	case WM_INITDIALOG: {
		RECT rect;
		MessageBeep(MB_ICONHAND);
		GetWindowRect(hwnd,&rect);
		SetWindowPos(hwnd,NULL,
			(GetSystemMetrics(SM_CXSCREEN)+rect.left-rect.right)/2,
			(GetSystemMetrics(SM_CYSCREEN)+rect.top-rect.bottom)/2,
			0,0,SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOREDRAW | SWP_NOZORDER);
		return TRUE;
	}
	// end case WM_INITDIALOG

	case WM_COMMAND: {
		char helpstr[_MAX_PATH+MAX_HELP];
		WORD wID = GET_WM_COMMAND_ID(wParam,lParam);
		switch (wID)
		{
		case IDOK:
			EndDialog(hwnd,TRUE);
			return TRUE;
		case IDD_MAIL:
			if ((WORD)ShellExecute(NULL, "open", MAILEXCEPTION, NULL, NULL, SW_SHOWNORMAL) <= 32)
				MessageBox(hwnd,"Failed to start mail application","XCALC",MB_OK);
			return TRUE;
		case IDHELP:
			strcpy(helpstr,HelpFile);
			strcat(helpstr,FATALHELP);
			HtmlHelp(GetParent(hwnd),helpstr,HH_DISPLAY_TOPIC,0);
			return TRUE;
		}
	}
	// end case WM_COMMAND

	} // switch (wMsg)
	return FALSE;
}

//------------------------------------------------------------------------
// About Dialog Box procedure
//------------------------------------------------------------------------

#ifdef __BORLANDC__
#pragma argsused
#endif

BOOL CALLBACK AboutDlgProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	switch (wMsg)
	{
	case WM_INITDIALOG: {
		RECT rect;
		HWND mailWnd, webWnd;
		DWORD dwStyle;

		GetWindowRect(hwnd,&rect);
		SetWindowPos(hwnd,NULL,
			(GetSystemMetrics(SM_CXSCREEN)+rect.left-rect.right)/2,
			(GetSystemMetrics(SM_CYSCREEN)+rect.top-rect.bottom)/2,
			0,0,SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOREDRAW | SWP_NOZORDER);

		// Subclass mailto and web buttons
		
		mailWnd = GetDlgItem(hwnd,IDD_MAIL);
		webWnd = GetDlgItem(hwnd,IDD_WEB);
		
		OldMailButtonProc = (WNDPROC)SetWindowLong(mailWnd,GWL_WNDPROC,(LONG)MailButtonProc);
		SetWindowLong(webWnd,GWL_WNDPROC,(LONG)MailButtonProc);
		
		// And let them get mouse clicks but no default cursor
		
		dwStyle = GetWindowLong(mailWnd, GWL_STYLE);
		SetWindowLong(mailWnd, GWL_STYLE, dwStyle | SS_NOTIFY);
		SetClassWord(mailWnd, GCL_HCURSOR, (WORD)NULL);
		
		dwStyle = GetWindowLong(webWnd, GWL_STYLE);
		SetWindowLong(webWnd, GWL_STYLE, dwStyle | SS_NOTIFY);
		SetClassWord(webWnd, GCL_HCURSOR, (WORD)NULL);
		
		// Load their cursor
		HandCursor = LoadCursor(hInst,"HANDCUR");
		
		return TRUE;
	}
	// end case WM_INITDIALOG:

	case WM_CTLCOLORSTATIC: {
		int id = GetDlgCtrlID((HWND)lParam);
		if (id == IDD_WEB || id == IDD_MAIL) {
			if (CommButtonIsPressed) {
				SetTextColor((HDC)wParam,URLPRESSEDCOL);
			}
			else {
				if (id == IDD_WEB && WebVisited || id == IDD_MAIL && MailVisited)
					SetTextColor((HDC)wParam,URLVISITEDCOL);
				else
					SetTextColor((HDC)wParam,URLCOL);
			}
			SetBkColor((HDC)wParam,GetSysColor(COLOR_BTNFACE));
			return (LONG)GetSysColorBrush(COLOR_BTNFACE);
		}
		return FALSE;
	}
	// end case WM_CTLCOLORSTATIC:

	case WM_MOUSEMOVE:
		// Take care of press/move on web/mail buttons
		if (CommButtonIsPressed) {
			CommButtonIsPressed = FALSE;
			InvalidateRect(GetDlgItem(hwnd,IDD_WEB),NULL,TRUE);
			InvalidateRect(GetDlgItem(hwnd,IDD_MAIL),NULL,TRUE);
		}
		return TRUE;
	// end case WM_MOUSEMOVE:
		
	case WM_COMMAND: {
		WORD wID = GET_WM_COMMAND_ID(wParam,lParam);
		switch (wID)
		{
		case IDOK:
		case IDCANCEL:
			EndDialog(hwnd,TRUE);
			return TRUE;
		}
	}
	// end case WM_COMMAND:

	} // switch (wMsg)

	return FALSE;
}

LRESULT CALLBACK MailButtonProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	switch (wMsg)
	{
	case WM_SETCURSOR:
		SetCursor(HandCursor);
		return 0;
		
	case WM_LBUTTONDOWN:
		CommButtonIsPressed = TRUE;
		InvalidateRect(hwnd,NULL,FALSE);
		return 0;
		
	case WM_LBUTTONUP:
		// Make me blue again
		CommButtonIsPressed = FALSE;
		InvalidateRect(hwnd,NULL,FALSE);
		
		// Find which button and either mail or web
		switch(GetDlgCtrlID(hwnd))
		{
		case IDD_WEB:
			WebVisited = TRUE;
			if ((WORD)ShellExecute(NULL, "open", WEBADDRESS, NULL, NULL, SW_SHOWNORMAL) <= 32)
				MessageBox(hwnd,"Failed to start web browser","XCALC",MB_OK);
			return 0;
			
		case IDD_MAIL:
			MailVisited = TRUE;
			if ((WORD)ShellExecute(NULL, "open", MAILTO, NULL, NULL, SW_SHOWNORMAL) <= 32)
				MessageBox(hwnd,"Failed to start mail application","XCALC",MB_OK);
			return 0;
		}
		return 0;
		
		// else call previous proc
	}
	return CallWindowProc(OldMailButtonProc,hwnd,wMsg,wParam,lParam);
}

/*
LRESULT CALLBACK BMPSubProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	char tmp[32];
	HDC hdc;
	COLORREF rgb;
	switch(wMsg) {
	case WM_MOUSEMOVE:
		hdc=GetDC(hwnd);
		rgb=GetPixel(hdc,LOWORD(lParam),HIWORD(lParam));
		ReleaseDC(hwnd,hdc);
		wsprintf(tmp,"%02x %02x %02x",GetRValue(rgb),GetGValue(rgb),GetBValue(rgb));
		SetDlgItemText(GetParent(hwnd),IDD_SELECTEDITEM,tmp);
		return 0;
	}
	return CallWindowProc(OldBMPProc,hwnd,wMsg,wParam,lParam);
}
*/

#ifdef __BORLANDC__
#pragma argsused
#endif

int CALLBACK EnumFontProc(LPLOGFONT logfont,LPTEXTMETRIC textmetric,
						  int fonttype,LPARAM lParam)
{
	if ((logfont->lfCharSet == ANSI_CHARSET || logfont->lfCharSet == OEM_CHARSET)) {
		SendMessage((HWND)lParam,PM_ADDFONTNAME,0,(LPARAM)(LPSTR)logfont->lfFaceName);
	}
	return TRUE;
}

//------------------------------------------------------------------------
// We need swap routine here
//------------------------------------------------------------------------
void swap(short *a,short *b)
{
	short t = *a;
	*a = *b;
	*b = t;
}
//------------------------------------------------------------------------
// Convert Dialog Box procedure - converts numbers on the fly,
// with option to paste result. (OK button)
//------------------------------------------------------------------------

// NCONVERSIONS is global and defined in XCALC.H
#define MAXSELECT 10   // Max number of conversion items

// Local variables
LD ConvertVal;   // Input number
LD ConvertedVal; // Result

// NOTE: Variables ConvType, From, To, SelectFrom[] and SelectTo[]
// are global and defined in VARIABLE.C

//------------------------------------------------------------------------
// Preset values
//------------------------------------------------------------------------

// Number of selections per converson type (cannot be const since we share with editable array)

static int fixed_nSelect[NFIXEDCONVERSIONS] = {3,10,9,10,10};

// Text for description field
static char * fixed_Descript[NFIXEDCONVERSIONS] = {
	"Temperature","Mass","Distance","Area","Volume"
};

// Text for radio buttons
static char * fixed_ConvText[NFIXEDCONVERSIONS][MAXSELECT] = {
	// Temperatures
	{"Celsius","Fahrenheit","Kelvin"},
	// Mass
	{"Gram","Kilogram","Metric tonne","Ounce","Pound","Troy ounce","Troy pound","Stone","Long ton (imp)","Short ton (US)"},
	// Distances
	{"Metre","Centimetre","Millimetre","Kilometre","Inch","Foot","Yard","Mile","Nautic mile"},
	// Area
	{"Sq metre","Sq centimetre","Sq millimetre","Sq kilometre","Sq inch","Sq foot","Sq yard","Sq mile","Sq nautic mile","Acre"},
	// Volume
	{"Litre","Fl.oz (imp)", "Fl.oz (US)","Metric cup","Cup (imp)","Cup (US)","Pint (imp)","Pint (US)","Gallon (imp)", "Gallon (US)"},
};

// Conversion constants for everything except temperatures
static LD fixed_ConvConst[NFIXEDCONVERSIONS][MAXSELECT] = {
	// Nothing for temperatures
	{ 0 },
	// Mass, conversion to grams
	{ 1.L, 1000.L, 1000000.L, 28.349523125L, 453.59237L, 31.1034768L, 373.2417216L, 6350.29318L, 1016046.9088L, 907184.74L },
	// Distances, conversion to metres
	{ 1.L, 0.01L, 0.001L, 1000.L, 0.0254L, 0.3048L, 0.9144L, 1609.344L, 1852.L },
	// Area, conversion to square metres
	{ 1.L, 0.0001L, 0.000001L, 1000000.L, 0.00064516L, 0.09290304L, 0.83612736L, 2589988.110336L, 3429904.L, 4046.8564224L },
	// Volume, conversion to litres
	{ 1.L, 0.0284130625L, 0.0295735295625L, 0.25L, 0.2273045L, 0.2365882365L, 0.56826125L, 0.473176473L, 4.54609L, 3.785411784L },
};

//------------------------------------------------------------------------
// Used values and titles (second half is user modifiable)
//------------------------------------------------------------------------

#define MAXDESCRIPTSIZE 32
#define MAXTEXTSIZE 32
#define MAXVALUESIZE 64

static int nSelect[NCONVERSIONS];
static char *Descript[NCONVERSIONS];
static char *ConvText[NCONVERSIONS][MAXSELECT];
static LD ConvConst[NCONVERSIONS][MAXSELECT];

// This is for edit box - must be known to others as well:
static int tmp_ConvType;

//------------------------------------------------------------------------
// InitConversion() must be called at startup
//------------------------------------------------------------------------

void InitConversion()
{
	int i,j,k;
	char keyname[MAXDESCRIPTSIZE];
	char defstring[MAXTEXTSIZE];
	char valuetext[MAXVALUESIZE];

	// First, set non-editable fields
	for (i=0; i<NFIXEDCONVERSIONS; i++) {
		nSelect[i] = fixed_nSelect[i];
		Descript[i] = fixed_Descript[i];
		for (j=0; j<MAXSELECT; j++) {
			ConvText[i][j] = fixed_ConvText[i][j];
			ConvConst[i][j] = fixed_ConvConst[i][j];
		}
	}
	// Then editable fields
	for (i=NFIXEDCONVERSIONS; i<NCONVERSIONS; i++) {
		// Always say we have 10 selectable fields - we won't show these where text is empty
		nSelect[i] = 10;
		k = i-NFIXEDCONVERSIONS+1;
		Descript[i] = (char*)calloc(MAXDESCRIPTSIZE,sizeof(char));
		sprintf(keyname,"User%dName",k);
		sprintf(defstring,"User%d",k);
		GetPrivateProfileString("Convert",keyname,defstring,Descript[i],MAXDESCRIPTSIZE,IniFile);
		for (j=0; j<MAXSELECT; j++) {
			ConvText[i][j] = (char*)calloc(MAXTEXTSIZE,sizeof(char));
			sprintf(keyname,"User%dString%d",k,j+1);
			GetPrivateProfileString("Convert",keyname,"",ConvText[i][j],MAXTEXTSIZE,IniFile);
			sprintf(keyname,"User%dValue%d",k,j+1);
			GetPrivateProfileString("Convert",keyname,"",valuetext,MAXTEXTSIZE,IniFile);
			ConvConst[i][j] = _strtold(valuetext,NULL);
			if (ConvConst[i][j] == 0.0L) ConvConst[i][j] = 1.0L;
		}
	}
	// Set default convert text (if quick convert called before convert)
	if (!*ConvText[ConvType-FIRSTCONVTYPE][From-IDD_FROM] || !ConvText[ConvType-FIRSTCONVTYPE][To-IDD_TO]) {
		lstrcpy(ConvMsg,"(No conversion)");
	}
	else {
		lstrcpy(ConvMsg,ConvText[ConvType-FIRSTCONVTYPE][From-IDD_FROM]);
		lstrcat(ConvMsg," to ");
		lstrcat(ConvMsg,ConvText[ConvType-FIRSTCONVTYPE][To-IDD_TO]);
	}
}

//------------------------------------------------------------------------
// EndConversion() must be called at shutdown - releases memory
//------------------------------------------------------------------------

void EndConversion()
{
	int i,j;
	for (i=NFIXEDCONVERSIONS; i<NCONVERSIONS; i++) {
		free(Descript[i]);
		for (j=0; j<MAXSELECT; j++) {
			free(ConvText[i][j]);
		}
	}
}

// ValidZero is used to allow start of inputs before valid numbers are entered.

BOOL ValidZero(char *s)
{
	return !lstrcmp(s,".") || !lstrcmp(s,"-") || !lstrcmp(s,"-.");
}

// NoAmpersand returns copy of string with '&' characters removed

char *NoAmpersand(char *text)
{
	static char cleantext[MAXDESCRIPTSIZE];
	char *p, *q;
	p = text;
	q = cleantext;
	while (*p) {
		if (*p != '&') {
			*q++ = *p;
		}
		p++;
	}
	*q = 0;
	return cleantext;
}

#ifdef __BORLANDC__
#pragma argsused
#endif

BOOL CALLBACK ConvertDlgProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	char stemp[32];
	int i;
	static short _ConvType,_From,_To;
	static short _SelectFrom[NCONVERSIONS],_SelectTo[NCONVERSIONS];
	
	switch (wMsg)
	{
	case WM_INITDIALOG:
		// Initialise number to be converted
		ConvertVal = getreal(&ST[Xreg]);
		// Update Input field
		sprintf(stemp,"%.15Lg",ConvertVal);
		SetDlgItemText(hwnd,IDD_INPUT,stemp);
		
		// We do an integrity check here, just in case. I have experienced
		// difficulties for some unknown reason that illegal values have
		// been set.
		if (ConvType<FIRSTCONVTYPE || ConvType>=FIRSTCONVTYPE+NCONVERSIONS)
			ConvType = FIRSTCONVTYPE;
		if (From<IDD_FROM0 || From>=IDD_FROM0+nSelect[ConvType-FIRSTCONVTYPE])
			From = IDD_FROM0;
		if (To<IDD_TO0 || To>=IDD_TO0+nSelect[ConvType-FIRSTCONVTYPE])
			To = IDD_TO0;
		for (i=0;i<NCONVERSIONS;i++) {
			if (SelectFrom[i]<IDD_FROM0 || SelectFrom[i]>=IDD_FROM0+nSelect[i])
				SelectFrom[i] = IDD_FROM0;
			if (SelectTo[i]<IDD_TO0 || SelectTo[i]>=IDD_TO0+nSelect[i])
				SelectTo[i] = IDD_TO0;
		}
		// Set temp settings
		_ConvType = ConvType;
		_From = From;
		_To = To;
		for (i=0;i<NCONVERSIONS;i++) {
			_SelectFrom[i] = SelectFrom[i];
			_SelectTo[i] = SelectTo[i];
		}
		
		// select current conversion
		SetDlgItemText(hwnd,IDD_CONVERTTEXT,NoAmpersand(Descript[ConvType-FIRSTCONVTYPE]));
		// don't use radiobuttons any more, so don't..
		// CheckRadioButton(hwnd,IDD_TEMPERATURE,IDD_VOLUME,_ConvType);
		// Set focus on button for selected conversion
		SendMessage(hwnd,PM_SETBUTTONS,0,0);
		SetFocus(GetDlgItem(hwnd,_ConvType));
		return FALSE;	// Return FALSE to stop win from setting focus
		
		// PM_SETBUTTONS activates and sets text to the correct radio buttons
		// based on global variable ConvType
		
    case PM_SETBUTTONS:
		// Activate and set text on active buttons
		// Note: From and To can never be outside valid range,
		// due to arrays SelectFrom and SelectTo.
		// New Feb02 (v2.5): user-modified selections means we also don't show any if description is empty
		for (i=0;i<nSelect[_ConvType-FIRSTCONVTYPE];i++) {
			if (*ConvText[_ConvType-FIRSTCONVTYPE][i]) {
				// We have a description
				sprintf(stemp,"&%c - %s",i+'0',ConvText[_ConvType-FIRSTCONVTYPE][i]);
				SetDlgItemText(hwnd,i+IDD_FROM0,stemp);
				sprintf(stemp,"&%c - %s",i+'A',ConvText[_ConvType-FIRSTCONVTYPE][i]);
				SetDlgItemText(hwnd,i+IDD_TO0,stemp);
				ShowWindow(GetDlgItem(hwnd,i+IDD_FROM0),SW_SHOWNOACTIVATE);
				ShowWindow(GetDlgItem(hwnd,i+IDD_TO0),SW_SHOWNOACTIVATE);
			}
			else {
				// Empty
				SetDlgItemText(hwnd,i+IDD_FROM0,"");
				SetDlgItemText(hwnd,i+IDD_TO0,"");
				ShowWindow(GetDlgItem(hwnd,i+IDD_FROM0),SW_HIDE);
				ShowWindow(GetDlgItem(hwnd,i+IDD_TO0),SW_HIDE);
			}
		}
		// and deactivate the rest
		for (;i<MAXSELECT;i++) {
			SetDlgItemText(hwnd,i+IDD_FROM0,"");
			SetDlgItemText(hwnd,i+IDD_TO0,"");
			ShowWindow(GetDlgItem(hwnd,i+IDD_FROM0),SW_HIDE);
			ShowWindow(GetDlgItem(hwnd,i+IDD_TO0),SW_HIDE);
		}
		
		// Name the user buttons (a bit of an overkill to do it here though)
		for (i=NFIXEDCONVERSIONS; i<NCONVERSIONS; i++) {
			SetDlgItemText(hwnd,i+FIRSTCONVTYPE,Descript[i]);
		}

		// Set current button
		CheckRadioButton(hwnd,IDD_FROM0,
			IDD_FROM0+nSelect[_ConvType-FIRSTCONVTYPE]-1,_From);
		CheckRadioButton(hwnd,IDD_TO0,
			IDD_TO0+nSelect[_ConvType-FIRSTCONVTYPE]-1,_To);
		// Do conversion
		SendMessage(hwnd,PM_CONVERT,0,0);
		return TRUE;
		
	case PM_CONVERT:
		// Cludge here: DoConversion() must rely on ConvType, while
		// current settings are in _ConvType. Do a temporary swap.
		swap(&ConvType,&_ConvType);
		swap(&From,&_From);
		swap(&To,&_To);
		if (DoConversion()) {
			sprintf(stemp,"%Lg",ConvertedVal);
			SetDlgItemText(hwnd,IDD_RESULT,stemp);
		}
		else {
			SetDlgItemText(hwnd,IDD_INPUT,"0");
			SendDlgItemMessage(hwnd,IDD_INPUT,EM_SETSEL,0,MAKELONG(0,-1));
			ConvertVal = 0;
			MessageBox(hwnd,
				"Number would be too large.\nInput set to zero.",
				"XCALC Convert",MB_OK);
		}
		swap(&ConvType,&_ConvType);
		swap(&From,&_From);
		swap(&To,&_To);
		
		return TRUE;
		
	case WM_COMMAND:
		{
			WORD wID = GET_WM_COMMAND_ID(wParam,lParam);
			switch (wID)
			{
			case IDOK:
				ConvType = _ConvType;
				From = _From;
				To = _To;
				for (i=0;i<NCONVERSIONS;i++) {
					SelectFrom[i] = _SelectFrom[i];
					SelectTo[i] = _SelectTo[i];
				}
				if (!*ConvText[ConvType-FIRSTCONVTYPE][From-IDD_FROM] || !ConvText[ConvType-FIRSTCONVTYPE][To-IDD_TO]) {
					lstrcpy(ConvMsg,"(No conversion)");
				}
				else {
					lstrcpy(ConvMsg,ConvText[ConvType-FIRSTCONVTYPE][From-IDD_FROM]);
					lstrcat(ConvMsg," to ");
					lstrcat(ConvMsg,ConvText[ConvType-FIRSTCONVTYPE][To-IDD_TO]);
				}
				EndDialog(hwnd,TRUE);
				return TRUE;
				
			case IDCANCEL:
				EndDialog(hwnd,FALSE);
				return TRUE;

			case IDD_EDIT:
				tmp_ConvType = _ConvType - FIRSTUSERCONVTYPE;
				if (DialogBox(hInst,"EDITCONVERT",hwnd,EditConvertDlgProc)) { // true="OK"
					// Update current values, but make sure we have no invalid current values (if possible)
					for (i=NFIXEDCONVERSIONS; i<NCONVERSIONS; i++) {
						int j;
						if (!*ConvText[i][_SelectFrom[i]-IDD_FROM]) {
							_SelectFrom[i] = IDD_FROM;
							for (j=0; j<MAXSELECT; j++) {
								if (*ConvText[i][j]) {
									_SelectFrom[i] = (short)(IDD_FROM + j);
									break;
								}
							}
							if (i == _ConvType - FIRSTCONVTYPE) {
								_From = _SelectFrom[i];
							}
						}
						if (!*ConvText[i][_SelectTo[i]-IDD_TO]) {
							_SelectTo[i] = IDD_TO;
							for (j=0; j<MAXSELECT; j++) {
								if (*ConvText[i][j]) {
									_SelectTo[i] = (short)(IDD_TO + j);
									break;
								}
							}
							if (i == _ConvType - FIRSTCONVTYPE) {
								_To = _SelectTo[i];
							}
						}
					}
					SetDlgItemText(hwnd,IDD_CONVERTTEXT,NoAmpersand(Descript[_ConvType-FIRSTCONVTYPE]));
					SendMessage(hwnd,PM_SETBUTTONS,0,0);
				}
				return TRUE;
				
			case IDD_INPUT:
				{
					static char okstr[32] = "";
					int i,nconv;
					char c=0;
					GetDlgItemText(hwnd,IDD_INPUT,stemp,32);
					// Input field is cleared
					if (!lstrlen(stemp)) {
						SetDlgItemText(hwnd,IDD_INPUT,"0");
						SendDlgItemMessage(hwnd,IDD_INPUT,EM_SETSEL,0,MAKELONG(0,-1));
					}
					// Convert ',' to '.'
					for(i=0;i<(int)lstrlen(stemp);i++) if(stemp[i]==',') stemp[i]='.';
					nconv = sscanf(stemp,"%Lg%c",&ConvertVal,&c);
					// Special consideration for valid "zeros" - leave untouched
					if (ValidZero(stemp))
						ConvertVal = 0;
					else {
						// Otherwise get rid of possible illegal characters
						if (nconv!=1) {
							DWORD dw = SendDlgItemMessage(hwnd,IDD_INPUT,EM_GETSEL,0,0);
							SetDlgItemText(hwnd,IDD_INPUT,okstr);
							if (lstrcmp(okstr,"0"))
								SendDlgItemMessage(hwnd,IDD_INPUT,EM_SETSEL,0,dw);
							else
								SendDlgItemMessage(hwnd,IDD_INPUT,EM_SETSEL,0,MAKELONG(0,-1));
						}
					}
					// We always save last valid entry in 'okstr'.
					GetDlgItemText(hwnd,IDD_INPUT,okstr,32);
					// Do not SendMessage here - it will for some reason fail.
					PostMessage(hwnd,PM_CONVERT,0,0);
					return FALSE;
				}
			case IDHELP: {
				char helpstr[_MAX_PATH+MAX_HELP];
				strcpy(helpstr,HelpFile);
				strcat(helpstr,CONVERSIONHELP);
				HtmlHelp(GetParent(hwnd),helpstr,HH_DISPLAY_TOPIC,0);
				return TRUE;
			}	
			case IDD_TEMPERATURE:
			case IDD_DISTANCE:
			case IDD_AREA:
			case IDD_VOLUME:
			case IDD_MASS:
			case IDD_USER1:
			case IDD_USER2:
			case IDD_USER3:
			case IDD_USER4:
			case IDD_USER5:
				_SelectFrom[_ConvType-FIRSTCONVTYPE] = _From;
				_SelectTo[_ConvType-FIRSTCONVTYPE] = _To;
				_ConvType = wID;
				_From = _SelectFrom[_ConvType-FIRSTCONVTYPE];
				_To = _SelectTo[_ConvType-FIRSTCONVTYPE];
				SetDlgItemText(hwnd,IDD_CONVERTTEXT,NoAmpersand(Descript[_ConvType-FIRSTCONVTYPE]));
				SendMessage(hwnd,PM_SETBUTTONS,0,0);
				return TRUE;
				
			case IDD_FROM0:
			case IDD_FROM1:
			case IDD_FROM2:
			case IDD_FROM3:
			case IDD_FROM4:
			case IDD_FROM5:
			case IDD_FROM6:
			case IDD_FROM7:
			case IDD_FROM8:
			case IDD_FROM9:
				_From = wID;
				SendMessage(hwnd,PM_CONVERT,0,0);
				return TRUE;
				
			case IDD_TO0:
			case IDD_TO1:
			case IDD_TO2:
			case IDD_TO3:
			case IDD_TO4:
			case IDD_TO5:
			case IDD_TO6:
			case IDD_TO7:
			case IDD_TO8:
			case IDD_TO9:
				_To = wID;
				SendMessage(hwnd,PM_CONVERT,0,0);
				return TRUE;
			}
	    } // case WM_COMMAND
	default:
		return FALSE;
	}
}

static char tmp_Descript[NUSERCONVERSIONS][MAXDESCRIPTSIZE];
static char tmp_ConvText[NUSERCONVERSIONS][MAXSELECT][MAXTEXTSIZE];
static char tmp_ConvConst[NUSERCONVERSIONS][MAXSELECT][MAXVALUESIZE];

BOOL CALLBACK EditConvertDlgProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	WORD wID;
	HWND hwndtab;
	int i,j,k;
	TCITEM tci;
	LPNMHDR pnmh;
	char keyname[MAXDESCRIPTSIZE];

	hwndtab = GetDlgItem(hwnd,IDC_TAB);
	tci.mask = TCIF_TEXT;

	switch (wMsg)
	{
	case WM_INITDIALOG:
		for (i=0; i<NUSERCONVERSIONS; i++) {
			k = i + NFIXEDCONVERSIONS;
			lstrcpy(tmp_Descript[i],Descript[k]);
			for (j=0; j<MAXSELECT; j++) {
				lstrcpy(tmp_ConvText[i][j],ConvText[k][j]);
				sprintf(tmp_ConvConst[i][j],"%.15Lg",ConvConst[k][j]);
			}
			tci.pszText = Descript[NFIXEDCONVERSIONS+i];
			SendMessage(hwndtab,TCM_INSERTITEM,(WPARAM)i,(LPARAM)&tci);
		}
		// Set current selected tab (no WM_NOTIFY is generated)
		if (tmp_ConvType < 0 || tmp_ConvType >= NUSERCONVERSIONS) tmp_ConvType = 0; // assert valid
		SendMessage(hwndtab,TCM_SETCURSEL,(WPARAM)tmp_ConvType,0);
		// Fill all edit fields
		SendMessage(hwnd,PM_SETVALUES,0,0);
		return TRUE;

	case PM_READVALUES:
		if (tmp_ConvType < 0 || tmp_ConvType >= NUSERCONVERSIONS) return FALSE; // only if valid
		GetDlgItemText(hwnd,IDC_GROUPNAME,tmp_Descript[tmp_ConvType],MAXDESCRIPTSIZE);
		for (i=0; i<MAXSELECT; i++) {
			GetDlgItemText(hwnd,IDC_TEXT0 + i,tmp_ConvText[tmp_ConvType][i],MAXTEXTSIZE);
			GetDlgItemText(hwnd,IDC_VALUE0 + i,tmp_ConvConst[tmp_ConvType][i],MAXVALUESIZE);
		}
		return TRUE;

	case PM_SETVALUES:
		if (tmp_ConvType < 0 || tmp_ConvType >= NUSERCONVERSIONS) tmp_ConvType = 0; // assert valid
		SetDlgItemText(hwnd,IDC_GROUPNAME,tmp_Descript[tmp_ConvType]);
		for (i=0; i<MAXSELECT; i++) {
			SetDlgItemText(hwnd,IDC_TEXT0 + i,tmp_ConvText[tmp_ConvType][i]);
			SetDlgItemText(hwnd,IDC_VALUE0 + i,tmp_ConvConst[tmp_ConvType][i]);
		}
		return TRUE;

	case WM_NOTIFY:
		pnmh = (LPNMHDR)lParam;
		if (pnmh->hwndFrom != hwndtab) return FALSE;
		if (pnmh->code == TCN_SELCHANGING) {
			tmp_ConvType = SendMessage(hwndtab,TCM_GETCURSEL,0,0);
			if (tmp_ConvType != -1) {
				SendMessage(hwnd,PM_READVALUES,0,0);
			}
			return FALSE; // allow change
		}
		else if (pnmh->code == TCN_SELCHANGE) {
			tmp_ConvType = SendMessage(hwndtab,TCM_GETCURSEL,0,0);
			if (tmp_ConvType != -1) {
				SendMessage(hwnd,PM_SETVALUES,0,0);
			}
			return TRUE;
		}
		return FALSE;

	case WM_COMMAND:
		wID = GET_WM_COMMAND_ID(wParam,lParam);
		switch (wID)
		{
		case IDOK:
			SendMessage(hwnd,PM_READVALUES,0,0);
			WritePrivateProfileString("Convert",NULL,NULL,IniFile);
			for (i=0; i<NUSERCONVERSIONS; i++) {
				k = i + NFIXEDCONVERSIONS;
				sprintf(keyname,"User%dName",i+1);
				lstrcpy(Descript[k],tmp_Descript[i]);
				WritePrivateProfileString("Convert",keyname,tmp_Descript[i],IniFile);
				for (j=0; j<MAXSELECT; j++) {
					lstrcpy(ConvText[k][j],tmp_ConvText[i][j]);
					ConvConst[k][j] = _strtold(tmp_ConvConst[i][j],NULL);
					if (ConvConst[k][j] == 0.0L) {
						lstrcpy(tmp_ConvConst[i][j],"1");
						ConvConst[k][j] = 1.0L;
					}
					if (*tmp_ConvText[i][j]) {
						sprintf(keyname,"User%dString%d",i+1,j+1);
						WritePrivateProfileString("Convert",keyname,tmp_ConvText[i][j],IniFile);
						sprintf(keyname,"User%dValue%d",i+1,j+1);
						WritePrivateProfileString("Convert",keyname,tmp_ConvConst[i][j],IniFile);
					}
				}
			}
			EndDialog(hwnd,TRUE);
			return TRUE;
			
		case IDCANCEL:
			EndDialog(hwnd,FALSE);
			return TRUE;

		case IDC_CLEARALL:
			if (tmp_ConvType < 0 || tmp_ConvType >= NUSERCONVERSIONS) tmp_ConvType = 0; // assert valid
			lstrcpy(tmp_Descript[tmp_ConvType],"");
			for (i=0; i<MAXSELECT; i++) {
				lstrcpy(tmp_ConvText[tmp_ConvType][i],"");
				lstrcpy(tmp_ConvConst[tmp_ConvType][i],"1");
			}
			SendMessage(hwnd,PM_SETVALUES,0,0);
			return TRUE;

		case IDHELP: {
			char helpstr[_MAX_PATH+MAX_HELP];
			strcpy(helpstr,HelpFile);
			strcat(helpstr,CONVSETUPHELP);
			HtmlHelp(GetParent(hwnd),helpstr,HH_DISPLAY_TOPIC,0);
			return TRUE;
		}
		case IDC_GROUPNAME:
			// Groupname may have changed - update tab
			GetDlgItemText(hwnd,IDC_GROUPNAME,keyname,MAXDESCRIPTSIZE);
			tci.pszText = keyname;
			SendMessage(hwndtab,TCM_SETITEM,(WPARAM)tmp_ConvType,(LPARAM)&tci);
			return TRUE;
		}
	}

	return FALSE;
}

//------------------------------------------------------------------------
// DoConversion() does the actual conversions. Called both from
// dialogue box (Convert) and main procedure (Quick Convert).
// Uses global variables
//   ConvType, To, From, ConvertVal, ConvertedVal
//------------------------------------------------------------------------

BOOL DoConversion()
{
	LD value = ConvertVal;
	// Convert From -> unit value
	if (ConvType==IDD_TEMPERATURE) {
		switch (From) {
		case IDD_FROM0: // Celsius;
			break;
		case IDD_FROM1: // Fahrenheit
			value = (value-32)*5./9.;
			break;
		case IDD_FROM2: // Kelvin
			value -=  273.15;
			break;
		}
	}
	else {
		value *= ConvConst[ConvType-FIRSTCONVTYPE][From-IDD_FROM0];
	}
	// Convert unit value -> To
	if (ConvType==IDD_TEMPERATURE) {
		switch (To) {
		case IDD_TO0: // Celsius;
			//value = value;
			break;
		case IDD_TO1: // Fahrenheit
			value = value*9./5.+32;
			break;
		case IDD_TO2: // Kelvin
			value = value + 273.15;
			break;
		}
	}
	else {
		value = value / ConvConst[ConvType-FIRSTCONVTYPE][To-IDD_TO0];
	}
	// Round small numbers to zero
	if (fabsl(value)<NUM_MIN) {
		ConvertedVal = 0;
	}
	// Update Result field
	if (fabsl(value)<=NUM_MAX) {
		ConvertedVal = value;
		return TRUE;
	}
	else {
		return FALSE;
	}
}

//------------------------------------------------------------------------
//	Constant Dialogue Box procedures
//	We keep as much as possible local to this file.
//------------------------------------------------------------------------

char ButtonText[NCONSTBTNS][32];
char ButtonNumber[NCONSTBTNS][32];
char ButtonTextUndo[NCONSTBTNS][32];
char ButtonNumberUndo[NCONSTBTNS][32];
BOOL Undoable[NCONSTBTNS];
char _ButtonText[NCONSTBTNS][32];
char _ButtonNumber[NCONSTBTNS][32];
char _ButtonTextUndo[NCONSTBTNS][32];
char _ButtonNumberUndo[NCONSTBTNS][32];
BOOL _Undoable[NCONSTBTNS];

BOOL bFirst = TRUE;
char NewButtonText[32],NewNumber[32];

WNDPROC OldButtonProc;
HMENU hpopup;

// Prototypes (ConstantDlgProc declared in XCALC.H)

LRESULT CALLBACK ButtonSubProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK EditDlgProc(HWND,UINT,WPARAM,LPARAM);

BOOL CALLBACK ConstantDlgProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	int i;
	char ach[10],str[65],DefButtonStr[65];
	char *p;
	POINT pt;
	int iID;
	HWND hMousewnd;
	
	switch (wMsg)
	{
	case WM_INITDIALOG:
		// Initialise buffers on first call
		if (bFirst) {
			bFirst = FALSE;
			for (i=0;i<NCONSTBTNS;i++) {
				lstrcpy(ButtonTextUndo[i],"");
				lstrcpy(ButtonNumberUndo[i],"");
				// Now read .INI file -
				// the default values are stored in a stringtable
				LoadString(hInst,IDD_STRINGS+i,DefButtonStr,65);
				GetPrivateProfileString("Constants",itoa(i,ach,10),DefButtonStr,str,65,IniFile);
				// Parse string
				p = strtok(str,";");
				if (p)
					lstrcpyn(ButtonText[i],p,32);
				else
					lstrcpy(ButtonText[i],"");
				p = strtok(NULL,";");
				if (p)
					lstrcpyn(ButtonNumber[i],p,32);
				else
					lstrcpy(ButtonNumber[i],"");
				Undoable[i] = FALSE;
			}
		}
		// Save old values in case IDCANCEL is pressed
		for (i=0;i<NCONSTBTNS;i++) {
			lstrcpy(_ButtonText[i],ButtonText[i]);
			lstrcpy(_ButtonNumber[i],ButtonNumber[i]);
			lstrcpy(_ButtonTextUndo[i],ButtonTextUndo[i]);
			lstrcpy(_ButtonNumberUndo[i],ButtonNumberUndo[i]);
			_Undoable[i] = Undoable[i];
		}
		// Initialise buttons
		for (i=0;i<NCONSTBTNS;i++) {
			SetDlgItemText(hwnd,IDD_FIRSTBUTTON+i,ButtonText[i]);
			SetDlgItemText(hwnd,IDD_FIRSTNUMBER+i,ButtonNumber[i]);
		}
		
		// Create popup menu
		hpopup = CreatePopupMenu();
		AppendMenu(hpopup,MF_ENABLED|MF_STRING,IDM_EDIT,"Edit");
		AppendMenu(hpopup,MF_ENABLED|MF_STRING,IDM_UNDO,"Undo");
		AppendMenu(hpopup,MF_ENABLED|MF_STRING,IDM_CLEAR,"Clear");
		AppendMenu(hpopup,MF_SEPARATOR,0,NULL);
		AppendMenu(hpopup,MF_ENABLED|MF_STRING,IDM_RESTORE,"Restore default value");
		AppendMenu(hpopup,MF_ENABLED|MF_STRING,IDM_RESTOREALL,"Restore all default values");
		AppendMenu(hpopup,MF_SEPARATOR,0,NULL);
		AppendMenu(hpopup,MF_STRING,IDM_VALUE,"");
		// Subclass child windows
		OldButtonProc = (WNDPROC)GetWindowLong(GetDlgItem(hwnd,IDD_FIRSTBUTTON),GWL_WNDPROC); // all share same proc..
		for (i=IDD_FIRSTBUTTON;i<IDD_FIRSTBUTTON+NCONSTBTNS;i++)
			if (SetWindowLong(GetDlgItem(hwnd,i),GWL_WNDPROC,(LONG)ButtonSubProc) != (LONG)OldButtonProc) {
				MessageBox(hwndXCALC,
					"Button subclassing failed.\nI never thought this would happen,\n"
					"so contact the author!","Error",MB_OK);
				EndDialog(hwnd,FALSE);
				return TRUE;
			}
			SendMessage(hwnd,PM_ACTIVATEBUTTONS,0,0);
			return TRUE;
			
	case WM_RBUTTONDOWN:
		pt.x = LOWORD(lParam);
		pt.y = HIWORD(lParam);
		hMousewnd = ChildWindowFromPoint(hwnd,pt);
		iID = GetDlgCtrlID(hMousewnd);
		ClientToScreen(hwnd,&pt);
		ScreenToClient(hMousewnd,&pt);
		if (iID>=IDD_FIRSTBUTTON && iID<IDD_FIRSTBUTTON+NCONSTBTNS)
			SendMessage(hMousewnd,WM_RBUTTONDOWN,wParam,
			MAKELONG((WORD)pt.x,(WORD)pt.y));
		return TRUE;
		
	case PM_ACTIVATEBUTTONS:
		for (i=0;i<NCONSTBTNS;i++) {
			SendMessage(GetDlgItem(hwnd,IDD_FIRSTBUTTON+i),BM_SETSTYLE,
				(WPARAM)BS_PUSHBUTTON,TRUE);
			EnableWindow(GetDlgItem(hwnd,IDD_FIRSTBUTTON+i),
				SendDlgItemMessage(hwnd,IDD_FIRSTBUTTON+i,WM_GETTEXTLENGTH,0,0) &&
				SendDlgItemMessage(hwnd,IDD_FIRSTNUMBER+i,WM_GETTEXTLENGTH,0,0));
		}
		// Set focus to first available constant, or CANCEL if all blank
		SetFocus(GetNextDlgTabItem(hwnd,GetDlgItem(hwnd,IDOK),FALSE));
		SendMessage(GetFocus(),BM_SETSTYLE,(WPARAM)BS_DEFPUSHBUTTON,TRUE);
		return TRUE;
		
    case PM_SAVEBUTTONS: // Used to share save sequence by two exits
		for (i=0;i<NCONSTBTNS;i++) {
			// Store values on actual buttons
			GetDlgItemText(hwnd,IDD_FIRSTBUTTON+i,ButtonText[i],32);
			GetDlgItemText(hwnd,IDD_FIRSTNUMBER+i,ButtonNumber[i],32);
			// Now write .INI file
			lstrcpy(str,ButtonText[i]);
			lstrcat(str,";");
			lstrcat(str,ButtonNumber[i]);
			WritePrivateProfileString("Constants",itoa(i,ach,10),str,IniFile);
		}
		return TRUE;
		
	case WM_COMMAND: {
		WORD wID = GET_WM_COMMAND_ID(wParam,lParam);
		
		if (wID>=IDD_FIRSTBUTTON && wID<IDD_FIRSTBUTTON+NCONSTBTNS) {
			// Prepare to paste number (copied to Xreg in XCWINPRO)
			ldConstant = _strtold(ButtonNumber[wID-IDD_FIRSTBUTTON],NULL);
			lstrcpy(ConstMsg,NoAmpersand(ButtonText[wID-IDD_FIRSTBUTTON]));
			// Save buttons
			SendMessage(hwnd,PM_SAVEBUTTONS,0,0);
			EndDialog(hwnd,TRUE);
			DestroyMenu(hpopup);
			return TRUE;
		}
		
		switch (wID)
		{
		case IDOK:
			// save buttons
			ldConstant = 0;
			SendMessage(hwnd,PM_SAVEBUTTONS,0,0);
			EndDialog(hwnd,FALSE);  // Return TRUE means paste number
			DestroyMenu(hpopup);
			return TRUE;
			
		case IDCANCEL:
			// Restore all old values
			for (i=0;i<NCONSTBTNS;i++) {
				lstrcpy(ButtonText[i],_ButtonText[i]);
				lstrcpy(ButtonNumber[i],_ButtonNumber[i]);
				lstrcpy(ButtonTextUndo[i],_ButtonTextUndo[i]);
				lstrcpy(ButtonNumberUndo[i],_ButtonNumberUndo[i]);
				Undoable[i] = _Undoable[i];
			}
			EndDialog(hwnd,FALSE);
			DestroyMenu(hpopup);
			return TRUE;
			
		case IDHELP: {
			char helpstr[_MAX_PATH+MAX_HELP];
			strcpy(helpstr,HelpFile);
			strcat(helpstr,CONSTANTSHELP);
			HtmlHelp(GetParent(hwnd),helpstr,HH_DISPLAY_TOPIC,0);
			return TRUE;
		}
		}
	}
	// end case WM_COMMAND
	}

	return FALSE;
}

LRESULT CALLBACK ButtonSubProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	POINT pt;
	UINT enable;
	int iID = GetDlgCtrlID(hwnd);
	int ix = iID-IDD_FIRSTBUTTON;
	HWND hwndDlg = GetParent(hwnd);
	char s[34];
	char DefButtonStr[65];
	char *p;

	switch (wMsg)
	{
	case WM_RBUTTONDOWN:
		pt.x = LOWORD(lParam);
		pt.y = HIWORD(lParam);
		ClientToScreen(hwnd, &pt);
		// Activate/deactivate Undo choice:
		GetDlgItemText(hwndDlg,iID,NewButtonText,32);
		GetDlgItemText(hwndDlg,iID+100,NewNumber,32);
		enable = Undoable[ix] || lstrlen(ButtonTextUndo[ix]) && lstrlen(ButtonNumberUndo[ix]) ? MF_ENABLED : MF_GRAYED;
		EnableMenuItem(hpopup,IDM_UNDO,MF_BYCOMMAND|enable);
		enable = lstrlen(ButtonText[ix]) && lstrlen(ButtonNumber[ix]) ? MF_ENABLED : MF_GRAYED;
		EnableMenuItem(hpopup,IDM_CLEAR,MF_BYCOMMAND|enable);
		if (lstrlen(ButtonNumber[ix])) {
			lstrcpy(s,"(Value: ");
			lstrcat(s,ButtonNumber[ix]);
			lstrcat(s,")");
		}
		else
			lstrcpy(s,"(empty)");
		ModifyMenu(hpopup,IDM_VALUE,MF_DISABLED|MF_STRING,IDM_VALUE,s);
		TrackPopupMenu(hpopup,TPM_LEFTALIGN|TPM_RIGHTBUTTON,
			pt.x,pt.y,0,hwnd,NULL);
		return 0;
		
	case WM_COMMAND: {
		WORD wID = GET_WM_COMMAND_ID(wParam,lParam);
		switch (wID) {
		case IDM_EDIT:
			GetDlgItemText(hwndDlg,iID,NewButtonText,32);
			GetDlgItemText(hwndDlg,iID+100,NewNumber,32);
			if (DialogBox(hInst,"EDITBUTTON",hwnd,EditDlgProc)) { // true="OK"
				Undoable[ix] = TRUE;
				GetDlgItemText(hwndDlg,iID,ButtonTextUndo[ix],32);
				GetDlgItemText(hwndDlg,iID+100,ButtonNumberUndo[ix],32);
				SetDlgItemText(hwndDlg,iID,NewButtonText);
				SetDlgItemText(hwndDlg,iID+100,NewNumber);
				lstrcpy(ButtonText[ix],NewButtonText);
				lstrcpy(ButtonNumber[ix],NewNumber);
			}
			SendMessage(hwndDlg,PM_ACTIVATEBUTTONS,0,0);
			return 0;
			
		case IDM_UNDO:
			GetDlgItemText(hwndDlg,iID,NewButtonText,32);
			GetDlgItemText(hwndDlg,iID+100,NewNumber,32);
			SetDlgItemText(hwndDlg,iID,ButtonTextUndo[ix]);
			SetDlgItemText(hwndDlg,iID+100,ButtonNumberUndo[ix]);
			lstrcpy(ButtonText[ix],ButtonTextUndo[ix]);
			lstrcpy(ButtonNumber[ix],ButtonNumberUndo[ix]);
			lstrcpy(ButtonTextUndo[ix],NewButtonText);
			lstrcpy(ButtonNumberUndo[ix],NewNumber);
			SendMessage(hwndDlg,PM_ACTIVATEBUTTONS,0,0);
			return 0;
			
		case IDM_CLEAR:
			GetDlgItemText(hwndDlg,iID,ButtonTextUndo[ix],32);
			GetDlgItemText(hwndDlg,iID+100,ButtonNumberUndo[ix],32);
			Undoable[ix] = TRUE;
			SetDlgItemText(hwndDlg,iID,"");
			SetDlgItemText(hwndDlg,iID+100,"");
			lstrcpy(ButtonText[ix],"");
			lstrcpy(ButtonNumber[ix],"");
			SendMessage(hwndDlg,PM_ACTIVATEBUTTONS,0,0);
			return 0;

		case IDM_RESTORE:
			GetDlgItemText(hwndDlg,iID,ButtonTextUndo[ix],32);
			GetDlgItemText(hwndDlg,iID+100,ButtonNumberUndo[ix],32);
			Undoable[ix] = TRUE;
			// the default values are stored in a stringtable
			LoadString(hInst,IDD_STRINGS+ix,DefButtonStr,65);
			// Parse string
			p = strtok(DefButtonStr,";");
			if (p)
				lstrcpyn(ButtonText[ix],p,32);
			else
				lstrcpy(ButtonText[ix],"");
			p = strtok(NULL,";");
			if (p)
				lstrcpyn(ButtonNumber[ix],p,32);
			else
				lstrcpy(ButtonNumber[ix],"");
			SetDlgItemText(hwndDlg,iID,ButtonText[ix]);
			SetDlgItemText(hwndDlg,iID+100,ButtonNumber[ix]);
			SendMessage(hwndDlg,PM_ACTIVATEBUTTONS,0,0);
			return 0;

		case IDM_RESTOREALL:
			for (ix=0;ix<NCONSTBTNS;ix++) {
				GetDlgItemText(hwndDlg,ix+IDD_FIRSTBUTTON,ButtonTextUndo[ix],32);
				GetDlgItemText(hwndDlg,ix+IDD_FIRSTNUMBER,ButtonNumberUndo[ix],32);
				Undoable[ix] = TRUE;
				// the default values are stored in a stringtable
				LoadString(hInst,IDD_STRINGS+ix,DefButtonStr,65);
				// Parse string
				p = strtok(DefButtonStr,";");
				if (p)
					lstrcpyn(ButtonText[ix],p,32);
				else
					lstrcpy(ButtonText[ix],"");
				p = strtok(NULL,";");
				if (p)
					lstrcpyn(ButtonNumber[ix],p,32);
				else
					lstrcpy(ButtonNumber[ix],"");
				SetDlgItemText(hwndDlg,ix+IDD_FIRSTBUTTON,ButtonText[ix]);
				SetDlgItemText(hwndDlg,ix+IDD_FIRSTNUMBER,ButtonNumber[ix]);
			}
			SendMessage(hwndDlg,PM_ACTIVATEBUTTONS,0,0);
			return 0;
		}
	}
	// end case WM_COMMAND:

    } // switch (wMsg)

	return CallWindowProc(OldButtonProc,hwnd,wMsg,wParam,lParam);
}

#ifdef __BORLANDC__
#pragma argsused
#endif

BOOL CALLBACK EditDlgProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	switch (wMsg)
	{
	case WM_INITDIALOG:
		SetDlgItemText(hwnd,IDD_NEWBUTTON,NewButtonText);
		SetDlgItemText(hwnd,IDD_NEWNUMBER,NewNumber);
		return TRUE;
	// end case WM_INITDIALOG:
		
	case WM_COMMAND: {
		WORD wID = GET_WM_COMMAND_ID(wParam,lParam);
		LD lddum;
		char *p;
		switch (wID)
		{
		case IDOK:
			GetDlgItemText(hwnd,IDD_NEWBUTTON,NewButtonText,32);
			GetDlgItemText(hwnd,IDD_NEWNUMBER,NewNumber,32);
			// Must check for valid entries:
			// Stop if no title, but number
			if (!lstrlen(NewButtonText) && lstrlen(NewNumber)) {
				SetFocus(GetDlgItem(hwnd,IDD_NEWBUTTON));
				return TRUE;
			}
			// Stop if a ";" in title
			if ((p=strchr(NewButtonText,';'))!=0) {
				SendDlgItemMessage(hwnd,IDD_NEWBUTTON,EM_SETSEL,
					p-NewButtonText,p-NewButtonText+1);
				SetFocus(GetDlgItem(hwnd,IDD_NEWBUTTON));
				return TRUE;
			}
			// Stop if not a number (or extra input)
			lddum = _strtold(NewNumber,&p);
			if (p-NewNumber<(int)lstrlen(NewNumber)) {
				SendDlgItemMessage(hwnd,IDD_NEWNUMBER,EM_SETSEL,
					p-NewNumber,lstrlen(NewNumber));
				SetFocus(GetDlgItem(hwnd,IDD_NEWNUMBER));
				return TRUE;
			}
			// Stop if number is too large or too small (abs value)
			if (fabsl(lddum)>NUM_MAX || fabsl(lddum)<NUM_MIN) {
				SendDlgItemMessage(hwnd,IDD_NEWNUMBER,EM_SETSEL,0,MAKELONG(0,-1));
				SetFocus(GetDlgItem(hwnd,IDD_NEWNUMBER));
				return TRUE;
			}
			EndDialog(hwnd,TRUE);
			return TRUE;
			
		case IDCANCEL:
			EndDialog(hwnd,FALSE);
			return TRUE;
		}
	}
	// end case WM_COMMAND:

	} // switch (wMsg)
	return FALSE;
}

#ifdef __BORLANDC__
#pragma argsused
#endif

BOOL CALLBACK EditKeyDlgProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	switch (wMsg)
	{
	case WM_INITDIALOG: {
		int i;
		int ix;
		int s=-1;
		WORD m=(WORD)CURR_RADIX;
		int x,y;
		RECT rc;
		for (i=0;i<NKEYID;i++) {
			if (Key[i].mode&m) {
				ix = SendDlgItemMessage(hwnd,IDC_KEYLIST,LB_ADDSTRING,0,(LPARAM)(LPSTR)Key[i].tip);
				SendDlgItemMessage(hwnd,IDC_KEYLIST,LB_SETITEMDATA,ix,i);
				if (sameKeys(Key[i].id,KeyToEdit)) s=ix;
			}
		}
		if (s>=0) {
			SendDlgItemMessage(hwnd,IDC_KEYLIST,LB_SETCURSEL,s,0);
		}
		// Position dialog just left (or right) of key being edited
		GetClientRect(hwnd,&rc);
		x = RectToEdit.left-rc.right-20;
		y = RectToEdit.top-rc.bottom/2;
		if (x<0) x = RectToEdit.right+20;
		if (y<0) y = 0;
		SetWindowPos(hwnd,NULL,x,y,0,0,SWP_NOZORDER|SWP_NOSIZE);
		return TRUE;
	}
	// end case WM_INITDIALOG:
		
	case WM_COMMAND: {
		int ix;
		WORD wID = GET_WM_COMMAND_ID(wParam,lParam);
		WORD wCMD = GET_WM_COMMAND_CMD(wParam,lParam);
		// DBLCLK on listbox: transform to an IDOK click
		if (wID==IDC_KEYLIST && wCMD==LBN_DBLCLK) wID = IDOK;
		switch (wID)
		{
		case IDOK:
			ix = SendDlgItemMessage(hwnd,IDC_KEYLIST,LB_GETCURSEL,0,0);
			KeyToEdit = SendDlgItemMessage(hwnd,IDC_KEYLIST,LB_GETITEMDATA,ix,0);
			EndDialog(hwnd,KeyToEdit!=LB_ERR);
			return TRUE;
			
		case IDCANCEL:
			EndDialog(hwnd,FALSE);
			return TRUE;
		case IDHELP: {
			char helpstr[_MAX_PATH+MAX_HELP];
			strcpy(helpstr,HelpFile);
			strcat(helpstr,EDITKEYHELP);
			HtmlHelp(GetParent(hwnd),helpstr,HH_DISPLAY_TOPIC,0);
			return TRUE;
		}
		}
	}
	// end case WM_COMMAND:

	} // switch (wMsg)
	return FALSE;
}

/* Later? - requires drag&drop
#ifdef __BORLANDC__
#pragma argsused
#endif

BOOL CALLBACK EditKeysProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	switch (wMsg)
	{
	case WM_INITDIALOG: {
		return TRUE;
	}
	// end case WM_INITDIALOG:
		
	case WM_COMMAND: {
		WORD wID = GET_WM_COMMAND_ID(wParam,lParam);
		switch (wID)
		{
		case IDOK:
			EndDialog(hwnd,TRUE);
			hwndEditKeys = NULL;
			return TRUE;
			
		case IDCANCEL:
			EndDialog(hwnd,FALSE);
			hwndEditKeys = NULL;
			return TRUE;

		case IDHELP: {
			char helpstr[_MAX_PATH+MAX_HELP];
			strcpy(helpstr,HelpFile);
			strcat(helpstr,EDITKEYSHELP);
			HtmlHelp(GetParent(hwnd),helpstr,HH_DISPLAY_TOPIC,0);
			return TRUE;
		}
		}
	}
	// end case WM_COMMAND:

	} // switch (wMsg)
	return FALSE;
}
*/

/* Ownerdraw button-like radiobutton doesn't work
#define ICON_HEIGHT 16
#define ICON_WIDTH 16
static HICON hOwnerDrawIcon = NULL;

static void PrepareImageRect(HWND hButtonWnd, BOOL bHasTitle, RECT* rpItem, RECT* rpTitle, BOOL bIsPressed, DWORD dwWidth, DWORD dwHeight, RECT* rpImage)
{
	RECT rBtn;

	CopyRect(rpImage, rpItem);

	GetClientRect(hButtonWnd, &rBtn);
	if (bHasTitle == FALSE)
	{
		// Center image horizontally
		LONG rpImageWidth = rpImage->right - rpImage->left;
		rpImage->left += ((rpImageWidth - (long)dwWidth)/2);
	}
	else
	{
		// Image must be placed just inside the focus rect
		LONG rpImageHeight;
		LONG rpTitleWidth = rpTitle->right - rpTitle->left;
		rpTitle->right = rpTitleWidth - dwWidth - 30;
		rpTitle->left = 30;
		rpImage->left = rBtn.right - dwWidth - 30;
		// Center image vertically
		rpImageHeight = rpImage->bottom - rpImage->top;
		rpImage->top += ((rpImageHeight - (long)dwHeight)/2);
	}
		
	// If button is pressed then press image also
	if (bIsPressed)
		OffsetRect(rpImage, 1, 1);
} // End of PrepareImageRect

static void DrawTheIcon(HWND hButtonWnd, HDC* dc, BOOL bHasTitle, RECT* rpItem, RECT* rpTitle, BOOL bIsPressed, BOOL bIsDisabled)
{
	RECT	rImage;
	PrepareImageRect(hButtonWnd, bHasTitle, rpItem, rpTitle, bIsPressed, ICON_WIDTH, ICON_HEIGHT, &rImage);

	// Ole'!
	DrawState(	*dc,
				NULL,
				NULL,
				(LPARAM)hOwnerDrawIcon,
				0,
				rImage.left,
				rImage.top,
				(rImage.right - rImage.left),
				(rImage.bottom - rImage.top), 
				(bIsDisabled ? DSS_DISABLED : DSS_NORMAL) | DST_ICON);
} // End of DrawTheIcon
*/

#ifdef __BORLANDC__
#pragma argsused
#endif

BOOL CALLBACK EditProgramProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	static int lastEdited = 0;
	static int _lastEdited = 0;
	static BOOL positioned = FALSE;
	static int x,y;
	static char *progs[12] = {NULL};
	switch (wMsg)
	{
	case WM_INITDIALOG: {
		int i;
		FILE *f;
		char fname[16];
		RECT rcXCALC,rcDlg;
		if (!positioned) {
			// Position dialog just left (or right) of XCALC
			GetWindowRect(hwndXCALC,&rcXCALC);
			GetWindowRect(hwnd,&rcDlg);
			x = rcXCALC.left-(rcDlg.right-rcDlg.left)-2;
			y = rcXCALC.top;
			if (x<0) x = rcXCALC.right+2;
			positioned = TRUE;
		}
		SetWindowPos(hwnd,NULL,x,y,0,0,SWP_NOZORDER|SWP_NOSIZE);
		// Read all 12 programs (if they exist) into memory
		for (i=0;i<12;i++) {
			sprintf(fname,"prog%d.txt",i+1);
			if ((f=fopen(fname,"rb"))!=NULL) {
				//char line[128];
				long l;
				fseek(f,0,SEEK_END);
				l=ftell(f);
				fseek(f,0,SEEK_SET);
				progs[i]=malloc(l+1);
				fread(progs[i],1,l,f);
				progs[i][l]=0;
				fclose(f);
			}
		}
		SendDlgItemMessage(hwnd,IDC_P1+lastEdited,BM_SETCHECK,BST_CHECKED,0);
		SetDlgItemText(hwnd,IDC_PROGEDIT,progs[lastEdited]);
		/*
		if(hOwnerDrawIcon == NULL)
			hOwnerDrawIcon = (HICON)LoadIcon(hInst,"RECORD");
		*/
		return TRUE;
	}
	// end case WM_INITDIALOG:

	// PM_ADDTOPROGRAM: translate input key to a program statement and insert in editor if recording
	case PM_ADDTOPROGRAM: {
		int i;
		char cmd[32];
		const char SPC[] = " ";
		if (SendDlgItemMessage(hwnd,IDC_RECORD,BM_GETCHECK,0,0)==BST_CHECKED) {
			// Comma/point fudge. Getting nasty!
			if (wRadix==10 && wParam==0 && (lParam=='.' || lParam==',')) {
				if (wDispMode&COMPLEX || lParam==k_useComma?',':'.') {
					cmd[0]=(char)lParam;
					cmd[1]=0;
					strcat(cmd,SPC);
					SendDlgItemMessage(hwnd,IDC_PROGEDIT,EM_REPLACESEL,TRUE,(LPARAM)cmd);
				}
			}
			else {
				for (i=0;i<NKEYID;i++) {
					if (Key[i].op[0]==wParam && Key[i].op[1]==lParam) {
						// Hyperbolic fudge. Better make a proper system now...
						if (bHyperbolic && (Key[i].id==KEY_SIN || Key[i].id==KEY_COS || Key[i].id==KEY_TAN || Key[i].id==KEY_ASIN || Key[i].id==KEY_ACOS || Key[i].id==KEY_ATAN)) {
							continue;
						}
						else if (bHyperbolic && (Key[i].id==KEY_SINH || Key[i].id==KEY_COSH || Key[i].id==KEY_TANH || Key[i].id==KEY_ASINH || Key[i].id==KEY_ACOSH || Key[i].id==KEY_ATANH)) {
							strcpy(cmd,Key[i].name);
							strcat(cmd,SPC);
							SendDlgItemMessage(hwnd,IDC_PROGEDIT,EM_REPLACESEL,TRUE,(LPARAM)cmd);
							break;
						}
						else if (Key[i].mode&wDispMode && *Key[i].name) {
							strcpy(cmd,Key[i].name);
							strcat(cmd,SPC);
							SendDlgItemMessage(hwnd,IDC_PROGEDIT,EM_REPLACESEL,TRUE,(LPARAM)cmd);
							break;
						}
					}
				}
			}
		}
		return TRUE;
	}
	// end case PM_ADDTOPROGRAM:

	// PM_UNDOPROGRAM: scan back in editor, remove word prior to insertion point
	case PM_UNDOPROGRAM: {
		int start=0;
		int end=0;
		char *edittext;
		int textlen;
		SendDlgItemMessage(hwnd,IDC_PROGEDIT,EM_GETSEL,(WPARAM)&start,(LPARAM)&end);
		if (start==end && start>1) { // need a space and a word prior
			// No selected text, so backtrack to find next space (or beginning-of-buffer)
			textlen=SendDlgItemMessage(hwnd,IDC_PROGEDIT,WM_GETTEXTLENGTH,0,0);
			edittext=malloc(textlen+1);
			SendDlgItemMessage(hwnd,IDC_PROGEDIT,WM_GETTEXT,(WPARAM)(textlen+1),(LPARAM)edittext);
			start-=2;
			while (start>=0 && edittext[start]!=' ') start--;
			start++;
			SendDlgItemMessage(hwnd,IDC_PROGEDIT,EM_SETSEL,start,end);
			SendDlgItemMessage(hwnd,IDC_PROGEDIT,EM_REPLACESEL,TRUE,(LPARAM)"");
			free(edittext);
		}
		// otherwise, do nothing
		return TRUE;
	}
	// end case PM_UNDOPROGRAM:

	case WM_MOVE: {
		RECT rc;
		GetWindowRect(hwnd,&rc);
		x = rc.left;
		y = rc.top;
		return TRUE;
	}
	// end case WM_MOVE:

	case WM_COMMAND: {
		WORD wID = GET_WM_COMMAND_ID(wParam,lParam);
		int len;
		int i;
		FILE *f;
		char fname[16];
		switch (wID)
		{
		case IDOK:
			SendMessage(hwnd,WM_COMMAND,IDC_APPLY,0);
			for (i=0; i<12; i++) if (progs[i]) free(progs[i]);
			EndDialog(hwnd,TRUE);
			hwndEditProgram = NULL;
			return TRUE;

		case IDCANCEL:
			for (i=0; i<12; i++) if (progs[i]) free(progs[i]);
			EndDialog(hwnd,FALSE);
			hwndEditProgram = NULL;
			return TRUE;

		case IDHELP: {
			char helpstr[_MAX_PATH+MAX_HELP];
			strcpy(helpstr,HelpFile);
			strcat(helpstr,PROGRAMHELP);
			HtmlHelp(GetParent(hwnd),helpstr,HH_DISPLAY_TOPIC,0);
			return TRUE;
		}

		case IDC_CODES: {
			char helpstr[_MAX_PATH+MAX_HELP];
			strcpy(helpstr,HelpFile);
			strcat(helpstr,CODESHELP);
			HtmlHelp(GetParent(hwnd),helpstr,HH_DISPLAY_TOPIC,0);
			return TRUE;
		}

		case IDC_APPLY:
			SendDlgItemMessage(hwnd,IDC_RECORD,BM_SETCHECK,BST_UNCHECKED,0);
			SendMessage(hwnd,WM_COMMAND,_lastEdited+IDC_P1,0);
			for (i=0;i<12;i++) {
				sprintf(fname,"prog%d.txt",i+1);
				if (progs[i]) {
					if ((f=fopen(fname,"wb"))!=NULL) {
						fwrite(progs[i],1,lstrlen(progs[i]),f);
						fclose(f);
					}
				} else {
					unlink(fname);
				}
			}
			lastEdited = _lastEdited;
			return TRUE;

		case IDC_RUN:
			SendDlgItemMessage(hwnd,IDC_RECORD,BM_SETCHECK,BST_UNCHECKED,0);
			SendMessage(hwnd,WM_COMMAND,IDC_APPLY,0);
			PostMessage(hwndXCALC,WM_COMMAND,AS_PROG1+_lastEdited,0);
			return TRUE;

		case IDC_P1:
		case IDC_P2:
		case IDC_P3:
		case IDC_P4:
		case IDC_P5:
		case IDC_P6:
		case IDC_P7:
		case IDC_P8:
		case IDC_P9:
		case IDC_P10:
		case IDC_P11:
		case IDC_P12:
			SendDlgItemMessage(hwnd,IDC_RECORD,BM_SETCHECK,BST_UNCHECKED,0);
			if (progs[_lastEdited]) {
				free(progs[_lastEdited]);
				progs[_lastEdited]=NULL;
			}
			len = SendDlgItemMessage(hwnd,IDC_PROGEDIT,WM_GETTEXTLENGTH,0,0);
			if (len) {
				progs[_lastEdited]=malloc(len+1);
				GetDlgItemText(hwnd,IDC_PROGEDIT,progs[_lastEdited],len+1);
			}
			_lastEdited = wID-IDC_P1;
			SetDlgItemText(hwnd,IDC_PROGEDIT,progs[_lastEdited]);
			// SetDlgItemText - sets edit control text
			return TRUE;
		}
		default:
			i=0;
			return FALSE;
	}
	// end case WM_COMMAND:
	/*
	case WM_DRAWITEM: {
		HDC dc;
		BOOL bIsChecked;
		BOOL bIsFocused;
		BOOL bIsDisabled;
		BOOL bDrawFocusRect;
		RECT itemRect;
		COLORREF crColor;
		HBRUSH brBackground;
		char sTitle[100];
		RECT captionRect;
		BOOL bHasTitle;

		// draw the owner draw button
		LPDRAWITEMSTRUCT lpDIS = (LPDRAWITEMSTRUCT) lParam;
		
		if(lpDIS->CtlID != IDC_RECORD)
			return 0;

		dc = lpDIS->hDC;

		// button state
		bIsChecked = ((lpDIS->itemState & ODS_SELECTED) || (lpDIS->itemState & ODS_CHECKED) || IsDlgButtonChecked(hwnd,IDC_RECORD));
		bIsFocused = (lpDIS->itemState & ODS_FOCUS);
		bIsDisabled = (lpDIS->itemState & ODS_DISABLED);
		bDrawFocusRect = !(lpDIS->itemState & ODS_NOFOCUSRECT);

		itemRect = lpDIS->rcItem;

		SetBkMode(dc, TRANSPARENT);

		// Prepare draw... paint button background

		if (bIsFocused) {
			HBRUSH br = CreateSolidBrush(RGB(0,0,0));  
			FrameRect(dc, &itemRect, br);
			InflateRect(&itemRect, -1, -1);
			DeleteObject(br);
		}

		crColor = GetSysColor(COLOR_BTNFACE);
		brBackground = CreateSolidBrush(crColor);
		FillRect(dc, &itemRect, brBackground);
		DeleteObject(brBackground);

		// Draw pressed button
		if (bIsChecked) {
			HBRUSH brBtnShadow = CreateSolidBrush(GetSysColor(COLOR_BTNSHADOW));
			FrameRect(dc, &itemRect, brBtnShadow);
			DeleteObject(brBtnShadow);
		}
		else { // ...else draw non pressed button
		    UINT uState = DFCS_BUTTONPUSH |
                  ((bIsChecked) ? DFCS_PUSHED : 0);
			DrawFrameControl(dc, &itemRect, DFC_BUTTON, uState);
		}
		// Read the button's title
		GetWindowText(GetDlgItem(hwnd, IDC_RECORD), sTitle, 100);

		captionRect = lpDIS->rcItem;

		// Draw the icon
		bHasTitle = (sTitle[0] != '\0');
		DrawTheIcon(GetDlgItem(hwnd, IDC_RECORD), &dc, bHasTitle, &lpDIS->rcItem, &captionRect, bIsChecked, bIsDisabled);

		// Write the button title (if any)
		if (bHasTitle) { // Draw the button's title
			RECT centerRect;
			LONG captionRectWidth;
			LONG captionRectHeight;
			LONG centerRectWidth;
			LONG centerRectHeight;
			// If button is pressed then "press" title also
			if (bIsChecked)
				OffsetRect(&captionRect, 1, 1);

			// Center text
			centerRect = captionRect;
			DrawText(dc, sTitle, -1, &captionRect, DT_WORDBREAK | DT_CENTER | DT_CALCRECT);
			captionRectWidth = captionRect.right - captionRect.left;
			captionRectHeight = captionRect.bottom - captionRect.top;
			centerRectWidth = centerRect.right - centerRect.left;
			centerRectHeight = centerRect.bottom - centerRect.top;
			OffsetRect(&captionRect, (centerRectWidth - captionRectWidth)/2, (centerRectHeight - captionRectHeight)/2);

			SetBkMode(dc, TRANSPARENT);

			if (bIsDisabled) {
				OffsetRect(&captionRect, 1, 1);
				SetTextColor(dc, GetSysColor(COLOR_3DHILIGHT));
				DrawText(dc, sTitle, -1, &captionRect, DT_WORDBREAK | DT_CENTER);
				OffsetRect(&captionRect, -1, -1);
				SetTextColor(dc, GetSysColor(COLOR_3DSHADOW));
				DrawText(dc, sTitle, -1, &captionRect, DT_WORDBREAK | DT_CENTER);
			}
			else {
				SetTextColor(dc, GetSysColor(COLOR_BTNTEXT));
				SetBkColor(dc, GetSysColor(COLOR_BTNFACE));
				DrawText(dc, sTitle, -1, &captionRect, DT_WORDBREAK | DT_CENTER);
			} // if
		}

		// Draw the focus rect
		if (bIsFocused && bDrawFocusRect)
			{
			RECT focusRect = itemRect;
			InflateRect(&focusRect, -3, -3);
			DrawFocusRect(dc, &focusRect);
			} // if
		return (TRUE);
	}
	break;
	*/
	// end case WM_COMMAND:
	} // switch (wMsg)
	return FALSE;
}
