//------------------------------------------------------------------------
// File for project XCALC
//------------------------------------------------------------------------
// Number Display Window procedure
// Actions:
//	- Update the various display cells
//  - Allow window move if caption is not shown
//	- Respond to (double-)click for copy/paste of numbers
//------------------------------------------------------------------------

#include "xcalc.h"

// These variables are global and declared in XCALC.H:

BOOL Moving=FALSE;
HWND hwndMoving;

LRESULT CALLBACK DisplayProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	int iWindow,n,i,SrcX,DstX;
	static int TimerID, BlinkWindow=0;
	static HPEN hPen=NULL;
	static HBRUSH hBrush;
	static POINT origpt;
	static RECT origrect;
	static int delayMove;
	
	// find which window this is, to use the right digit string (cST)
	iWindow = GetWindowLong(hwnd,GWL_ID);
	
	switch (wMsg) {
		
	case PM_UPDATE:		// is sent from main window to update display
		{
			InvalidateRect(hwnd,NULL,0);
			break;
		}
		
	case WM_LBUTTONDOWN:
		{
			POINTS mousepts = MAKEPOINTS(lParam);
			if (Title || !EnableMouseMove) break;

			origpt.x = mousepts.x;
			origpt.y = mousepts.y;

			ClientToScreen(hwnd,&origpt);

			GetWindowRect(GetParent(hwnd),&origrect);
			SetCapture(hwnd);

			delayMove = 2;		// do not start moving after 2 mouse move messages
			Moving = TRUE;			// checked in MouseMove
			break;
		}
		
	case WM_MOUSEMOVE:
		{
			POINTS mousepts = MAKEPOINTS(lParam);
			POINT mousept;
			RECT rect;
			int xoff,yoff;
			mousept.x = mousepts.x;
			mousept.y = mousepts.y;
			
			if (Title || !EnableMouseMove) break;
			if (Moving) {
				ClientToScreen(hwnd,&mousept);
				rect = origrect;
				if (delayMove) {			// this will avoid a blink on DBLCLK
					if (delayMove--) break;
				}
				xoff = mousept.x - origpt.x;
				yoff = mousept.y - origpt.y;
				OffsetRect(&rect,xoff,yoff);
				MoveWindow(GetParent(hwnd),rect.left,rect.top,rect.right-rect.left,rect.bottom-rect.top,TRUE);
				WindowX=rect.left;
				WindowY=rect.top;
				// Need to track Mini settings
				if (Mini) {
					m_WindowX=WindowX;
					m_WindowY=WindowY;
				}
				else {
					n_WindowX=WindowX;
					n_WindowY=WindowY;
				}
			}
			break;
		}
		
	case WM_LBUTTONUP:
		{
			if (Title || !EnableMouseMove) break;
			if (Moving) {
				Moving = FALSE;
				ReleaseCapture();
			}
			break;
		}
		
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			RECT rect;
			POINT points[3];
			HPEN holdpen;
			HBRUSH holdbrush;
			BOOL BinDot = FALSE;
			int DotYPos = Border+BitmapY-1;
			
			BeginPaint(hwnd,&ps);
			
			GetClientRect(hwnd,&rect);                   // draw window
			FillRect(ps.hdc,&rect,hBackBrush);
			
			// copy digits to display, from left to right
			// index destination (display) by n, source (bitmap) by i
			// dec96: leave StatusX pixels blank on left
			
			for (n=0;n<NDIGITS;n++) {
				char c = cST[iWindow][n];
				// Flip "." and "," if so required:
				if (d_useComma) {
					if (c=='.') c=',';
					else if (c==',') c='.';
				}
				i = strindex(c,Printable);
				SrcX = i*BitmapX;
				DstX = n*BitmapX;
				BitBlt(ps.hdc,DstX+Border+StatusX,Border,BitmapX,BitmapY,
					hDigitsDC,SrcX,0,SRCCOPY);
				// Make 4-digit separators in Binary mode. Primitive, but ...
				// July 97 - also for Hex (4-digit) / Octal (3-digit)
				if (
					ThousandSep	&& BinDot &&
					(    (wRadix==2 || wRadix==16) && !((NDIGITS-n)%4)
					|| (wRadix==8) && !((NDIGITS-n)%3)
					)
					) {
					SetPixel(ps.hdc,DstX+Border+StatusX,  DotYPos,GetColour(RGBDIGITS));
					SetPixel(ps.hdc,DstX+Border+StatusX-1,DotYPos+1,GetColour(RGBDIGITS));
					SetPixel(ps.hdc,DstX+Border+StatusX,  DotYPos+1,GetColour(RGBDIGITS));
					SetPixel(ps.hdc,DstX+Border+StatusX+1,DotYPos+1,GetColour(RGBDIGITS));
				}
				BinDot = i!=0;
			}
			
			// if hPen != NULL, a timer is running. Add a blot to the right of the number.
			// hPen/hBrush will contain the right colour for the blot, BlinkWindow
			// specifies which window shall receive the blink.
			
			if (hPen && iWindow==BlinkWindow) {
				points[0].x = rect.right-Border;
				points[0].y = rect.bottom/2-Border;
				points[1].x = rect.right-Border*2;
				points[1].y = rect.bottom/2;
				points[2].x = rect.right-Border;
				points[2].y = rect.bottom/2+Border;
				holdpen = (HPEN)SelectObject(ps.hdc,hPen);
				holdbrush = (HBRUSH)SelectObject(ps.hdc,hBrush);
				Polygon(ps.hdc,points,3);
				SelectObject(ps.hdc,holdpen);
				SelectObject(ps.hdc,holdbrush);
			}
			
			EndPaint(hwnd,&ps);
			
			break;
		}
		
	case WM_RBUTTONDBLCLK:				// paste number from clipboard, if possible
		{
			HANDLE hData;
			LPSTR lpSTR;
			WORD NumberType = (WORD)CURR_RADIX;
			char sData[NDIGITS+3],c;	// need (possibly) room for '0b'
			LD real=0.0,imag=0.0;
			long bin=0;
			long num=0,den=0;
			ContentType ct = ctREAL;
			BOOL Decimal=FALSE;
			
			
			// We have to do an UNDO save here, since mouse clicks only
			// gets here. Therefore, XCWINPRO does not save undo information
			// for keyboard-triggered paste sequences.
			
			SaveUndo(UT_PASTE,FALSE);
			
			if (IsClipboardFormatAvailable(CF_TEXT)) {
				LPSTR p;
				
				if (!OpenClipboard(hwnd)) break;
				hData = GetClipboardData(CF_TEXT);
				lpSTR = (LPSTR)GlobalLock(hData);
				c = *lpSTR;
				while (c==' ' || c=='\t' || c==0x0d || c==0x0a) c = *++lpSTR;
				lstrcpyn(sData,lpSTR,NDIGITS+2);
				GlobalUnlock(hData);
				CloseClipboard();
				// Delete any comma or point that is NOT the decimal character
				c = (char)(c_useComma ? '.' : ',');
				p = sData;
				while (*p) {
					if (*p == c) {
						LPSTR q = p;
						do {
							*q = *(q+1);
							q++;
						} while (*q);
					}
					if (*p != c) p++;
				}
				// Change commas (,) to points (.) if current setting is to
				// use comma to/from clipboard
				if (c_useComma) {
					p = sData;
					while (*p) {
						if (*p==',') *p='.';
						p++;
					}
				}
				// Change any " or ' to 
				p = sData;
				while (*p) {
					if (*p=='\'' || *p=='"') *p='';
					p++;
				}
				
				// Check for a decimal point
				p = sData;
				while (*p) {
					if (*p++=='.') Decimal = TRUE;
				}
				
				// Now figure out what kind of number format we have and read it:
				// - complex (contains i)
				// - fraction (contains /)
				// - DMS (contains  or ' or ")
				// - hex (0x... or &H...)
				// - binary (0b...)
				// - octal (0..., but no decimal point)
				// - decimal
				
				if (!strnicmp(sData,"0x",2) || !strnicmp(sData,"&h",2)) {	// hex
					if (sscanf(sData+2,"%lx",&bin)<1)
						return 1;
					real = (LD)bin;
				}
				else if (!strnicmp(sData,"0b",2) || !strnicmp(sData,"&b",2)) {	// binary
					long i = 0;
					BOOL OK = FALSE;
					
					for(n=2;n<NDIGITS+2;n++) {
						if (sData[n]!='0' && sData[n]!='1') break;
						i<<=1;
						i+=(sData[n]-'0');
						OK = TRUE;
					}
					if (OK)
						real = (LD)i;
					else
						return 1;
				}
				else if (!strnicmp(sData,"&o",2)) {		// octal
					if (sscanf(sData+2,"%lo",&bin)<1)
						return 1;
					real = (LD)bin;
				}
				else if (*sData=='0' && !Decimal) {		// octal
					if (sscanf(sData,"%lo",&bin)<1)
						return 1;
					real = (LD)bin;
				}
				else if (strchr(sData,'i')) {			// complex
					LD Re,Im;
					if ((n=sscanf(sData,"%Lg%Lg",&Re,&Im))<1)
						return 1;
					real = Re;
					imag = n==2? Im:0.0;
				}
				else if (strchr(sData,'/')) {			// fraction
					long w;
					if (sscanf(sData,"%d %d/%d",&w,&num,&den) == 3) {	// a b/c
						if (num<0 || den<=0)
							return 1;
						if (w<0) num = -num;
						num += w*den;
						ct = ctFRACTION;
					}
					else if (sscanf(sData,"%d/%d",&num,&den) == 2) {	// a/b
						if (den<=0)
							return 1;
						ct = ctFRACTION;
					}
					else
						return 1;
				}
				else if (strchr(sData,'')) {			// DMS
					long d,m;
					LD s=0;
					int sign;
					if (sscanf(sData,"%d%d%Lg",&d,&m,&s) >= 2) {	// dm[s]
						if (m<0 || s<0)
							return 1;
						sign = d<0?-1:1;
						real = d + sign*m/60.0L + sign*s/3600.0L;
						imag = 0;
						ct = ctDMS;
					}
					else
						return 1;
				}
				else {									// plain number
					if (sscanf(sData,"%Lg",&real)<1)
						return 1;
				}
				
				// Come here, number is pasted to real/imag or num/den
				
				switch (NumberType)
				{
				case DECIMAL:
					imag = 0.0;
					break;
				case COMPLEX:
					if (num != 0) {
						real = (LD)num / (LD)den;
						imag = 0.0;
						num = den = 0;
					}
					break;
				case BINARY: case OCTAL: case HEX:
					imag = 0.0;
					if (num != 0) {
						bin = num/den;
						num = den = 0;
					}
					else {
						if (real>(LD)LONG_MAX || real<(LD)LONG_MIN)
							return 1;
						bin = (long)real;
					}
					break;
				}
				
				// We have a paste. Commit Undo.
				
				CommitUndo();
				
				// Special considerations needed for Xreg. If the paste operation
				// to this register is keyboard-triggered, both wParam and lParam
				// will contain -1. Use this to correctly handle state information
				// in iState and bSL.
				
				if (iWindow==Xreg) {
					if (wParam==(WPARAM)-1 && lParam==(LPARAM)-1) {
						// Special consideration if paste is done by ^INS
						if (iState != 0) bSL = TRUE;
						if (bSL) lift();
					}
					iState = 0;	// destroy partially entered number
					// if mouse paste. Otherwise, lift is done already.
					bSL = TRUE;
				}
				
				// Update stack number
				ST[iWindow].ct   = ct;
				ST[iWindow].real = real;
				ST[iWindow].imag = imag;
				ST[iWindow].i	 = bin&CURR_MASK;
				ST[iWindow].num	 = num;
				ST[iWindow].den	 = den;
				
				MakeString(iWindow);      // create cST[] for window
				
				// visual feedback (do not activate if timer already running)
				if (!hPen) {
					if ((TimerID = SetTimer(hwnd,ID_BLINKTIMER,BLINKMS,NULL)) != 0) {
						hPen = hCaptionPen;
						hBrush = hCaptionBrush;
						BlinkWindow = iWindow;
					}
				}
				
				SendMessage(hwnd,PM_UPDATE,0,0); // update the display in any case
				
			} // if isClipboardAvailable
			
			break;
		}
		
		case WM_LBUTTONDBLCLK:       	// copy number/bitmap to clipboard
			{
				HANDLE hMem;
				HDC hdc,hmemdc;
				HBITMAP hBitMap,hbmpold;
				LPSTR lpMem,s,d;
				RECT rect;
				BOOL skipzero = FALSE;
				WORD NumberType = (WORD)CURR_RADIX;
				
				if ((hMem = GlobalAlloc(GHND,NDIGITS+3)) == NULL) break;
				
				if (iWindow==Xreg) {
					iState = 0; 						// end keyboard input, if any
					bSL = TRUE;
					MakeString(iWindow);				// format number for display
					SendMessage(hwnd,PM_UPDATE,0,0);	// update the display
				}
				
				// lock memory and get far pointer to it,
				// trim unnecessary spaces and transfer data
				
				lpMem=(LPSTR)GlobalLock(hMem);
				
				s = cST[iWindow];                       // source string
				d = lpMem;                              // destination
				
				if (NumberType == HEX) {				// prefix '0x' or '&H'
					if (Cstyle) {
						*d++ = '0';
						*d++ = 'x';
					}
					else {
						*d++ = '&';
						*d++ = 'H';
					}
				}
				else if (NumberType == OCTAL) {			// prefix '0' or '&O'
					if (Cstyle) {
						*d++ = '0';
						skipzero = TRUE;
					}
					else {
						*d++ = '&';
						*d++ = 'O';
					}
				}
				else if (NumberType == BINARY) {		// prefix '0b'
					if (Cstyle) {
						*d++ = '0';
						*d++ = 'b';
					}
					else {
						*d++ = '&';
						*d++ = 'B';
					}
				}
				while (*s==' ') s++;					// trim leading space
				if (skipzero && *s=='0') s++;			// skip zero on octal
				while (*s!=' ' && *s!=0) {				// copy first number,
					if (*s==',')						// ignoring 1000 commas
						s++;							// if any,
					else if (*s=='.') {					// adhering to point/comma settings for clipboard,
						*d++=(char)(c_useComma?',':'.');
						s++;
					}
					else								// else
						*d++=*s++;						// just copy.
				}
				
				if (*s!=0) {							// spaces -> complex or fraction, or decimal part with spaces
					if (NumberType==DECIMAL && ST[iWindow].ct==ctREAL) {
						while (*s) {					// copy rest without spaces
							while (*s==' ') s++;
							*d++=*s++;
						}
					}
					else {
						while (*s==' ') s++;				// skip space
						if (*s!=0) {						// possible incomplete: stop
							if (*s!='-') {
								if (NumberType == COMPLEX)	// add '+' if imag not -ve
									*d++='+';				// (if not fraction)
								else
									*d++=' ';
							}
							while (*s!=' ' && *s!=0) {
								if (*s=='.') {				// follow point/comma setting
									*d++=(char)(c_useComma?',':'.');
									s++;
								}
								else
									*d++=*s++;				// copy rest of digits
							}
						}
					}
				}
				
				GlobalUnlock(hMem);
				
				// open clipboard
				
				if (OpenClipboard(hwnd)) {
					EmptyClipboard();
					
					// transfer OEM text to the clipboard and close it
					
					SetClipboardData(CF_TEXT,hMem);
					
					// transfer bitmap image to clipboard
					
					hdc = GetDC(hwnd);
					hmemdc = CreateCompatibleDC(hdc);
					hBitMap = CreateCompatibleBitmap(hdc,BitmapX*NDIGITS+StatusX,BitmapY);
					hbmpold = (HBITMAP)SelectObject(hmemdc,hBitMap);
					BitBlt(hmemdc,0,0,BitmapX*NDIGITS+StatusX,BitmapY,hdc,Border,Border,SRCCOPY);
					SelectObject(hmemdc,hbmpold);
					DeleteDC(hmemdc);
					ReleaseDC(hwnd,hdc);
					
					SetClipboardData(CF_BITMAP,hBitMap);
					
					CloseClipboard();
					
					// visual feedback (do not activate if timer already running)
					if (!hPen) {
						if ((TimerID = SetTimer(hwnd,ID_BLINKTIMER,BLINKMS,NULL)) != 0) {
							hPen = hSelectedPen;
							hBrush = hSelectedBrush;
							BlinkWindow = iWindow;
							GetClientRect(hwnd,&rect);
							rect.left = rect.right-Border*2;
							InvalidateRect(hwnd,&rect,FALSE);
						}
					}
					
				} // if OpenClipboard
				
				break;
		}
		
		case WM_TIMER:
			{
				RECT rect;
				KillTimer(hDisplay[BlinkWindow],ID_BLINKTIMER);
				GetClientRect(hDisplay[BlinkWindow],&rect);
				rect.left = rect.right-Border*2;
				InvalidateRect(hDisplay[BlinkWindow],&rect,FALSE);
				hPen = NULL;
				break;
			}
			
		default:
			return DefWindowProc(hwnd, wMsg, wParam, lParam);
	}
	return 0L;
}

//------------------------------------------------------------------------
// strindex(char *,char)
// Finds index of character in string.
// If not found, return zero.
//------------------------------------------------------------------------
int strindex(char c, const char *const s)
{
	int i;
	for (i=0;*(s+i);i++) if (*(s+i)==c) return i;
	return 0;
}

//-----------------------------------------------------------------------
// MakeString converts number in register iReg to string, using
// global mode variables (wDispMode, sFix ..)
//-----------------------------------------------------------------------

void MakeString(int iReg)
{
	char bstr[NDIGITS+1];
	memset(cST[iReg],' ',NDIGITS);
	if(wRadix!=10) {
		ultoa(ST[iReg].i,bstr,wRadix);
		strupr(bstr);
		lstrcpy(&cST[iReg][NDIGITS-lstrlen(bstr)],bstr);
	}
	else {
		if (CURR_RADIX == DECIMAL) {
			// Separate routine to print fractions and DMS
			if (isproperfrac(&ST[iReg]))
				mkfstr(ST[iReg].num,ST[iReg].den,bstr);
			else if (isdms(&ST[iReg]))
				mkdmsstr(ST[iReg].real,bstr);
			else
				mkstr(ST[iReg].real,DECOUTWIDTH,bstr);
			memcpy(&cST[iReg][NDIGITS-lstrlen(bstr)],bstr,lstrlen(bstr));
		}
		else {
			mkstr(ST[iReg].real,CPLXOUTWIDTH,bstr);
			memcpy(&cST[iReg][CPLXOUTWIDTH-lstrlen(bstr)],bstr,lstrlen(bstr));
			mkstr(ST[iReg].imag,CPLXOUTWIDTH,bstr);
			memcpy(&cST[iReg][NDIGITS-lstrlen(bstr)-1],bstr,lstrlen(bstr));
			cST[iReg][NDIGITS-1]='i';
		}
	}
}

//-----------------------------------------------------------------------
// mkfstr converts fractional value to string bstr
//-----------------------------------------------------------------------

void mkfstr(long n,long d,char *bstr)
{
	long i;
	int sign;
	char *start = bstr;
	if (!bCleanFrac) {
		sign = mkifrac(&i,&n,&d);
		if (i) {
			ltoa(i,bstr,10);
			while (*++bstr);
			
			if (ThousandSep) {
				char *lastmant = bstr;
				// lastmant now points one right of last mantissa digit
				lastmant -= 3;
				while (lastmant>start) {
					memmove(lastmant+1,lastmant,bstr-lastmant);
					*lastmant = ',';
					bstr++;
					lastmant -= 3;
				}
			}
			
			*bstr++=' ';
		}
		else {
			if (sign)
				*bstr++ = '-';
		}
	}
	ltoa(n,bstr,10);
	while (*++bstr);
	*bstr++='/';
	ltoa(d,bstr,10);
}

//-----------------------------------------------------------------------
// mkdmsstr converts DMS value to string bstr
//-----------------------------------------------------------------------

void mkdmsstr(LD v,char *bstr)
{
	long d,m;
	LD s;
	int sign;
	int localFix;
	char *start = bstr;
	LD dmsdiv = 1.0L;
	LD dmsmul = 1.0L;
	int i;
	sign = v<0?-1:1;
	v = fabsl(v);
	d = (long)floorl(v); // TODO - check range?
	v = (v-d)*60.0L;
	m = (long)floorl(v);
	s = (v-m)*60.0L;
	// Inaccuracies here - round s to max DMSMAXFIX decimal places (dependent on Fix setting) and add carry if needed
	if (sFix==FLOAT)
		localFix = DMSMAXFIX;
	else
		localFix = sFix;
	for (i=0;i<localFix;i++) {
		dmsdiv *= 0.1L;
		dmsmul *= 10.0L;
	}
	s = dmsdiv * floorl(s*dmsmul+0.5L);
	if (s>=60.0L) {
		s -= 60.0L;
		m++;
		if (m>=60) {
			m -= 60;
			d++;
		}
	}
	if (sign==-1) *bstr++ = '-';
	ltoa(d,bstr,10);
	while (*++bstr);
	if (ThousandSep) {
		char *lastmant = bstr;
		// lastmant now points one right of last mantissa digit
		lastmant -= 3;
		while (lastmant>start) {
			memmove(lastmant+1,lastmant,bstr-lastmant);
			*lastmant = ',';
			bstr++;
			lastmant -= 3;
		}
	}
	*bstr++='';
	sprintf(bstr,"%02ld",m);
	while (*++bstr);
	*bstr++='\'';
	/*
	if (sFix==FLOAT)
		localFix = DMSMAXFIX;
	else
		localFix = sFix;
	*/
	sprintf(bstr,"%#0*.*Lf",localFix+3,localFix,s);
	// Taylor for FLOAT by removing trailing 0 and "."
	if (sFix==FLOAT) {
		BOOL point;
		int i = lstrlen(bstr) - 1;
		while ((i>1) && (bstr[i]=='0'||bstr[i]=='.')) {
			point=(bstr[i]=='.');
			lstrcpy(bstr+i,bstr+i+1);
			i--;
			if (point) break;
		}
	}
	while (*++bstr);
	*bstr++='"';
	*bstr=0;
}

//-----------------------------------------------------------------------
// mkstr converts value to string bstr at maximum maxsize characters
//-----------------------------------------------------------------------

// Define a small value to add to numbers before display

#define LDBL_EPSILONL        1.E-18L

// Turn off fast floating point here to avoid rounding errors.

#ifdef __BORLANDC__
#pragma option -ff-
#endif

void mkstr(LD value,int maxsize,char *bstr)
{
	BOOL zero,sign,UseExp,ForceExp,ForceEng;
	int i,Exp,EngExp,EngExtra,localFix,prec;
	LD Mant;
	char tmp[NDIGITS+1],etmp[5];
	
	if (sFix==FLOAT)
		localFix = DBL_DIG-1;  // DBL_DIG = 15
	else
		localFix = sFix;
	
	ForceExp = CURR_FIX==SCI;
	ForceEng = CURR_FIX==ENG;
	
	zero = (value==0);
	sign = (value<0);
	value = fabsl(value);
	
	if (zero) {
		Exp = 0;
		Mant = 0;
	}
	
	else {
		Exp=(int)floorl(log10l(value));		// Exp is true exponent of ten
		Mant=value*powl(10.0L,-Exp-1);		// Mant now 0.1 <= mant < 1
		// add a small value to Mant so that e.g. 2.445 is displayed as "2.45"
		// using two decimals. This is the only way I can make it work with
		// LD numbers...
		Mant+=LDBL_EPSILONL;
		// We some times have rounding errors here. A few numbers might
		// end up with the exponent one off, and the mantissa equal to 1.
		// So far I know no better solution than this:
		if (Mant>=1.0L) {
			Mant *= 0.1L;
			Exp++;
		}
	}
	
	EngExp = 3*(int)(floor((double)Exp/3));
	EngExtra = ForceEng*(Exp-EngExp);
	// Use Exponential notation if
	UseExp = (localFix+Exp-sign<0 /*&& Mant<0.5L*/) // number too small
		// (or too wide with "-"),
		// (removed -sign: doesn't seem right!)
		// Ahh... Put that one back again...
		// It failed on "-4e-14" in float.
		// added 0.5L test to display "0.5" as "1."
		// Removed it again. Need more thinking.
		|| (sign+1+Exp+1>maxsize) // too big to display,
		|| (Exp>=DBL_DIG)         // more than sign. digits,
		|| ForceExp || ForceEng;  // or forced display.
	
	prec = min(localFix+1+EngExtra,DBL_DIG);			// restrict prec. for accuracy
	
	if (UseExp) {
		prec = min(prec,maxsize-sign-1-5);				// reduce prec. for "[-]e+000"
	}
	else {
		prec += Exp;
		if (sFix==FLOAT) prec-=sign;
		prec = min(min(prec,maxsize-1-sign),DBL_DIG);	// allow for - and .
	}
	
	sprintf(tmp,"%*.*Lf",prec,prec,Mant);
	
	// fix possible rounding error if mant~<1 (Not elegant!!!)
	
	if ((Mant<1.0L)&&(tmp[0]=='1')) {
		Exp++;
		Mant*=0.1L;
		EngExp = 3*(int)(floor((double)Exp/3));
		EngExtra = ForceEng*(Exp-EngExp);
		
		UseExp = (localFix+Exp-sign<0)
			|| (sign+1+Exp+1>maxsize)
			|| (Exp>=DBL_DIG)
			|| ForceExp || ForceEng;
		
		prec = min(localFix+1+EngExtra,DBL_DIG);
		
		if (UseExp) {
			prec = min(prec,maxsize-sign-1-5);				// reduce prec. for "e+000"
		}
		else {
			prec += Exp;
			if (sFix==FLOAT) prec-=sign;
			prec = min(min(prec,maxsize-1-sign),DBL_DIG);	// allow for - and .
		}
		sprintf(tmp,"%*.*Lf",prec,prec,Mant);
		
	}
	memset(bstr,0,NDIGITS+1);
	
	if (sign) bstr[0] = '-';
	
	if ((!UseExp)&&(Exp<0)) {
		lstrcat(bstr,"0.");
		for (i=Exp+1;i<0;i++) lstrcat(bstr,"0");
		lstrcat(bstr,tmp+2);
	}
	else if (ForceEng) {
		strncat(bstr,tmp+2,1+EngExtra);
		lstrcat(bstr,".");
		lstrcat(bstr,tmp+3+EngExtra);
		lstrcat(bstr,"e");
		sprintf(etmp,"%+3.3d",EngExp);
		lstrcat(bstr,etmp);
	}
	else if (UseExp) {
		strncat(bstr,tmp+2,1);
		lstrcat(bstr,".");
		lstrcat(bstr,tmp+3);
		lstrcat(bstr,"e");
		sprintf(etmp,"%+3.3d",Exp);
		lstrcat(bstr,etmp);
	}
	else {
		strncat(bstr,tmp+2,Exp+1);
		lstrcat(bstr,".");
		lstrcat(bstr,tmp+3+Exp);
	}
	
	// Format is now fixed - taylor for FLOAT by removing trailing 0 and "."
	
	if (sFix==FLOAT) {
		BOOL point;
		i = lstrlen(bstr) - 1 - 5*UseExp;
		while ((i>0) && ((bstr[i]=='0')||(bstr[i]=='.'))) {
			point=(bstr[i]=='.');
			lstrcpy(bstr+i,bstr+i+1);
			i--;
			if (point) break;
		}
	}
	
	// Nov96: add thousands separator if decimal (but not complex)
	
	if (ThousandSep && CURR_RADIX == DECIMAL) {
		char *lastmant;
		if ((lastmant=strchr(bstr,'.'))==0 && (lastmant=strchr(bstr,'e'))==0)
			lastmant = bstr+lstrlen(bstr);
		// lastmant now points one right of last mantissa digit
		lastmant -= 3;
		while (lastmant>bstr && isdigit(*(lastmant-1))) {
			memmove(lastmant+1,lastmant,lstrlen(lastmant)+1);
			*lastmant = ',';
			lastmant -= 3;
		}
	}

	// Mar08: add space separator in fraction if decimal (but not complex)
	
	if (ThousandthSep && CURR_RADIX == DECIMAL) {
		char *dot;
		int i;
		if ((dot=strchr(bstr,'.'))!=0) {
			// have decimal separator
			while (dot<bstr+lstrlen(bstr)-4) {
				for (i=1;i<=4;i++)
					if (dot[i]=='e') 
						goto nomore;
				dot+=4;
				memmove(dot+1,dot,lstrlen(dot)+1);
				*dot=' ';
			}
nomore:;
		}
	}
}

