//------------------------------------------------------------------------
// File for project XCALC
//------------------------------------------------------------------------
// User draw button utility procedures:
//	DrawButton:     draws the 3d-looking button up or down
//	DrawButtonText: draws the text in specified colour, with underline
//  LButtonDown:    general mouse press routine
//  LButtonUp:      general mouse release routine
//  MouseMove:      general mouse move routine
//	DrawKey:		draws keyboard buttons with text, up or down
//	KeyLButtonDown:	(HWND hwnd, LPARAM lParam, BOOL *ButtonDown, BOOL *MouseDown)
//	KeyLButtonUp:	(HWND hwnd, LPARAM lParam, BOOL *ButtonDown, BOOL *MouseDown)
//	KeyMouseDownMove:	(HWND hwnd, LPARAM lParam, BOOL *ButtonDown)
//------------------------------------------------------------------------

#include "xcalc.h"

void DrawButton(HDC hdc, BOOL ButtonDown)
{
	if (!ButtonDown) {
		SelectObject(hdc,hBrightPen);
		MoveTo(hdc,Border*2,ButHeight-Border*2);
		LineTo(hdc,Border*2,Border*2);
		LineTo(hdc,ButHeight-Border*2,Border*2);
		
		SelectObject(hdc,hDarkPen);
		LineTo(hdc,ButHeight-Border*2,ButHeight-Border*2);
		LineTo(hdc,Border*2-1,ButHeight-Border*2);
	}
	else {
		SelectObject(hdc,hDarkPen);
		MoveTo(hdc,Border*2+1,ButHeight-Border*2-1);
		LineTo(hdc,Border*2+1,Border*2+1);
		LineTo(hdc,ButHeight-Border*2-1,Border*2+1);
		
		SelectObject(hdc,hBrightPen);
		LineTo(hdc,ButHeight-Border*2-1,ButHeight-Border*2-1);
		LineTo(hdc,Border*2,ButHeight-Border*2-1);
	}
}

void DrawButtonText(HDC hdc,RECT *rect, COLORREF colour,
					LPCSTR text,int underline,int len)
{
	MoveTo(hdc,ButHeight,(ButHeight-BigFontY)/2);
	SetTextAlign(hdc,TA_UPDATECP);
	SelectObject(hdc,hBigFont);
	SetBkColor(hdc,GetColour(RGBBACK));
	SetTextColor(hdc,colour);
	if (underline!=-1) {
		ExtTextOut(hdc,0,0,ETO_CLIPPED,rect,text,underline,NULL);
		SelectObject(hdc,hBigFont_);
		ExtTextOut(hdc,0,0,ETO_CLIPPED,rect,text+underline,1,NULL);
		SelectObject(hdc,hBigFont);
	}
	ExtTextOut(hdc,0,0,ETO_CLIPPED,rect,text+underline+1,len-underline-1,NULL);
}

void LButtonDown(HWND hwnd, LPARAM lParam, HWND *ButtonDownHWND, BOOL *MouseDown)
{
	RECT rect;
	POINTS pt = MAKEPOINTS(lParam);
	
	if ((pt.x>Border)&&(pt.x<ButHeight-Border)&&(pt.y>Border)&&(pt.y<ButHeight-Border)) {
		SetCapture(hwnd);
		*ButtonDownHWND = hwnd;
		*MouseDown = TRUE;
		SetRect(&rect,Border,Border,ButHeight-Border,ButHeight-Border);
		InvalidateRect(hwnd,&rect,0);
	}
}

BOOL LButtonUp(HWND hwnd, LPARAM lParam, HWND *ButtonDownHWND, BOOL *MouseDown)
{
	RECT rect;
	POINTS pt = MAKEPOINTS(lParam);
	
	ReleaseCapture();
	*ButtonDownHWND = NULL;
	*MouseDown = FALSE;
	
	if ((pt.x>Border)&&(pt.x<ButHeight-Border)&&(pt.y>Border)&&(pt.y<ButHeight-Border)) {
		SetRect(&rect,Border,Border,ButHeight-Border,ButHeight-Border);
		InvalidateRect(hwnd,&rect,0);
		return TRUE;
	}
	return FALSE;
}

void MouseMove(HWND hwnd, LPARAM lParam, HWND *ButtonDownHWND)
{
	RECT rect;
	POINTS pt = MAKEPOINTS(lParam);
	SetRect(&rect,Border,Border,ButHeight-Border,ButHeight-Border);
	
	if ((pt.x>Border)&&(pt.x<ButHeight-Border)&&(pt.y>Border)&&(pt.y<ButHeight-Border)) {
		if (hwnd != *ButtonDownHWND) {
			*ButtonDownHWND = hwnd;
			InvalidateRect(hwnd,&rect,0);
		}
	}
	else {
		if (hwnd == *ButtonDownHWND) {
			*ButtonDownHWND = NULL;
			InvalidateRect(hwnd,&rect,0);
		}
	}
}

void DrawKey(HDC hdc,RECT *rect, BOOL ButtonDown, COLORREF colour,
			 LPCSTR text,int symbol,int len)
{
	if (!ButtonDown) {
		SelectObject(hdc,hBrightPen);
		MoveTo(hdc,Border,ButHeight-Border);
		LineTo(hdc,Border,Border);
		LineTo(hdc,KeyWidth-Border,Border);
		
		SelectObject(hdc,hDarkPen);
		LineTo(hdc,KeyWidth-Border,ButHeight-Border);
		LineTo(hdc,Border-1,ButHeight-Border);
	}
	
	SetBkColor(hdc,GetColour(RGBBACK));
	SetTextColor(hdc,colour);

	if (symbol == -1) {
		MoveTo(hdc,1+KeyWidth/2+ButtonDown,(ButHeight-BigFontY)/2+ButtonDown); // 1+ looks better
		SetTextAlign(hdc,TA_CENTER|TA_UPDATECP);
		SelectObject(hdc,hBigFont);
		ExtTextOut(hdc,0,0,ETO_CLIPPED,rect,text,len,NULL);
	}
	else {
		SIZE siz;
		int width;
		SelectObject(hdc,hSymbolFont);
		GetTextExtentPoint32(hdc,text+symbol,1,&siz);
		width  = siz.cx;
		SelectObject(hdc,hBigFont);
		GetTextExtentPoint32(hdc,text,symbol,&siz);
		width += siz.cx;
		GetTextExtentPoint32(hdc,text+symbol+1,len-symbol-1,&siz);
		width += siz.cx;
		SetTextAlign(hdc,TA_UPDATECP);
		MoveTo(hdc,(KeyWidth-width)/2+ButtonDown,(ButHeight-BigFontY)/2+ButtonDown);
		ExtTextOut(hdc,0,0,ETO_CLIPPED,rect,text,symbol,NULL);
		SelectObject(hdc,hSymbolFont);
		ExtTextOut(hdc,0,0,ETO_CLIPPED,rect,text+symbol,1,NULL);
		SelectObject(hdc,hBigFont);
		ExtTextOut(hdc,0,0,ETO_CLIPPED,rect,text+symbol+1,len-symbol-1,NULL);
	}
}

void KeyLButtonDown(HWND hwnd, LPARAM lParam, HWND *ButtonDownHWND, BOOL *MouseDown)
{
	RECT rect;
	POINTS pt = MAKEPOINTS(lParam);
	
	if ((pt.x>Border)&&(pt.x<KeyWidth-Border)&&(pt.y>Border)&&(pt.y<ButHeight-Border)) {
		SetCapture(hwnd);
		*ButtonDownHWND = hwnd;
		*MouseDown = TRUE;
		SetRect(&rect,Border,Border,KeyWidth-Border,ButHeight-Border);
		InvalidateRect(hwnd,NULL,FALSE);
	}
}

BOOL KeyLButtonUp(HWND hwnd, LPARAM lParam, HWND *ButtonDownHWND, BOOL *MouseDown)
{
	POINTS pt = MAKEPOINTS(lParam);
	
	ReleaseCapture();
	*ButtonDownHWND = NULL;
	*MouseDown = FALSE;
	
	if ((pt.x>Border)&&(pt.x<KeyWidth-Border)&&(pt.y>Border)&&(pt.y<ButHeight-Border)) {
		InvalidateRect(hwnd,NULL,FALSE);
		return TRUE;
	}
	return FALSE;
}

void KeyMouseDownMove(HWND hwnd, LPARAM lParam, HWND *ButtonDownHWND)
{
	POINTS pt = MAKEPOINTS(lParam);
	
	if ((pt.x>Border)&&(pt.x<KeyWidth-Border)&&(pt.y>Border)&&(pt.y<ButHeight-Border)) {
		if (hwnd != *ButtonDownHWND) {
			*ButtonDownHWND = hwnd;
			InvalidateRect(hwnd,NULL,FALSE);
		}
	}
	else {
		if (hwnd == *ButtonDownHWND) {
			*ButtonDownHWND = NULL;
			InvalidateRect(hwnd,NULL,FALSE);
		}
	}
}
