//------------------------------------------------------------------------
// File for project XCALC
//------------------------------------------------------------------------
// Error display procedure
// Actions:
//	- on update, check global Error and modify display
//  - can also click on error window to remove message
//------------------------------------------------------------------------

#include "xcalc.h"

LRESULT CALLBACK ErrorProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	switch (wMsg) {
		
	case PM_UPDATE:		// is sent from main window to update display
		if (!Pane5 && MessagePopup) {
			ShowWindow(hError,Error?SW_SHOWNOACTIVATE:SW_HIDE);
		}
		InvalidateRect(hwnd,NULL,0);
		break;
		
	case WM_LBUTTONDOWN:
		PostMessage(GetParent(hwnd),WM_CHAR,' ',0);
		break;
		
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			RECT rect;
			SIZE sz;
			
			BeginPaint(hwnd,&ps);
			
			GetClientRect(hwnd,&rect);
			
			FillRect(ps.hdc,&rect,hBackBrush);
			
			// Draw text if error or message
			
			if (Error) {
				int x = Border;
				int y = Pane5?rect.bottom-Border:Border-2;
				SelectObject(ps.hdc,Pane5?hBigFont:hToolTipFont);
				SetBkColor(ps.hdc,GetColour(RGBBACK));
				SetTextColor(ps.hdc,GetColour(RGBMESSAGE));
				SetTextAlign(ps.hdc,Pane5?TA_BASELINE|TA_LEFT:TA_TOP|TA_LEFT);
				if (Error == __UNDONE) {
					char undotext[64] = "Undone ";
					lstrcat(undotext,UndoMsg[Undone]);
					GetTextExtentPoint32(ps.hdc,undotext,lstrlen(undotext),&sz);
					TextOut(ps.hdc,x,y,undotext,lstrlen(undotext));
				}
				else if (Error == __MESSAGE) {
					GetTextExtentPoint32(ps.hdc,Message,lstrlen(Message),&sz);
					TextOut(ps.hdc,x,y,Message,lstrlen(Message));
				}
				else {
					GetTextExtentPoint32(ps.hdc,ErrMsg[Error],lstrlen(ErrMsg[Error]),&sz);
					TextOut(ps.hdc,x,y,ErrMsg[Error],lstrlen(ErrMsg[Error]));
				}
				if (!Pane5 && MessagePopup) {
					SetWindowPos(hError,NULL,0,0,sz.cx+Border*2+1,sz.cy,SWP_NOMOVE|SWP_NOZORDER|SWP_NOACTIVATE);
					ShowWindow(hError,SW_SHOWNA);
				}
			}
			else {
				if (!Pane5) ShowWindow(hError,SW_HIDE);
			}
			
			EndPaint(hwnd,&ps);
			break;
		}
		
	default:
		return DefWindowProc(hwnd, wMsg, wParam, lParam);
	}
	return 0L;
}

