//------------------------------------------------------------------------
// File for project XCALC
//------------------------------------------------------------------------
// Fix Window procedure
// Actions:
//	- draw border at window edge
//	- update "LED's" to show status
//	- generate PM's on keyklicks
//------------------------------------------------------------------------

#include "xcalc.h"

const char aFixText[][2] = {
	"F","0","1","2","3","4","5","6","7","8","9"
};

// ToolTip added July 00

char *aFixDesc = "Number of fractional digits to display\n[Alt-0 - Alt-9; Alt-F or left/right arrows]";

LRESULT CALLBACK FixProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	static rgt, aDotXPos[11];
	RECT rect;
	
	// Relay to XCALC ToolTip window
	GetClientRect(hwnd,&rect);
	rect.left += Border;
	rect.top += Border;
	rect.right -= Border;
	rect.bottom -= Border;
	ToolTipEvent(hwnd,wMsg,wParam,lParam,&rect,aFixDesc);

	switch (wMsg) {
		
	case PM_SIZEWINDOW:
		{
			RECT rect;
			int i;
			GetClientRect(hwnd,&rect);
			rgt = rect.right;
			for (i=0;i<11;i++) {
				aDotXPos[i] = rgt - MulDiv(rgt,i+1,20);
			}
			break;
		}
		
	case PM_UPDATE:
		{
			RECT rect;
			SetRect(&rect,Border,Border,rgt-Border,ButHeight-Border);
			InvalidateRect(hwnd,&rect,0);
			break;
		}
		
	case WM_LBUTTONDOWN:
		{
			int x,i;
			if (IsBin) break;
			x = (int)LOWORD(lParam);
			for (i=0;i<11;i++) {
				if (abs(x-aDotXPos[i])<=rgt/40) {
					// We cannot send PM_SETFIX, since that cannot be undone
					PostMessage(GetParent(hwnd),WM_COMMAND,AS_FLOAT+i,0);
					break;
				}
			}
			break;
		}
		
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			RECT rect;
			int i;
			char cFix[] = "Fix: ";
			
			BeginPaint(hwnd,&ps);
			
			GetClientRect(hwnd,&rect);
			FillRect(ps.hdc,&rect,hBackBrush);
			MakeBorder(ps.hdc,&rect);  // makeborder will also update rect to point inside
			
			// Draw explaining text ("Fix: " or "Float") using BigFont:
			SelectObject(ps.hdc,hBigFont);
			
			if (sFix==FLOAT)
				lstrcpy(cFix,"Float");
			else
				cFix[4] = (char)(sFix+'0');
			
			SetBkColor(ps.hdc,GetColour(RGBBACK));
			SetTextColor(ps.hdc,GetColour(IsBin?RGBGRAYBUTTON:RGBBUTTON));
			
			ExtTextOut(ps.hdc,ButHeight,(ButHeight-BigFontY)/2,ETO_CLIPPED,
				&rect,cFix,lstrlen(cFix),NULL);
			
			// Draw text to show possibilities
			SelectObject(ps.hdc,hSmallFont);
			SetTextAlign(ps.hdc,TA_CENTER);
			for (i=0;i<11;i++) {
				ExtTextOut(ps.hdc,aDotXPos[i],(ButHeight-SmallFontY)/2,ETO_CLIPPED,
					&rect,aFixText[i],1,NULL);
			}
			
			// Draw selected item RED
			SetTextColor(ps.hdc,GetColour(RGBSELECTED));
			ExtTextOut(ps.hdc,aDotXPos[sFix+1],(ButHeight-SmallFontY)/2,ETO_CLIPPED,
				&rect,aFixText[sFix+1],1,NULL);
			
			EndPaint(hwnd,&ps);
			break;
		}
		
	default:
		return DefWindowProc(hwnd, wMsg, wParam, lParam);
	}
	return 0L;
}

