//------------------------------------------------------------------------
// File for project XCALC
//------------------------------------------------------------------------
// Memory Window procedure
// Actions:
//	- update digits to show memory status
//	- takes complete care of move/resize of popup window
//	- generate PM's on keyklicks to notify main program
//------------------------------------------------------------------------

#include "xcalc.h"

// global variables in this file

static const char	aMemText[] = "0123456789";
static const char	aMemDesc1[] = "Cell [n]: ";
#define				MEMDESC_DIGITPOS 6
static const char	aMemDesc2[] = "\n(Left click to store, Left doubleclick to clear, right click to recall)";
static const char	aPopupMemDesc[] = "Left doubleclick to store, right doubleclick to recall";
int					PopupBorder;
int					MemWidth,MemHeight;
HFONT				hPopupFont = NULL;

//--------Functions----------------------------------------------------------

//------------------------------------------------------------------------
// Memory Indicator Window procedure
// Actions:
//  - shows small memory cell indicators
//  - processes mouse clicks for save/clear/restore 
//  - relays to tooltip procedure
//------------------------------------------------------------------------

LRESULT CALLBACK MemoryProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	static int DigitXPos[MEMSIZ];
	static int rgt;
	static int cellwidth;
	static int digwidth;
	static char DisplayText[TOOLTIP_MAXLENGTH];
	static char DisplayNumber[NDIGITS+20];
	RECT rect;
	int x,i;
	
	// Relay to XCALC ToolTip window
	x = (int)LOWORD(lParam); // may be very incorrect for non-mouse messages...
	SetRect(&rect,0,0,0,0);
	*DisplayText = 0;
	for (i=0;i<MEMSIZ;i++) {
		if ((x>DigitXPos[i])&&(x<DigitXPos[i]+digwidth)) {
			lstrcpy(DisplayText,aMemDesc1);
			DisplayText[MEMDESC_DIGITPOS] = (char)(i + '0');
			FormatMemoryNumber(&Memory[i],DisplayNumber);
			lstrcat(DisplayText,DisplayNumber);
			lstrcat(DisplayText,aMemDesc2);
			SetRect(&rect,DigitXPos[i],0,DigitXPos[i]+digwidth,ButHeight);
		}
	}
	ToolTipEvent(hwnd,wMsg,wParam,lParam,&rect,DisplayText);
	
	switch (wMsg) {
		
	case PM_SIZEWINDOW:
		{
			GetClientRect(hwnd,&rect);
			rgt = rect.right;
			cellwidth = (rgt-Border)/MEMSIZ;
			digwidth = cellwidth - Border;
			for (i=0;i<MEMSIZ;i++) {
				DigitXPos[i] = Border + i*cellwidth;
			}
			break;
		}
		
	case WM_LBUTTONDOWN:
		{
			x = (int)LOWORD(lParam);
			for (i=0;i<MEMSIZ;i++) {
				if ((x>DigitXPos[i])&&(x<DigitXPos[i]+digwidth)) {
					SendMessage(GetParent(hwnd),WM_COMMAND,AS_STORE0+i,0);
				}
			}
			break;
		}
		
	case WM_LBUTTONDBLCLK:
		{
			x = (int)LOWORD(lParam);
			for (i=0;i<MEMSIZ;i++) {
				if ((x>DigitXPos[i])&&(x<DigitXPos[i]+digwidth)) {
					SendMessage(GetParent(hwnd),WM_COMMAND,AS_CLEAR0+i,0);
				}
			}
			break;
		}
		
	case WM_RBUTTONDOWN:
		{
			x = (int)LOWORD(lParam);
			for (i=0;i<MEMSIZ;i++) {
				if ((x>DigitXPos[i])&&(x<DigitXPos[i]+digwidth)) {
					SendMessage(GetParent(hwnd),WM_COMMAND,AS_RECALL0+i,0);
				}
			}
			break;
		}
		
	case PM_UPDATE:
		{
			WORD wReg = (WORD)wParam;
			SetRect(&rect,DigitXPos[wReg],0,DigitXPos[wReg]+digwidth,ButHeight);
			InvalidateRect(hwnd,&rect,FALSE);
			break;
		}
		
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			RECT rect;
			int i;
			
			BeginPaint(hwnd,&ps);
			
			GetClientRect(hwnd,&rect);
			FillRect(ps.hdc,&rect,hBackBrush);
			//Rectangle(ps.hdc,rect.left,rect.top,rect.right,rect.bottom);
			
			// Draw text to show occupied memory cells
			SelectObject(ps.hdc,hSmallFont);
			SetBkColor(ps.hdc,GetColour(RGBBACK));
			
			for (i=0;i<MEMSIZ;i++) {
				BOOL Empty = 
					(Memory[i].ct==ctREAL && Memory[i].real==0 && Memory[i].imag==0) ||
					(Memory[i].ct==ctINTEGER && Memory[i].i==0);
				SetTextColor(ps.hdc,GetColour(Empty?RGBBUTTON:RGBSELECTED));
				ExtTextOut(ps.hdc,DigitXPos[i]+(Border>>1)|1,(ButHeight-SmallFontY)/2,ETO_CLIPPED,&rect,aMemText+i,1,NULL);
			}
			
			EndPaint(hwnd,&ps);
			break;
		}
		
	default:
		return DefWindowProc(hwnd, wMsg, wParam, lParam);
	}
	return 0L;
}

//------------------------------------------------------------------------
// Memory Popup Window procedure
// Actions:
//  - creates font for and sizes popup
//  - parent for memory display windows 
//------------------------------------------------------------------------

LRESULT CALLBACK MemoryPopupProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	static POINT origpt;
	static RECT  origrect;
	static int   origPopupSize;
	static int   origPopupX, origPopupRight;
	static BOOL	 InRgtCorner,InLftCorner;
	static BOOL  Sizing=FALSE,Moving=FALSE;
	static int   delayMove;
	static HRGN	 hRightRgn = NULL,hLeftRgn = NULL;
	static int	 CrnrX,CrnrY;
	HDC hdc;
	HFONT hOldFont;
	
	switch (wMsg) {
		
	case WM_DESTROY:
		DeleteObject(hRightRgn);
		DeleteObject(hLeftRgn);
		DeleteObject(hPopupFont);
		break;
		
	case WM_CREATE:
    case PM_SIZEWINDOW:
		{
			char PopupSizer[NDIGITS+4];     // To accomodate "n: ddddd...."
			POINT Corner[6];
			SIZE siz;
			int n;
			int oldPopupDX = PopupDX;
			int oldPopupDY = PopupDY;
			
			if (hPopupFont) DeleteObject(hPopupFont);
			hPopupFont = CreateFont(-MulDiv(14,PopupSize,100),0,0,0,0,0,0,0,
				ANSI_CHARSET,
				OUT_DEFAULT_PRECIS,
				CLIP_DEFAULT_PRECIS,
				DEFAULT_QUALITY,
				DEFAULT_PITCH | FF_SWISS,
				"Arial");
			
			hdc = GetDC(hwnd);
			hOldFont = (HFONT)SelectObject(hdc,hPopupFont);
			
			// Make some nice crude guesswork on border size:
			
			PopupBorder = 4;
			if (PopupSize<100) PopupBorder = 3;
			if (PopupSize<50) PopupBorder = 2;
			//if (PopupSize<25) PopupBorder = 1;
			
			// Here the popup font is selected - now find size of popup window
			// which is filled with two columns of 5 numbers (max NDIGITS wide)
			// or one column of 10 numbers. Add borders on all sides.
			
			lstrcpy(PopupSizer,"0: ");
			for (n=0;n<NDIGITS;n++) lstrcat(PopupSizer,"0");
			GetTextExtentPoint32(hdc,PopupSizer,NDIGITS+3,&siz);
			
			MemWidth  = siz.cx + 2 + 2*PopupBorder; // +3 since "n:"
			MemHeight = siz.cy + 2*PopupBorder; // at pos 3, and digits left 2 positions too.
			
			PopupDX = MemCols*(MemWidth+1)+PopupBorder*(2+(MemCols==2)) +1;
			PopupDY = MemRows*(MemHeight+1)+PopupBorder*2 +1;

			// Do not process further if no change
			if (PopupDX == oldPopupDX && PopupDY == oldPopupDY)
				return 0;

			if (InLftCorner) {
				PopupX = origPopupRight-PopupDX;
				// Track Mini settings
				if (Mini) {
					m_PopupX=PopupX;
				}
				else {
					n_PopupX=PopupX;
				}
			}

			SetWindowPos(hwnd,NULL,
				PopupX,PopupY,
				PopupDX,PopupDY,
				SWP_NOZORDER|SWP_NOACTIVATE);
			
			CrnrX = GetSystemMetrics(SM_CXCURSOR);
			CrnrY = GetSystemMetrics(SM_CYCURSOR);
			
			Corner[0].x = PopupDX-2;				Corner[0].y = PopupDY-2;
			Corner[1].x = PopupDX-CrnrX-2;			Corner[1].y = PopupDY-2;
			Corner[2].x = PopupDX-CrnrX-2;			Corner[2].y = PopupDY-PopupBorder-2;
			Corner[3].x = PopupDX-PopupBorder-2;	Corner[3].y = PopupDY-PopupBorder-2;
			Corner[4].x = PopupDX-PopupBorder-2;	Corner[4].y = PopupDY-CrnrY-2;
			Corner[5].x = PopupDX-2;				Corner[5].y = PopupDY-CrnrY-2;
			if (hRightRgn) DeleteObject(hRightRgn);
			hRightRgn = CreatePolygonRgn(Corner,6,WINDING);
			
			Corner[0].x = 0;						Corner[0].y = PopupDY-2;
			Corner[1].x = CrnrX;					Corner[1].y = PopupDY-2;
			Corner[2].x = CrnrX;					Corner[2].y = PopupDY-PopupBorder-2;
			Corner[3].x = PopupBorder;				Corner[3].y = PopupDY-PopupBorder-2;
			Corner[4].x = PopupBorder;				Corner[4].y = PopupDY-CrnrY-2;
			Corner[5].x = 0;						Corner[5].y = PopupDY-CrnrY-2;
			if (hLeftRgn) DeleteObject(hLeftRgn);
			hLeftRgn = CreatePolygonRgn(Corner,6,WINDING);
			
			SelectObject(hdc,hOldFont);
			ReleaseDC(hwnd,hdc);
			return 1;
		}
		
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			RECT rect;
			
			BeginPaint(hwnd,&ps);
			
			GetClientRect(hwnd,&rect);
			FillRect(ps.hdc,&rect,hBackBrush);
			
			SelectObject(ps.hdc,hDarkPen);
			MoveTo(ps.hdc,PopupDX-CrnrX-2,	PopupDY-PopupBorder-2); // right bottom size corner
			LineTo(ps.hdc,PopupDX-CrnrX-2,	PopupDY-2);
			MoveTo(ps.hdc,PopupDX-PopupBorder-2, PopupDY-CrnrY-2);
			LineTo(ps.hdc,PopupDX-2,		PopupDY-CrnrY-2);
			
			MoveTo(ps.hdc,CrnrX,  		PopupDY-PopupBorder-2);      // left bottom size corner
			LineTo(ps.hdc,CrnrX,  		PopupDY-2);
			MoveTo(ps.hdc,0,      		PopupDY-CrnrY-2);
			LineTo(ps.hdc,PopupBorder,	PopupDY-CrnrY-2);
			
			EndPaint(hwnd,&ps);
			if (GetActiveWindow()==hwnd) SetFocus(GetParent(hwnd));
			break;
		}
		
	case WM_LBUTTONDOWN:
		{
			POINTS mousepts = MAKEPOINTS(lParam);
			origpt.x = mousepts.x;
			origpt.y = mousepts.y;

			InRgtCorner = PtInRegion(hRightRgn,origpt.x,origpt.y);
			InLftCorner = PtInRegion(hLeftRgn,origpt.x,origpt.y);
			ClientToScreen(hwnd,&origpt);

			GetWindowRect(hwnd,&origrect);
			SetCapture(hwnd);

			if (InRgtCorner || InLftCorner) {
				origPopupSize = PopupSize;
				origPopupX = PopupX;
				origPopupRight = origrect.right;
				Sizing = TRUE;			// checked in MouseMove
			}
			else {
				delayMove = 2;			// do not start moving after 2 mouse move messages
				Moving = TRUE;			// checked in MouseMove
			}
			break;
		}
		
	case WM_MOUSEMOVE:
		{
			POINTS mousepts = MAKEPOINTS(lParam);
			POINT mousept;
			RECT rect;
			int xoff,yoff;
			mousept.x = mousepts.x;
			mousept.y = mousepts.y;
			
			if (Moving) {
				ClientToScreen(hwnd,&mousept);
				rect = origrect;
				if (delayMove) {			// this will avoid a blink on DBLCLK
					if (delayMove--) break;
				}
				xoff = mousept.x - origpt.x;
				yoff = mousept.y - origpt.y;
				OffsetRect(&rect,xoff,yoff);
				MoveWindow(hwnd,rect.left,rect.top,rect.right-rect.left,rect.bottom-rect.top,1);
				PopupX = rect.left;
				PopupY = rect.top;
				// Track Mini settings
				if (Mini) {
					m_PopupX=PopupX;
					m_PopupY=PopupY;
				}
				else {
					n_PopupX=PopupX;
					n_PopupY=PopupY;
				}
			}
			else if (Sizing) {
				int i;
				int DX0 = origrect.right-origrect.left;
				ClientToScreen(hwnd,&mousept);
				xoff = mousept.x - origpt.x;
				if (InRgtCorner) {
					int DX = DX0+xoff;
					DX = max(DX,GetSystemMetrics(SM_CXMIN));
					PopupSize = MulDiv(origPopupSize,DX,DX0);
				}
				else {
					int DX = DX0-xoff;
					DX = max(DX,GetSystemMetrics(SM_CXMIN));
					PopupSize = MulDiv(origPopupSize,DX,DX0);
					PopupX = origPopupX+xoff;
				}
				if (SendMessage(hwnd,PM_SIZEWINDOW,0,0)) { // nonzero if any changes
					for (i=0;i<MEMSIZ;i++)
						SendMessage(hMemoryDisplay[i],PM_SIZEWINDOW,0,0);
					// Track Mini settings
					if (Mini) {
						m_PopupSize=PopupSize;
						m_PopupX=PopupX;
					}
					else {
						n_PopupSize=PopupSize;
						n_PopupX=PopupX;
					}
					InvalidateRect(hwnd,NULL,FALSE);
				}
			}
			else { // neither move nor size
				if (PtInRegion(hRightRgn,mousept.x,mousept.y))
					SetCursor(LoadCursor(NULL,IDC_SIZENWSE));
				else if (PtInRegion(hLeftRgn,mousept.x,mousept.y))
					SetCursor(LoadCursor(NULL,IDC_SIZENESW));
				else
					SetCursor(LoadCursor(NULL,IDC_ARROW));
			}
			break;
		}
		
	case WM_LBUTTONUP:
		{
			if (Moving || Sizing) {
				Moving = FALSE;
				Sizing = FALSE;
				ReleaseCapture();
			}
			SetFocus(GetParent(hwnd));
			break;
		}
		
	case WM_RBUTTONUP:
    case WM_RBUTTONDBLCLK:
		{
			SetFocus(GetParent(hwnd));
			break;
		}
		
	case WM_SHOWWINDOW:
		{
			if (ShowMemoryPopup) return DefWindowProc(hwnd, wMsg, wParam, lParam);
			break;
		}
		
    default:
		return DefWindowProc(hwnd, wMsg, wParam, lParam);
	}
	return 0;
}

//------------------------------------------------------------------------
// Popup Memory Display procedure (subwindows in memory popup)
// Actions:
//	- update number to show memory content
//  - allow move by click/drag 
//	- generate PM's on dblclicks to notify main program
//------------------------------------------------------------------------

#ifdef __BORLANDC__
#pragma warn -8084
#endif

LRESULT CALLBACK MemoryDisplayProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	int iWindow;
	HFONT hOldFont;
	RECT rect;
	
	// find which window this is, to use the right memory register
	iWindow = GetWindowLong(hwnd,GWL_ID);
	
	// Relay to XCALC ToolTip window
	GetClientRect(hwnd,&rect);
	ToolTipEvent(hwnd,wMsg,wParam,lParam,&rect,aPopupMemDesc);
	switch (wMsg) {
		
	// set correct placement
	case PM_SIZEWINDOW:
		SetWindowPos(hwnd,NULL,
			(iWindow/MemRows)*(MemWidth+1)+PopupBorder*(1+iWindow/MemRows),
			(iWindow%MemRows)*(MemHeight+1)+PopupBorder,
			MemWidth,
			MemHeight,
			SWP_NOZORDER|SWP_NOACTIVATE);
		break;
		
	case PM_UPDATE:
		InvalidateRect(hwnd,NULL,FALSE);
		break;
		
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			SIZE siz;
			char DisplayText[NDIGITS+20];
			
			BeginPaint(hwnd,&ps);
			
			GetClientRect(hwnd,&rect);
			FillRect(ps.hdc,&rect,hBackBrush);
			SetTextColor(ps.hdc,GetColour(RGBDIGITS));
			SetBkColor(ps.hdc,GetColour(RGBBACK));
			hOldFont = (HFONT)SelectObject(ps.hdc,hPopupFont);
			
			lstrcpy(DisplayText,"n: ");
			DisplayText[0] = (char)(iWindow+'0');
			TextOut(ps.hdc,3,PopupBorder,DisplayText,lstrlen(DisplayText));

			// Format displayed number
			FormatMemoryNumber(&Memory[iWindow],DisplayText);

			GetTextExtentPoint32(ps.hdc,DisplayText,lstrlen(DisplayText),&siz);
			TextOut(ps.hdc,rect.right-siz.cx-3,PopupBorder,DisplayText,lstrlen(DisplayText));
			
			// Border draw move after textout nov96; small windows would
			// overwrite border. (Probably not any more. Dec96)
			
			SelectObject(ps.hdc,hBrightPen);
			MoveTo(ps.hdc,rect.left,rect.bottom-1);
			LineTo(ps.hdc,rect.right-1,rect.bottom-1);
			LineTo(ps.hdc,rect.right-1,rect.top);
			SelectObject(ps.hdc,hDarkPen);
			LineTo(ps.hdc,rect.left,rect.top);
			LineTo(ps.hdc,rect.left,rect.bottom-1);
			SelectObject(ps.hdc,hOldFont);
			
			EndPaint(hwnd,&ps);
			break;
		}
		
	case WM_LBUTTONDOWN:				// transfer mouseclick to
		{								// popup window for move operation
			WORD xPos = LOWORD(lParam);
			WORD yPos = HIWORD(lParam);
			RECT ThisRect,ParentRect;
			GetWindowRect(hwnd,&ThisRect);
			GetWindowRect(GetParent(hwnd),&ParentRect);
			PostMessage(GetParent(hwnd),wMsg,wParam,
				MAKELONG(xPos-ParentRect.left+ThisRect.left,
				yPos-ParentRect.top+ThisRect.top));
			break;
		}
		
	case WM_LBUTTONDBLCLK:
		PostMessage(hwndXCALC,PM_STORE,iWindow,0);
		SetFocus(hwndXCALC);
		break;
		
	case WM_RBUTTONUP:
		SetFocus(hwndXCALC);
		break;
		
	case WM_RBUTTONDBLCLK:
		PostMessage(hwndXCALC,PM_RECALL,iWindow,0);
		SetFocus(hwndXCALC);
		break;
		
	default:
		return DefWindowProc(hwnd, wMsg, wParam, lParam);
	}
	return 0L;
}

void FormatMemoryNumber(Register *mem, char *text)
{
	// select decimal, fraction, DMS, complex or binary number to display
	
	if (mem->ct==ctINTEGER) {
		// hex, octal, binary content
		ltoa(mem->i,text,wRadix);
		strupr(text);
		if (wRadix==2) lstrcat(text," B");
		else if (wRadix==8) lstrcat(text," Octal");
		else if (wRadix==16) lstrcat(text," Hex");
	}
	else if (mem->ct==ctFRACTION) { 
		// fraction content
		int i = mem->num/mem->den;
		if (bCleanFrac || i==0)
			sprintf(text,"%d/%d",mem->num,mem->den);
		else
		{
			int n = abs(mem->num) % mem->den;
			sprintf(text,"%d %d/%d",i,n,mem->den);
		}
	}
	else if (mem->ct==ctDMS) {
		LD v = mem->real;
		int sign = v<0?-1:1;
		long d,m;
		LD s;
		BOOL point;
		int localFix;
		LD dmsdiv = 1.0L;
		LD dmsmul = 1.0L;
		int i;
		v = fabsl(v);
		d = (long)floorl(v); // TODO - check range?
		v = (v-d)*60.0L;
		m = (long)floorl(v);
		s = (v-m)*60.0L;
		// Inaccuracies here - round s to DMSMAXFIX decimal places and add carry if needed
		localFix = DMSMAXFIX;
		for (i=0;i<localFix;i++) {
			dmsdiv *= 0.1L;
			dmsmul *= 10.0L;
		}
		s = dmsdiv * floorl(s*dmsmul+0.5L);
		if (s>=60.0L) {
			s -= 60.0L;
			m++;
			if (m>=60) {
				m -= 60;
				d++;
			}
		}
		if (sign==-1) *text++ = '-';
		ltoa(d,text,10);
		while (*++text);
		*text++='';
		sprintf(text,"%02ld",m);
		while (*++text);
		*text++='\'';
		sprintf(text,"%#0*.*Lf",localFix+3,localFix,s);
		// Remove trailing 0 and "."
		i = lstrlen(text) - 1;
		while ((i>1) && (text[i]=='0'||text[i]=='.')) {
			point=(text[i]=='.');
			lstrcpy(text+i,text+i+1);
			i--;
			if (point) break;
		}
		while (*++text);
		*text++='"';
		*text=0;
	}
	else if (mem->imag == 0) {
		// decimal content
		// July 00: new format
		sprintf(text,"%.14Lg",mem->real);
	}
	else {
		// complex content
		// Correction nov96: 10.10 in print - otherwise too long strings
		// might result.
		// June 97: only %+.10 for imaginary. Looks better.
		// July 00: new format again
		sprintf(text,"%.10Lg%+.10Lgi",mem->real,mem->imag);
	}
}

