//------------------------------------------------------------------------
// File for project XCALC
//------------------------------------------------------------------------
// User draw Mode button procedure
// Used for both D/C/H/O/B modes, F/S/E and D/R modes.
// Actions:
//	- draw border at window edge
//	- draw LED light
//	- fixed text
//  - post accelerator message to parent on button release
//------------------------------------------------------------------------

#include "xcalc.h"

// Sep06: useful 2nd row for bin modes as well (index [0][*])

const unsigned AccelCode[2][NMODES] = {
	{
	AS_DECIMAL,AS_COMPLEX,AS_HEX,AS_OCTAL,AS_BINARY,
	AS_32BIT,AS_16BIT,AS_8BIT,AS_DEG,AS_RAD
	},
	{
	AS_DECIMAL,AS_COMPLEX,AS_HEX,AS_OCTAL,AS_BINARY,
	AS_FIX,AS_SCI,AS_ENG,AS_DEG,AS_RAD
	},
};

const char * const aModeName[2][NMODES] = {
	{
	"Decimal","Complex","Hex","Octal","Binary",
	"32 bit","16 bit","8 bit","Degree","Radian"
	},
	{
	"Decimal","Complex","Hex","Octal","Binary",
	"Fix","Sci","Eng","Degree","Radian"
	},
};


const int aModeUnder[2][NMODES] = {      // specifies which letter is underlined
	{
	0,0,0,0,0,
	0,0,0,0,0
	},
	{
	0,0,0,0,0,
	1,0,0,2,0
	},
};

// ToolTip added July 00

char *aModeDesc[2][NMODES] = {
	{
	"Decimal mode [Alt-D]",
	"Complex mode [Alt-C]",
	"Hex mode [Alt-H]",
	"Octal mode [Alt-O]",
	"Binary mode [Alt-B]",
	"32 bit words [Alt-3 or Alt-2]",
	"16 bit words [Alt-1 or Alt-6]",
	"8 bit words [Alt-8]",
	"Degree mode [Alt-G]",
	"Radian mode [Alt-R]",
	},
	{
	"Decimal mode [Alt-D]",
	"Complex mode [Alt-C]",
	"Hex mode [Alt-H]",
	"Octal mode [Alt-O]",
	"Binary mode [Alt-B]",
	"Normal fix/float [Alt-I]",
	"Scientific [Alt-S]",
	"Engineering [Alt-E]",
	"Degree mode [Alt-G]",
	"Radian mode [Alt-R]",
	}
};

LRESULT CALLBACK ModeProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	int iWindow,enabled,active;
	int rowsel = IsBin ? 0 : 1;
	static HWND ButtonDownHWND = NULL; // the onscreen button
	static BOOL MouseDown = FALSE;	// TRUE when mousebutton down
	RECT rect;

	// find which window this is
	iWindow = GetWindowLong(hwnd,GWL_ID);
	enabled = GetWindowWord(hwnd,0); // window enabled
	active  = GetWindowWord(hwnd,2); // window active (red dot on)

	// Relay to XCALC ToolTip window
	GetClientRect(hwnd,&rect);
	rect.left += Border;
	rect.top += Border;
	rect.right -= Border;
	rect.bottom -= Border;
	ToolTipEvent(hwnd,wMsg,wParam,lParam,&rect,aModeDesc[rowsel][iWindow]);

	switch (wMsg) {

		case PM_UPDATE:		// is sent from main window to update display
		{
			RECT rect;
			SetRect(&rect,Border,Border,ButHeight-Border,ButHeight-Border);
			InvalidateRect(hwnd,&rect,0);
			break;
		}

		case WM_LBUTTONDOWN:
    	if (!enabled) break;
			LButtonDown(hwnd,lParam,&ButtonDownHWND,&MouseDown);
			break;

		case WM_LBUTTONUP:
			if (!enabled) break;
			if (LButtonUp(hwnd,lParam,&ButtonDownHWND,&MouseDown))
				PostMessage(GetParent(hwnd),WM_COMMAND,AccelCode[rowsel][iWindow],0);
			break;

		case WM_MOUSEMOVE:
			if (!enabled) break;
			if (!MouseDown) break;
			MouseMove(hwnd,lParam,&ButtonDownHWND);
			break;

		case WM_PAINT:
		{
			PAINTSTRUCT ps;
			RECT rect;
			int i;

			BeginPaint(hwnd,&ps);

			GetClientRect(hwnd,&rect);
			FillRect(ps.hdc,&rect,hBackBrush);
			MakeBorder(ps.hdc,&rect);

			//3d look around red dot
			DrawButton(ps.hdc,ButtonDownHWND==hwnd);

			// draw red rectangle if enabled && active
			if (enabled && active) {
				RECT redrect;
				int reduce = ButtonDownHWND==hwnd? 1 : 0;
				SetRect(&redrect,
							Border*2+2+reduce,
							Border*2+2+reduce,
							ButHeight-Border*2-1-reduce,
							ButHeight-Border*2-1-reduce);
				if (redrect.bottom>redrect.top && redrect.right>redrect.left)
					FillRect(ps.hdc,&redrect,hSelectedBrush);
			}

			// Draw fixed text

			i = lstrlen(aModeName[rowsel][iWindow]);
			DrawButtonText(ps.hdc,&rect,GetColour(enabled?RGBBUTTON:RGBGRAYBUTTON),
								aModeName[rowsel][iWindow],aModeUnder[rowsel][iWindow],i);

			EndPaint(hwnd,&ps);

			break;
		}

		default:
			return DefWindowProc(hwnd, wMsg, wParam, lParam);
	}
	return 0L;
}
