//------------------------------------------------------------------------
// File for project XCALC
//-----------------------------------------------------------------------
// Memory Register save/restore routines.
// Now also saves stack and conversion settings. Mar97
// Utility routine to make path from dir/filename added Jan99
//------------------------------------------------------------------------

#include "xcalc.h"

const char MagicString[] = "XCALC save state file 2.0\x0d\x0a\x1a";

//-----------------------------------------------------------------------
// ReadRegisters restores memory from file
//-----------------------------------------------------------------------
void ReadRegisters(const char *const SaveFile)
{
	FILE *f;
	char CheckString[sizeof(MagicString)];
	BOOL readOK = FALSE;
	
	if ((f=fopen(SaveFile,"rb"))==NULL) return;
	// Check file integrity
	fseek(f,0,SEEK_END);
	if (ftell(f)==(long)(sizeof(MagicString)
				+sizeof(Register)*(MEMSIZ+MAXNUMREGS)
				+(3+2*NCONVERSIONS)*sizeof(short))) {
		fseek(f,0,SEEK_SET);
		fread(CheckString,sizeof(MagicString),1,f);
		if (!memcmp(MagicString,CheckString,sizeof(CheckString))) {
			fread(Memory,sizeof(Register),MEMSIZ,f);
			fread(ST,sizeof(Register),MAXNUMREGS,f);
			fread(&ConvType,sizeof(short),1,f);
			fread(&From,sizeof(short),1,f);
			fread(&To,sizeof(short),1,f);
			fread(SelectFrom,sizeof(short),NCONVERSIONS,f);
			fread(SelectTo,sizeof(short),NCONVERSIONS,f);
			readOK = TRUE;
		}
	}
	
	if (!readOK) {
		MessageBeep(MB_ICONHAND);
		MessageBox(NULL,
			"Registers were saved by a different version of XCALC. "
			"They will not be restored this time.",
			"XCALC warning",MB_OK);
	}
	
	fclose(f);
}

//-----------------------------------------------------------------------
// SaveRegisters saves memory in file
//-----------------------------------------------------------------------
void SaveRegisters(const char *const SaveFile)
{
	FILE *f;
	
	if ((f=fopen(SaveFile,"wb"))==NULL) return;
	fwrite(MagicString,sizeof(MagicString),1,f);
	fwrite(Memory,sizeof(Register),MEMSIZ,f);
	fwrite(ST,sizeof(Register),MAXNUMREGS,f);
	fwrite(&ConvType,sizeof(short),1,f);
	fwrite(&From,sizeof(short),1,f);
	fwrite(&To,sizeof(short),1,f);
	fwrite(SelectFrom,sizeof(short),NCONVERSIONS,f);
	fwrite(SelectTo,sizeof(short),NCONVERSIONS,f);
	fclose(f);
}

//-----------------------------------------------------------------------
// MakeFullPath to concatenate dir/filename
//-----------------------------------------------------------------------
void MakeFullPath(char *path, const char *file, const char *dir)
{
	lstrcpy(path,dir);
	lstrcat(path,file);
}

//-----------------------------------------------------------------------
// chext - in-place extension change (no size checks)
//-----------------------------------------------------------------------
char *chext(char *path, const char *newext)
{
	char *pdot = strrchr(path,'.');
	char *pbsl = strrchr(path,'\\');
	if (pdot>pbsl) {
		lstrcpy(pdot,newext);
	} else {
		lstrcat(path,newext);
	}
	return path;
}
