/*------------------------------------------------------------------------
XCALC.C
RPN complex/hex/oct/bin calculator for Windows 32 bit

	Ver 2.8.5
  
	March 2008
	Bernt Ribbum
	
	Version history:
	  
	V1.0beta	- September 1992; Windows version beta release.
	V1.0		- September 1992; a few bugs corrected (thanks to the beta testers).
	V1.0b		- October 1992; added Borland Dialogue Boxes (needs BWCC.DLL).
	V1.01a/
	V1.01b		- November 1992; added Clipboard functions.
	V1.02b		- November 1992; separate VGA/1024x768 Borland dialogue boxes
	V1.1		- December 1992; keep in front, display only some elements on screen,
				hide/show title bar/system menu, config on system menu,
				memory popup window, cleaned up font handling.
	V1.2		- June 1993; removed silly bug that made 500! crash.
				Added keyboard onscreen. Grayed out inaccessible buttons
				based on mode. Cleaned up source quite a bit.
				New functions % (percent) and % (remainder).
	V1.2 r.2	- April 94: inserted possible decimal COMMA (,) inst.o. (.),
				new keyboard mappings, new keyboard look,  better mouse resize.
	V1.3		- Screen colours.
	V2.0		- Feb 97: Added optional thousands separator for decimal display.
				Visible feedback when selecting display bitmaps. Colour editing.
				Display also from system fonts. Finally on WIN95 taskbar
				when caption is off. Some nice touches to dialogue boxes.
				Added space to the display, to make sure the caption
				circle does not conceal a digit. Sounds now play through
				standard windows device. Separate resizing of main and popup
				window. Long double numbers; better maths.
				Different normal/minimum settings.
				Added some new maths functions - now need medium model.
	V2.01		- Feb 97: Added conversions. Finally proper keyboard handling for
				dialogue boxes! (SetWindowPos must simply include
				SWP_NOZORDER|SWP_NOACTIVATE.)
	V2.02		- Mar 97: Using ShowWindow instead of SetWindowPos does the same
				thing. This is also OK for 32-bit Windows. More accurate value
				for PI. Entering PI now also lifts stack properly.
				Portable 16/32 bit source. Proper stacklift/State handling
				for copy/paste. Stack registers and conversion settings also
				saved between sessions. Can input numbers in convert box.
				Undo also restores conversion settings. More functions.
				Proper state/stack lift handling for LastX and Pi.
				Pi stored in binary representation. BC5.0 does not parse the
				textual information properly.
				ESC can now be pressed to cancel move/size operations.
	V2.03		- June 97: Constants dialogue. Keyboard right mouse menu.
				NumLock toggle. New memory popup format specifier.
				Keyboard paste bug fixed (undo was not complete).
	V2.04		- July 97: Hyperbolic functions. NumLock toggle now also works
				in dialogue boxes. Fixed bug that made entering too big or
				too small numbers sometimes possible.
	V2.05		- August 97: Separate comma/point settings for
				keyboard/display/clipboard
	V2.1		- May 98: 32-bit only. Minor 32-bitness added. Fractions.
				C and Basic style copy/paste. Better clipboard functions.
	V2.1.1		- Dec 98: Bug fixes for fractions, large negative numbers.
	V2.2		- July 00: atan(0.15)->complex fixed. ClStk also clears LastX.
				.ini and .dat files now in the directory XCALC starts in.
	V2.3		- July 00: Dialogs more standard;
				Silly 8 year old CS_DBLCLK removed from window types.
				Tooltips (not dialogs yet).
				Bug recalling ZERO when Xreg is fraction fixed.
	V2.4		- Aug 00-Jan 01: Now called XCALC.exe again.
				More detailed tooltip help. When resizing, check is
				made to assure window is within screen work area.
				Hyperbolic mode also reset by arrow keys.
				Proper separation of normal and mini settings (done twice - where
				did that source go??)
	V2.4.1		- Jan 02: F5 -- F8 for radix selection
	V2.5		- Sep 02: User editable conversions. Fixed minor bug with configure
				dialogue and topmost settings (it did not set the flag properly on exit).
				Will now allow only one XCALC. A second copy will just show first,
				then exit. (May later be made optional with "-MULTIPLE" cmdline flag.)
	V2.6		- Oct 03: XCALC now on "tray" instead of taskbar
				Also now using Borland C++ Builder 5.5.1
	V2.6.1		- Oct 03: Mouse resize with immediate feedback
	V2.6.2		- Sep 04: Bugfix: Fractions div by zero gave exception
	V2.6.3		- Dec 04: Back to taskbar, except for in Mini mode; one more bitmap
				More choices on system menu; about box. Lclick on "dot"
	V2.6.4		- Oct 05: Bugfix - 3rd root of -27 produced complex; now up to 16 stack regs;
				tray/mini separated.
	V2.6.5		- Oct 05: Close button does not exit when running as tray application.
	V2.7		- Nov 05: Tabbed config box; topmost handled properly when turned on/off.
				Tray settings separate for normal/mini.
	V2.7.1		- Sep 06: 32/16/8 bit binary word size.
				Mini mode fully configurable.
				Fixed wrong constant values (contained IDC_STATIC) introduced in 2.7
	V2.7.2		- Oct 06: Config dialog starts with Normal or Mini page according to current mode
				(if General or Colours was active, no change).
				Separate stack size setting (also depends on display stack sizes).
				Stack size now max 64.
				Option not to duplicate top on stack drop.
				Bugfix: convert to bin mode did not crop negative numbers to current mask.
	V2.7.3		- Feb 07: corrected conversion values for US Gallon and US Pint
	V2.7.4		- March 07: added a few checks for im<<re (avoid switch to complex).
				8- and 16-bit operations now truly signed (division/remainder).
				Using VC8 for debugging.
	V2.8.0		- September 07: added proper degrees/minutes/seconds support.
	V2.8.1		- November 07: cast to long failed for BCC in isint(). Fixed.
	V2.8.2		- February 08: Message popup when error window is hidden. HTML help.
	V2.8.3		- February 08: Bugfix: entering a/b when result was integer.
	V2.8.4		- March 08: R and $ for round.
				Help file update.
				1/1000'th separators (spaces).
	V2.8.5		- March 08: Help file must be in same dir as XCALC.EXE.
				Rudimentary editing of keyboard layout.
				(Never published)
	V2.9.0		- March 2008: Some programming.
	V2.9.1		- June 2008: 6 kb rows as default, not 7.
	V2.9.2		- July 2008: Bugfix: % did / in 32 bit bin modes; remainder renamed
				to modulo (also in programs: MOD not %); added float modulo and div (integer parts)
				Some keyboard changes to accomodate this: Shift-D now DIV, Shift-M is modulo,
				Ctrl-D is ToDMS, % is still percent and modulo in bin modes.
				Added icons to the config box; fixed help (something has changed in Windows I think).
	V2.9.3		- July 2008: Bugfix: the ENG button did not show as activated when selected

------------------------------------------------------------------------*/

#include "xcalc.h"

//--Global Variables for this file only-----------------------------------
const char DisplayName[] =		"XDISPLAY";
const char DotName[] = 			"XDOT";
const char DotCursor[] =		"RIGHTARR";
const char ModeName[] =			"XMODE";
const char ButtonName[] =		"XBUTTON";
const char FixName[] =			"XFIX";
const char KeyName[] =			"XKEYS";
const char ErrorName[] =		"XERROR";
const char MemoryName[] =		"XMEMORY";
const char MemoryPopupName[]=	"XMEMORYPOPUP";
const char MemoryDisplayName[]=	"XMEMORYDISPLAY";
const char ToolTipName[] =		"XTOOLTIP";
const char AccelName[] = 		"XCALCACCEL";
const char m_DefaultBitmap[]=	"DIGITS8";
const char m_DefaultFont[]=		"Courier New";

const char MenuItemHide[] =		"&Hide\tEsc";
const char MenuItemConfig[] = 	"&Configure...\tAlt+N";
const char MenuItemMini[] = 	"&Toggle mini\tAlt+M";
const char MenuItemHelp[] = 	"&Help\tF1";
const char MenuItemKeyHelp[] = 	"&Keyhelp\tCtrl+F1";
const char MenuItemAbout[] = 	"&About XCALC";
const char MenuItemExit[] =	 	"E&xit\tAlt+X";

HCURSOR		hRightArrow;

//--Fatal exit longjump buffer--------------------------------------------

jmp_buf fatal_jmp;

//--radix_lin(mode)-------------------------------------------------------
// returns linear radix: DEC=0, CPX=1,...
int radix_lin(int mode)
{
	int m = mode&RADIXMASK;
	int r = 0;
	do {
		r++;
		m>>=1;
	} while (m);
	return r-1;
}

//--fix_lin(mode)-------------------------------------------------------
// returns linear fix: FIX=0, SCI=1, ENG=2
int fix_lin(int mode)
{
	int m = (mode&FIXMASK)>>8;
	int r = 0;
	do {
		r++;
		m>>=1;
	} while (m);
	return r-1;
}

//--angle_lin(mode)-------------------------------------------------------
// returns linear angle: RAD=0, DEG=1
int angle_lin(int mode)
{
	int m = (mode&ANGLEMASK)>>12;
	int r = 0;
	do {
		r++;
		m>>=1;
	} while (m);
	return r-1;
}

//--Windows Entry Point---------------------------------------------------
#ifdef __BORLANDC__
#pragma argsused
#pragma warn -8084
#endif
int PASCAL WinMain(HINSTANCE hInstance,HINSTANCE hPrev,LPSTR lpszCmdLine,int show)
{
	MSG msg;
	WNDCLASS wndclass;
	HACCEL hAccel;
	HMENU hSystemMenu;
	int i;
	//int ExtraX,ExtraY;
	RECT screenrect;
	char ach[16];
	HWND prevHwnd;
	long exstyle;
	NOTIFYICONDATA nid;
	MENUITEMINFO mii;
	char *pc0,*pc1;
	LPSTR cmdline;
	int iniVer;
	int keyboard;
	int keyno;

	hInst = hInstance;  // global hInst
	hwndEditKeys = NULL;
	hwndEditProgram = NULL;

	// New Sep02 - check for previous instances using the class name. If a window exists, activate it
	if ((prevHwnd = FindWindow(ClassName,NULL)) != NULL) {
		ShowWindow(prevHwnd,SW_RESTORE);
		SetForegroundWindow(prevHwnd);
		// TODO: also show memory popup window if active
		return 0;
	}

	// Find directory containing the XCALC.EXE file. CHM is expected there.
	cmdline = GetCommandLine(); // Seems to be always enclosed in ""
	pc0 = cmdline;
	if (*pc0=='"') pc0++;
	if ((pc1=strchr(pc0,'"'))!=0) *pc1 = 0;
	lstrcpy(HelpFile,pc0);
	chext(HelpFile,HelpExt);

	// Find CWD when executable starts. INI and DAT files now stored there by default.
	getcwd(XCALCDir,_MAX_DIR);
	if (XCALCDir[lstrlen(XCALCDir)-1] != '\\')
		lstrcat(XCALCDir,"\\");
	MakeFullPath(IniFile,IniFile0,XCALCDir);
	MakeFullPath(DatFile,DatFile0,XCALCDir);
	
	// Get screen size
	if (SystemParametersInfo(SPI_GETWORKAREA,0,&screenrect,0)) {
		ScreenX0 = screenrect.left;
		ScreenX1 = screenrect.right;
		ScreenY0 = screenrect.top;
		ScreenY1 = screenrect.bottom;
	}
	else {
		ScreenX0 = 0;
		ScreenX1 = GetSystemMetrics(SM_CXFULLSCREEN);
		ScreenY0 = 0;
		ScreenY1 = GetSystemMetrics(SM_CYFULLSCREEN);
	}

	if (!hPrev) { // Old inheritance from 16-bit Win, and always true...
		
		// register main window
		wndclass.style = CS_DBLCLKS;
		wndclass.lpfnWndProc = XCALCMainProc;
		wndclass.cbClsExtra = 0;
		wndclass.cbWndExtra = 0;
		wndclass.hInstance = hInstance;
		wndclass.hIcon = LoadIcon(hInstance, ClassName);
		wndclass.hCursor = NULL;
		wndclass.hbrBackground = NULL;
		wndclass.lpszMenuName = NULL;
		wndclass.lpszClassName = ClassName;
		if (!RegisterClass(&wndclass)) return FALSE;
		
		// register display child window
		wndclass.style = CS_DBLCLKS;
		wndclass.lpfnWndProc = DisplayProc;
		wndclass.cbClsExtra = 0;
		wndclass.cbWndExtra = 0;
		wndclass.hInstance = hInstance;
		wndclass.hIcon = NULL;
		wndclass.hCursor = LoadCursor(NULL,IDC_ARROW);
		wndclass.hbrBackground = NULL;
		wndclass.lpszMenuName = NULL;
		wndclass.lpszClassName = DisplayName;
		if (!RegisterClass(&wndclass)) return FALSE;
		
		// register "active dot" child window
		wndclass.style = CS_DBLCLKS;
		wndclass.lpfnWndProc = DotProc;
		wndclass.cbClsExtra = 0;
		wndclass.cbWndExtra = 0;
		wndclass.hInstance = hInstance;
		wndclass.hIcon = NULL;
		wndclass.hCursor = hRightArrow = LoadCursor(hInstance,DotCursor);
		wndclass.hbrBackground = NULL;
		wndclass.lpszMenuName = NULL;
		wndclass.lpszClassName = DotName;
		if (!RegisterClass(&wndclass)) return FALSE;
		
		// register mode child window
		wndclass.style = 0;//CS_DBLCLKS;
		wndclass.lpfnWndProc = ModeProc;
		wndclass.cbClsExtra = 0;
		wndclass.cbWndExtra = 4;
		wndclass.hInstance = hInstance;
		wndclass.hIcon = NULL;
		wndclass.hCursor = LoadCursor(NULL,IDC_ARROW);
		wndclass.hbrBackground = NULL;
		wndclass.lpszMenuName = NULL;
		wndclass.lpszClassName = ModeName;
		if (!RegisterClass(&wndclass)) return FALSE;
		
		// register fix child window
		wndclass.style = 0;//CS_DBLCLKS;
		wndclass.lpfnWndProc = FixProc;
		wndclass.cbClsExtra = 0;
		wndclass.cbWndExtra = 0;
		wndclass.hInstance = hInstance;
		wndclass.hIcon = NULL;
		wndclass.hCursor = LoadCursor(NULL,IDC_ARROW);
		wndclass.hbrBackground = NULL;
		wndclass.lpszMenuName = NULL;
		wndclass.lpszClassName = FixName;
		if (!RegisterClass(&wndclass)) return FALSE;
		
		// register button child window
		wndclass.style = 0;//CS_DBLCLKS;
		wndclass.lpfnWndProc = ButtonProc;
		wndclass.cbClsExtra = 0;
		wndclass.cbWndExtra = 0;
		wndclass.hInstance = hInstance;
		wndclass.hIcon = NULL;
		wndclass.hCursor = LoadCursor(NULL,IDC_ARROW);
		wndclass.hbrBackground = NULL;
		wndclass.lpszMenuName = NULL;
		wndclass.lpszClassName = ButtonName;
		if (!RegisterClass(&wndclass)) return FALSE;
		
		// register memory child window
		wndclass.style = CS_DBLCLKS;
		wndclass.lpfnWndProc = MemoryProc;
		wndclass.cbClsExtra = 0;
		wndclass.cbWndExtra = 0;
		wndclass.hInstance = hInstance;
		wndclass.hIcon = NULL;
		wndclass.hCursor = LoadCursor(NULL,IDC_ARROW);
		wndclass.hbrBackground = NULL;
		wndclass.lpszMenuName = NULL;
		wndclass.lpszClassName = MemoryName;
		if (!RegisterClass(&wndclass)) return FALSE;
		
		// register memory popup window
		wndclass.style = 0;
		wndclass.lpfnWndProc = MemoryPopupProc;
		wndclass.cbClsExtra = 0;
		wndclass.cbWndExtra = 0;
		wndclass.hInstance = hInstance;
		wndclass.hIcon = NULL;
		wndclass.hCursor = NULL;
		wndclass.hbrBackground = NULL;
		wndclass.lpszMenuName = NULL;
		wndclass.lpszClassName = MemoryPopupName;
		if (!RegisterClass(&wndclass)) return FALSE;
		
		// register memory display window
		wndclass.style = CS_DBLCLKS;
		wndclass.lpfnWndProc = MemoryDisplayProc;
		wndclass.cbClsExtra = 0;
		wndclass.cbWndExtra = 0;
		wndclass.hInstance = hInstance;
		wndclass.hIcon = NULL;
		wndclass.hCursor = LoadCursor(NULL,IDC_ARROW);
		wndclass.hbrBackground = NULL;
		wndclass.lpszMenuName = NULL;
		wndclass.lpszClassName = MemoryDisplayName;
		if (!RegisterClass(&wndclass)) return FALSE;
		
		// register error child window
		wndclass.style = 0;
		wndclass.lpfnWndProc = ErrorProc;
		wndclass.cbClsExtra = 0;
		wndclass.cbWndExtra = 0;
		wndclass.hInstance = hInstance;
		wndclass.hIcon = NULL;
		wndclass.hCursor = LoadCursor(NULL,IDC_ARROW);
		wndclass.hbrBackground = NULL;
		wndclass.lpszMenuName = NULL;
		wndclass.lpszClassName = ErrorName;
		if (!RegisterClass(&wndclass)) return FALSE;
		
		// register keyboard child window
		wndclass.style = 0;//CS_DBLCLKS;
		wndclass.lpfnWndProc = KeyProc;
		wndclass.cbClsExtra = 0;
		wndclass.cbWndExtra = 0;
		wndclass.hInstance = hInstance;
		wndclass.hIcon = NULL;
		wndclass.hCursor = LoadCursor(NULL,IDC_ARROW);
		wndclass.hbrBackground = NULL;
		wndclass.lpszMenuName = NULL;
		wndclass.lpszClassName = KeyName;
		if (!RegisterClass(&wndclass)) return FALSE;
		
		// register tool tip window
		wndclass.style = 0;
		wndclass.lpfnWndProc = ToolTipProc;
		wndclass.cbClsExtra = 0;
		wndclass.cbWndExtra = 0;
		wndclass.hInstance = hInstance;
		wndclass.hIcon = NULL;
		wndclass.hCursor = LoadCursor(NULL,IDC_ARROW);
		wndclass.hbrBackground = NULL; //GetStockObject(BLACK_BRUSH);
		wndclass.lpszMenuName = NULL;
		wndclass.lpszClassName = ToolTipName;
		if (!RegisterClass(&wndclass)) return FALSE;
		
	}
	
	// get profile from INI file
	// Version - used when changing things...
	{
		FILE *f = fopen(IniFile,"r");
		if (f) {
			fclose(f);
			iniVer = GetPrivateProfileInt(ClassName,"IniVersion",0,IniFile);
		}
		else {
			iniVer = 0x7fffffff;
		}
	}
	// Minimum display?
	Mini		= (BOOL)GetPrivateProfileInt(ClassName,"Mini",FALSE,IniFile);
	// Common to mini and normal
	wDispMode	= (WORD)GetPrivateProfileInt(ClassName,"Mode",DECIMAL|FIX|RAD,IniFile);
	sWordSize	= (WORD)GetPrivateProfileInt(ClassName,"WordSize",0,IniFile);
	wDecDR		= (WORD)GetPrivateProfileInt(ClassName,"DR",RAD,IniFile);
	sFix		= (short)GetPrivateProfileInt(ClassName,"Fix",FLOAT,IniFile);
	SoundOn		= GetPrivateProfileInt(ClassName,"Sound",TRUE,IniFile);
	SaveRegs	= GetPrivateProfileInt(ClassName,"SaveRegs",TRUE,IniFile); // save # between sessions?
	ThousandSep	= GetPrivateProfileInt(ClassName,"ThousandSep",1,IniFile);
	ThousandthSep= GetPrivateProfileInt(ClassName,"ThousandthSep",0,IniFile);
	ToolTip		= GetPrivateProfileInt(ClassName,"ToolTip",TRUE,IniFile);
	ZeroOverflow= GetPrivateProfileInt(ClassName,"ZeroOverflow",FALSE,IniFile);
	CopyTop		= GetPrivateProfileInt(ClassName,"CopyTop",TRUE,IniFile);
	InvTrigAsDMS= GetPrivateProfileInt(ClassName,"InvTrigAsDMS",FALSE,IniFile);
	MessagePopup= GetPrivateProfileInt(ClassName,"MessagePopup",TRUE,IniFile);
	AllowEdit	= GetPrivateProfileInt(ClassName,"AllowEdit",TRUE,IniFile);
	bCleanFrac	= GetPrivateProfileInt(ClassName,"CleanFrac",FALSE,IniFile);
	Cstyle		= GetPrivateProfileInt(ClassName,"CStyle",TRUE,IniFile);
	MinStkSiz	= GetPrivateProfileInt(ClassName,"MinStkSiz",4,IniFile);
	nKeyRows	= GetPrivateProfileInt(ClassName,"KeyRows",nKeyRows,IniFile);
	nKeyCols	= GetPrivateProfileInt(ClassName,"KeyCols",nKeyCols,IniFile);
	// Normal display settings
	n_InTray	= (BOOL)GetPrivateProfileInt(ClassName,"InTray",FALSE,IniFile);
	n_ShowTitle	= GetPrivateProfileInt(ClassName,"ShowTitle",TRUE,IniFile); // main title
	n_ShowPane1	= GetPrivateProfileInt(ClassName,"ShowPane1",TRUE,IniFile); // Keyboard
	n_ShowPane2	= GetPrivateProfileInt(ClassName,"ShowPane2",TRUE,IniFile); // DCHOB
	n_ShowPane3	= GetPrivateProfileInt(ClassName,"ShowPane3",TRUE,IniFile); // FSEDR
	n_ShowPane4	= GetPrivateProfileInt(ClassName,"ShowPane4",TRUE,IniFile); // Float/fix...
	n_ShowPane5	= GetPrivateProfileInt(ClassName,"ShowPane5",TRUE,IniFile); // Logo et al
	n_WindowX	= GetPrivateProfileInt(ClassName,"X",10,IniFile);
	n_WindowY	= GetPrivateProfileInt(ClassName,"Y",10,IniFile);
	n_Size		= GetPrivateProfileInt(ClassName,"Size",100,IniFile);
	n_ShowMemoryPopup =	GetPrivateProfileInt(ClassName,"ShowMemoryPopup",FALSE,IniFile); // Popups
	n_MemCols 	= GetPrivateProfileInt(ClassName,"MemCols",2,IniFile); // Popup format
	n_PopupX	= GetPrivateProfileInt(ClassName,"PopupX",10,IniFile);
	n_PopupY	= GetPrivateProfileInt(ClassName,"PopupY",393,IniFile);
	n_PopupSize = GetPrivateProfileInt(ClassName,"PopupSize",100,IniFile);
	n_IconStart	= GetPrivateProfileInt(ClassName,"IconStart",FALSE,IniFile);
	n_KeepInFront = GetPrivateProfileInt(ClassName,"KeepInFront",FALSE,IniFile);
	n_EnableMouseMove = GetPrivateProfileInt(ClassName,"EnableMouseMove",1,IniFile);
	n_DisplayStkSiz	= GetPrivateProfileInt(ClassName,"DisplayStkSiz",4,IniFile);
	n_UseFonts	= GetPrivateProfileInt(ClassName,"UseFonts",FALSE,IniFile);
	n_Bold		= GetPrivateProfileInt(ClassName,"Bold",FALSE,IniFile);
	GetPrivateProfileString(ClassName,"BitmapName",DefaultBitmap,n_BitmapName,31,IniFile);
	GetPrivateProfileString(ClassName,"FontName",DefaultFont,n_FontName,63,IniFile);
	// Minimum display settings
	// Position is initially based on bitmap DIGITS8 - popup position just below
	m_InTray	= (BOOL)GetPrivateProfileInt(ClassName,"m_InTray",TRUE,IniFile);
	m_ShowTitle	= GetPrivateProfileInt(ClassName,"m_ShowTitle",FALSE,IniFile); // main title
	m_ShowPane1	= GetPrivateProfileInt(ClassName,"m_ShowPane1",FALSE,IniFile); // Keyboard
	m_ShowPane2	= GetPrivateProfileInt(ClassName,"m_ShowPane2",FALSE,IniFile); // DCHOB
	m_ShowPane3	= GetPrivateProfileInt(ClassName,"m_ShowPane3",FALSE,IniFile); // FSEDR
	m_ShowPane4	= GetPrivateProfileInt(ClassName,"m_ShowPane4",FALSE,IniFile); // Float/fix...
	m_ShowPane5	= GetPrivateProfileInt(ClassName,"m_ShowPane5",FALSE,IniFile); // Logo et al
	m_WindowX	= GetPrivateProfileInt(ClassName,"m_X",ScreenX1-3*GetSystemMetrics(SM_CXSIZE)-345,IniFile);
	m_WindowY	= GetPrivateProfileInt(ClassName,"m_Y",/*-1*/0,IniFile);
	m_Size 		= GetPrivateProfileInt(ClassName,"m_Size",100,IniFile);
	m_ShowMemoryPopup =	GetPrivateProfileInt(ClassName,"m_ShowMemoryPopup",FALSE,IniFile); // Popups
	m_MemCols 	=	GetPrivateProfileInt(ClassName,"m_MemCols",2,IniFile); // Popup format
	m_PopupX	= GetPrivateProfileInt(ClassName,"m_PopupX",ScreenX1-3*GetSystemMetrics(SM_CXSIZE)-466,IniFile);
	m_PopupY	= GetPrivateProfileInt(ClassName,"m_PopupY",22,IniFile);
	m_PopupSize = GetPrivateProfileInt(ClassName,"m_PopupSize",70,IniFile);
	m_IconStart	= GetPrivateProfileInt(ClassName,"m_IconStart",FALSE,IniFile);
	m_KeepInFront = GetPrivateProfileInt(ClassName,"m_KeepInFront",TRUE,IniFile);
	m_EnableMouseMove = GetPrivateProfileInt(ClassName,"m_EnableMouseMove",FALSE,IniFile);
	m_DisplayStkSiz	= GetPrivateProfileInt(ClassName,"m_DisplayStkSiz",1,IniFile);
	m_UseFonts	= GetPrivateProfileInt(ClassName,"m_UseFonts",FALSE,IniFile);
	m_Bold		= GetPrivateProfileInt(ClassName,"m_Bold",FALSE,IniFile);
	GetPrivateProfileString(ClassName,"m_BitmapName",m_DefaultBitmap,m_BitmapName,31,IniFile);
	GetPrivateProfileString(ClassName,"m_FontName",m_DefaultFont,m_FontName,63,IniFile);
	
	// Decide which of '.' and ',' is decimal point
	GetProfileString("intl","sDecimal",".",ach,5);
	k_useComma	= (BOOL)GetPrivateProfileInt(ClassName,"k_useComma",lstrcmp(ach,",")?0:1,IniFile);
	d_useComma	= (BOOL)GetPrivateProfileInt(ClassName,"d_useComma",lstrcmp(ach,",")?0:1,IniFile);
	c_useComma	= (BOOL)GetPrivateProfileInt(ClassName,"c_useComma",lstrcmp(ach,",")?0:1,IniFile);
	
	// Change wDispMode if iniVer<1: bit patterns changed
	if (iniVer<1) {
		WORD mode = (WORD)(wDispMode&0x000f);
		if      (mode==0x1) mode=0x01;
		else if (mode==0x2) mode=0x02;
		else if (mode==0x3) mode=0x04;
		else if (mode==0x4) mode=0x08;
		else if (mode==0x5) mode=0x10;
		wDispMode = (WORD)(((wDispMode&0x0ff0)<<4) | mode);
	}
	// Read the screen colours off the .ini file
	InitColour();

	// Correct possible errors in the INI file (might cause nasty errors):
	if ((sFix<FLOAT)||(sFix>9)) sFix = FLOAT;
	if ( (CURR_RADIX > BINARY)
		|| (CURR_FIX   > ENG)
		|| (CURR_ANGLE > RAD)
		|| (CURR_RADIX == 0)
		|| (CURR_FIX   == 0)
		|| (CURR_ANGLE == 0)
		) wDispMode = DECIMAL|FIX|DEG;
	if (wDecDR!=DEG && wDecDR!=RAD) wDecDR = RAD;
	if (DisplayStkSiz<1 || DisplayStkSiz>MAXSTKSIZ) DisplayStkSiz = 4;
	if (m_DisplayStkSiz<1 || m_DisplayStkSiz>MAXSTKSIZ) m_DisplayStkSiz = 1;
	
	// Here we need to set correct radix (for stack restore on startup)
	wRadix = aRadix[RADIX_LIN];
	
	// initialise math registers & flags
	bSL = TRUE;						// Stack Lift
	iState = 0;						// Digit input state
	Dirty = FALSE;					// Whole stack needs update?
	bShift = FALSE;
	for (i=0;i<MAXNUMREGS;i++) clrstkreg(i);
	for (i=0;i<MEMSIZ;i++) {
		Memory[i].real=Memory[i].imag=0.0;
		Memory[i].i=Memory[i].num=Memory[i].den=0;
	}
	
	// Initialise undo information stack
	UndoIndex = 0;		// no undoinformation
	//RedoMaxIndex = 0;	// no redo information

	// read saved registers if SaveRegs is TRUE
	
	if (SaveRegs) ReadRegisters(DatFile);
	
	// initialise unit conversions & tooltip

	InitConversion();
	QConvText(ConvMsg);
	QConstText("(no constant)");

	// Make sure required DLLs are loaded
	InitCommonControls();

	// install floating point error signal handler and longjmp state
	signal(SIGFPE,(fptr)FPE_Handler);
	if (setjmp(fatal_jmp)) {
		Restart = FALSE;
		goto FatalExit;
	}
	
	Restart = FALSE;
	
Restart_Position:               // loop around to here if restart
	
	// Mar08: check internal setup of keyboards
	if (!initKeys()) return 1;

	// Read current keyboards
	if (!NewKB) {
		// Check version -- maybe key definitions have changed
		if (iniVer<2) {
			int m;
			char caption[64];
			sprintf(caption,"XCALC %s: New keyboard definitions",VERSION);
			if  (MessageBox(NULL,
				"Keyboard definitions have changed in this version of XCALC.\n"
				"Any modifications you have made to the keyboard layout will\n"
				"unfortunately be lost. Sorry about this.",caption,MB_OKCANCEL) == IDCANCEL)
				return 1;
			for (m=0;m<NRADIX;m++) SetDefaultKeyboard(m);
		}
		else {
			for (keyboard=0;keyboard<5;keyboard++) {
				char kb[16];
				sprintf(kb,"Keyboard%1d",keyboard);
				for (keyno=0;keyno<nKeyRows*nKeyCols;keyno++) {
					char kn[8];
					itoa(keyno,kn,10);
					Keys[keyboard][keyno] = GetPrivateProfileInt(kb,kn,Keys[keyboard][keyno],IniFile);
				}
			}
		}
	}
	NewKB = FALSE;
	// Apply fudge to keyboard ..   
	KeyboardDecimal(k_useComma);
	
	if (Mini) {
		DisplayStkSiz = m_DisplayStkSiz;
		Title=m_ShowTitle;
		Pane1=m_ShowPane1;
		Pane2=m_ShowPane2;
		Pane3=m_ShowPane3;
		Pane4=m_ShowPane4;
		Pane5=m_ShowPane5;
		Size=m_Size;
		WindowX=m_WindowX;
		WindowY=m_WindowY;
		PopupSize=m_PopupSize;
		PopupX=m_PopupX;
		PopupY=m_PopupY;
		MemCols=m_MemCols;
		ShowMemoryPopup=m_ShowMemoryPopup;
		EnableMouseMove=m_EnableMouseMove;
		KeepInFront=m_KeepInFront;
		UseFonts=m_UseFonts;
		Bold=m_Bold;
		lstrcpy(FontName,m_FontName);
		lstrcpy(BitmapName,m_BitmapName);
	}
	else {
		DisplayStkSiz = n_DisplayStkSiz;
		Title=n_ShowTitle;
		Pane1=n_ShowPane1;
		Pane2=n_ShowPane2;
		Pane3=n_ShowPane3;
		Pane4=n_ShowPane4;
		Pane5=n_ShowPane5;
		Size=n_Size;
		WindowX=n_WindowX;
		WindowY=n_WindowY;
		PopupSize=n_PopupSize;
		PopupX=n_PopupX;
		PopupY=n_PopupY;
		MemCols=n_MemCols;
		ShowMemoryPopup=n_ShowMemoryPopup;
		EnableMouseMove=n_EnableMouseMove;
		KeepInFront=n_KeepInFront;
		UseFonts=n_UseFonts;
		Bold=n_Bold;
		lstrcpy(FontName,n_FontName);
		lstrcpy(BitmapName,n_BitmapName);
	}
	
	if (MemCols!=1) MemCols=2;
	MemRows = MEMSIZ/MemCols;
	StkSiz = max(MinStkSiz,max(m_DisplayStkSiz,n_DisplayStkSiz));
	if (StkSiz<4) StkSiz = 4;
	for (i=StkSiz;i<MAXSTKSIZ;i++) clrstkreg(i);
	
	Error = 0;
	
	// For a restart, main window and memory popup needs to recalculate properly
	TotalWidth = 0;
	TotalHeight = 0;
	PopupDX = 0;
	PopupDY = 0;

	// Create main window to get information on bitmap size. (WM_CREATE)

	hwndXCALC = CreateWindowEx(
		0,
		ClassName,
		ClassName,
		Title? WS_OVERLAPPED|WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX
			 : WS_POPUP|WS_BORDER|WS_SYSMENU,
		WindowX,
		WindowY,
		0,
		0,
		NULL,
		NULL,
		hInstance,
		NULL);

	if (!hwndXCALC) return FALSE;

	// Are we really in the tray?
	InTray = Mini&&m_InTray || !Mini&&n_InTray;

	// Oct05: separate mini mode and taskbar/tray ownership
	if (InTray) {
		exstyle = GetWindowLong(hwndXCALC,GWL_EXSTYLE);
		SetWindowLong(hwndXCALC,GWL_EXSTYLE,exstyle | WS_EX_TOOLWINDOW & ~WS_EX_APPWINDOW);
		// make tray icon visible and active
		nid.cbSize=sizeof(nid);
		nid.hWnd=hwndXCALC;
		nid.uID=0;
		nid.uFlags=NIF_TIP|NIF_ICON|NIF_MESSAGE;
		nid.uCallbackMessage=PM_ICONMESSAGE;
		nid.hIcon=LoadIcon(hInstance, ClassName);
		lstrcpy(nid.szTip,ClassName);
		Shell_NotifyIcon(NIM_ADD,&nid);
	}

	// remove size/maximize option in system menu; insert "Configure","Mini","Help","About"
	hSystemMenu = GetSystemMenu(hwndXCALC,FALSE);
	RemoveMenu(hSystemMenu,SC_SIZE,MF_BYCOMMAND);
	RemoveMenu(hSystemMenu,SC_MAXIMIZE,MF_BYCOMMAND);
	InsertMenu(hSystemMenu,3,MF_BYPOSITION|MF_STRING,SC_ABOUT,MenuItemAbout);
	InsertMenu(hSystemMenu,3,MF_BYPOSITION|MF_STRING,SC_KEYHELP,MenuItemKeyHelp);
	InsertMenu(hSystemMenu,3,MF_BYPOSITION|MF_STRING,SC_HELP,MenuItemHelp);
	InsertMenu(hSystemMenu,3,MF_BYPOSITION|MF_SEPARATOR,0,0);
	InsertMenu(hSystemMenu,3,MF_BYPOSITION|MF_STRING,SC_MINI,MenuItemMini);
	InsertMenu(hSystemMenu,3,MF_BYPOSITION|MF_STRING,SC_CONFIG,MenuItemConfig);
	InsertMenu(hSystemMenu,3,MF_BYPOSITION|MF_SEPARATOR,0,0);
	// As tray app, change SC_CLOSE item (and then also caption "X") to SC_MINIMIZE.
	// In Xcwinpro SC_MINIMIZE will then hide the window.
	// Insert new SC_EXIT item to close the application.
	if (InTray) {
		memset(&mii,0,sizeof(mii));
		mii.cbSize = sizeof(mii);

		// "Undefault" old Close item
		mii.fMask  = MIIM_STATE;
		mii.fState = MFS_ENABLED;
		SetMenuItemInfo(hSystemMenu,SC_CLOSE,FALSE,&mii);

		mii.fMask  = MIIM_STATE|MIIM_ID|MIIM_STRING;
		mii.fState = MFS_ENABLED;
		mii.wID    = SC_EXIT;
		mii.dwTypeData = (char*)MenuItemExit;
		SetMenuItemInfo(hSystemMenu,SC_CLOSE,FALSE,&mii);

		mii.fMask  = MIIM_STATE|MIIM_ID|MIIM_STRING;
		mii.fState = MFS_ENABLED;
		mii.wID    = SC_CLOSE;
		mii.dwTypeData = (char*)MenuItemHide;
		SetMenuItemInfo(hSystemMenu,SC_MINIMIZE,FALSE,&mii);
		SetMenuDefaultItem(hSystemMenu,SC_EXIT,FALSE);
	}
	else {
		// change caption on close item
		ModifyMenu(hSystemMenu,SC_CLOSE,MF_BYCOMMAND|MF_STRING,SC_CLOSE,MenuItemExit);
		SetMenuDefaultItem(hSystemMenu,SC_CLOSE,FALSE);
	}

	// Send message to size main window based on global variables
	// wParam is nonzero to actually do the move/size
	SendMessage(hwndXCALC,PM_SIZEWINDOW,TRUE,0);
	
	// If required, make window topmost
	if (KeepInFront)
		SetWindowPos(hwndXCALC,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
	
	// Everything registered and set up, now create child windows.
	for (i=0;i<DisplayStkSiz;i++) {
		hDisplay[i] = CreateWindow(
			DisplayName,				// lspzClassName
			NULL,						// window title
			WS_CHILD | WS_VISIBLE | WS_CLIPCHILDREN,	// style
			0,0,0,0,
			hwndXCALC,					// parent
			(HMENU)i,					// child ID
			hInstance,					// creator
			NULL);						// parameters
		if (!hDisplay[i]) return FALSE;
	}
	
	if (!Title) {
		hDot = CreateWindow(
			DotName,					// lspzClassName
			NULL,						// window title
			WS_CHILD | WS_VISIBLE,		// style
			0,0,0,0,
			hDisplay[DisplayStkSiz-1],	// parent
			NULL,						// class menu
			hInstance,					// creator
			NULL);						// parameters
		if (!hDot) return FALSE;
	}
	
	if (Pane1) {
		hwndKey = calloc(nKeyRows*nKeyCols,sizeof(HWND));
		for (i=0;i<nKeyRows*nKeyCols;i++) {
			hwndKey[i] = CreateWindow(
				KeyName,				// lspzClassName
				NULL,					// window title
				WS_CHILD | WS_VISIBLE,	// style
				0,0,0,0,
				hwndXCALC,				// parent
				(HMENU)i,				// child ID
				hInstance,				// creator
				NULL);					// parameters
			
			if (!hwndKey[i]) return FALSE;
		}
	}
	
	if (Pane2) {
		for (i=DECWND;i<=BINWND;i++) {
			hDispMode[i] = CreateWindow(
				ModeName,				// lspzClassName
				NULL,					// window title
				WS_CHILD,				// style
				0,0,0,0,
				hwndXCALC,				// parent
				(HMENU)i,				// child ID
				hInstance,				// creator
				NULL);					// parameters
			if (!hDispMode[i]) return FALSE;
			SetWindowWord(hDispMode[i],0,TRUE);	// Enabled?
			SetWindowWord(hDispMode[i],2,FALSE);// Active (red dot on)?
			ShowWindow(hDispMode[i],SW_SHOW);
		}
	}
	
	if (Pane3) {
		for (i=FIXWND;i<=RADWND;i++) {
			hDispMode[i] = CreateWindow(
				ModeName,				// lspzClassName
				NULL,					// window title
				WS_CHILD,				// style
				0,0,0,0,
				hwndXCALC,				// parent
				(HMENU)i,				// child ID
				hInstance,				// creator
				NULL);					// parameters
			if (!hDispMode[i]) return FALSE;
			SetWindowWord(hDispMode[i],0,TRUE);	// Enabled?
			SetWindowWord(hDispMode[i],2,FALSE);// Active (red dot on)?
			ShowWindow(hDispMode[i],SW_SHOW);
		}
	}
	
	if (Pane4) {
		hFix = CreateWindow(
			FixName,					// lspzClassName
			NULL,						// window title
			WS_CHILD | WS_VISIBLE,		// style
			0,0,0,0,
			hwndXCALC,					// parent
			NULL,						// class menu
			hInstance,					// creator
			NULL);						// parameters
		if (!hFix) return FALSE;
		
		for (i=0;i<NBUTTONS;i++) {
			hButton[i] = CreateWindow(
				ButtonName,				// lspzClassName
				NULL,					// window title
				WS_CHILD | WS_VISIBLE,	// style
				0,0,0,0,
				hwndXCALC,				// parent
				(HMENU)i,				// child ID
				hInstance,				// creator
				NULL);					// parameters
			if (!hButton[i]) return FALSE;
		}
	}
	
	if (Pane5) {
		hError = CreateWindow(
			ErrorName,					// lspzClassName
			NULL,						// window title
			WS_CHILD | WS_VISIBLE,		// style
			0,0,0,0,
			hwndXCALC,					// parent
			NULL,						// class menu
			hInstance,					// creator
			NULL);						// parameters
		if (!hError) return FALSE;
		
		hMemory = CreateWindow(
			MemoryName,					// lspzClassName
			NULL,						// window title
			WS_CHILD | WS_VISIBLE,		// style
			0,0,0,0,
			hwndXCALC,					// parent
			NULL,						// class menu
			hInstance,					// creator
			NULL);						// parameters
		if (!hMemory) return FALSE;
	}
	else {
		// 27sep07: popup error window (below main window)
		hError = CreateWindow(
			ErrorName,					// lspzClassName
			NULL,						// window title
			WS_POPUP | WS_VISIBLE,		// style
			0,0,0,0,
			hwndXCALC,					// parent
			NULL,						// class menu
			hInstance,					// creator
			NULL);						// parameters
		if (!hError) return FALSE;
	}
	
	// Memory windows do their own sizing and positioning in reply
	// to PM_SIZEWINDOW
	
	hMemoryPopup = CreateWindow(
		MemoryPopupName,				// lspzClassName
		NULL,							// window title
		WS_POPUP | WS_BORDER,			// style
		CW_USEDEFAULT,0,
		CW_USEDEFAULT,0,
		hwndXCALC,						// parent
		NULL,							// class menu
		hInstance,						// creator
		NULL);							// parameters
	if (!hMemoryPopup) return FALSE;
	// This positions and sizes the window
	SendMessage(hMemoryPopup,PM_SIZEWINDOW,0,0);

	for (i=0;i<MEMSIZ;i++) {
		hMemoryDisplay[i] = CreateWindow(
			MemoryDisplayName,			// lspzClassName
			NULL,						// window title
			WS_CHILD | WS_VISIBLE,		// style
			CW_USEDEFAULT,0,
			CW_USEDEFAULT,0,
			hMemoryPopup,				// parent
			(HMENU)i,					// child ID
			hInstance,					// creator
			NULL);						// parameters
		if (!hMemoryDisplay[i]) return FALSE;
		// This positions and sizes the window
		SendMessage(hMemoryDisplay[i],PM_SIZEWINDOW,0,0);
	}
	
	// Create hidden ToolTip window (always present)
	hwndToolTip = CreateWindow(
		ToolTipName,
		"XCALC ToolTip",
		WS_POPUP,
		0,0,0,0,
		hwndXCALC,NULL,hInst,NULL);

	// Position and size all child windows
	SendMessage(hwndXCALC,PM_SIZECHILDREN,FALSE,0);

	if (Restart) {
		// Make sure window is fully onscreen (if possible)
		RECT rect;
		int d;
		GetWindowRect(hwndXCALC,&rect);
		d = rect.right-ScreenX1;
		if (d>0) {
			rect.left -= d;
		}
		if (rect.left<ScreenX0) rect.left = ScreenX0;
		d = rect.bottom-ScreenY1;
		if (d>0) {
			rect.top -= d;
		}
		// Allow one pixel above screen area - it is better for mini...
		if (rect.top<ScreenY0) rect.top = ScreenY0-1;
		SetWindowPos(hwndXCALC,NULL,rect.left,rect.top,0,0,SWP_NOZORDER|SWP_NOSIZE);
		Restart = FALSE;
		show = SW_SHOWNORMAL;
	}
	else {
		// Start minimised/hidden if IconStart
		if (Mini?m_IconStart:n_IconStart) {
			if (InTray) show = SW_HIDE;
			else show = SW_MINIMIZE;
		}
	}
	
	ShowWindow(hwndXCALC,show);
	if (ShowMemoryPopup) {
		if (show!=SW_HIDE) ShowWindow(hMemoryPopup,SW_SHOWNORMAL);
	}
	if (!Pane5) {
		ShowWindow(hError,show);
	}
	
	if (GetActiveWindow() == hwndXCALC) SetFocus(hwndXCALC);
	
	hAccel = LoadAccelerators(hInstance,AccelName);
	
	// run main routine
	
	while (GetMessage(&msg, NULL, 0, 0)) {
		if (   (!IsWindow(hwndEditKeys) || !IsDialogMessage(hwndEditKeys, &msg))
			&& (!IsWindow(hwndEditProgram) || !IsDialogMessage(hwndEditProgram, &msg))) {
			if (!TranslateAccelerator(hwndXCALC,hAccel,&msg)) {
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
	}
	
	// Destination for fatal exit after exception caught
	
FatalExit:
	
	// Version - used when changing things...
	WritePrivateProfileString(ClassName,"IniVersion",itoa(2,ach,10),IniFile);
	// Write profile to INI file
	WritePrivateProfileString(ClassName,"Mini",itoa(Mini,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"Mode",itoa(wDispMode,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"WordSize",itoa(sWordSize,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"DR",itoa(wDecDR,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"Fix",itoa(sFix,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"X",itoa(n_WindowX,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"Y",itoa(n_WindowY,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"Size",itoa(n_Size,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"PopupX",itoa(n_PopupX,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"PopupY",itoa(n_PopupY,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"PopupSize",itoa(n_PopupSize,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"m_X",itoa(m_WindowX,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"m_Y",itoa(m_WindowY,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"m_Size",itoa(m_Size,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"m_PopupX",itoa(m_PopupX,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"m_PopupY",itoa(m_PopupY,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"m_PopupSize",itoa(m_PopupSize,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"CleanFrac",itoa(bCleanFrac,ach,10),IniFile);
	WritePrivateProfileString(ClassName,"Cstyle",itoa(Cstyle,ach,10),IniFile);
	
	// Write current keyboards unless restart
	//if (!Restart) {
		for (keyboard=0;keyboard<5;keyboard++) {
			char kb[16];
			sprintf(kb,"Keyboard%1d",keyboard);
			WritePrivateProfileSection(kb,"0=0\0",IniFile);
			for (keyno=0;keyno<nKeyRows*nKeyCols;keyno++) {
				char kn[8];
				itoa(keyno,kn,10);
				WritePrivateProfileString(kb,kn,itoa(Keys[keyboard][keyno],ach,10),IniFile);
			}
		}
	//}
	// Free keyboard buttons
	if (hwndKey) {
		free(hwndKey);
		hwndKey=NULL;
	}
	freeKeys();

	if (Restart) {
		show = SW_SHOWNORMAL;
		goto Restart_Position;
	}
	
	// Free memory for conversions

	EndConversion();

	DestroyCursor(hRightArrow);
	
	return msg.wParam;
}
