//-----------------------------------------------------------------------
//
//  Main header file for XCALC 32-bit calculator.
//  Bernt Ribbum
//
//-----------------------------------------------------------------------

#ifndef _XCALC_H
#define _XCALC_H

#define STRICT

///*
#define _WIN32_WINNT 0x500
//*/

//--All include files----------------------------------------------------

#include <windows.h>
#include <commctrl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <float.h>
#include <math.h>
#include <limits.h>
#include <commdlg.h>
#include <time.h>
#include <setjmp.h>
#include <signal.h>
#include <setjmp.h>
#include <direct.h>
#include <htmlhelp.h>

//--Message Crackers-----------------------------------------------------

#include <windowsx.h>

#pragma hdrstop

//--Dialogue box #include------------------------------------------------

#include "xcalc.rh"

//--#defines to help portability-----------------------------------------

#define MoveTo(hdc,x,y) MoveToEx(hdc,x,y,NULL)

#ifndef __BORLANDC__
#define hypotl hypot
#define _strtold strtod
#define itoa _itoa
#define strnicmp _strnicmp
#define stricmp _stricmp
#define ultoa _ultoa
#define strupr _strupr
#define ltoa _ltoa
#define hypot _hypot
#define getcwd _getcwd
#define unlink _unlink
#endif

//--Display Constants----------------------------------------------------

#define NDIGITS				34			// Total number of characters in display
#define NPRINTABLE			27			// Number of printable chars in bitmap

//--Global constants for everything--------------------------------------

#define ID_BLINKTIMER		100
#define BLINKMS				500

#define TOOLTIP_SHOWTIMER	101
#define TOOLTIP_SHOWMS		500
#define TOOLTIP_FOLLOWSHOWMS	10
#define TOOLTIP_HIDETIMER	102
#define TOOLTIP_HIDEMS		30000
#define TOOLTIP_FOLLOWTIMER	103
#define TOOLTIP_FOLLOWMS	200
#define TOOLTIP_MAXLENGTH	256

#define MAXUNDO 			256

#define NBUTTONS			3
#define NMODEROWS			2
#define NMODECOLS			5
#define NMODES				(NMODEROWS*NMODECOLS)
#define NRADIX				5

#define MAXSTKSIZ			64
#define MAXNUMREGS			(MAXSTKSIZ+1)
#define MEMSIZ				10					// 5 or 10 only
#define FLOAT				(-1)
#define MINSIZE				20					// arbitrary values
#define MAXSIZE				1000

#define Xreg				0
#define Yreg				1
#define Lreg				MAXSTKSIZ

#define DECMAXDIGITS		LDBL_DIG			// for input
#define CPLXMAXDIGITS		DBL_DIG
#define DMSMAXINT			65535				// max integer for DMS display
#define DMSMAXFIX			10					// max decimal digits to display
#define HEXMAXDIGITS		8
#define OCTMAXDIGITS		11
#define BINMAXDIGITS		32
#define HEXMAXDIGITS16		4
#define OCTMAXDIGITS16		6
#define BINMAXDIGITS16		16
#define HEXMAXDIGITS8		2
#define OCTMAXDIGITS8		3
#define BINMAXDIGITS8		8

#ifdef __BORLANDC__
#define FACT_MAX			449			 		// max input x for x!
#else
#define FACT_MAX			166			 		// max input x for x!
#endif

#ifdef __BORLANDC__
#define NUM_MAX 			9.499999999999999e999L	// odd value to avoid 1.e+1000
#define NUM_MIN				1e-999L					// Avoid smaller numbers on display
#define MAXMANT				9.499999999999999L		// MAXMANT E MAXEXP == NUM_MAX
#define MAXEXP				999
#else
#define NUM_MAX 			9.999999999999999e300L	// We make Micro$oft happy
#define NUM_MIN				1e-300L					// since they do not
#define MAXMANT				9.999999999999999L		// support long double...
#define MAXEXP				300
#endif

#define DECOUTWIDTH			22					// for output, before possible 1000s separator
#define CPLXOUTWIDTH		16					// for complex display output

//--For keys and programs-----------------------------------------------

#define _AS 255

//--For unit conversions------------------------------------------------

#define NFIXEDCONVERSIONS	5
#define NUSERCONVERSIONS	5
#define NCONVERSIONS		(NFIXEDCONVERSIONS+NUSERCONVERSIONS) // Number of conversion types (added 5 user types Feb02)

//--Math structures------------------------------------------------------

typedef enum ContentType {
	ctREAL,
	ctINTEGER,
	ctFRACTION,
	ctDMS
} ContentType;

typedef long double LD; // for convenience

typedef struct Register {        // normal register structure
	ContentType ct;
	LD real,imag;
	long i;
	long num,den;
} Register;

//--Undo types-----------------------------------------------------------

typedef int UNDOTYPE;

enum UndoTypes {
	UT_NONE,	UT_INKEY,	UT_ENTER,
	UT_ADD,		UT_SUB,		UT_MUL,		UT_DIV,
	UT_STORE,	UT_RECALL,	UT_CLRMEM,	UT_PASTE,
	UT_SIN,		UT_COS,		UT_TAN,		UT_ASIN,	UT_ACOS,	UT_ATAN,
	UT_SINH,	UT_COSH,	UT_TANH,	UT_ASINH,	UT_ACOSH,	UT_ATANH,
	UT_FIX,		UT_SCI,		UT_ENG,		UT_FLOAT,	UT_DEG,		UT_RAD,
	UT_WORD,
	UT_DEC,		UT_CPX,		UT_HEX,		UT_OCT,		UT_BIN,
	UT_CLR,		UT_LASTX,	UT_XY,		UT_REIM,	UT_CHS,		UT_PERC,	UT_MOD,		UT_DIVF,
	UT_IJ,		UT_PI,		UT_CONJG,	UT_ABS,		UT_POW,		UT_ROOT,	UT_FACT,
	UT_SQRT,	UT_SQR,		UT_EXP,		UT_LN,		UT_TEN,		UT_LOG,		UT_RCP,
	UT_CLSTK,	UT_RUP,		UT_RDN,
	UT_OR,		UT_AND,		UT_XOR,		UT_NOT,
	UT_LROT,	UT_LSH,		UT_LASH,
	UT_RROT,	UT_RSH,		UT_RASH,
	UT_QROOT,	UT_RAND,	UT_NIJ,		UT_CART,	UT_POLAR,	UT_FRAC,	UT_INT,
	UT_CONV,	UT_DMS,		UT_RND,		UT_CONST,	UT_CUBE,
	UT_CLEARMEM,
};

//--Undo structure-------------------------------------------------------

typedef struct UndoInfoStruct {
	UNDOTYPE	UndoType;
	Register	Memory[MEMSIZ];
	Register	Stack[MAXNUMREGS];
	BOOL		StackLift;
	short		Fix;
	short		sWordSize;
	WORD		DispMode;
	WORD		Radix;
	WORD		DecDR;
	short		ConvType;
	short		From;
	short		To;
	short 		SelectFrom[NCONVERSIONS];
	short 		SelectTo[NCONVERSIONS];
	LD			ldConstant;
} UndoInfoStruct;

//Note: use this structure also for register save!

enum keyid {
	KEY_NONE, /* 0 is a good synonym */
	KEY_0,		KEY_1,		KEY_2,		KEY_3,		KEY_4,		KEY_5,		KEY_6,		KEY_7,		KEY_8,		KEY_9,	
	KEY_A,		KEY_B,		KEY_C,		KEY_D,		KEY_E,		KEY_F,		KEY_POINTP,	KEY_POINTC,	KEY_CHS,	KEY_EEX,	
	KEY_BACK,	KEY_ENTER,	KEY_ADD,	KEY_SUB,	KEY_MUL,	KEY_DIV,	KEY_DIVF,	KEY_MOD,	KEY_MODF,	KEY_LASTX,	KEY_UNDO,	
	KEY_CLSTK,	KEY_IJ,		KEY_FRAC,	KEY_DMS,	KEY_TODMS,	KEY_ROUND,	KEY_XY,		KEY_RUP,	KEY_RDN,	KEY_PERC,
	KEY_CONV,	KEY_QCONV,	KEY_CONS,	KEY_QCONS,	KEY_RAND,	KEY_TOFRAC,	KEY_TOINT,	KEY_CART,	KEY_POLAR,	KEY_EXP,	
	KEY_TEN,	KEY_ROOT,	KEY_SQRT,	KEY_QROOT,	KEY_LN,		KEY_LOG,	KEY_POW,	KEY_SQR,	KEY_CUBE,	KEY_ABS,	
	KEY_SIN,	KEY_COS,	KEY_TAN,	KEY_ASIN,	KEY_ACOS,	KEY_ATAN,	
	KEY_SINH,	KEY_COSH,	KEY_TANH,	KEY_ASINH,	KEY_ACOSH,	KEY_ATANH,	
	KEY_HYP,	KEY_NONHYP,	KEY_RCP,	KEY_PI,		KEY_FACT,
	// Additional complex
	KEY_IMAGP,	KEY_IMAGC,	KEY_NIJ,	KEY_REIM,	KEY_CONJ,	
	// Additional Hex/Oct/Bin
	KEY_AND,	KEY_OR,		KEY_XOR,	KEY_NOT,	KEY_SHL,	KEY_SHR,	KEY_ASHL,	KEY_ASHR,	KEY_ROTL,	KEY_ROTR,	
	// Keys for other operations
	KEY_STO0,	KEY_STO1,	KEY_STO2,	KEY_STO3,	KEY_STO4,	KEY_STO5,	KEY_STO6,	KEY_STO7,	KEY_STO8,	KEY_STO9,	
	KEY_RCL0,	KEY_RCL1,	KEY_RCL2,	KEY_RCL3,	KEY_RCL4,	KEY_RCL5,	KEY_RCL6,	KEY_RCL7,	KEY_RCL8,	KEY_RCL9,	
	KEY_CLR0,	KEY_CLR1,	KEY_CLR2,	KEY_CLR3,	KEY_CLR4,	KEY_CLR5,	KEY_CLR6,	KEY_CLR7,	KEY_CLR8,	KEY_CLR9,	
	KEY_FIX,	KEY_SCI,	KEY_ENG,	KEY_FLOAT,	
	KEY_FIX0,	KEY_FIX1,	KEY_FIX2,	KEY_FIX3,	KEY_FIX4,	KEY_FIX5,	KEY_FIX6,	KEY_FIX7,	KEY_FIX8,	KEY_FIX9,	
	KEY_DEG,	KEY_RAD,	KEY_DECIMAL,KEY_COMPLEX,KEY_HEX,	KEY_OCTAL,	KEY_BINARY,	KEY_32BIT,	KEY_16BIT,	KEY_8BIT,	
	KEY_CONFIG,	KEY_EXIT,	KEY_HELP,	KEY_KEYHELP,KEY_COPY,	KEY_PASTE,	KEY_POPUP,	KEY_MINI,	
	// Program keys
	KEY_PROG1,	KEY_PROG2,	KEY_PROG3,	KEY_PROG4,	KEY_PROG5,	KEY_PROG6,	
	KEY_PROG7,	KEY_PROG8,	KEY_PROG9,	KEY_PROG10,	KEY_PROG11,	KEY_PROG12,	
	KEY_EDITPR,	KEY_CLRMEM,
	// Fix +/-
	KEY_FIXP,	KEY_FIXM,
};

#define NKEYID 172

// Struct defining look & operation of keyboard key
typedef struct key {
	int id;					// key id - mainly for debug
	char *name;				// command name - for programming
	int mode;				// in which modes key is allowed: 1=DEC, 2=CPX, 4=HEX, 8=OCT, 16=BIN
	char *caption;			// key caption
	int symbolPos;			// symbol font character position
	unsigned char op[2];	// operation
	char *tip;				// tooltip
} key;

//--Function prototype for fatal error handler---------------------------

typedef void (*fptr)(int);

//--Global variables available to all routines---------------------------

extern int			nKeyRows, nKeyCols;
extern const int	aMaxInputDigits[][NMODECOLS];
extern const int	aBits[];
extern const int	aWordMask[];
extern const LD		aBinMax[];
extern const LD		aBinMin[];

extern const WORD	aRadix[];
extern const char   *ModeNames[NMODECOLS];

extern const char	IniFile0[];
extern const char	DatFile0[];
extern const char	HelpExt[_MAX_EXT];
extern       char	XCALCDir[_MAX_PATH];
extern       char	IniFile[_MAX_PATH];
extern       char	DatFile[_MAX_PATH];
extern       char	HelpFile[_MAX_PATH];
extern const char	ClassName[];
extern const char	DefaultBitmap[];
extern const char	DefaultFont[];
extern const char	VersionBitmap[];
extern const char	StatusBitmap[];

extern const char	Printable[];

extern int			BitmapX; 			// width of digit bitmaps
extern int			BitmapY; 			// height of same
extern int			Border;				// border width
extern int			LogoX, LogoY;		// size of "Version" Logo

extern HWND			hwndXCALC,
					hwndEditKeys,
					hwndEditProgram,
					hDisplay[MAXSTKSIZ],
					hDot,
					hFix,
					hDispMode[NMODES],
					hButton[NBUTTONS],
					hError,
					hMemory,
					*hwndKey,
					hMemoryPopup,
					hMemoryDisplay[MEMSIZ],
					hwndToolTip;
extern HDC			hDigitsDC;
extern HPEN			hBrightPen,
					hDarkPen,
					hSelectedPen,
					hCaptionPen;
extern HBRUSH		hBorderBrush,
					hBackBrush,
					hSelectedBrush,
					hCaptionBrush;
extern HFONT		hBigFont,
					hBigFont_,
					hSmallFont,
					hSmallFont_,
					hSymbolFont,
					hToolTipFont;
extern HINSTANCE	hInst;

// These bitmap handles are used with hDigitsDC only. Must be preserved.
extern HBITMAP		hbOld,hbDigits,hbVersion,hbStatus;

extern int			ScreenX0,ScreenX1,ScreenY0,ScreenY1;
extern int			StatusX,StatusY;
extern int 			DisplayX,DisplayY,BrdWidth,BrdHeight;
extern int			ClientWidth,ClientHeight,TotalWidth,TotalHeight;
extern int			WindowX,WindowY;
extern int			ButWidth,ButHeight,FixWidth,DCHYPos,FSEYPos,FixYPos,DecoYPos,KeyYPos;
extern int			KeyWidth,KeyXOffset,KeyXOffset2;
extern int     		BigFontY,SmallFontY,ToolTipFontY;
extern int			SizeX,SizeY;
extern int			StkSiz;
extern int			MinStkSiz;
extern int			DisplayStkSiz;

extern BOOL			SoundOn,ToolTip,ZeroOverflow,CopyTop,InvTrigAsDMS,MessagePopup,AllowEdit,SaveRegs,IsBin,ThousandSep,ThousandthSep,Restart,NewKB,RefreshNow;
extern BOOL			Moving, Sizing, ESC_Pressed;
extern HWND			hwndMoving;

// We keep separate copies of these variables for Norm/Mini, and one in use:
extern BOOL			Mini;

extern BOOL			m_InTray;
extern int			m_Size,m_WindowX,m_WindowY,m_PopupX,m_PopupY,m_EnableMouseMove;
extern int			m_PopupSize,m_DisplayStkSiz;
extern BOOL			m_UseFonts,m_Bold,m_KeepInFront;
extern char			m_FontName[];
extern char			m_BitmapName[];
extern BOOL			m_IconStart;
extern BOOL			m_ShowMemoryPopup;
extern int			m_MemCols;
extern BOOL			m_ShowTitle,m_ShowPane1,m_ShowPane2,m_ShowPane3,m_ShowPane4,m_ShowPane5;

extern BOOL			n_InTray;
extern int			n_Size,n_WindowX,n_WindowY,n_PopupX,n_PopupY,n_EnableMouseMove;
extern int			n_PopupSize,n_DisplayStkSiz;
extern BOOL			n_UseFonts,n_Bold,n_KeepInFront;
extern char			n_FontName[];
extern char			n_BitmapName[];
extern BOOL			n_IconStart;
extern BOOL			n_ShowMemoryPopup;
extern int			n_MemCols;
extern BOOL			n_ShowTitle,n_ShowPane1,n_ShowPane2,n_ShowPane3,n_ShowPane4,n_ShowPane5;

extern BOOL			InTray;
extern int			Size,EnableMouseMove;
extern int			PopupX,PopupY;
extern int			PopupDX,PopupDY;
extern int			PopupSize;
extern BOOL			UseFonts,Bold,KeepInFront;
extern char			FontName[];
extern char			BitmapName[];
extern BOOL			ShowMemoryPopup;
extern int			MemRows,MemCols;
extern BOOL 		Title,Pane1,Pane2,Pane3,Pane4,Pane5;

//--Math Global Variables------------------------------------------------

extern int			Error;						// Is set by matherr() and other routines
extern int			iState;           			// 0 for normal, !=0 for inkey operations
extern int			iMaxDigits;         		// For current mode of operation
extern UNDOTYPE		Undone;						// Last undone operation
extern WORD			wDispMode;        			// DCHOB | FSE | DR
extern short		sWordSize;					// 0:32bit, 1:16bit, 2:8bit
extern WORD			wRadix;						// 10 for decimal etc..
extern WORD			wDecDR;						// Saves D/R mode for Decimal setting
extern char			cST[MAXSTKSIZ][NDIGITS+1];	// Display output buffer
extern char			STR[NDIGITS+2];				// Input key buffer
extern short		sFix;						// -1 for float
extern BOOL			bCleanFrac;					// TRUE for 'd/c' not 'a b/c'
extern BOOL 		bSL;              			// Stacklift true/false
extern BOOL			bShift;						// TRUE for shift key down
extern BOOL			Dirty;						// TRUE if all stack needs update
extern BOOL			k_useComma;					// TRUE for keyboard decimal comma
extern BOOL			d_useComma;					// TRUE for display decimal comma
extern BOOL			c_useComma;					// TRUE for clipboard decimal comma
extern BOOL			Cstyle;						// TRUE to use 0x.. clipboard copy
extern Register		ST[MAXNUMREGS];   			// Topmost element is LastX
extern Register		SaveReg;            		// Temp register for LastX
extern Register		Memory[MEMSIZ];				// Memory registers
extern LD			ldConstant;					// Temporary used to paste constants
extern BOOL			bHyperbolic;				// Hyperboli key fudge
//--Accutate PI across versions-------------------------------------------
#ifdef __BORLANDC__
extern const LD		* const M_PI_L_PTR;			// PI defined as bit pattern in variable.c
#define M_PI_L		(*M_PI_L_PTR)
#else
//--Micro$oft not supplied constants-------------------------------------
#define M_PI_L		3.14159265358979
#define M_LN10		2.30258509299405
#endif
//--Undo/redo information------------------------------------------------

extern UndoInfoStruct UndoInfo[MAXUNDO];
extern int			UndoIndex;
//extern int			RedoMaxIndex;
extern const char	* const UndoMsg[];

//--Error message text---------------------------------------------------

extern const char	* const ErrMsg[];
extern char			* Message;

//--Remember selections for each conversion type-------------------------
extern short		ConvType;
extern short		From;
extern short		To;
extern short		SelectFrom[];
extern short		SelectTo[];

//--Function Prototypes by file------------------------------------------

//--xcalc.c--------------------------------------------------------------
int PASCAL WinMain(HINSTANCE, HINSTANCE, LPSTR, int);

//--xcwinpro.c-----------------------------------------------------------
LRESULT CALLBACK XCALCMainProc(HWND, UINT, WPARAM, LPARAM);
BOOL Undo(UNDOTYPE*);
void SaveUndo(UNDOTYPE,BOOL);
void CommitUndo(void);
void CreateRgns(void);
void DeleteRgns(void);
void CreatePensBrushes(void);
void DeletePensBrushes(void);
void CreateFonts(void);
void DeleteFonts(void);
void CreateDigitBitmap(HWND); // stretch and colour digit display bitmap
void DeleteDigitBitmap(void); // delete it...

//--sound.c--------------------------------------------------------------
void SoundReg(int reg);

//--dialogs.c------------------------------------------------------------
BOOL CALLBACK GeneralDlgProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK NormalDlgProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK MiniDlgProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK ColoursDlgProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam);
int DoPropertySheet(HWND hwndOwner);
int CALLBACK EnumFontProc(LPLOGFONT,LPTEXTMETRIC,int,LPARAM);
BOOL CALLBACK FatalDlgProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK ConvertDlgProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK EditConvertDlgProc(HWND, UINT, WPARAM, LPARAM);
void InitConversion(void);
void EndConversion(void);
BOOL DoConversion(void);
BOOL CALLBACK ConstantDlgProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK AboutDlgProc(HWND, UINT, WPARAM, LPARAM);
extern char ConstMsg[128];
extern char ConvMsg[128];
BOOL CALLBACK EditKeyDlgProc(HWND, UINT, WPARAM, LPARAM);
//BOOL CALLBACK EditKeysProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK EditProgramProc(HWND, UINT, WPARAM, LPARAM);

//--buttonp.c------------------------------------------------------------
LRESULT CALLBACK ButtonProc(HWND, UINT, WPARAM, LPARAM);

//--convert.c------------------------------------------------------------
BOOL Convert(BOOL);

//--displayp.c-----------------------------------------------------------
LRESULT CALLBACK DisplayProc(HWND, UINT, WPARAM, LPARAM);
//void DisplayDecimal(BOOL);
int  strindex(char, const char * const);
void MakeString(int iReg);
void mkstr(LD value,int maxsize,char *bstr);
void mkfstr(long n,long d,char *bstr);
//void mkdmsstr(long d,long m,LD s,char *bstr);
void mkdmsstr(LD v,char *bstr);

//--dotproc.c------------------------------------------------------------
LRESULT CALLBACK DotProc(HWND, UINT, WPARAM, LPARAM);

//--drawbut.c------------------------------------------------------------
void DrawButton(HDC, BOOL ButtonDown);
void DrawButtonText(HDC, RECT*, COLORREF, LPCSTR, int underline,int len);
void LButtonDown(HWND, LPARAM, HWND *ButtonDownHWND, BOOL *MouseDown);
int  LButtonUp(HWND, LPARAM, HWND *ButtonDownHWND, BOOL *MouseDown);
void MouseMove(HWND, LPARAM, HWND *ButtonDownHWND);

void DrawKey(HDC hdc,RECT *rect, BOOL ButtonDown, COLORREF colour,LPCSTR text,int symbol,int len);
void KeyLButtonDown(HWND hwnd, LPARAM lParam, HWND *ButtonDownHWND, BOOL *MouseDown);
BOOL KeyLButtonUp(HWND hwnd, LPARAM lParam, HWND *ButtonDownHWND, BOOL *MouseDown);
void KeyMouseDownMove(HWND hwnd, LPARAM lParam, HWND *ButtonDownHWND);

//--errorpro.c-----------------------------------------------------------
LRESULT CALLBACK ErrorProc(HWND, UINT, WPARAM, LPARAM);

//--fixproc.c------------------------------------------------------------
LRESULT CALLBACK FixProc(HWND, UINT, WPARAM, LPARAM);

//--getcolou.c-----------------------------------------------------------
void InitColour(void);        // reads colours off xcalc.ini
void WriteColour(void);       // writes colours to xcalc.ini
COLORREF GetColour(int);      // retrieves colour from index
void SetColour(int,COLORREF); // sets colour

//--keyproc.c------------------------------------------------------------
BOOL initKeys(void);
void freeKeys(void);
int sameKeys(int a, int b);
extern int KeyToEdit;
extern int ModeToEdit;
extern int WindowToEdit;
extern HWND HwndToEdit;
extern RECT RectToEdit;
extern key Key[NKEYID];
//extern int Keys[5][NKEYS];
extern int *Keys[5];
LRESULT CALLBACK KeyProc(HWND, UINT, WPARAM, LPARAM);
void KeyboardDecimal(BOOL);
void HyperbolicKeys(BOOL);
void QConvText(char *text);
void QConstText(char *text);
void SetDefaultKeyboard(int mode);

//--tooltip.c------------------------------------------------------------
void ToolTipEvent(HWND hwnd,UINT wMsg,WPARAM wParam,LPARAM lParam,RECT *prect,const char *text);
void HideToolTip(void);
LRESULT CALLBACK ToolTipProc(HWND, UINT, WPARAM, LPARAM);

//--makebrdr.c-----------------------------------------------------------
void MakeBorder(HDC, RECT*);

//--memory.c-------------------------------------------------------------
LRESULT CALLBACK MemoryProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK MemoryPopupProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK MemoryDisplayProc(HWND, UINT, WPARAM, LPARAM);
void FormatMemoryNumber(Register *mem, char *text);

//--readwrit.c-----------------------------------------------------------
void ReadRegisters(const char *const SaveFile);
void SaveRegisters(const char *const SaveFile);
void MakeFullPath(char *path, const char *file, const char *dir);
char *chext(char *path, const char *newext);

//--modeproc.c-----------------------------------------------------------
LRESULT CALLBACK ModeProc(HWND, UINT, WPARAM, LPARAM);

//--program.c------------------------------------------------------------
void RunProgram(int p);

//--mathfunc.c-----------------------------------------------------------
void FPE_Handler(void);
BOOL Evaluate(void);
void Reduce(Register *r);
void mkfrac(long *i,long *n,long *d);
int mkifrac(long *i,long *n,long *d);
void mkreal(Register *z);
BOOL isproperfrac(Register *z);
BOOL isint(Register *z);
BOOL isdms(Register *z);
int todms(Register *z);
LD getreal(Register *z);
void null(Register *z);
void set(char);		void app(char);		void delend(void);
void lift(void);	void drop(void);	void clrstkreg(int);
void clr(Register *);void insbeg(char);	void delbeg(void);
void toint(void);	void tofloat(void);	void xchg(void);
void mask(Register *r);
void setwl(void);
void rollup(void);	void rolldn(void);	void X__Y(void);	void R__I(void);
void CHS(void);

int Add(void);		int Sub(void);		int Mul(void);		int Div(void);
int Mod(void);		int Divf(void);		int Percent(void);
int Conjg(void);	int Abs(void); 		int Rcp(void);
int Cos(void); 		int Sin(void); 		int Tan(void);
int Acos(void);		int Asin(void); 	int Atan(void);
int Cosh(void); 	int Sinh(void); 	int Tanh(void);
int Acosh(void);	int Asinh(void); 	int Atanh(void);
int Power(void);	int Root(void);		int Sqrt(void); 	int Sqr(void);	int Cube(void);
int Exp(void);		int Ln(void); 		int Ten(void); 		int Log(void);
int NOT(void);		int AND(void);		int OR(void);		int XOR(void);
int LShift(void);	int LRot(void);		int LAShift(void);
int RShift(void);	int RRot(void);		int RAShift(void);
int Fact(void);		int Cart(void);		int Polar(void);
int Qroot(void);	int Rand(void);		int Frac(void);	int Int(void);
int DMS(void);	/*	int DDD(void);	*/	int Round(void);

int ypowx	(Register *,Register *,Register *);
int yrootx	(Register *,Register *,Register *);
int ymulx	(Register *,Register *,Register *);
int ydivx	(Register *,Register *,Register *);
int ymodxf	(Register *,Register *,Register *);
int ydivxf	(Register *,Register *,Register *);
int rcpx	(Register *,Register *);
int etox	(Register *,Register *);
int lnx		(Register *,Register *);
int sqrtx	(Register *,Register *);
int sqrx	(Register *,Register *);
int cubex	(Register *,Register *);
int cosx	(Register *,Register *);
int sinx	(Register *,Register *);
int tanx	(Register *,Register *);
int acosx	(Register *,Register *);
int asinx	(Register *,Register *);
int atanx	(Register *,Register *);
int coshx	(Register *,Register *);
int sinhx	(Register *,Register *);
int tanhx	(Register *,Register *);
int acoshx	(Register *,Register *);
int asinhx	(Register *,Register *);
int atanhx	(Register *,Register *);

//--Error constants------------------------------------------------------

#define	__NOERROR		0		// No error
#define	__DIVBYZERO		1		// Division by zero
#define	__RANGE			2		// Overflow
#define	__DOMAIN		3		// log 0 and such
#define	__SINGULARITY	4		// singulariy
#define	__TLOSS			5		// total loss of significant digits
#define	__OTHER			6		// anything else
#define	__NOUNDO		7		// no more undo information
#define	__UNDONE		8		// undone (this is not an error)
//#define __NOREDO		9		// no more redo information
//#define __REDONE		10		// redone (also not an error)
#define __MESSAGE		11		// message (also not an error)

//--Indices for the display colours - see getcolou.c---------------------

#define NCOLOURS		11		// Number of different customisable colours
#define RGBBACK		    0		// background
#define RGBBORDER	    1		// border
#define RGBBRIGHT	    2		// 3d bright
#define RGBDARK		    3		// 3d dark
#define RGBBUTTON	    4		// buttontext
#define RGBGRAYBUTTON	5		// graybuttontext
#define RGBSELECTED		6		// selectedbutton
#define RGBMESSAGE		7		// message
#define RGBCAPTION		8		// caption
#define RGBKEYCAPS		9		// keycaps
#define RGBDIGITS		10		// digits

//--Constants for various windows----------------------------------------

#define DECWND			0
#define CPXWND			1
#define HEXWND			2
#define OCTWND			3
#define BINWND			4
#define FIXWND			5
#define SCIWND			6
#define ENGWND			7
#define DEGWND			8
#define RADWND			9

#define RADIXMASK		0x00ff	// All modes are ORed and kept
#define DECIMAL			0x0001	// in wDispMode
#define COMPLEX			0x0002
#define HEX				0x0004
#define OCTAL			0x0008
#define BINARY			0x0010

#define FIXMASK			0x0f00
#define FIX				0x0100
#define SCI				0x0200
#define ENG				0x0400

#define ANGLEMASK		0xf000
#define DEG				0x1000
#define	RAD				0x2000

//--Convenience macros---------------------------------------------------

int radix_lin(int mode);
int fix_lin(int mode);
int angle_lin(int mode);

#define RADIX_LIN		(radix_lin(wDispMode))	// linear radix: DEC=0, CPX=1,...
#define CURR_RADIX		(wDispMode&RADIXMASK)
#define CURR_FIX		(wDispMode&FIXMASK)
#define CURR_ANGLE		(wDispMode&ANGLEMASK)
#define CURR_MASK		(aWordMask[sWordSize])
#define CURR_BITS		(aBits[sWordSize])
#define CURR_MAX		(aBinMax[sWordSize])
#define CURR_MIN		(aBinMin[sWordSize])

//--Text for help contexts-----------------------------------------------

#define MAX_HELP		64
#define FATALHELP		"::/html/Fatal.htm>Main"
#define GENERALHELP		"::/html/General.htm>Main"
#define NORMALHELP		"::/html/Normal.htm>Main"
#define MINIHELP		"::/html/Mini.htm>Main"
#define COLOURSHELP		"::/html/Colours.htm>Main"
#define CONSTANTSHELP	"::/html/Constants.htm>Main"
#define CONVERSIONHELP	"::/html/UnitConversion.htm>Main"
#define CONVSETUPHELP	"::/html/ConversionSetup.htm>Main"
#define	EDITKEYHELP		"::/html/KeyboardEdit.htm>Main"
#define EDITKEYSHELP	"::/html/EditKeys.htm>Main"
#define PROGRAMHELP		"::/html/Programming.htm>Main"
#define DECIMALHELP		">QuickDecimal"
#define COMPLEXHELP		">QuickComplex"
#define BINARYHELP		">QuickBinary"
#define CODESHELP		">ProgramCodes"

#endif // _XCALC_H
