//------------------------------------------------------------------------
// File for projects XCALC_A and XCALC_B
//------------------------------------------------------------------------
// Routines in this file:
//	XCALCMainProc
//	Undo
//	SaveUndo
//	CreatePensBrushes
//	DeletePensBrushes
//	CreateDigitBitmap
//	DeleteDigitBitmap
//------------------------------------------------------------------------

#include "xcalc.h"

#ifndef SY
#define SY "Symbol"
#endif
#ifndef TT
#define TT "MS Sans Serif"
#define TTSIZE -8
#else
#define TTSIZE -12
#endif

//--Variables defined and used only in this file--------------------------

// ErrSound defines one sound per error condition:
static UINT ErrSound[] =
{
	0,					// no error
	MB_ICONEXCLAMATION, // __DIVBYZERO
	MB_ICONASTERISK,	// __RANGE
	MB_ICONQUESTION,	// __DOMAIN
	MB_ICONASTERISK,  	// __SINGULARITY
	MB_ICONQUESTION,	// __TLOSS
	MB_ICONQUESTION,	// __OTHER
	MB_OK, 				// __NOUNDO
	0,					// __UNDONE
	//MB_OK,				// __NOREDO
	//0,					// __REDONE
	0					// __MESSAGE
};

// Used to switch keyboard maps
static const char dPoint[] = {'.',','};
static const char Comma[] = {',','.'};

// Sizing corners
static HRGN hRightRgn,hLeftRgn;

//--Undo function---------------------------------------------------------
// Actions:
// - If undo information exists, change back to previous state.
//   The item-by-item check saves unnecessary screen updates.
//------------------------------------------------------------------------

BOOL Undo(UNDOTYPE *Undone)
{
	int i;
	
	if (UndoIndex>0) {
		// Undo information exists - proceed
		UndoIndex--;
		// Get what is undone
		*Undone = UndoInfo[UndoIndex].UndoType;
		// Always restore previous stacklift content
		bSL = UndoInfo[UndoIndex].StackLift;
		// Check stack changes, and update if any
		for (i=0;i<MAXNUMREGS;i++) {
			if (memcmp(&ST[i],&UndoInfo[UndoIndex].Stack[i],sizeof(Register))) {
				ST[i] = UndoInfo[UndoIndex].Stack[i];
				// Only update visible registers
				if (i<DisplayStkSiz) SendMessage(hwndXCALC,PM_UPDATE1,i,0);
			}
		}
		// If new fix settings, update that
		if (sFix != UndoInfo[UndoIndex].Fix) {
			sFix = UndoInfo[UndoIndex].Fix;
			SendMessage(hwndXCALC,PM_SETFIX,sFix,0);
		}
		// Always restore Decimal Radix
		wDecDR = UndoInfo[UndoIndex].DecDR;
		// If any mode setting change, update all of them
		if (wDispMode != UndoInfo[UndoIndex].DispMode
			||wRadix != UndoInfo[UndoIndex].Radix
			||sWordSize != UndoInfo[UndoIndex].sWordSize) {
			wRadix = UndoInfo[UndoIndex].Radix;
			wDispMode = UndoInfo[UndoIndex].DispMode;
			sWordSize = UndoInfo[UndoIndex].sWordSize;
			SendMessage(hwndXCALC,PM_SETMODE,wDispMode,0); // All modes
		}
		// Check memory changes, and update if any
		for (i=0;i<MEMSIZ;i++) {
			if (memcmp(&Memory[i],&UndoInfo[UndoIndex].Memory[i],sizeof(Register))) {
				Memory[i] = UndoInfo[UndoIndex].Memory[i];
				SendMessage(hMemoryDisplay[i],PM_UPDATE,0,0);
				SendMessage(hMemory,PM_UPDATE,i,0);
			}
		}
		// Update conversion settings
		ConvType = UndoInfo[UndoIndex].ConvType;
		From = UndoInfo[UndoIndex].From;
		To = UndoInfo[UndoIndex].To;
		memcpy(SelectFrom,UndoInfo[UndoIndex].SelectFrom,NCONVERSIONS*sizeof(short));
		memcpy(SelectTo,UndoInfo[UndoIndex].SelectTo,NCONVERSIONS*sizeof(short));
		// Update last constant
		ldConstant = UndoInfo[UndoIndex].ldConstant;
		return TRUE;
	}
	else {
		// No undo information - return FALSE to set error
		return FALSE;
	}
}

//--Redo function---------------------------------------------------------
// Actions:
// - If an Undo is the last action, redo by changing to that state again.
//   The item-by-item check saves unnecessary screen updates.
//------------------------------------------------------------------------
/*
BOOL Redo(UNDOTYPE *Undone)
{
	int i;
	
	if (UndoIndex>0) {
		// Undo information exists - proceed
		UndoIndex--;
		// Get what is undone
		*Undone = UndoInfo[UndoIndex].UndoType;
		// Always restore previous stacklift content
		bSL = UndoInfo[UndoIndex].StackLift;
		// Check stack changes, and update if any
		for (i=0;i<MAXNUMREGS;i++) {
			if (memcmp(&ST[i],&UndoInfo[UndoIndex].Stack[i],sizeof(Register))) {
				ST[i] = UndoInfo[UndoIndex].Stack[i];
				// Only update visible registers
				if (i<DisplayStkSiz) SendMessage(hwndXCALC,PM_UPDATE1,i,0);
			}
		}
		// If new fix settings, update that
		if (sFix != UndoInfo[UndoIndex].Fix) {
			sFix = UndoInfo[UndoIndex].Fix;
			SendMessage(hwndXCALC,PM_SETFIX,sFix,0);
		}
		// Always restore Decimal Radix
		wDecDR = UndoInfo[UndoIndex].DecDR;
		// If any mode setting change, update all of them
		if (wDispMode != UndoInfo[UndoIndex].DispMode
			||wRadix != UndoInfo[UndoIndex].Radix) {
			wRadix = UndoInfo[UndoIndex].Radix;
			wDispMode = UndoInfo[UndoIndex].DispMode;
			SendMessage(hwndXCALC,PM_SETMODE,wDispMode,0); // All modes
		}
		// Check memory changes, and update if any
		for (i=0;i<MEMSIZ;i++) {
			if (memcmp(&Memory[i],&UndoInfo[UndoIndex].Memory[i],sizeof(Register))) {
				Memory[i] = UndoInfo[UndoIndex].Memory[i];
				SendMessage(hMemoryDisplay[i],PM_UPDATE,0,0);
				SendMessage(hMemory,PM_UPDATE,i,0);
			}
		}
		// Update conversion settings
		ConvType = UndoInfo[UndoIndex].ConvType;
		From = UndoInfo[UndoIndex].From;
		To = UndoInfo[UndoIndex].To;
		memcpy(SelectFrom,UndoInfo[UndoIndex].SelectFrom,NCONVERSIONS*sizeof(short));
		memcpy(SelectTo,UndoInfo[UndoIndex].SelectTo,NCONVERSIONS*sizeof(short));
		// Update last constant
		ldConstant = UndoInfo[UndoIndex].ldConstant;
		
		return TRUE;
	}
	else {
		// No undo information - return FALSE to set error
		return FALSE;
	}
}
*/
//--SaveUndo function-----------------------------------------------------
// Save all information to temporary storage, and commit if Commit=TRUE.
//------------------------------------------------------------------------

UndoInfoStruct UndoTemp;

void SaveUndo(UNDOTYPE type,BOOL Commit)
{
	UndoTemp.UndoType = type;
	memcpy(UndoTemp.Memory,Memory,MEMSIZ*sizeof(Register));
	memcpy(UndoTemp.Stack,ST,MAXNUMREGS*sizeof(Register));
	UndoTemp.StackLift = bSL;
	UndoTemp.Fix = sFix;
	UndoTemp.sWordSize = sWordSize;
	UndoTemp.DispMode = wDispMode;
	UndoTemp.Radix = wRadix;
	UndoTemp.DecDR = wDecDR;
	// Update conversion settings
	UndoTemp.ConvType = ConvType;
	UndoTemp.From = From;
	UndoTemp.To = To;
	memcpy(UndoTemp.SelectFrom,SelectFrom,NCONVERSIONS*sizeof(short));
	memcpy(UndoTemp.SelectTo,SelectTo,NCONVERSIONS*sizeof(short));
	// Update last constant
	UndoTemp.ldConstant = ldConstant;
	if (Commit) CommitUndo();
}

//--CommitUndo function---------------------------------------------------
// Actions:
// - Save undo information on stack, if necessary push stack downwards
// 	 to make room.
//------------------------------------------------------------------------

void CommitUndo()
{
	if (UndoIndex==MAXUNDO)
		memmove(&UndoInfo[0],&UndoInfo[1],(MAXUNDO-1)*sizeof(UndoInfoStruct));
	else
		UndoIndex++;

	UndoInfo[UndoIndex-1] = UndoTemp;
}

//--DeleteRgns------------------------------------------------------------
// Actions:
// - deletes global size regions
//	 Used when size changes
//------------------------------------------------------------------------

void DeleteRgns()
{
	DeleteObject(hRightRgn);
	DeleteObject(hLeftRgn);
}

//--CreateRgns------------------------------------------------------------
// Actions:
// - creates global size regions
//	 Used when size changes
//------------------------------------------------------------------------

void CreateRgns()
{
	POINT Corner[6];
	Corner[0].x = ClientWidth;			Corner[0].y = ClientHeight;
	Corner[1].x = ClientWidth-SizeX;	Corner[1].y = ClientHeight;
	Corner[2].x = ClientWidth-SizeX;	Corner[2].y = ClientHeight-Border;
	Corner[3].x = ClientWidth-Border;	Corner[3].y = ClientHeight-Border;
	Corner[4].x = ClientWidth-Border;	Corner[4].y = ClientHeight-SizeY;
	Corner[5].x = ClientWidth;			Corner[5].y = ClientHeight-SizeY;
	hRightRgn = CreatePolygonRgn(Corner,6,WINDING);
	
	Corner[0].x = 0;					Corner[0].y = ClientHeight;
	Corner[1].x = SizeX;				Corner[1].y = ClientHeight;
	Corner[2].x = SizeX;		 		Corner[2].y = ClientHeight-Border;
	Corner[3].x = Border;				Corner[3].y = ClientHeight-Border;
	Corner[4].x = Border;				Corner[4].y = ClientHeight-SizeY;
	Corner[5].x = 0;					Corner[5].y = ClientHeight-SizeY;
	hLeftRgn = CreatePolygonRgn(Corner,6,WINDING);
}

//--DeletePensBrushes-----------------------------------------------------
// Actions:
// - deletes all globally defined pen and brush objects.
//	 Used when colours change.
//------------------------------------------------------------------------

void DeletePensBrushes()
{
	DeleteObject(hDarkPen);
	DeleteObject(hBrightPen);
	DeleteObject(hSelectedPen);
	DeleteObject(hCaptionPen);
	
	DeleteObject(hBackBrush);
	DeleteObject(hBorderBrush);
	DeleteObject(hSelectedBrush);
	DeleteObject(hCaptionBrush);
}

//--CreatePensBrushes-----------------------------------------------------
// Actions:
// - defines global pen and brush objects. Used when colours change.
//------------------------------------------------------------------------

void CreatePensBrushes()
{
	hDarkPen = CreatePen(PS_SOLID,1,GetColour(RGBDARK));
	hBrightPen = CreatePen(PS_SOLID,1,GetColour(RGBBRIGHT));
	hSelectedPen = CreatePen(PS_SOLID,1,GetColour(RGBSELECTED));
	hCaptionPen = CreatePen(PS_SOLID,1,GetColour(RGBCAPTION));
	
	hBackBrush = CreateSolidBrush(GetColour(RGBBACK));
	hBorderBrush = CreateSolidBrush(GetColour(RGBBORDER));
	hSelectedBrush = CreateSolidBrush(GetColour(RGBSELECTED));
	hCaptionBrush = CreateSolidBrush(GetColour(RGBCAPTION));
}

//--DeleteFonts-----------------------------------------------------------
// Actions:
// - deletes global fonts. Used when size changes and at shutdown
//------------------------------------------------------------------------

void DeleteFonts()
{
	DeleteObject(hSmallFont);
	DeleteObject(hSmallFont_);
	DeleteObject(hBigFont);
	DeleteObject(hBigFont_);
	DeleteObject(hSymbolFont);
	DeleteObject(hToolTipFont);
}

//--CreateFonts-----------------------------------------------------------
// Actions:
// - defines global fonts. Used when size changes.
//------------------------------------------------------------------------

void CreateFonts()
{
	int BigY = BitmapY - Border;
	int SmallY = BitmapY - 2*Border;
	hBigFont = CreateFont(BigY,0,0,0,0,0,0,0,
		ANSI_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		DEFAULT_PITCH | FF_SWISS,
		"Arial");
	hBigFont_ = CreateFont(BigY,0,0,0,0,0,1,0,
		ANSI_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		DEFAULT_PITCH | FF_SWISS,
		"Arial");
	hSmallFont = CreateFont(SmallY,0,0,0,0,0,0,0,
		ANSI_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		DEFAULT_PITCH | FF_SWISS,
		"Arial");
	hSmallFont_ = CreateFont(SmallY,0,0,0,0,0,1,0,
		ANSI_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		DEFAULT_PITCH | FF_SWISS,
		"Arial");
	// At least on my system, the arial font is suddenly made
	// bulkier at size 21. The symbol font is not. Therefore,
	// we do it manually.
	hSymbolFont = CreateFont(BigY,0,0,0,BigY>20?FW_BOLD:0,0,0,0,
		SYMBOL_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		DEFAULT_PITCH,
		SY);
	hToolTipFont = CreateFont(TTSIZE,0,0,0,0,0,0,0,
		ANSI_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		DEFAULT_PITCH | FF_SWISS,
		TT);
}

//--DeleteDigitBitmap-----------------------------------------------------
// Actions:
// - unselects bitmap from existing hDigitsDC and deletes the bitmap object
//------------------------------------------------------------------------

void DeleteDigitBitmap(void)
{
	SelectObject(hDigitsDC,hbOld);
	DeleteObject(hbDigits);
}

//--CreateDigitBitmap-----------------------------------------------------
// Actions:
// - gets DC from hwnd (the handle is always hwndXCALC, but global variable
//   not yet set correctly when rebuilding; therefore parameter)
// - Selects bitmap or font for digits
// - stretches and colours the selected digit bitmap into existing hDigitsDC
// - releases DC for hwndXCALC
//------------------------------------------------------------------------

void CreateDigitBitmap(HWND hwnd)
{
	// Uses global extern variables:
	// hInst, BitmapName, DefaultBitmap, BitmapX, BitmapY, Size, hDigitsDC
	// Printable
	// Global variables in this file:
	// hbOld, hbDigits
	
	BITMAP  bitmap;
	HBITMAP hTempBitmap, hOldTempBitmap;
	HDC     hdc, hTempDC;
	int i;
	
	// this dc used to create bitmap images
	hdc = GetDC(hwnd);
	
	hTempDC = CreateCompatibleDC(hdc);
	
	if (UseFonts) {
		SIZE siz;
		HFONT hFont, hOldFont;
		RECT rect;
		BOOL Found;
		COLORREF oldtextcolour,oldbkcolour;
		int MapModePrevious;
		int FontSize;
		int x,y, DigitTop,DigitBottom;
		char OneChar[2] = "0";
		
		FontSize = MulDiv(Size,18,100); // 100% means 18pt font
		
		MapModePrevious = SetMapMode(hTempDC, MM_TEXT);
		hFont = CreateFont(
			-FontSize,     // height - negative to ignore leading
			0,             // width
			0,             // escapement
			0,             // orientation
			Bold?FW_BOLD:FW_NORMAL, // weight
			0,             // italic
			0,             // underline
			0,             // strikeout
			ANSI_CHARSET,  // character set
			OUT_DEFAULT_PRECIS,  // output precision
			CLIP_DEFAULT_PRECIS, // clip precision
			PROOF_QUALITY,       // output quality
			0,                   // pitch and family
			FontName);      // typeface name
		
		hOldFont = (HFONT)SelectObject(hTempDC, hFont);
		
		BitmapX = 0;
		BitmapY = 0;
		for (i=0;i<NPRINTABLE;i++) {
			OneChar[0] = Printable[i];
			GetTextExtentPoint32(hTempDC,OneChar,1,&siz);
			BitmapX = max((long)BitmapX,siz.cx);
			BitmapY = max((long)BitmapY,siz.cy);
		}
		
		hTempBitmap = CreateBitmap(BitmapX*NPRINTABLE,BitmapY,1,1,NULL);
		hOldTempBitmap = (HBITMAP)SelectObject(hTempDC,hTempBitmap);
		SetRect(&rect,0,0,BitmapX*NPRINTABLE+1,BitmapY+1);
		
		// Draw White-on-Black text; clip each character within its rectangle -
		// some bold fonts seem to cheat!
		FillRect(hTempDC,&rect,(HBRUSH)GetStockObject(BLACK_BRUSH));
		oldtextcolour = SetTextColor(hTempDC,RGB(255,255,255));
		oldbkcolour = SetBkColor(hTempDC,RGB(0,0,0));
		for (i=0;i<NPRINTABLE;i++) {
			SetRect(&rect,i*BitmapX,0,(i+1)*BitmapX,BitmapY);
			ExtTextOut(hTempDC,i*BitmapX,0,ETO_CLIPPED,&rect,Printable+i,1,NULL);
		}
		SetTextColor(hTempDC,oldtextcolour);
		SetBkColor(hTempDC,oldbkcolour);
		
		// now find the first line in the bitmap != black
		
		Found = FALSE;
		for (y=0;!Found;y++) {
			for (x=0;x<BitmapX*NPRINTABLE;x++) {
				if (GetPixel(hTempDC,x,y)==RGB(255,255,255)) {
					Found = TRUE;
					break;
				}
			}
		}
		DigitTop = y-1; // y++ is executed once too many in for structure
		
		// now find the last line in the bitmap != black
		
		Found = FALSE;
		for (y=BitmapY-1;!Found;y--) {
			for (x=0;x<BitmapX*NPRINTABLE;x++) {
				if (GetPixel(hTempDC,x,y)==RGB(255,255,255)) {
					Found = TRUE;
					break;
				}
			}
		}
		DigitBottom = y+1; // same comment about for loop
		
		BitmapY = DigitBottom-DigitTop+1; // fit exactly
		
		hbDigits = CreateCompatibleBitmap(hdc,BitmapX*NPRINTABLE,BitmapY);
		hbOld = (HBITMAP)SelectObject(hDigitsDC,hbDigits);
		
		SetBkColor(hDigitsDC,GetColour(RGBDIGITS));  // use global background for digits
		SetTextColor(hDigitsDC,GetColour(RGBBACK));  // use global digit colour
		
		BitBlt(hDigitsDC,0,0,BitmapX*NPRINTABLE,BitmapY,hTempDC,0,DigitTop,SRCCOPY);
		
		SetMapMode(hTempDC, MapModePrevious);
		SelectObject(hTempDC, hOldFont);
		SelectObject(hTempDC, hOldTempBitmap);
		DeleteObject(hFont);
		DeleteObject(hTempBitmap);
	}
	else { // not UseFonts	
		// load display bitmap and stretch to fit dimensions
		if ((hTempBitmap = LoadBitmap(hInst,BitmapName)) == NULL) {
			hTempBitmap = LoadBitmap(hInst,DefaultBitmap);
		}
		GetObject(hTempBitmap,sizeof(bitmap),(LPVOID)&bitmap);
		BitmapX = MulDiv(Size,bitmap.bmWidth/NPRINTABLE,100);
		BitmapY = MulDiv(Size,bitmap.bmHeight,100);
		
		hbDigits = CreateCompatibleBitmap(hdc,BitmapX*NPRINTABLE,BitmapY);
		
		hOldTempBitmap = (HBITMAP)SelectObject(hTempDC,hTempBitmap);
		hbOld = (HBITMAP)SelectObject(hDigitsDC,hbDigits);
		
		SetStretchBltMode(hTempDC,STRETCH_ORSCANS);
		
		// note: here we have an inverse colourisation - the bitmaps should really have
		// been black digits on a white background. Then we would have set BkColor to
		// RGBBACK and TextColor to RBGDIGITS. However, this works...
		
		SetBkColor(hDigitsDC,GetColour(RGBDIGITS));  // use global background for digits
		SetTextColor(hDigitsDC,GetColour(RGBBACK));  // use global digit colour
		StretchBlt(hDigitsDC,0,0,BitmapX*NPRINTABLE,BitmapY,
			hTempDC,0,0,bitmap.bmWidth,bitmap.bmHeight,
			SRCCOPY);
		
		SelectObject(hTempDC,hOldTempBitmap);
		DeleteObject(hTempBitmap);
	}
	
	// set border width based on fairly crude guesswork
	
	Border = 4;
	if (BitmapY<25) Border = 3;
	if (BitmapY<20) Border = 2;
	if (BitmapY<15) Border = 1;
			
	DeleteDC(hTempDC);
	ReleaseDC(hwnd,hdc);
}

//--Convenience declarations and routines---------------------------------

void update(void)
{
	PostMessage(hwndXCALC,Dirty?PM_UPDATEALL:PM_UPDATE1,Xreg,0);
	Dirty=FALSE;
}

void ClearErr(void)
{
	Error = FALSE;
	PostMessage(hError,PM_UPDATE,0,0);
}

#define eval() {if (Evaluate()) break; else update();}
#define SetLastX() {ST[Lreg] = ST[Xreg];}

//------------------------------------------------------------------------
// Main Window procedure
// Actions:
//	- process input and display results
//    This routine contains the keyboard input.
//	- start Help screen on demand
//------------------------------------------------------------------------

// These variables are global and declared extern in XCALC.H:

BOOL Sizing=FALSE, ESC_Pressed;

LRESULT CALLBACK XCALCMainProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	static BOOL		InRgtCorner,InLftCorner;
	static POINT	origpt;
	static POINTS	lastpts;
	static RECT		origrect;
	static int		origSize;

	switch (wMsg)
	{
	/* Later? - frequires drag&drop
	case PM_EDITKEYS:
		if (!IsWindow(hwndEditKeys)) {
			hwndEditKeys = CreateDialog(hInst, "XCALCEDITKEYS", hwnd, (DLGPROC)EditKeysProc); 
			ShowWindow(hwndEditKeys, SW_SHOW); 
		} 
		break; 
	*/
	case WM_SETTINGCHANGE:
	{
		// TODO - react to screen size changes (and probably other settings)
	}
    break;
	case PM_EDITPROGRAM:
		if (!IsWindow(hwndEditProgram)) {
			hwndEditProgram = CreateDialog(hInst, "XCALCEDITPROGRAM", hwnd, (DLGPROC)EditProgramProc); 
			ShowWindow(hwndEditProgram, SW_SHOW); 
		} 
		break; 
	case WM_NCRBUTTONDOWN:
		// Do nothing -- prevents system from displaying menu on WM_NCRBUTTONUP
		break;
	case WM_NCRBUTTONUP:
		if (wParam == HTCAPTION) {
			POINT pt;
			GetCursorPos(&pt);
			PostMessage(hwndXCALC,PM_SYSMENU,(WPARAM)pt.x,(LPARAM)pt.y);
		}
		break;
	case PM_ICONMESSAGE:
		{
			POINT pt;
			// Message is in lParam in this case
			switch (lParam)
			{
			case WM_LBUTTONDOWN:
				ShowWindow(hwndXCALC,SW_SHOWNORMAL);
				SetForegroundWindow(hwndXCALC);
				if (ShowMemoryPopup) ShowWindow(hMemoryPopup,SW_SHOWNORMAL);
				break;
			case WM_RBUTTONDOWN:
				GetCursorPos(&pt);
				PostMessage(hwndXCALC,PM_SYSMENU,(WPARAM)pt.x,(LPARAM)pt.y);
				break;
			}
			break;
		}

	case WM_DESTROY:
	case WM_QUERYENDSESSION:
		{
			NOTIFYICONDATA nid;

			// save registers if SaveRegs is TRUE
			if (SaveRegs) SaveRegisters(DatFile);
			
			if (!Restart) HtmlHelp(NULL,NULL,HH_CLOSE_ALL,0);
			
			DeletePensBrushes();
			DeleteDigitBitmap();
			
			DeleteObject(hbVersion);
			DeleteObject(hbStatus);
			DeleteDC(hDigitsDC);

			DeleteFonts();
			
			DeleteRgns();

			// Hide tray icon
			nid.cbSize=sizeof(nid);
			nid.hWnd=hwndXCALC;
			nid.uID=0;
			Shell_NotifyIcon(NIM_DELETE,&nid);

			PostQuitMessage(0);

			// We make both these messages stop the application (thereby saving all settings),
			// but the return value from the Winproc must be different.
			if (wMsg==WM_QUERYENDSESSION)
				return TRUE;
			else
				return 0;

			//break;
		}
		
	case WM_CREATE:		// set up all static variables
		{
			//int BigY,SmallY;
			BITMAP bitmap;
			HDC hdc;
			
			// create global hDigitsDC to store digit bitmap
			hdc = GetDC(hwndXCALC);
			hDigitsDC = CreateCompatibleDC(hdc);
			ReleaseDC(hwndXCALC,hdc);
			
			// create scaled, coloured bitmap for digit display
			CreateDigitBitmap(hwnd);
			
			// create resize regions
			CreateRgns();
			
			// load version number bitmap and get dimensions
			hbVersion = LoadBitmap(hInst,VersionBitmap);
			GetObject(hbVersion,sizeof(bitmap),(LPVOID)&bitmap);
			LogoX = bitmap.bmWidth;
			LogoY = bitmap.bmHeight;
			
			// load status bitmap and get dimensions
			hbStatus = LoadBitmap(hInst,StatusBitmap);
			GetObject(hbStatus,sizeof(bitmap),(LPVOID)&bitmap);
			StatusY = bitmap.bmHeight;
			StatusX = bitmap.bmWidth/11; // 11 subbmps in bitmap
			
			// create global pens and brushes
			CreatePensBrushes();
			
			// Create fonts for use in other windows
			CreateFonts();

			break;
		}
		
	// move window to saved position:
		
	case PM_SIZEWINDOW:
		{
			HDC hdc;
			SIZE siz;
			BOOL showChanges = wParam!=0;
			int oldTotalWidth = TotalWidth;
			int oldTotalHeight = TotalHeight;
	
			int ExtraX,ExtraY;

			// Calculate the sizes of borders and possible caption, based on
			// both selected display and the system. The extra is needed for both
			// Win95 and WinNT.
			
			if (Title) {
				ExtraX = 2*GetSystemMetrics(SM_CXFRAME)-2;
				//ExtraY = GetSystemMetrics(SM_CYSMCAPTION)+2*GetSystemMetrics(SM_CYFRAME)-2;
				//ExtraY = GetSystemMetrics(SM_CYCAPTION)+2*GetSystemMetrics(SM_CYFRAME)-2;
				ExtraY = GetSystemMetrics(InTray?SM_CYSMCAPTION:SM_CYCAPTION)+2*GetSystemMetrics(SM_CYFRAME)-2;
			}
			else {
				ExtraX = 2*GetSystemMetrics(SM_CXBORDER);
				ExtraY = 2*GetSystemMetrics(SM_CYBORDER);
			}

			DeleteDigitBitmap();
			CreateDigitBitmap(hwnd);

			// find the correct size for the display windows and borders
			
			DisplayX	= NDIGITS*BitmapX+2*Border+StatusX; // Everything based on bitmap size -
			DisplayY	= BitmapY+2*Border;         		// each display adds border around.
			
			BrdHeight	= DisplayY*DisplayStkSiz+Border*2; // Border around digit display.
			ButHeight	= BitmapY+Border;           // Height of any button.
			
			BrdWidth	= DisplayX+2*Border;        // Display border - initial value.
			ButWidth	= BrdWidth/NMODECOLS;       // Distribute mode buttons across
			if (ButWidth*NMODECOLS!=BrdWidth) {		// screen,
				ButWidth++;							// adjust total width to integer
				BrdWidth = NMODECOLS*(ButWidth);    // multiple of button width.
			}
			FixWidth 	= ButWidth					// The Fix window gets whatever is
				*(NMODECOLS-NBUTTONS);				// left on second button row, after
													// the rest of the buttons are
													// placed there.
			DisplayX	= BrdWidth-2*Border;        // Re-adjust display width.
			
			// KeyWidth is now adjusted so NKEYCOLS buttons fit inside border
			KeyWidth	= DisplayX/nKeyCols;
			KeyXOffset	= (BrdWidth-DisplayX)/2;
			KeyXOffset2 = (BrdWidth-Border-KeyWidth*nKeyCols-KeyXOffset)/2;
			KeyYPos		= DisplayY*DisplayStkSiz+Border*2; // Y position for keyboard
			DCHYPos		= KeyYPos+(nKeyRows*ButHeight+Border*2)*Pane1; // Y position for DCHOB buttons
			FSEYPos		= DCHYPos+ButHeight*Pane2;	// Y position for FSE and DR buttons
			FixYPos		= FSEYPos+ButHeight*Pane3;	// Y position for fix & other buttons
			DecoYPos	= FixYPos+ButHeight*Pane4;	// Y position for bottom window.
			
			ClientWidth	= BrdWidth;
			ClientHeight = BrdHeight+
				(ButHeight*nKeyRows+Border*2)*Pane1+
				ButHeight*(Pane2+Pane3+Pane4+2*Pane5);
			
			TotalWidth	= ClientWidth+ExtraX;		// Including Windows borders
			TotalHeight	= ClientHeight+ExtraY;

			// Skip further processing if no change
			if (TotalWidth == oldTotalWidth && TotalHeight == oldTotalHeight) return 0;

			DeleteFonts();
			CreateFonts();

			DeleteRgns();
			CreateRgns();

			// Find measures for fonts
			
			hdc = GetDC(hwnd);

			SelectObject(hdc,hBigFont);
			GetTextExtentPoint32(hdc,"X",1,&siz);
			BigFontY = siz.cy;
			SelectObject(hdc,hSmallFont);
			GetTextExtentPoint32(hdc,"X",1,&siz);
			SmallFontY = siz.cy;
			SelectObject(hdc,hToolTipFont);
			GetTextExtentPoint32(hdc,"X",1,&siz);
			ToolTipFontY = siz.cy;
			
			ReleaseDC(hwnd,hdc);
			
			SizeX = min(GetSystemMetrics(SM_CXCURSOR),ButHeight);
			SizeY = min(GetSystemMetrics(SM_CYCURSOR),ButHeight);

			// If called from WM_MOUSEMOVE (sizing), don't move it - it will happen there

			if (showChanges) SetWindowPos(hwnd,0,WindowX,WindowY,TotalWidth,TotalHeight,SWP_NOZORDER|SWP_NOACTIVATE);
			
			PostMessage(hwnd,PM_SETMODE,wDispMode,0); // All modes

			return 1;
		}

	// This message sent after active child windows are created and PM_SIZEWINDOW has returned

	case PM_SIZECHILDREN:
		{
			int i;
			BOOL defer = (BOOL)wParam;
			HDWP hdwp = 0;

			if (defer) {
				hdwp = BeginDeferWindowPos(100);
			}

			for (i=0;i<DisplayStkSiz;i++) {
				if (defer)
					hdwp = DeferWindowPos(hdwp,hDisplay[i],NULL,Border,Border+DisplayY*(DisplayStkSiz-1-i),DisplayX,DisplayY,SWP_NOZORDER);
				else
					SetWindowPos(hDisplay[i],NULL,Border,Border+DisplayY*(DisplayStkSiz-1-i),DisplayX,DisplayY,SWP_NOZORDER);
			}
			if (!Title) {
				if (defer)
					hdwp = DeferWindowPos(hdwp,hDot,NULL,0,0,StatusX,StatusY,SWP_NOZORDER);
				else
					SetWindowPos(hDot,NULL,0,0,StatusX,StatusY,SWP_NOZORDER);
			}
			if (Pane1) {
				for (i=0;i<nKeyRows*nKeyCols;i++) {
					if (defer)
						hdwp = DeferWindowPos(hdwp,hwndKey[i],NULL,
							KeyXOffset+KeyXOffset2*(i%nKeyCols>=nKeyCols/2)+KeyWidth*(i%nKeyCols),
							Border+KeyYPos+ButHeight*(i/nKeyCols),
							KeyWidth,ButHeight,SWP_NOZORDER);
					else
						SetWindowPos(hwndKey[i],NULL,
							KeyXOffset+KeyXOffset2*(i%nKeyCols>=nKeyCols/2)+KeyWidth*(i%nKeyCols),
							Border+KeyYPos+ButHeight*(i/nKeyCols),
							KeyWidth,ButHeight,SWP_NOZORDER);
				}
			}
			if (Pane2) {
				for (i=DECWND;i<=BINWND;i++) {
					if (defer)
						hdwp = DeferWindowPos(hdwp,hDispMode[i],NULL,ButWidth*(i%NMODECOLS),DCHYPos,ButWidth,ButHeight,SWP_NOZORDER);
					else
						SetWindowPos(hDispMode[i],NULL,ButWidth*(i%NMODECOLS),DCHYPos,ButWidth,ButHeight,SWP_NOZORDER);
				}
			}
			if (Pane3) {
				for (i=FIXWND;i<=RADWND;i++) {
					if (defer)
						hdwp = DeferWindowPos(hdwp,hDispMode[i],NULL,ButWidth*(i%NMODECOLS),FSEYPos,ButWidth,ButHeight,SWP_NOZORDER);
					else
						SetWindowPos(hDispMode[i],NULL,ButWidth*(i%NMODECOLS),FSEYPos,ButWidth,ButHeight,SWP_NOZORDER);
				}
			}
			if (Pane4) {
				if (defer) {
					hdwp = DeferWindowPos(hdwp,hFix,NULL,0,FixYPos,FixWidth,ButHeight,SWP_NOZORDER);
					for (i=0;i<NBUTTONS;i++) {
						hdwp = DeferWindowPos(hdwp,hButton[i],NULL,ButWidth*(i+NMODECOLS-NBUTTONS),FixYPos,ButWidth,ButHeight,SWP_NOZORDER);
					}
				}
				else {
					SetWindowPos(hFix,NULL,0,FixYPos,FixWidth,ButHeight,SWP_NOZORDER);
					for (i=0;i<NBUTTONS;i++) {
						SetWindowPos(hButton[i],NULL,ButWidth*(i+NMODECOLS-NBUTTONS),FixYPos,ButWidth,ButHeight,SWP_NOZORDER);
					}
				}
				PostMessage(hFix,PM_SIZEWINDOW,0,0);
			}
			if (Pane5) {
				if (defer) {
					hdwp = DeferWindowPos(hdwp,hError,NULL,Border,DecoYPos+ButHeight,TotalWidth/2,ButHeight-Border,SWP_NOZORDER);
					hdwp = DeferWindowPos(hdwp,hMemory,NULL,2*Border,DecoYPos+Border,ButWidth+(ButWidth>>2),ButHeight-Border,SWP_NOZORDER);
				}
				else {
					SetWindowPos(hError,NULL,Border,DecoYPos+ButHeight,TotalWidth/2,ButHeight-Border,SWP_NOZORDER);
					SetWindowPos(hMemory,NULL,2*Border,DecoYPos+Border,ButWidth+(ButWidth>>2),ButHeight-Border,SWP_NOZORDER);
				}
				PostMessage(hMemory,PM_SIZEWINDOW,0,0);
			}
			else {
				RECT rect;
				GetWindowRect(hwnd,&rect);
				if (defer) {
					hdwp = DeferWindowPos(hdwp,hError,NULL,rect.left,rect.bottom+0,rect.right-rect.left,ToolTipFontY+3,SWP_NOZORDER);
				}
				else {
					SetWindowPos(hError,NULL,rect.left,rect.bottom+0,rect.right-rect.left,ToolTipFontY,SWP_NOZORDER);
				}
			}
			if (defer) {
				EndDeferWindowPos(hdwp);
			}
			break;
		}

	// Paint: Background, border around display, keyboard and lower board,
	//        and Version logo.

	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			RECT rect;
			HGDIOBJ holdbitmap,hMonoBitmap,hOldMonoBitmap;
			HDC hMonoDC;
			COLORREF oldbk,oldtext;

			BeginPaint(hwnd,&ps);

			SetRect(&rect,0,0,BrdWidth,BrdHeight);
			MakeBorder(ps.hdc,&rect);
			if (Pane1) {
				SetRect(&rect,0,KeyYPos,BrdWidth,DCHYPos);
				FillRect(ps.hdc,&rect,hBackBrush);
				MakeBorder(ps.hdc,&rect);
			}
			if (Pane5) {
				// New bitmap strategy:
				// NOTE: for future enhancements, do the computation of new bitmap
				// only during WM_CREATE and save bitmap for later. Then only one
				// BitBlt is needed in WM_PAINT. 
				hMonoDC = CreateCompatibleDC(ps.hdc); // memDC with BW bitmap
				hMonoBitmap = CreateCompatibleBitmap(hMonoDC,LogoX,LogoY);
				hOldMonoBitmap = SelectObject(hMonoDC,hMonoBitmap);
				holdbitmap = SelectObject(hDigitsDC,hbVersion);

				// To get correct colours on screen, we first need to set pixels
				// to be colured by the logo BLACK. Then we can OR in the correct
				// logo colours. The logo is colours on black background.

				// Copying from colour bitmap to monochrome. Using the source DC
				// colour selection, background colour pixels->white, others->black.
				oldbk = SetBkColor(hDigitsDC,RGB(0,0,0));
				BitBlt(hMonoDC,0,0,LogoX,LogoY,hDigitsDC,0,0,SRCCOPY);
				SetBkColor(hDigitsDC,oldbk);
				// Can now fill background on screen with colour, and AND in
				// mono bitmap
				SetRect(&rect,0,DecoYPos,BrdWidth,DecoYPos+ButHeight*2);
				FillRect(ps.hdc,&rect,hBackBrush);
				MakeBorder(ps.hdc,&rect);
				// Copying from monochrome bitmap to colour. Using the destination DC
				// colour selection, white->background, black->foreground.
				oldbk = SetBkColor(ps.hdc,GetColour(RGBBACK));
				oldtext = SetTextColor(ps.hdc,RGB(0,0,0));
				BitBlt(ps.hdc,rect.right-1-Border-LogoX,rect.bottom-1-LogoY,
					LogoX,LogoY,hMonoDC,0,0,SRCAND);
				SetBkColor(ps.hdc,oldbk);
				SetTextColor(ps.hdc,oldtext);
				// Can now OR in the logo (on black background)
				BitBlt(ps.hdc,rect.right-1-Border-LogoX,rect.bottom-1-LogoY,
					LogoX,LogoY,hDigitsDC,0,0,SRCPAINT);
				
				SelectObject(hDigitsDC,holdbitmap);
				SelectObject(hMonoDC,hOldMonoBitmap);
				DeleteObject(hMonoBitmap);
				DeleteDC(hMonoDC);
			}
			SelectObject(ps.hdc,hDarkPen);
			// right bottom size corner:
			MoveTo(ps.hdc,ClientWidth-SizeX, ClientHeight);
			LineTo(ps.hdc,ClientWidth-SizeX, ClientHeight-Border);
			MoveTo(ps.hdc,ClientWidth,       ClientHeight-SizeY);
			LineTo(ps.hdc,ClientWidth-Border,ClientHeight-SizeY);
			
			// left bottom size corner:
			MoveTo(ps.hdc,SizeX,	ClientHeight);
			LineTo(ps.hdc,SizeX,	ClientHeight-Border);
			MoveTo(ps.hdc,0,    	ClientHeight-SizeY);
			LineTo(ps.hdc,Border,	ClientHeight-SizeY);
			
			EndPaint(hwnd,&ps);
			
			break;
		}
		
	// Display little dot (or radix) top left if active and no caption (or bin modes and size if!=32)
	// Also, make sure NumLock is turned on if so is requested in setup.
	// Save old settings for when XCALC is again disabled.
	// 7/7/97: Similar logic is installed also in all dialogues.
	// 8/7/97: This is NOT allowed in 32-bit XCALC. Disabled.
		
	case WM_ACTIVATE:
		{
			WPARAM active = GET_WM_ACTIVATE_STATE(wParam,lParam);
			if (active) {
				SetFocus(hwnd);
			}
			if (!Title) InvalidateRect(hDot,NULL,FALSE);
			break;
		}
		
	// Double click on the "Verx.x" icon starts Help
		
	case WM_LBUTTONDBLCLK:
		{
			WORD xpos = LOWORD(lParam);
			WORD ypos = HIWORD(lParam);
			if (!Pane5) break;
			if (xpos>ClientWidth-Border-LogoX && ypos>ClientHeight-LogoY)
				SendMessage(hwnd,PM_STARTHELP,0,0);
			break;
		}
		
	// Click/drag in lower corners to resize, unless caption is off and mousemove disabled
		
	case WM_LBUTTONDOWN:
		{
			POINTS mousepts = MAKEPOINTS(lParam);
			lastpts = mousepts;

			origpt.x = mousepts.x;
			origpt.y = mousepts.y;

			if (!Title && !EnableMouseMove) break;

			InRgtCorner = PtInRegion(hRightRgn,origpt.x,origpt.y);
			InLftCorner = PtInRegion(hLeftRgn,origpt.x,origpt.y);
			if (!InRgtCorner && !InLftCorner) break;

			ClientToScreen(hwnd,&origpt);

			ClearErr();

			GetWindowRect(hwnd,&origrect);
			SetCapture(hwnd);

			origSize = Size;
			Sizing = TRUE;			// checked in MouseMove
			break;
		}
		
	case WM_MOUSEMOVE:
		{
			POINTS mousepts = MAKEPOINTS(lParam);
			POINT mousept;
			HDWP hdwp;
			mousept.x = mousepts.x;
			mousept.y = mousepts.y;

			if (!Title && !EnableMouseMove) break;
			if (mousepts.x == lastpts.x && mousepts.y == lastpts.y) break; // stop unnecessary processing
			lastpts = mousepts;

			if (Sizing) {
				int oldSize = Size;
				int DX0 = origrect.right-origrect.left;
				int DX;
				int xoff;
				ClientToScreen(hwnd,&mousept);
				xoff = mousept.x - origpt.x;
				if (InRgtCorner) {
					DX = DX0+xoff;
					DX = max(DX,GetSystemMetrics(SM_CXMIN));
					Size = MulDiv(origSize,DX,DX0);
				}
				else {
					DX = DX0-xoff;
					DX = max(DX,GetSystemMetrics(SM_CXMIN));
					Size = MulDiv(origSize,DX,DX0);
				}
				// wParam in send is zero - window is then not updated there
				if (SendMessage(hwnd,PM_SIZEWINDOW,FALSE,0)) { // nonzero if change
					if (InLftCorner) {
						WindowX = origrect.right-TotalWidth;
					}
					hdwp = BeginDeferWindowPos(1);
					hdwp = DeferWindowPos(hdwp,hwnd,NULL,WindowX,WindowY,TotalWidth,TotalHeight,SWP_NOZORDER);
					//SetWindowPos(hwnd,NULL,WindowX,WindowY,TotalWidth,TotalHeight,SWP_NOZORDER);
					EndDeferWindowPos(hdwp);
					//SendMessage(hwnd,PM_SIZECHILDREN,TRUE,0); // defer does not work when no caption..!
					SendMessage(hwnd,PM_SIZECHILDREN,0,0);
					// Track Mini settings
					if (Mini) {
						m_Size=Size;
						m_WindowX=WindowX;
					}
					else {
						n_Size=Size;
						n_WindowX=WindowX;
					}
					InvalidateRect(hwnd,NULL,FALSE);
				}
				else {
					Size = oldSize;
				}
			}
			else { // not sizing
				if (PtInRegion(hRightRgn,mousept.x,mousept.y))
					SetCursor(LoadCursor(NULL,IDC_SIZENWSE));
				else if (PtInRegion(hLeftRgn,mousept.x,mousept.y))
					SetCursor(LoadCursor(NULL,IDC_SIZENESW));
				else
					SetCursor(LoadCursor(NULL,IDC_ARROW));
			}
			break;
		}
		
	case WM_LBUTTONUP:
		{
			if (Sizing) {
				Sizing = FALSE;
				ReleaseCapture();
			}
			break;
		}
		
	case WM_MOUSEWHEEL:
		{
			HyperbolicKeys(bHyperbolic=FALSE);
			SaveUndo(UT_RUP,TRUE);
			iState = 0;
			bSL = TRUE;
			if (wParam&0x80000000)  {
				if (IsWindow(hwndEditProgram)) {
					SendMessage(hwndEditProgram,PM_ADDTOPROGRAM,VK_DOWN,0);
				}
				rolldn(); 
			} else {
				if (IsWindow(hwndEditProgram)) {
					SendMessage(hwndEditProgram,PM_ADDTOPROGRAM,VK_UP,0);
				}
				rollup();
			}
			Dirty = TRUE;
			update();
			break;
		}
	// Follow WM_MOVE so we can update position on exit
		
	case WM_MOVE:
		{
			RECT rect;
			if (IsIconic(hwnd)) break;
			
			GetWindowRect(hwnd,&rect);
			WindowX = rect.left; // Global position variables
			WindowY = rect.top;
			// We need to update positions based on Mini here also. Bummer
			if (Mini) {
				m_WindowX=WindowX; 
				m_WindowY=WindowY;
			}
			else {
				n_WindowX=WindowX;
				n_WindowY=WindowY;
			}
			// Track error popup window if present
			if (!Pane5) {
				SetWindowPos(hError,NULL,rect.left,rect.bottom+0,rect.right-rect.left,ToolTipFontY+3,SWP_NOZORDER|SWP_NOACTIVATE);
			}
			break;
		}
		
	// show memory popup window, if enabled
		
	case WM_SIZE:
		{
			WPARAM sizetype = wParam;
			if (sizetype!=SIZE_MINIMIZED && ShowMemoryPopup)
				ShowWindow(hMemoryPopup,SW_SHOWNOACTIVATE);
			break;
		}

	case PM_SYSMENU:
		{
			HMENU hSystemMenu = GetSystemMenu(hwndXCALC,FALSE);
			int syscmd;
			SetForegroundWindow(hwndXCALC);
			syscmd = TrackPopupMenuEx(hSystemMenu,TPM_RETURNCMD,(int)wParam,(int)lParam,hwndXCALC,NULL);
			if (syscmd) {
				PostMessage(hwndXCALC,WM_SYSCOMMAND,syscmd,0);
			}
			break;
		}

	case PM_UPDATEALL:  // sends PM_UPDATE1 to all displays
		{
			int i;
			for (i=0;i<DisplayStkSiz;i++)
				SendMessage(hwnd,PM_UPDATE1,i,0);
			break;
		}
		
	case PM_UPDATE1:	// update display given in wParam
		// Special considerations for register 0 (x)
		// at keybord input (iState!=0)
		// Nov96: added thousands separator on input
		{
			WORD wReg = (WORD)wParam;
			WORD NumberType = wDispMode & (WORD)RADIXMASK;
			if (wReg==Xreg) {	// Update error window on every X update
								// and sound beep if error condition
				if (Error && SoundOn && Error!=__UNDONE /*&& Error!=__REDONE*/ && Error!=__MESSAGE) MessageBeep(ErrSound[Error]);
				PostMessage(hError,PM_UPDATE,0,0);
			}
			
			if ((wReg==Xreg)&&(iState)) {	// use input string (not complete input)
				memset(cST[0],' ',NDIGITS);
				
				if (NumberType==COMPLEX) {	// share screen real&cmplx parts
					if (iState>0) {			// real part being input
						memcpy(&cST[0][CPLXOUTWIDTH-lstrlen(STR)],STR,lstrlen(STR));
					}
					else {					// imaginary part
						MakeString(wReg);
						memset(&cST[0][CPLXOUTWIDTH+1],' ',CPLXOUTWIDTH+1);
						memcpy(&cST[0][NDIGITS-lstrlen(STR)-1],STR,lstrlen(STR));
					}
				}
				else {		// copy STR to TSTR, maybe insert 1000s separator, clear first '/', and copy
					char TSTR[NDIGITS+1];
					char *lastmant;
					lstrcpy(TSTR,STR);
					//if (ThousandSep && NumberType==DECIMAL && iState<4) {
					if (ThousandSep && NumberType==DECIMAL) {
						if ((lastmant=strchr(TSTR,'.'))==0 && (lastmant=strchr(TSTR,'e'))==0 && (lastmant=strchr(TSTR,'/'))==0 && (lastmant=strchr(TSTR,''))==0)
							lastmant = TSTR+lstrlen(TSTR);
						// lastmant now points one right of last mantissa digit
						lastmant -= 3;
						while (lastmant>TSTR && *(lastmant-1)>=0 && isdigit(*(lastmant-1))) { // test for >=0 because ms debug isdigit asserts on negative chars ()
							memmove(lastmant+1,lastmant,lstrlen(lastmant)+1);
							*lastmant = ',';
							lastmant -= 3;
						}
					}
					if (iState==4 || iState==5) {
						char *f1;
						if ((f1=strchr(TSTR,'/')) != NULL)
							if (strchr(f1+1,'/'))
								*f1=' ';
					}
					memcpy(&cST[0][NDIGITS-lstrlen(TSTR)],TSTR,lstrlen(TSTR));
				}
			}
			else {
				MakeString(wReg);      // writes string to cST[wReg]
			}
			
			SendMessage(hDisplay[wReg],PM_UPDATE,0,0);
			break;
		}
		
	// PM_SETFIX sets float/fix settings.
	// We cannot save undo information here. Must be done in calling routine.
	// (If we set undo info here, initial fix could be undone)
	// We use SendMessage() to get information.
	case PM_SETFIX:
		{
			if (IsBin) return TRUE;  // no change
			sFix = (short)wParam;
			PostMessage(hwnd,PM_UPDATEALL,0,0);
			if (Pane4) PostMessage(hFix,PM_UPDATE,0,0);
			iState = 0;
			return FALSE;            // change OK
		}
	// PM_SETBITS adjusts word size for binary modes
	case PM_SETBITS:
		{
			int i;
			if (!IsBin) return TRUE;  // no change
			sWordSize = (short)wParam;
			SetWindowWord(hDispMode[FIXWND],2,(WORD)(sWordSize==0));
			SetWindowWord(hDispMode[SCIWND],2,(WORD)(sWordSize==1));
			SetWindowWord(hDispMode[ENGWND],2,(WORD)(sWordSize==2));
			iMaxDigits = aMaxInputDigits[sWordSize][RADIX_LIN];
			// Strip excess bits in stack registers
			for (i=0;i<StkSiz;i++) {
				ST[i].i &= CURR_MASK;
			}
			ST[Lreg].i &= CURR_MASK;
			for (i=FIXWND;i<=ENGWND;i++) {
				InvalidateRect(hDispMode[i],NULL,FALSE);  // draw whole window
				PostMessage(hDispMode[i],PM_UPDATE,0,0);
			}
			if (!Title) InvalidateRect(hDot,NULL,FALSE); // Update status
			PostMessage(hwnd,PM_UPDATEALL,0,0);
			iState = 0;
			return FALSE;            // change OK
		}
// Convenience macros that get window numbers from wDispMode
#define WindowFromRADIX(x) (DECWND+radix_lin(x))
#define WindowFromFIX(x)   (FIXWND+fix_lin(x))
#define WindowFromANGLE(x) (DEGWND+angle_lin(x))
		
	// PM_SETMODE called with new wDispMode in wParam (may contain all modes)
	// We cannot save undo information here. Must be done in calling routine.
	case PM_SETMODE:
		{
			int i,iWindow;
			WORD newRadix;
			WORD NewMode;
			WORD wMode = (WORD)wParam;
			WORD NumberType = wDispMode & (WORD)RADIXMASK;
			WORD AngleType = wDispMode & (WORD)ANGLEMASK;
			int retval = FALSE; // Ok by default
				// retval stores return value. This is only used in single mode
				// settings to trigger undo store. Here, FALSE is returned for OK.
				// In case of multivalue mode, the return value will be ignored.
			bSL = TRUE;
			
			// Change radix
			
			if (wMode&RADIXMASK) {
				iWindow = WindowFromRADIX(wMode);	// a bit dirty, uses structure of mode to find button
				NewMode = (WORD)(wMode&RADIXMASK);
				newRadix = aRadix[iWindow];
				
				if ((newRadix!=10)&&(wRadix==10)) {	// change from radix 10
					if (NumberType==DECIMAL) wDecDR = AngleType;
					toint();
					if (Error) {
						retval = TRUE;      // No change
						goto UpdateAndBeep;
					}
				}
				else if ((newRadix==10)&&(wRadix!=10)) {	// change to radix 10
					tofloat();
				}
				
				if ((NewMode==DECIMAL)) {						// -> decimal
					SetWindowWord(hDispMode[DEGWND],0,TRUE);	// enable Degree
					// 24/1/97:
					// It looks like InvalidateRect is not needed here. In addition,
					// having it means we get full screen update if mode is changed
					// to either DEC or CPX and XCALC is in minimum mode.
					// Reason not known.
					// Leave it out. Same applies below (complex).
					// InvalidateRect(hDispMode[DEGWND],NULL,FALSE);       // draw new text colour
					PostMessage(hwnd,PM_SETMODE,wDecDR,0);
					for (i=0;i<StkSiz;i++) {
						ST[i].imag = 0;
					}
					ST[Lreg].imag = 0;
				}
				else if ((NewMode==COMPLEX)) {		// -> complex
					// Convert from fraction to decimal where needed
					for (i=0;i<StkSiz;i++) {
						mkreal(&ST[i]);
					}
					mkreal(&ST[Lreg]);
					SetWindowWord(hDispMode[DEGWND],0,FALSE);	// disable Degree
					// InvalidateRect(hDispMode[DEGWND],NULL,FALSE);       // draw new text colour
					if (NumberType==DECIMAL) wDecDR = AngleType;
					PostMessage(hwnd,PM_SETMODE,RAD,0);
				}
				
				wDispMode &= ~RADIXMASK;
				wDispMode |= NewMode;
				wRadix = newRadix;
				IsBin = (wRadix!=10);
				iMaxDigits = aMaxInputDigits[sWordSize][iWindow];
				iState = 0;
				if (Pane1) {	// change keyboard
					for (i=0;i<nKeyRows*nKeyCols;i++) {
						InvalidateRect(hwndKey[i],NULL,FALSE);
					}
				}

				if (Pane3) {
					// Update fix/sci/eng (or 32/16/8 bit)
					if (IsBin) {
						// New mode binary
						SendMessage(hwnd,PM_SETBITS,sWordSize,0);
					}
					else {
						// New mode not binary
						if (!(wMode&FIXMASK)) SendMessage(hwnd,PM_SETMODE,wDispMode&FIXMASK,0);
					}
					for (i=DEGWND;i<=RADWND;i++) {
						SetWindowWord(hDispMode[i],0,(WORD)(!IsBin && (i==DEGWND?NewMode!=COMPLEX:TRUE))); //special for degree
						InvalidateRect(hDispMode[i],NULL,FALSE);  // draw whole window (text changes colour)
					}
					for (i=FIXWND;i<=RADWND;i++) {
						InvalidateRect(hDispMode[i],NULL,FALSE);  // draw whole window
						PostMessage(hDispMode[i],PM_UPDATE,0,0);
					}
				}

				if (Pane4) {
					InvalidateRect(hFix,NULL,FALSE);		// update FIX: display
				}
				
				if (Pane2) {
					for (i=DECWND;i<=BINWND;i++) {
						SetWindowWord(hDispMode[i],2,(WORD)(i==iWindow));	// set active flags (always enabled)
						PostMessage(hDispMode[i],PM_UPDATE,0,0);
					}
				}
				
UpdateAndBeep:							// goto if error in converting dec -> bin (toint())
				PostMessage(hwnd,PM_UPDATEALL,0,0);
				for (i=0;i<MEMSIZ;i++) SendMessage(hMemoryDisplay[i],PM_UPDATE,0,0);
			}
			
			// Change Fix/Sci/Eng
			
			if (wMode&FIXMASK) {
				if (!IsBin) {
					iWindow = WindowFromFIX(wMode); // macro that extracts window number
					NewMode = (WORD)(wMode&FIXMASK);
					wDispMode &= ~FIXMASK;
					wDispMode |= NewMode;
					if (Pane3) {
						for (i=FIXWND;i<=ENGWND;i++) {
							SetWindowWord(hDispMode[i],2,(WORD)(i==iWindow));
							PostMessage(hDispMode[i],PM_UPDATE,1,0);
						}
					}
					PostMessage(hwnd,PM_UPDATEALL,0,0);
				}
				else
					retval = TRUE;
			}
			
			// Change Deg/Rad
			
			if (wMode&ANGLEMASK) {
				if (!IsBin) {
					iWindow = WindowFromANGLE(wMode);
					NewMode = (WORD)(wMode&ANGLEMASK);
					if (NumberType==DECIMAL) wDecDR = NewMode;
					if (!((NumberType==COMPLEX)&&(NewMode==DEG))) {// not (complex and degree)
						wDispMode &= (WORD)~ANGLEMASK;
						wDispMode |= NewMode;
						if (Pane3) {
							for (i=DEGWND;i<=RADWND;i++) {
								SetWindowWord(hDispMode[i],2,(WORD)(i==iWindow));
								InvalidateRect(hDispMode[i],NULL,FALSE);  // draw whole window (text changes colour)
								PostMessage(hDispMode[i],PM_UPDATE,0,0);
							}
						}
					}
					else
						retval = TRUE;
				}
				else
					retval = TRUE;
			}
			
			if (!Title) InvalidateRect(hDot,NULL,FALSE); // Update status
			
			return retval; // not break. Now, SendMessage() gets results
		}
	case PM_CONFIG:
		{
			DoPropertySheet(hwnd);
			// Could be done better - update active version of some settings
			if (Mini) {
				EnableMouseMove=m_EnableMouseMove;
				KeepInFront=m_KeepInFront;
			}
			else {
				EnableMouseMove=n_EnableMouseMove;
				KeepInFront=n_KeepInFront;
			}
			if (Restart)
				DestroyWindow(hwnd);
			else {
				if (RefreshNow) SendMessage(hwnd,PM_UPDATEALL,0,0);
				// If required, make window topmost
				if (KeepInFront)
					SetWindowPos(hwndXCALC,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
				else
					SetWindowPos(hwndXCALC,HWND_NOTOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
				InvalidateRect(hwnd,NULL,FALSE);
			}
			break;
		}
	case PM_MINI:
		Mini = !Mini;
		Restart = TRUE;
		DestroyWindow(hwnd);
		break;
		
	case PM_HELP:
	case PM_STARTHELP:
		HtmlHelp(hwnd,HelpFile,HH_DISPLAY_TOPIC,0);
		break;
		
	case PM_KEYHELP: {
		char helpstr[_MAX_PATH+MAX_HELP];
		strcpy(helpstr,HelpFile);
		switch (CURR_RADIX) {
			case DECIMAL: strcat(helpstr,DECIMALHELP); break;
			case COMPLEX: strcat(helpstr,COMPLEXHELP); break;
			case HEX:
			case OCTAL:
			case BINARY: strcat(helpstr,BINARYHELP); break;
		}
		HtmlHelp(hwnd,helpstr,HH_DISPLAY_TOPIC,0);
		break;
	}		
	case PM_EXIT:
		DestroyWindow(hwnd);
		break;
		
		// memory store/recall/clear
		
	case PM_STORE:
		{
			WORD wReg = (WORD)wParam;
			// remove error message if any
			ClearErr();
			SaveUndo(UT_STORE,TRUE);
			bSL = TRUE;
			iState = 0;
			update();
			Memory[wReg] = ST[Xreg];
			if (wRadix==10)
				Memory[wReg].i = 0;
			else {
				Memory[wReg].real = Memory[wReg].imag = 0;
				Memory[wReg].num = Memory[wReg].den = 0;
			}
			if (Pane5) PostMessage(hMemory,PM_UPDATE,wReg,0);
			PostMessage(hMemoryDisplay[wReg],PM_UPDATE,0,0);
			break;
		} // case PM_STORE

	case PM_RECALL:
		{
			BOOL FromFloat,FromFrac,FromDMS;
			WORD wReg = (WORD)wParam;
			
			// remove error message if any
			ClearErr();
			
			SaveUndo(UT_RECALL,FALSE);
			
			if (iState != 0) bSL = TRUE;
			iState = 0;
			
			// check which memory part is used
			
			FromFloat = Memory[wReg].ct==ctREAL;
			FromFrac = isproperfrac(Memory+wReg);
			FromDMS = isdms(Memory+wReg);
			
			if (wRadix==10)	{		// convert to float
				if (bSL) lift();
				if (FromFloat || FromFrac || FromDMS) {
					ST[Xreg] = Memory[wReg];
					if (CURR_RADIX==COMPLEX && (FromFrac || FromDMS))
						mkreal(&ST[Xreg]);
					if (CURR_RADIX==DECIMAL && ST[Xreg].imag!=0)
						PostMessage(hwnd,PM_SETMODE,COMPLEX,0);
				}
				else {
					clrstkreg(Xreg);
					ST[Xreg].real = Memory[wReg].i;
				}
				CommitUndo();
			}
			else {
				LD v;
				if (FromFloat||FromFrac||FromDMS) {
					if (FromFloat||FromDMS) v = Memory[wReg].real;
					if (FromFrac) v = Memory[wReg].num/Memory[wReg].den;
					// Round towards zero to mimic long cast
					if (v<0)
						v = ceill(v);
					else
						v = floorl(v);
					if (v>CURR_MAX||v<CURR_MIN) {
						Error = __RANGE;
					}
					else {
						if (bSL) lift();
						clrstkreg(Xreg);
						ST[Xreg].i=(long)v&CURR_MASK;
						CommitUndo();
					}
				}
				else {
					if (bSL) lift();
					ST[Xreg] = Memory[wReg];
					ST[Xreg].i &= CURR_MASK;
					CommitUndo();
				}
			}
			bSL = TRUE;
			update();
			break;
		} // case PM_RECALL
				
	case PM_CLEARMEM:
		{
			WORD wReg = (WORD)wParam;
			Register zero = {ctREAL,0};
			// remove error message if any
			ClearErr();
			SaveUndo(UT_CLRMEM,TRUE);
			Memory[wReg] = zero;
			if (Pane5) PostMessage(hMemory,PM_UPDATE,wReg,0);
			PostMessage(hMemoryDisplay[wReg],PM_UPDATE,0,0);
			break;
		}
		
	case PM_CLEARMEMORY:
		{
			WORD wReg;
			Register zero = {ctREAL,0};
			// remove error message if any
			ClearErr();
			SaveUndo(UT_CLEARMEM,TRUE);
			for (wReg=0;wReg<10;wReg++) {
				Memory[wReg] = zero;
				if (Pane5) PostMessage(hMemory,PM_UPDATE,wReg,0);
				PostMessage(hMemoryDisplay[wReg],PM_UPDATE,0,0);
			}
			break;
		}
		
	case PM_COPY:
		{
			// remove error message if any
			ClearErr();
			PostMessage(hDisplay[0],WM_LBUTTONDBLCLK,0,0); // copy done in display proc.
			break;
		}
		
	case PM_PASTE:
		// This is a special PASTE done via keyboard. Most of the logic
		// is kept under WM_RBUTTONDBLCLK in display.c. To signify this
		// paste operation as keyboard triggered, both wParam and lParam
		// are set to (-1). This will make sure the logic will lift the
		// stack in the right circumstances.
		{
			int oldstate = iState;	// save state in case paste fails
			int oldSL = bSL;
			// remove error message if any
			ClearErr();
			if (!SendMessage(hDisplay[0],WM_RBUTTONDBLCLK,(WPARAM)-1,(LPARAM)-1)) { // return 0 if paste OK
				update();			// paste OK.
			}
			else {
				bSL = oldSL;		// paste failed - restore originals
				iState = oldstate;
			}
			break;
		}
				
	case WM_SYSCOMMAND:
		{
			WPARAM command = wParam;
			switch (command&0xFFF0)
			{
			case SC_CONFIG:
				PostMessage(hwnd,PM_CONFIG,0,0);
				return 0;
			case SC_MINI:
				PostMessage(hwnd,PM_MINI,0,0);
				return 0;
			case SC_MINIMIZE:
				if (InTray) {
					ShowWindow(hwndXCALC,SW_HIDE);
					ShowWindow(hMemoryPopup,SW_HIDE);
				}
				else {
					ShowWindow(hwndXCALC,SW_MINIMIZE);
				}
				return 0;
			case SC_CLOSE: // same as SC_MINIMIZE if InTray; same as SC_EXIT if not
				if (InTray) {
					ShowWindow(hwndXCALC,SW_HIDE);
					ShowWindow(hMemoryPopup,SW_HIDE);
				}
				else {
					PostMessage(hwnd,PM_EXIT,0,0);
				}
				return 0;
			case SC_HELP:
				PostMessage(hwnd,PM_STARTHELP,0,0);
				return 0;
			case SC_KEYHELP:
				PostMessage(hwnd,PM_KEYHELP,0,0);
				return 0;
			case SC_ABOUT:
				DialogBox(hInst,"XCALCABOUT",hwnd,AboutDlgProc);
				return 0;
			case SC_EXIT: // always exit
				PostMessage(hwnd,PM_EXIT,0,0);
				return 0;
			default:
				return DefWindowProc(hwnd,wMsg,wParam,lParam);
			}
		} // case WM_SYSCOMMAND
				
	case PM_UNDO:
		ClearErr();
		// Clear last command if recording
		if (IsWindow(hwndEditProgram)) {
			SendMessage(hwndEditProgram,PM_UNDOPROGRAM,0,0);
		}
		if (iState) {									// send VK_BACK if
			SendMessage(hwndXCALC,WM_CHAR,VK_BACK,0);	// keying digits
		}
		else if (!Undo(&Undone)) {						// no undo information
			Error = __NOUNDO;
			update();
		}
		else {
			Error = __UNDONE;   // signal undo. undo type in global var Undone
			update();
		}
		break;
	/*	
	case PM_REDO:
		ClearErr();
		if (!Redo(&Undone)) {						// no undo information
			Error = __NOREDO;
			update();
		}
		else {
			Error = __REDONE;   // signal undo. undo type in global var Undone
			update();
		}
		break;
	*/

	// keyboard accelerators:
				
	case WM_COMMAND:
		{
			// Use message crackers to extract parameters:
			WORD wID = GET_WM_COMMAND_ID(wParam,lParam);
			// Undo information is saved in PM_xxx responses where appropriate,
			// otherwise it is saved here in case of a change.
			
			// Send key to program window if open
			if (IsWindow(hwndEditProgram)) {
				SendMessage(hwndEditProgram,PM_ADDTOPROGRAM,_AS,wID);
			}

			// remove error message if any
			ClearErr();
			// Always update Hyperbolic settings on WM_COMMAND
			HyperbolicKeys(bHyperbolic=FALSE);
			
			switch (wID)
			{
			case AS_MINI:
				PostMessage(hwnd,PM_MINI,0,0);
				break;
			case AS_KEYHELP:
				SendMessage(hwnd,PM_KEYHELP,0,0);
				break;
			case AS_UNDO:
				SendMessage(hwnd,PM_UNDO,0,0);
				break;
			case AS_COPY:
				SendMessage(hwnd,PM_COPY,0,0);
				break;
			case AS_PASTE:
				SendMessage(hwnd,PM_PASTE,0,0);
				break;
			case AS_STORE0:
				SendMessage(hwnd,PM_STORE,0,0);
				break;
			case AS_STORE1:
				SendMessage(hwnd,PM_STORE,1,0);
				break;
			case AS_STORE2:
				SendMessage(hwnd,PM_STORE,2,0);
				break;
			case AS_STORE3:
				SendMessage(hwnd,PM_STORE,3,0);
				break;
			case AS_STORE4:
				SendMessage(hwnd,PM_STORE,4,0);
				break;
			case AS_STORE5:
				SendMessage(hwnd,PM_STORE,5,0);
				break;
			case AS_STORE6:
				SendMessage(hwnd,PM_STORE,6,0);
				break;
			case AS_STORE7:
				SendMessage(hwnd,PM_STORE,7,0);
				break;
			case AS_STORE8:
				SendMessage(hwnd,PM_STORE,8,0);
				break;
			case AS_STORE9:
				SendMessage(hwnd,PM_STORE,9,0);
				break;
				
			case AS_RECALL0:
				SendMessage(hwnd,PM_RECALL,0,0);
				break;
			case AS_RECALL1:
				SendMessage(hwnd,PM_RECALL,1,0);
				break;
			case AS_RECALL2:
				SendMessage(hwnd,PM_RECALL,2,0);
				break;
			case AS_RECALL3:
				SendMessage(hwnd,PM_RECALL,3,0);
				break;
			case AS_RECALL4:
				SendMessage(hwnd,PM_RECALL,4,0);
				break;
			case AS_RECALL5:              
				SendMessage(hwnd,PM_RECALL,5,0);
				break;
			case AS_RECALL6:
				SendMessage(hwnd,PM_RECALL,6,0);
				break;
			case AS_RECALL7:
				SendMessage(hwnd,PM_RECALL,7,0);
				break;
			case AS_RECALL8:
				SendMessage(hwnd,PM_RECALL,8,0);
				break;
			case AS_RECALL9:
				SendMessage(hwnd,PM_RECALL,9,0);
				break;
				
			case AS_CLEAR0:
				SendMessage(hwnd,PM_CLEARMEM,0,0);
				break;
			case AS_CLEAR1:
				SendMessage(hwnd,PM_CLEARMEM,1,0);
				break;
			case AS_CLEAR2:                
				SendMessage(hwnd,PM_CLEARMEM,2,0);
				break;
			case AS_CLEAR3:
				SendMessage(hwnd,PM_CLEARMEM,3,0);
				break;
			case AS_CLEAR4:
				SendMessage(hwnd,PM_CLEARMEM,4,0);
				break;
			case AS_CLEAR5:
				SendMessage(hwnd,PM_CLEARMEM,5,0);
				break;
			case AS_CLEAR6:
				SendMessage(hwnd,PM_CLEARMEM,6,0);
				break;
			case AS_CLEAR7:
				SendMessage(hwnd,PM_CLEARMEM,7,0);
				break;
			case AS_CLEAR8:
				SendMessage(hwnd,PM_CLEARMEM,8,0);
				break;
			case AS_CLEAR9:
				SendMessage(hwnd,PM_CLEARMEM,9,0);
				break;
			case AS_CLEARMEM:
				SendMessage(hwnd,PM_CLEARMEMORY,0,0);
				break;
				
			case AS_FIX0:
				SaveUndo(UT_FLOAT,FALSE);
				if (!SendMessage(hwnd,PM_SETFIX,0,0))
					CommitUndo();
				break;
			case AS_FIX1: // case AS_16BIT
				if (wRadix == 10) {
					SaveUndo(UT_FLOAT,FALSE);
					if (!SendMessage(hwnd,PM_SETFIX,1,0))
						CommitUndo();
				}
				else {
					SaveUndo(UT_WORD,FALSE);
					if (!SendMessage(hwnd,PM_SETBITS,1,0))
						CommitUndo();
				}
				break;
			case AS_FIX2: // case AS_32BIT
				if (wRadix == 10) {
					SaveUndo(UT_FLOAT,FALSE);
					if (!SendMessage(hwnd,PM_SETFIX,2,0))
						CommitUndo();
				}
				else {
					SaveUndo(UT_WORD,FALSE);
					if (!SendMessage(hwnd,PM_SETBITS,0,0))
						CommitUndo();
				}
				break;
			case AS_FIX3: // case AS_32BIT
				if (wRadix == 10) {
					SaveUndo(UT_FLOAT,FALSE);
					if (!SendMessage(hwnd,PM_SETFIX,3,0))
						CommitUndo();
				}
				else {
					SaveUndo(UT_WORD,FALSE);
					if (!SendMessage(hwnd,PM_SETBITS,0,0))
						CommitUndo();
				}
				break;
			case AS_FIX4:
				SaveUndo(UT_FLOAT,FALSE);
				if (!SendMessage(hwnd,PM_SETFIX,4,0))
					CommitUndo();
				break;
			case AS_FIX5:
				SaveUndo(UT_FLOAT,FALSE);
				if (!SendMessage(hwnd,PM_SETFIX,5,0))
					CommitUndo();
				break;
			case AS_FIX6: // case AS_16BIT
				if (wRadix == 10) {
					SaveUndo(UT_FLOAT,FALSE);
					if (!SendMessage(hwnd,PM_SETFIX,6,0))
						CommitUndo();
				}
				else {
					SaveUndo(UT_WORD,FALSE);
					if (!SendMessage(hwnd,PM_SETBITS,1,0))
						CommitUndo();
				}
				break;
			case AS_FIX7:
				SaveUndo(UT_FLOAT,FALSE);
				if (!SendMessage(hwnd,PM_SETFIX,7,0))
					CommitUndo();
				break;
			case AS_FIX8: // case AS_8BIT
				if (wRadix == 10) {
					SaveUndo(UT_FLOAT,FALSE);
					if (!SendMessage(hwnd,PM_SETFIX,8,0))
						CommitUndo();
				}
				else {
					SaveUndo(UT_WORD,FALSE);
					if (!SendMessage(hwnd,PM_SETBITS,2,0))
						CommitUndo();
				}
				break;
			case AS_FIX9:
				SaveUndo(UT_FLOAT,FALSE);
				if (!SendMessage(hwnd,PM_SETFIX,9,0))
					CommitUndo();
				break;
			case AS_FLOAT:
				SaveUndo(UT_FLOAT,FALSE);
				if (!SendMessage(hwnd,PM_SETFIX,(WPARAM)FLOAT,0))
					CommitUndo();
				break;
			case AS_F32BIT:
				if (wRadix != 10) {
					SaveUndo(UT_WORD,FALSE);
					if (!SendMessage(hwnd,PM_SETBITS,0,0))
						CommitUndo();
					}
				break;
			case AS_F16BIT:
				if (wRadix != 10) {
					SaveUndo(UT_WORD,FALSE);
					if (!SendMessage(hwnd,PM_SETBITS,1,0))
						CommitUndo();
				}
				break;
			case AS_F8BIT:
				if (wRadix != 10) {
					SaveUndo(UT_WORD,FALSE);
					if (!SendMessage(hwnd,PM_SETBITS,2,0))
						CommitUndo();
				}
				break;
			case AS_DECIMAL:
				SaveUndo(UT_DEC,TRUE);
				SendMessage(hwnd,PM_SETMODE,DECIMAL,0);
				break;
			case AS_COMPLEX:
				SaveUndo(UT_CPX,TRUE);
				SendMessage(hwnd,PM_SETMODE,COMPLEX,0);
				break;
			case AS_HEX:
				SaveUndo(UT_HEX,FALSE);
				if (!SendMessage(hwnd,PM_SETMODE,HEX,0))
					CommitUndo();
				break;
			case AS_OCTAL:
				SaveUndo(UT_OCT,FALSE);
				if (!SendMessage(hwnd,PM_SETMODE,OCTAL,0))
					CommitUndo();
				break;
			case AS_BINARY:
				SaveUndo(UT_BIN,FALSE);
				if (!SendMessage(hwnd,PM_SETMODE,BINARY,0))
					CommitUndo();
				break;
			case AS_FIX:
				SaveUndo(UT_FIX,FALSE);
				if (!SendMessage(hwnd,PM_SETMODE,FIX,0))
					CommitUndo();
				break;
			case AS_SCI:
				SaveUndo(UT_SCI,FALSE);
				if (!SendMessage(hwnd,PM_SETMODE,SCI,0))
					CommitUndo();
				break;
			case AS_ENG:
				SaveUndo(UT_ENG,FALSE);
				if (!SendMessage(hwnd,PM_SETMODE,ENG,0))
					CommitUndo();
				break;
			case AS_DEG:
				SaveUndo(UT_DEG,FALSE);
				if (!SendMessage(hwnd,PM_SETMODE,DEG,0))
					CommitUndo();
				break;
			case AS_RAD:
				SaveUndo(UT_RAD,FALSE);
				if (!SendMessage(hwnd,PM_SETMODE,RAD,0))
					CommitUndo();
				break;
			case AS_CONFIG:
				SendMessage(hwnd,PM_CONFIG,0,0);
				break;
			case AS_EXIT:
				PostMessage(hwnd,PM_EXIT,0,0);
				break;
			case AS_HELP:
				SendMessage(hwnd,PM_STARTHELP,0,0);
				break;
			case AS_POPUP:
				ShowMemoryPopup = !ShowMemoryPopup;
				ShowWindow(hMemoryPopup,ShowMemoryPopup?/*SW_RESTORE*/SW_SHOWNOACTIVATE:SW_HIDE);
				break;
			case AS_CTRL_LEFT:
				if (wRadix!=10) {
					SaveUndo(UT_LROT,TRUE);
					iState = 0;
					bSL = TRUE;
					LRot();
					update();
				}
				break;
			case AS_CTRL_RIGHT:
				if (wRadix!=10) {
					SaveUndo(UT_RROT,TRUE);
					iState = 0;
					bSL = TRUE;
					RRot();
					update();
				}
				break;
			case AS_PROG1:
			case AS_PROG2:
			case AS_PROG3:
			case AS_PROG4:
			case AS_PROG5:
			case AS_PROG6:
			case AS_PROG7:
			case AS_PROG8:
			case AS_PROG9:
			case AS_PROG10:
			case AS_PROG11:
			case AS_PROG12:
				RunProgram(wID-AS_PROG1+1);
				break;
			case AS_EDITPROGRAM:
				PostMessage(hwndXCALC,PM_EDITPROGRAM,0,0);
				break;
			} // end switch wID

			break;

		} // end case WM_COMMAND

	// WM_CHAR needs a complex decoding dependent on state:
	//
	// iState=0: normal operation, may change to any other mode based on key
	// iState=1: keying in integer part of number
	// iState=2: keying in fractional part of number
	// iState=3: keying in exponent part
	//
	// A negative iState means we are keying in the complex part.
	// iState=-4: special mode similar to 0 for complex part
	//
	// iState=4: keying in first fractional part (x in a/x)
	// iState=5: keying in second fractional part (x in a/b/x)
	//
	// iState=6: keying in minutes (x in ax)
	// iState=7: keying in integer part of seconds (x in ab'x)
	// iState=8: keying in fractional part of seconds (x in ab's.x)
	//
	// If the state input found no valid input key, a check is made for
	// function keys (such as +, -, c for cosine etc...)
	//
	// Undo information is saved explicitly where needed.
	//
	// Hyperbolic key is also treated specially here. We do a hard-coded
	// check, and update unless any of 'sctSCT' is pressed.

	case WM_CHAR:
		{
			char c = (char)wParam;
			char C = (char)toupper(c);
			int iSign = iState>=0?1:-1;
			int L;
			int val;
			
			// Fudges to allow key synonyms (this is needed for the program Record function)
			if (wParam=='>') wParam='<';
			if (wParam=='X') wParam='Z';
			if (wRadix==10 && wParam=='j') wParam='i';
			if (wRadix==10 && wParam=='J') wParam='I';
			if (wRadix==10 && wParam=='r') wParam='&';
			if (wRadix==10 && wParam=='o') wParam='^';
			if (wRadix==10 && wParam=='f') wParam=' ';
			if (wRadix!=10 && wParam=='&') wParam='A';
			if (wRadix!=10 && wParam=='|') wParam='o';
			if (wRadix!=10 && wParam=='^') wParam='x';

			// remove error message if any
			ClearErr();

			// Send key to program window if open
			if (IsWindow(hwndEditProgram)) {
				SendMessage(hwndEditProgram,PM_ADDTOPROGRAM,0,wParam);
			}
			// Update Hyperbolic keys if needed:
			if (wRadix==10 && c!='s' && c!='c' && c!='t' && c!='S' && c!='C' && c!='T')
				HyperbolicKeys(bHyperbolic=FALSE);
			// set up parameters
			L = lstrlen(STR);
			val = 100;
			if ((c>='0')&&(c<='9')) val = c-'0';
			if ((c>='a')&&(c<='f')) val = c-'a'+10;
			
			// do state-by-state check
			
			if (iState==0) {
				if (val<wRadix) {
					SaveUndo(UT_INKEY,TRUE);
					if (bSL) {bSL=0; lift();}   // lift() moves & sets Dirty flag
					set(C);                     // set() sets STR to c alone
					eval();                     // eval - checks for error
					//  deletes char if error
					//  never returns if error
					//  updates display and ST[0].
					iState=1;
					break;
				}
				if (wRadix==10) {
					if (c==dPoint[k_useComma]) {
						SaveUndo(UT_INKEY,TRUE);
						if (bSL) {bSL=0; lift();}
						set('.');
						eval();
						iState=2;
						break;
					}
					else if (c=='e') {
						SaveUndo(UT_INKEY,TRUE);
						if (bSL) {bSL=0; lift();}
						set('1');
						app('e');
						app('+');
						eval();
						iState=3;
						break;
					}
					else if ((c=='\'' || c=='"') && CURR_RADIX!=COMPLEX) {
						SaveUndo(UT_INKEY,TRUE);
						if (bSL) {bSL=0; lift();}
						set('0');
						app('');
						eval();
						iState=6;
						break;
					}
					else if ((c==Comma[k_useComma]) && CURR_RADIX==COMPLEX) {
						SaveUndo(UT_INKEY,TRUE);
						if (bSL) {bSL=0; lift();}
						set('0');
						eval();
						iState=-4;
						set ('0');
						eval();
						break;
					}
				}
			}
			
			else if (iState==-4) {
				if (c==VK_BACK) {
					iState = 0;
					bSL = 0;
					ST[Xreg].real = 0.0;
					update();
					break;
				}
				if (val<wRadix) {
					set(C);
					eval();
					iState=-1;
					break;
				}
				if (c==dPoint[k_useComma]) {
					set('.');
					eval();
					iState=-2;
					break;
				}
				else if (c=='e') {
					set('1');
					app('e');
					app('+');
					eval();
					iState=-3;
					break;
				}
			}
			
			else if (abs(iState)==1) {
				if (c==VK_BACK) {
					if ((L==1)||(L==2)&&(STR[0]=='-')) {
						set('0');
						eval();
						iState = iState>0 ? 0:-4;
						break;
					}
					else {
						STR[L-1] = 0;
						eval();
						break;
					}
				}
				else if (val<wRadix) {
					if (L>=iMaxDigits) break;
					if (wRadix==10) {
						if ((iState>0?ST[Xreg].real:ST[Xreg].imag)==0)
							set(C);
						else
							app(C);
						eval();
						break;
					}
					else {
						if ((ST[Xreg].i==0))
							set(C);
						else
							app(C);
						eval();
						break;
					}
				}
				else if (wRadix==10) {
					if (c==dPoint[k_useComma]) {
						if (L>=iMaxDigits) break;
						if ((iState>0?ST[Xreg].real:ST[Xreg].imag)==0)
							set('.');
						else
							app('.');
						eval();
						iState = 2*iSign;
						break;
					}
					else if (c=='e') {
						if (L>=iMaxDigits-3) break; // make sure there is room for
						app('e');                      // at least "e+1"
						app('+');
						eval();
						iState = 3*iSign;
						break;
					}
					else if (c=='m') {
						if (STR[0]=='-') delbeg(); else insbeg('-');
						eval();
						break;
					}
					else if ((c==Comma[k_useComma])&&(CURR_RADIX==COMPLEX)) {
						eval();
						iState=-4;
						set ('0');
						eval();
						break;
					}
					else if (iState==1 
							&& (c=='f' || c == ' ') 
							&& CURR_RADIX != COMPLEX
							&& isint(&ST[Xreg])) {
						app('/');
						eval();
						iState = 4;
						break;
					}
					else if (iState==1 
							&& (c=='\'' || c == '"') 
							&& CURR_RADIX != COMPLEX
							&& isint(&ST[Xreg])) {
						app('');
						eval();
						iState = 6;
						break;
					}
				}
			}
			
			else if (abs(iState)==2) {
				if (c==VK_BACK) {
					if (L==1) {
						set('0');
						eval();
						iState=iState>0 ? 0:-4;
						break;
					}
					else if (STR[L-1]=='.') {
						if ((L==2)&&(STR[0]=='-')) {
							set ('0');
							eval();
							iState=iState>0? 0:-4;
							break;
						}
						STR[L-1]=0;
						eval();
						iState=1*iSign;
						break;
					}
					else {
						STR[L-1]=0;
						eval();
						break;
					}
				}
				else if (val<wRadix) {
					if (L>=iMaxDigits) break;
					app(C);
					eval();
					break;
				}
				else if (c=='e') {
					if (L>=iMaxDigits-3) break;
					app('e');
					app('+');
					eval();
					iState=3*iSign;
					break;
				}
				else if (c=='m') {
					if (STR[0]=='-') delbeg(); else insbeg('-');
					eval();
					break;
				}
				else if ((c==Comma[k_useComma])&&(CURR_RADIX==COMPLEX)) {
					eval();
					iState=-4;
					set ('0');
					eval();
					break;
				}
			}
			
			else if (abs(iState)==3) {
				if (c==VK_BACK) {
					if ((STR[L-1]=='+')||(STR[L-1]=='-')) {
						int i;
						BOOL Dot=FALSE;
						STR[L-2]=0;
						for (i=L-3;i;i--) Dot|=STR[i]=='.';
						eval();
						iState = (Dot?2:1)*iSign;
						break;
					}
					
					STR[L-1]=0;
					eval();
					break;
				}
				else if (val<wRadix) {
					char ctemp[NDIGITS];
					int m;
					LD value;
					
					if (L>=iMaxDigits) break;
					app(C);
					for (m=L+1;STR[m]!='e';m--);            // find pos. of "e"
					lstrcpy(ctemp,&STR[m+1]);
					if (abs(atoi(ctemp))>MAXEXP) {
						STR[L] = 0;     // Exponent too big
						break;       
					}
					lstrcpyn(ctemp,STR,m);
					ctemp[m]=0;
					sscanf(STR,"%Lg",&value);
					if (value<0) value = -value;
					
					if (value>NUM_MAX) {
						STR[L] = 0;			// total number too big
						break;
					}
					if (value<NUM_MIN) {
						STR[L] = 0;			// total number too small
						break;
					}
					eval();
					break;
				}
				else if (c=='m') {
					int i=L-1;
					while ((STR[i]!='+')&&(STR[i]!='-')) i--;
					STR[i]=(char)(STR[i]=='-'?'+':'-');
					eval();
					break;
				}
				else if ((c==Comma[k_useComma])&&(CURR_RADIX==COMPLEX)) {
					eval();
					iState=-4;
					set ('0');
					eval();
					break;
				}
			}
			
			else if (iState==4)	{ // after first '/'
				if (c==VK_BACK) {
					if (STR[L-1]=='/')
					{
						STR[L-1]=0;
						eval();
						iState=1;
						break;
					}
					else {
						STR[L-1]=0;
						eval();
						break;
					}
				}
				else if (val<wRadix) {
					// if (L>=iMaxDigits) break;
					// Not needed - a check for overflow is done in Evaluate()
					app(C);
					eval();
					break;
				}
				else if (c=='m') {
					if (STR[0]=='-') delbeg(); else insbeg('-');
					eval();
					break;
				}
				else if (c=='f' || c==' ') {
					app('/');
					eval();
					iState=5;
					break;
				}
			}
			
			else if (iState==5) { // after second '/'
				if (c==VK_BACK) {
					if (STR[L-1]=='/') {
						STR[L-1]=0;
						eval();
						iState=4;
						break;
					}
					else {
						STR[L-1]=0;
						eval();
						break;
					}
				}
				else if (val<wRadix) {
					if (L>=iMaxDigits) break;
					app(C);
					eval();
					break;
				}
				else if (c=='m') {
					if (STR[0]=='-') delbeg(); else insbeg('-');
					eval();
					break;
				}
			}
			else if (iState==6) { // after first '
				if (c==VK_BACK) {
					if (STR[L-1]=='') {
						STR[L-1]=0;
						eval();
						iState=1;
						break;
					}
					else {
						STR[L-1]=0;
						eval();
						break;
					}
				}
				else if (val<wRadix) {
					if (L>=iMaxDigits) break;
					app(C);
					eval();
					break;
				}
				else if (c=='m') {
					if (STR[0]=='-') delbeg(); else insbeg('-');
					eval();
					break;
				}
				else if (c=='\'' || c=='"') {
					if (STR[L-1]=='') app('0');
					app('\'');
					eval();
					iState=7;
					break;
				}
			}
			else if (iState==7) { // after second '
				if (c==VK_BACK) {
					if (STR[L-1]=='\'') {
						STR[L-1]=0;
						eval();
						iState=6;
						break;
					}
					else {
						STR[L-1]=0;
						eval();
						break;
					}
				}
				else if (c==dPoint[k_useComma]) {
					if (L>=iMaxDigits) break;
					if (STR[L-1]=='\'') app('0');
					app('.');
					eval();
					iState=8;
					break;
				}
				else if (val<wRadix) {
					if (L>=iMaxDigits) break;
					app(C);
					eval();
					break;
				}
				else if (c=='m') {
					if (STR[0]=='-') delbeg(); else insbeg('-');
					eval();
					break;
				}
			}
			else if (iState==8) { // after second ' and .
				if (c==VK_BACK) {
					if (STR[L-1]=='.') {
						STR[L-1]=0;
						eval();
						iState=7;
						break;
					}
					else {
						STR[L-1]=0;
						eval();
						break;
					}
				}
				else if (val<wRadix) {
					if (L>=iMaxDigits) break;
					app(C);
					eval();
					break;
				}
				else if (c=='m') {
					if (STR[0]=='-') delbeg(); else insbeg('-');
					eval();
					break;
				}
			} // end if state...
			
			// **** Function Keys ****
			
			// Save X for those unary functions that cannot save it on stack -
			// to update LastX on success.
			
			SaveReg = ST[Xreg];
			
			switch (c)
			{
			case 'F':									// Switch clean/unclean fraction
				if (wRadix!=10) break;
				if (iState != 0) bSL = TRUE;
				iState = 0;
				bCleanFrac = !bCleanFrac;
				Dirty = TRUE;
				update();
				break;
				
			case 'k':									// Constants
				if (wRadix!=10) break;
				SaveUndo(UT_CONST,FALSE);
				if (DialogBox(hInst,"XCALCCONSTANT",hwndXCALC,ConstantDlgProc)) { // true=number to paste
					CommitUndo();
					if (iState != 0) bSL = TRUE;
					iState = 0;
					if (bSL) lift();
					bSL = TRUE;
					ST[Xreg].real = ldConstant;
					ST[Xreg].imag = 0;
					null(&ST[Xreg]);
					ST[Xreg].ct = ctREAL;
					Message = ConstMsg;
				}
				else {
					Message = "(no constant)";
				}
				QConstText(Message);
				Error = __MESSAGE;
				update();
				break;
				
			case 'K':									// Quick last constant
				if (wRadix!=10) break;
				if (ldConstant!=0) {
					SaveUndo(UT_CONST,TRUE);
					if (iState != 0) bSL = TRUE;
					iState = 0;
					if (bSL) lift();
					bSL = TRUE;
					ST[Xreg].real = ldConstant;
					ST[Xreg].imag = 0;
					null(&ST[Xreg]);
					ST[Xreg].ct = ctREAL;
					Message = ConstMsg;
				}
				else {
					Message = "(no constant)";
				}
				QConstText(Message);
				Error = __MESSAGE;
				update();
				break;
				
			case VK_ESCAPE:						// Minimise, or cancel sizing operation
				if (Sizing) {
					ESC_Pressed = TRUE;
					SendMessage(hwnd,WM_LBUTTONUP,0,0);
				}
				else if (Moving) {
					ESC_Pressed = TRUE;
					SendMessage(hwndMoving,WM_LBUTTONUP,0,0);
				}
				else {
				  PostMessage(hwnd,WM_SYSCOMMAND,SC_MINIMIZE,0);
				}
				break;
				
			case VK_RETURN:           // ENTER
				SaveUndo(UT_ENTER,TRUE);
				iState = 0;
				bSL = FALSE;
				lift();
				update();
				break;
				
			case VK_BACK:             // Clear
				SaveUndo(UT_CLR,TRUE);
				clrstkreg(Xreg);
				bSL = FALSE;
				update();
				break;
				
				// 'V' creates some weird zombie state on the keyboard.
				// Also, must get the quick convert to set proper error
				// on overflow.
			case 'v':case 'V':				// Convert
				if (wRadix!=10) break;
				SaveUndo(UT_CONV,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Convert(c=='V')) {
					QConvText(ConvMsg);
					ST[Xreg].ct=ctREAL;
					ST[Lreg]=SaveReg;
					Message=ConvMsg;
					Error=__MESSAGE;
					CommitUndo();
				}
				update();
				break;
				
			case 4:                 // D.ms
				if (wRadix!=10 || CURR_RADIX==COMPLEX) break;
				SaveUndo(UT_DMS,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!DMS()) {
					ST[Lreg]=SaveReg;
					CommitUndo();
				}
				update();
				break;
			case 'l':                 // LastX
				SaveUndo(UT_LASTX,TRUE);
				if (iState != 0) bSL = TRUE;
				iState = 0;
				if (bSL) lift();
				bSL = TRUE;
				ST[Xreg] = ST[Lreg];
				update();
				break;
				
			case '<':case '>':			 // X<>Y
				SaveUndo(UT_XY,TRUE);
				iState = 0;
				bSL = TRUE;
				X__Y();
				Dirty = TRUE;
				update();
				break;
				
			case '"':                // RE<>IM
				if (CURR_RADIX!=COMPLEX) break;
				SaveUndo(UT_REIM,TRUE);
				iState = 0;
				bSL = TRUE;
				R__I();
				update();
				break;
				
			case 'm':
				SaveUndo(UT_CHS,TRUE);
				iState = 0;
				bSL = TRUE;
				CHS();
				update();
				break;
				
			case '+':
				SaveUndo(UT_ADD,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Add()) {    // error return is nonzero from math routines
					SetLastX();
					drop();
					CommitUndo();
				}
				update();
				break;
				
			case '-':
				SaveUndo(UT_SUB,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Sub()) {
					SetLastX();
					drop();
					CommitUndo();
				}
				update();
				break;
				
			case '*':
				SaveUndo(UT_MUL,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Mul()) {
					SetLastX();
					drop();
					CommitUndo();
				}
				update();
				break;
				
			case '/':
				SaveUndo(UT_DIV,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Div()) {
					SetLastX();
					drop();
					CommitUndo();
				}
				update();
				break;
				
			case '%':                 // percent/modulo
				SaveUndo(wRadix==10?UT_PERC:UT_MOD,FALSE);
				iState = 0;
				bSL = TRUE;
				if (wRadix==10) {		// percent
					Register temp = ST[Xreg]; 
					if (!Percent()) {
						ST[Lreg] = temp;
						CommitUndo();
					}
				}
				else {                   // modulo
					if (!Mod()) {
						SetLastX();
						drop();
						CommitUndo();
					}
				}
				update();
				break;
				
			case 'M':                 // decimal modulo
				if (wRadix!=10) break;
				SaveUndo(UT_MOD,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Mod()) {
					SetLastX();
					drop();
					CommitUndo();
				}
				update();
				break;
				
			case 'D':                 // decimal integer division
				if (wRadix!=10) break;
				SaveUndo(UT_DIVF,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Divf()) {
					SetLastX();
					drop();
					CommitUndo();
				}
				update();
				break;
				
			case 'i':case'j':
				if (wRadix!=10) break;
				if (iState) {
					if (Evaluate()) break;
					iState = 0;
				}
				SaveUndo(UT_IJ,TRUE);
				bSL = TRUE;
				ST[Yreg].real = getreal(&ST[Yreg]);
				ST[Yreg].imag = getreal(&ST[Xreg]);
				null(&ST[Yreg]);
				if (CURR_RADIX!=COMPLEX) PostMessage(hwnd,PM_SETMODE,COMPLEX,0);
				drop();
				update();
				break;
				
			case 'I':case 'J':
				if (CURR_RADIX!=COMPLEX) break;
				if (iState) {
					if (Evaluate()) break;
					iState = 0;
				}
				SaveUndo(UT_NIJ,TRUE);
				bSL = TRUE;
				lift();
				ST[Yreg].real = ST[Xreg].real;
				ST[Yreg].imag = 0;
				ST[Xreg].real = ST[Xreg].imag;
				ST[Xreg].imag = 0;
				update();
				break;
				
			case 'p':
				if (wRadix!=10) break;
				SaveUndo(UT_PI,TRUE);
				if (iState != 0) bSL = TRUE;
				iState = 0;
				if (bSL) lift();
				bSL = TRUE;
				ST[Xreg].real = M_PI_L;
				ST[Xreg].imag = 0;
				ST[Xreg].ct = ctREAL;
				null(&ST[Xreg]);
				update();
				break;
				
			case '?':
				if (wRadix!=10) break;
				SaveUndo(UT_RAND,TRUE);
				iState = 0;
				if (bSL) lift();
				bSL = TRUE;
				Rand();
				update();
				break;
				
			case 'h':
				if (wRadix!=10) break;
				HyperbolicKeys(bHyperbolic=TRUE);
				return 0; // A "break" would force hyperbolic off again! (not any more)
				
			case 's':
				if (wRadix!=10) break;
				SaveUndo(bHyperbolic?UT_SINH:UT_SIN,FALSE);
				iState = 0;
				bSL = TRUE;
				if (bHyperbolic?!Sinh():!Sin()) {
					ST[Lreg]=SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 'c':
				if (wRadix!=10) break;
				SaveUndo(bHyperbolic?UT_COSH:UT_COS,FALSE);
				iState = 0;
				bSL = TRUE;
				if (bHyperbolic?!Cosh():!Cos()) {
					ST[Lreg]=SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 't':
				if (wRadix!=10) break;
				SaveUndo(bHyperbolic?UT_TANH:UT_TAN,FALSE);
				iState = 0;
				bSL = TRUE;
				if (bHyperbolic?!Tanh():!Tan()) {
					ST[Lreg]=SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 'S':
				if (wRadix!=10) break;
				SaveUndo(bHyperbolic?UT_ASINH:UT_ASIN,FALSE);
				iState = 0;
				bSL = TRUE;
				if (bHyperbolic?!Asinh():!Asin()) {
					ST[Lreg]=SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 'C':
				if (wRadix!=10) break;
				SaveUndo(bHyperbolic?UT_ACOSH:UT_ACOS,FALSE);
				iState = 0;
				bSL = TRUE;
				if (bHyperbolic?!Acosh():!Acos()) {
					ST[Lreg]=SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 'T':
				if (wRadix!=10) break;
				SaveUndo(bHyperbolic?UT_ATANH:UT_ATAN,FALSE);
				iState = 0;
				bSL = TRUE;
				if (bHyperbolic?!Atanh():!Atan()) {
					ST[Lreg]=SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case ':':
				if (wRadix!=10) break;
				SaveUndo(UT_CART,FALSE);
				iState = 0;
				bSL = TRUE;
				Dirty = TRUE;
				if (!Cart()) {
					ST[Lreg]=SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case ';':
				if (wRadix!=10) break;
				SaveUndo(UT_POLAR,FALSE);
				iState = 0;
				bSL = TRUE;
				Dirty = TRUE;
				if (!Polar()) {
					ST[Lreg]=SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 6: // Frac
				if (wRadix!=10) break;
				SaveUndo(UT_FRAC,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Frac()) {
					ST[Lreg]=SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 9: // Int
				if (wRadix!=10) break;
				SaveUndo(UT_INT,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Int()) {
					ST[Lreg]=SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case '$': case 'R': // Round
				if (wRadix!=10) break;
				SaveUndo(UT_RND,FALSE);
				iState = 0;
				bSL = TRUE;
				SendMessage(hwndXCALC,PM_UPDATE1,Xreg,0);
				if (!Round()) {
					ST[Lreg]=SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case '\'':
				if (CURR_RADIX!=COMPLEX) break;
				SaveUndo(UT_CONJG,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Conjg()) {
					ST[Lreg]=SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case '|':
				SaveUndo(wRadix==10?UT_ABS:UT_OR,FALSE);
				iState = 0;
				bSL = TRUE;
				if (wRadix==10) {         // Absolute value
					if (!Abs()) {
						ST[Lreg]=SaveReg;
						CommitUndo();
					}
				}
				else {                    // OR
					OR();
					SetLastX();
					drop();
					CommitUndo();
				}
				update();
				break;
				
			case 'o':                 // Power or OR
				SaveUndo(wRadix==10?UT_POW:UT_OR,FALSE);
				iState = 0;
				bSL = TRUE;
				if (wRadix==10) {        // Power
					if (!Power()) {
						SetLastX();
						drop();
						CommitUndo();
					}
				}
				else {                   // OR
					OR();
					SetLastX();
					drop();
					CommitUndo();
				}
				update();
				break;
				
			case 'r':                 // Root
				if (wRadix!=10) break;
				// Fall through: ROOT
				/*
				SaveUndo(UT_ROOT,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Root()) {
					SetLastX();
					drop();
					CommitUndo();
				}
				update();
				break;
				*/
			case '&':
				SaveUndo(wRadix==10?UT_ROOT:UT_AND,FALSE);
				iState = 0;
				bSL = TRUE;
				if (wRadix==10) {        // Root
					if (!Root()) {
						SetLastX();
						drop();
						CommitUndo();
					}
				}
				else {                   // AND
					AND();
					SetLastX();
					drop();
					CommitUndo();
				}
				update();
				break;
				
			case 'A':                 //  AND
				if (wRadix==10) break;
				SaveUndo(UT_AND,TRUE);
				iState = 0;
				bSL = TRUE;
				AND();
				SetLastX();
				drop();
				update();
				break;
				
			case '^':
				SaveUndo(wRadix==10?UT_POW:UT_XOR,FALSE);
				iState = 0;
				bSL = TRUE;
				if (wRadix==10) {        // Power
					if (!Power()) {
						SetLastX();
						drop();
						CommitUndo();
					}
				}
				else {                   // XOR
					XOR();
					SetLastX();
					drop();
					CommitUndo();
				}
				update();
				break;
				
			case 'x':                 //  XOR
				if (wRadix==10) break;
				SaveUndo(UT_XOR,TRUE);
				iState = 0;
				bSL = TRUE;
				XOR();
				SetLastX();
				drop();
				update();
				break;
				
			case '~': case 'n':       // NOT
				if (wRadix==10) break;
				SaveUndo(UT_NOT,TRUE);
				iState = 0;
				bSL = TRUE;
				NOT();
				ST[Lreg] = SaveReg;
				update();
				break;
				
			case '!':
				if (wRadix!=10) break;
				SaveUndo(UT_FACT,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Fact()) {
					ST[Lreg] = SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 'q':
				if (wRadix!=10) break;
				SaveUndo(UT_SQRT,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Sqrt()) {
					ST[Lreg] = SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 'Q':
				if (wRadix!=10) break;
				SaveUndo(UT_SQR,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Sqr()) {
					ST[Lreg] = SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 'u':
				if (wRadix!=10) break;
				SaveUndo(UT_QROOT,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Qroot()) {
					ST[Lreg] = SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 'U':
				if (wRadix!=10) break;
				SaveUndo(UT_CUBE,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Cube()) {
					ST[Lreg] = SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 'E':       //binary modes already trapped
				SaveUndo(UT_EXP,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Exp()) {
					ST[Lreg] = SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 'L':
				if (wRadix!=10) break;
				SaveUndo(UT_LN,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Ln()) {
					ST[Lreg] = SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 'N':
				if (wRadix!=10) break;
				SaveUndo(UT_TEN,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Ten()) {
					ST[Lreg] = SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 'G':
				if (wRadix!=10) break;
				SaveUndo(UT_LOG,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Log()) {
					ST[Lreg] = SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case '\\':
				if (wRadix!=10) break;
				SaveUndo(UT_RCP,FALSE);
				iState = 0;
				bSL = TRUE;
				if (!Rcp()) {
					ST[Lreg] = SaveReg;
					CommitUndo();
				}
				update();
				break;
				
			case 'X':case 'Z':
				{
					int i;
					SaveUndo(UT_CLSTK,TRUE);
					for (i=Xreg;i<StkSiz;i++) clrstkreg(i);
					// 3May99: Clear LastX also...
					clrstkreg(Lreg);
					iState = 0;
					bSL = FALSE;
					Dirty = TRUE;
					update();
					break;
				}
				
		} // end switch c
		
		// Update Hyperbolic keys if needed after keypress is handled
		HyperbolicKeys(bHyperbolic=FALSE);
		break;
		
	} // end case WM_CHAR

	case WM_KEYUP:
		{
			int virtkey = (int)wParam;
			if (Error) return 0;
			if (virtkey==VK_SHIFT) {
				bShift = FALSE;    // Needed for Shift-Left/Right handling;
				goto DefaultProc;  // For correct handling of Translate -> WM_CHAR
			}
			break;
		} // end case WM_KEYUP

	case WM_KEYDOWN:
		{
			int virtkey = (int)wParam;
			// remove error message if any
			ClearErr();
			// remove tooltip if any
			HideToolTip();
			
			// Jul 00: Hyperbolic mode was not reset when the arrows were pressed.
			switch (virtkey)
			{
			case VK_SHIFT:
				bShift = TRUE;		// Needed for Shift-Left/Right handling;
				goto DefaultProc;	// For correct handling of Translate -> WM_CHAR
				
			case VK_UP:
				HyperbolicKeys(bHyperbolic=FALSE);
				// Send key to program window if open
				if (IsWindow(hwndEditProgram)) {
					SendMessage(hwndEditProgram,PM_ADDTOPROGRAM,bShift?VK_SHIFT:virtkey,bShift?virtkey:0);
				}
				SaveUndo(UT_RUP,TRUE);
				iState = 0;
				bSL = TRUE;
				rollup();
				Dirty = TRUE;
				update();
				break;
				
			case VK_DOWN:
				HyperbolicKeys(bHyperbolic=FALSE);
				// Send key to program window if open
				if (IsWindow(hwndEditProgram)) {
					SendMessage(hwndEditProgram,PM_ADDTOPROGRAM,bShift?VK_SHIFT:virtkey,bShift?virtkey:0);
				}
				SaveUndo(UT_RDN,TRUE);
				iState = 0;
				bSL = TRUE;
				rolldn();
				Dirty = TRUE;
				update();
				break;
				
			case VK_LEFT:
				HyperbolicKeys(bHyperbolic=FALSE);
				// Send key to program window if open
				if (IsWindow(hwndEditProgram)) {
					SendMessage(hwndEditProgram,PM_ADDTOPROGRAM,bShift?VK_SHIFT:virtkey,bShift?virtkey:0);
				}
				if (wRadix==10) {
					SaveUndo(UT_FLOAT,FALSE);
					iState = 0;
					bSL = TRUE;
					if (sFix<9) {
						SendMessage(hwnd,PM_SETFIX,sFix+1,0);
						CommitUndo();
						update();
					}
				}
				else {
					if (bShift) {
						SaveUndo(UT_LASH,TRUE);
						iState = 0;
						bSL = TRUE;
						LAShift();
					}
					else {
						SaveUndo(UT_LSH,TRUE);
						iState = 0;
						bSL = TRUE;
						LShift();
					}
					update();
				}
				break;
				
			case VK_RIGHT:
				HyperbolicKeys(bHyperbolic=FALSE);
				// Send key to program window if open
				if (IsWindow(hwndEditProgram)) {
					SendMessage(hwndEditProgram,PM_ADDTOPROGRAM,bShift?VK_SHIFT:virtkey,bShift?virtkey:0);
				}
				if (wRadix==10) {
					SaveUndo(UT_FLOAT,FALSE);
					iState = 0;
					bSL = TRUE;
					if (sFix>FLOAT) {
						SendMessage(hwnd,PM_SETFIX,sFix-1,0);
						CommitUndo();
						update();
					}
				}
				else {
					if (bShift) {
						SaveUndo(UT_RASH,TRUE);
						iState = 0;
						bSL = TRUE;
						RAShift();
					}
					else {
						SaveUndo(UT_RSH,TRUE);
						iState = 0;
						bSL = TRUE;
						RShift();
					}
					update();
				}
				break;
				
			} // end switch virtkey
			
		} // end case WM_KEYDOWN

	default:
DefaultProc:
		return DefWindowProc(hwnd,wMsg,wParam,lParam);

	} // end switch wMsg

	return 0L;
}

